#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# iteration_mfac_pwa_case_ii.pl:
#	This Perl script calculates the portion of the probability for a local PWA
#	contributed by the local histories each of which consists of up to a specified number of indel events,
#	under fixed model parameters and a fixed value of the time lapse (t_F - t_I).
#	The local PWAs examined in this script are in cases (i) and (ii).
#	In case (i), there are neither ancestral sites nor descendant sites 
#	in between a pair of preserved ancestral sites (PASs) that are next to each other.
#	In case (ii), each of them has some ancestral sites and no descendant sites in between the PASs.
#
#	Because the calculation is done via an ITERATION, the script also outputs the
#	multiplication factors in the intermediate steps and along sub-intervals.
#
#	[CREATION STARTED on May 5 (Tue), 2015, ENDED on May 7 (Thu), 2015.]
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext tfm_labels2indices_gen
	);
use MyArrayManipu qw( diff_array );

=moved to 'MyPertProb_locPWA_itr_dawg'.
use MyPertProb_locPWA_dawg qw(
	mfac_asites_vs_dgaps_1event
	mfac_asites_vs_dgaps_2events
	);
=cut

use MyPertProb_locPWA_itr_dawg qw(
	initialize_probs_case_ii
	composite_probs_since_1st_mut_case_ii
	finish_iter_calc_probs_case_ii
	);


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 3;

##$MyPertProb_locPWA_itr_dawg::TEST_LEVEL1 = 1;
$MyPertProb_locPWA_itr_dawg::TEST_LEVEL1 = 2;
##$MyPertProb_locPWA_itr_dawg::TEST_LEVEL1 = 3;
##$MyPertProb_locPWA_itr_dawg::TEST_LEVEL1 = 5;

	### Parameters important for the iterative calculation. ###

our $TIME_INIT = 0; # Initial time.
our $TIME_FIN = 1.0; # Final time.
##our $TIME_FIN = 0.2; # Final time.

##our @TIME_LAPSES = (0.05, 0.2, 0.5, 1.0, 2.0); # Values of the time lapse (= t_F - t_I) to be examined.

=for the full-scale analysis (1,2)
#our $UB_CT_ANCSITES = 1000; # Upper-bound of the number of ancestral sites in between the PASs.
=cut
=for the full-scale analysis (3)
#our $UB_CT_ANCSITES = 500; # Upper-bound of the number of ancestral sites in between the PASs.
=cut
##=for the full-scale analysis 
our $UB_CT_ANCSITES = 300; # Upper-bound of the number of ancestral sites in between the PASs.
##=cut
=for TESTING
our $UB_CT_ANCSITES = 100; # Upper-bound of the number of ancestral sites in between the PASs.
=cut

##=for the full-scale analysis.
our $CT_SB_INTVLS = 100; # Number of sub-intervals for the numerical integration.
##our $CT_SB_INTVLS = 50; # Number of sub-intervals for the numerical integration.
##=cut
=for TESTING
our $CT_SB_INTVLS = 10; # Number of sub-intervals for the numerical integration.
=cut

##=for the full-scale analysis
our $INIT_CT_EVENTS = 2; # Initial iteration step (i.e., the number of indels up to which each initial probability accommodates.)
##=cut
=for Test0a & Test1a
our $INIT_CT_EVENTS = 1; # Initial iteration step (i.e., the number of indels up to which each initial probability accommodates.) # For 'Test0' & 'Test1' .
=cut

##=for the full-scale analysis.
our $UB_CT_EVENTS = 50; # The upper-bound of the number of indel events in a history.
#our $UB_CT_EVENTS = 20; # The upper-bound of the number of indel events in a history.
#our $UB_CT_EVENTS = 10; # The upper-bound of the number of indel events in a history.
##=cut
=for TESTING
our $UB_CT_EVENTS = 5; # The upper-bound of the number of indel events in a history.
#our $UB_CT_EVENTS = 2; # The upper-bound of the number of indel events in a history. # For 'Test0' & 'Test1' .
=cut



	### MODEL PARAMETERS ###

our $TOTRATE_INS = 0.1; # Total insertion rate (per unit time). (lambda_I in Dawg's indel model.)
our $TOTRATE_DEL = 0.1; # Total deletion rate (per unit time). (lambda_D in Dawg's indel model.)

our $EXPONENT_INS = 1.6; # Exponent for the power-law distribution of insertion lengths: f_I (l) is proportional to l^{-$EXPONENT_INS}.
our $EXPONENT_DEL = 1.6; # Exponent for the power-law distribution of deletion lengths: f_D (l) is proportional to l^{-$EXPONENT_DEL}.

=for the full-scale analysis (1)
#our $CUTOFF_INS_LEN = 5000; # Cutoff value for the insertion length.
#our $CUTOFF_DEL_LEN = 5000; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (2)
#our $CUTOFF_INS_LEN = 1000; # Cutoff value for the insertion length.
#our $CUTOFF_DEL_LEN = 1000; # Cutoff value for the deletion length.
=cut
=for the full-scale analysis (3lng)
our $CUTOFF_INS_LEN = 500; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 500; # Cutoff value for the deletion length.
=cut
##=for the full-scale analysis (3sht)
our $CUTOFF_INS_LEN = 100; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 100; # Cutoff value for the deletion length.
##=cut
=for TESTING
our $CUTOFF_INS_LEN = 200; # Cutoff value for the insertion length.
our $CUTOFF_DEL_LEN = 200; # Cutoff value for the deletion length.
=cut

	# The cut-off indel lengths for initializing the probabilities. #
our $INIT_CUTOFF_INS_LEN = $UB_CT_ANCSITES + 2*$CUTOFF_DEL_LEN;
our $INIT_CUTOFF_DEL_LEN = $UB_CT_ANCSITES + 2*$CUTOFF_INS_LEN;
($INIT_CUTOFF_INS_LEN < $CUTOFF_INS_LEN) and ($INIT_CUTOFF_INS_LEN = $CUTOFF_INS_LEN);
($INIT_CUTOFF_DEL_LEN < $CUTOFF_DEL_LEN) and ($INIT_CUTOFF_DEL_LEN = $CUTOFF_DEL_LEN);

my $nfac_ins = my $nfac_del = 0;
for (my $l=1; $l<= $INIT_CUTOFF_INS_LEN; $l++) { $nfac_ins += exp (-$EXPONENT_INS * log ($l)); }
for (my $l=1; $l<= $INIT_CUTOFF_DEL_LEN; $l++) { $nfac_del += exp (-$EXPONENT_DEL * log ($l)); }
##for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $nfac_ins += exp (-$EXPONENT_INS * log ($l)); }
##for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $nfac_del += exp (-$EXPONENT_DEL * log ($l)); }


our @FREQS_INS = (0); # Relative frequencies of the insertion lengths.
our @FREQS_DEL = (0); # Relative frequencies of the deletion lengths.

for (my $l=1; $l<= $INIT_CUTOFF_INS_LEN; $l++) { $FREQS_INS[$l] = exp (-$EXPONENT_INS * log ($l)) /$nfac_ins; }
for (my $l=1; $l<= $INIT_CUTOFF_DEL_LEN; $l++) { $FREQS_DEL[$l] = exp (-$EXPONENT_DEL * log ($l)) /$nfac_del; }
##for (my $l=1; $l<= $CUTOFF_INS_LEN; $l++) { $FREQS_INS[$l] = exp (-$EXPONENT_INS * log ($l)) /$nfac_ins; }
##for (my $l=1; $l<= $CUTOFF_DEL_LEN; $l++) { $FREQS_DEL[$l] = exp (-$EXPONENT_DEL * log ($l)) /$nfac_del; }

our $TOTRATE_INDEL = $TOTRATE_INS + $TOTRATE_DEL ;

	### Output files. ###

	# $outfile = "$outdir/$out_pref$itr_step$out_suff"

my $outdir = "Tinit${TIME_INIT}_Tfin${TIME_FIN}_Sbdiv$CT_SB_INTVLS.lngco"; ## For the full-scale analysis.
##my $outdir = 'TESTs/Test2'; ## For TESTING the script.
##my $outdir = 'TESTs/Test2.large_cutoff'; ## For TESTING the script.
##my $outdir = 'TESTs/Test0c.DETAIL2'; ## For TESTING the script.

my $out_pref = 'out_probs.up_to_';
my $out_suff = '_events.txt';

##my $out_summ = "$outdir/out_ratios_for_varied_tlapse.txt";


	### Create the output directory. ###

unless ( -d $outdir ) {
    unless (0 == system ("mkdir -p $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}
unless (( -r $outdir ) and ( -w $outdir )) {
    unless (0 == system ("chmod u+rw $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
    }
}


	### Declare the Beginning of the Script ###

##print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";

##print STDOUT "Summary output file: $out_summ .\n\n";

print STDOUT "Output files: {out_dir}/$out_pref", '{maxcount_events}', "$out_suff,\n";
print STDOUT "       where {out_dir} = $outdir ,\n";
print STDOUT "             {maxcount_events} = $INIT_CT_EVENTS, ..., $UB_CT_EVENTS .\n\n";

print STDOUT "\n", '$TEST_LEVEL = ', $TEST_LEVEL, " ,\n";
print STDOUT '$MyPertProb_locPWA_itr_dawg::TEST_LEVEL1 = ', $MyPertProb_locPWA_itr_dawg::TEST_LEVEL1, " .\n\n";


print STDOUT "\n<<  Parameters important for the iteration >>\n\n";

print STDOUT 'Time interval:  [$TIME_INIT, $TIME_FIN] = [', $TIME_INIT, ', ', $TIME_FIN, ']', " .\n\n";

print STDOUT '$UB_CT_ANCSITES = ', $UB_CT_ANCSITES, " ,\n";
print STDOUT '$CT_SB_INTVLS   = ', $CT_SB_INTVLS, " .\n\n";
print STDOUT '$INIT_CT_EVENTS = ', $INIT_CT_EVENTS, " ,\n";
print STDOUT '$UB_CT_EVENTS   = ', $UB_CT_EVENTS, " .\n\n";

print STDOUT "\n<< Model Parameters >>\n\n";

print STDOUT '$TOTRATE_INS = ', $TOTRATE_INS, " ,\n";
print STDOUT '$TOTRATE_DEL = ', $TOTRATE_DEL, " ,\n";
print STDOUT '($TOTRATE_INDEL = ', $TOTRATE_INDEL, ") .\n\n";

print STDOUT '@FREQS_INS[l] prpt l^{', -$EXPONENT_INS, "}   (l=1,...,$INIT_CUTOFF_INS_LEN) ,\n";
print STDOUT '@FREQS_DEL[l] prpt l^{', -$EXPONENT_DEL, "}   (l=1,...,$INIT_CUTOFF_DEL_LEN) .\n\n";
##print STDOUT '@FREQS_INS[l] prpt l^{', -$EXPONENT_INS, "}   (l=1,...,$CUTOFF_INS_LEN) ,\n";
##print STDOUT '@FREQS_DEL[l] prpt l^{', -$EXPONENT_DEL, "}   (l=1,...,$CUTOFF_DEL_LEN) .\n\n";


if ($TEST_LEVEL>2) {

   print STDOUT "\n< Indel length frequencies >\n\n";

   print STDOUT join ("\t", 'Length', 'Ins', 'Del'), "\n\n";

   my $cutoff = ($INIT_CUTOFF_INS_LEN > $INIT_CUTOFF_DEL_LEN) ? $INIT_CUTOFF_INS_LEN : $INIT_CUTOFF_DEL_LEN ;
   ##my $cutoff = ($CUTOFF_INS_LEN > $CUTOFF_DEL_LEN) ? $CUTOFF_INS_LEN : $CUTOFF_DEL_LEN ;

   for (my $l=1; $l<= $cutoff; $l++) {
	my $f_ins = ($l > $INIT_CUTOFF_INS_LEN) ? 'undef' : $FREQS_INS[$l] ;
	my $f_del = ($l > $INIT_CUTOFF_DEL_LEN) ? 'undef' : $FREQS_DEL[$l] ;
	##my $f_ins = ($l > $CUTOFF_INS_LEN) ? 'undef' : $FREQS_INS[$l] ;
	##my $f_del = ($l > $CUTOFF_DEL_LEN) ? 'undef' : $FREQS_DEL[$l] ;

	print STDOUT join ("\t", $l, $f_ins, $f_del), "\n";
   }
   print STDOUT "\n\n";

}  # END of "if ($TEST_LEVEL>2) {...}". #


print STDOUT "\n<<< Preparation >>>\n\n";


our @TIME_POINTS = ();
for (my $i=0; $i<= $CT_SB_INTVLS; $i++) { $TIME_POINTS[$i] = $TIME_INIT + $i*($TIME_FIN - $TIME_INIT)/$CT_SB_INTVLS; }

print STDOUT "Time points: ", join (', ', @TIME_POINTS), " .\n\n";

our @LABELS_FLDS = ('#{Anc_sites}');
for (my $i=0; $i<= $CT_SB_INTVLS; $i++) {
	my $t_init = $TIME_POINTS[$i];
	push @LABELS_FLDS, "[$t_init,$TIME_FIN]";
}


print STDOUT "\n<<< Main Analysis >>>\n\n\n";

print STDOUT "<< Initializing the probabilities >>\n\n"; 

	# Initialize the upper-bound of #{ancestral sites}. #

my $ub_ct_ancsites = $INIT_CUTOFF_DEL_LEN;  # For LONG-cutoffs
#my $ub_ct_ancsites = $CUTOFF_DEL_LEN;  # For SHORT-cutoffs
##my $ub_ct_ancsites = 2; # For Test0c.

print STDOUT "{Upper-bound of #{ancestral sites}} = $ub_ct_ancsites ,\n";
print STDOUT "{Initial upper-bound of #{events per history}} = $INIT_CT_EVENTS .\n\n";

my @probs = (); # $probs[$i_tp]->[$n_A] ($i_tp = 0, ..., $CT_SB_INTVLS; $n_A = 0, ..., $ub_ct_ancsites) is the probability that the sequence had $n_A sites in between the PASs at time $TIME_POINTS[$i_tp], and that it has no sites in between the PASs at time $TIME_FIN, calculated in each iteration step.

	# For LONG-cutoffs
unless ( initialize_probs_case_ii (@probs, @TIME_POINTS, $ub_ct_ancsites, $INIT_CT_EVENTS, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, @FREQS_DEL, $INIT_CUTOFF_INS_LEN, $INIT_CUTOFF_DEL_LEN) ) {
	# For SHORT-cutoffs
#unless ( initialize_probs_case_ii (@probs, @TIME_POINTS, $ub_ct_ancsites, $INIT_CT_EVENTS, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, @FREQS_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN) ) { 
	print STDOUT "[ERROR '$program_name'] The subroutine 'initialize_probs_case_ii' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

	## Print out the initial probabilities. ##

my $outfile0 = "$outdir/$out_pref$INIT_CT_EVENTS$out_suff";
my $fh_out0 = open_output_file ($outfile0, 1);
unless (defined $fh_out0) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_out0 join ("\t", @LABELS_FLDS), "\n\n";

for (my $n_A=0; $n_A <= $ub_ct_ancsites; $n_A++) { # Preliminary for-loop (over $n_A's). #
    my @output = ($n_A);
    for (my $i_tp=0; $i_tp <= $CT_SB_INTVLS; $i_tp++) { push @output, $probs[$i_tp]->[$n_A]; }
    print $fh_out0 join ("\t", @output), "\n";
}					# End of the preliminary for-loop (over $n_A's). #
close ($fh_out0);

print STDOUT "\n\n<< Iterative calculations >>\n\n";

my $last_ub_ct_ancsites = $ub_ct_ancsites;

for (my $ct_events = $INIT_CT_EVENTS + 1; $ct_events <= $UB_CT_EVENTS; $ct_events++) { # Main outer for-loop (over the upper-bound of #{events per history}). # 

    print STDOUT "\n<< Incorporating histories each of which consists of $ct_events indels... >>\n\n";

    #$ub_ct_ancsites = $CUTOFF_DEL_LEN ; # For SHORT-cutoffs
    $ub_ct_ancsites = $INIT_CUTOFF_DEL_LEN ; # For LONG-cutoffs

=A time- and memory-saving measure (now obsolete).
    my $cand_ub_ctas = $UB_CT_ANCSITES + ($UB_CT_EVENTS - $ct_events) * $CUTOFF_INS_LEN ;
    $ub_ct_ancsites = ($INIT_CUTOFF_DEL_LEN < $cand_ub_ctas) ?  $INIT_CUTOFF_DEL_LEN : $cand_ub_ctas ;
=cut

    print STDOUT "{Upper-bound of #{ancestral sites}} = $ub_ct_ancsites .\n\n";

    print STDOUT "\n< Calculate the intermediate \"composite probability densities\"... >\n\n";

=for SHORT-cutoffs
    my $cmp_probs = composite_probs_since_1st_mut_case_ii (@probs, $last_ub_ct_ancsites, $ub_ct_ancsites, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, @FREQS_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN);
=cut

#=for LONG-cutoffs
    my $cmp_probs = composite_probs_since_1st_mut_case_ii (@probs, $last_ub_ct_ancsites, $ub_ct_ancsites, $TOTRATE_INS, $TOTRATE_DEL, @FREQS_INS, @FREQS_DEL, $INIT_CUTOFF_INS_LEN, $INIT_CUTOFF_DEL_LEN);
#=cut

    unless (defined $cmp_probs) {
	print STDOUT "[ERROR '$program_name'] 'composite_probs_since_1st_mut_case_ii' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    print STDOUT "\n< Finish calculating the probabilities with new histories incorporated... >\n\n";

    my $new_probs = finish_iter_calc_probs_case_ii (@{$cmp_probs}, $ub_ct_ancsites, @TIME_POINTS, $TOTRATE_INDEL);

    unless (defined $new_probs) {
	print STDOUT "[ERROR '$program_name'] 'finish_iter_calc_probs_case_ii' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }

    print STDOUT "\n< Print out the new probabilities just calculated... >\n\n";

    my $outfile = "$outdir/$out_pref$ct_events$out_suff";
    my $fh_out = open_output_file ($outfile, 1);

    print $fh_out join ("\t", @LABELS_FLDS), "\n\n";

    for (my $n_A=0; $n_A <= $ub_ct_ancsites; $n_A++) { # Inner for-loop (over #{ancestral sites}). #
	my @output = ($n_A);
	for (my $i_tp=0; $i_tp <= $CT_SB_INTVLS; $i_tp++) { push @output, $new_probs->[$i_tp][$n_A]; }
	print $fh_out join ("\t", @output), "\n";
    }				# End of the inner for-loop (over #{ancestral sites}). #

    close ($fh_out);

	# Update the main variables. #

    @probs = @{$new_probs};
    $last_ub_ct_ancsites = $ub_ct_ancsites;

}				# End of the main outer for-loop (over the upper-bound of #{events per history}). #


print STDOUT "\n\n=== All Processes FINISHED !! ===\n\n";



