#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# bch_serial_br_infer_indels_dawg.hs.pl:
#	This modified version of 'bch_serial_br_infer_indels_dawg.pl' serially performs
#	'br_infer_psm_indel_events_spt_odr_dawg.hs.pl' to analyze a batch of
#	multiple sequence alignments simulated via Dawg (and pre-processed via 'preprocess_msa_dawg.pl').
#
#	[CREATED on May 18 (Mon), 2015.]
#
# bch_serial_br_infer_indels_dawg.pl:
#	This modified version of 'bch_serial_br_infer_indels.pl' can analyze a batch of
#	multiple sequence alignments simulated via Dawg.
#	(CREATED on Feb 28 (Fri), 2014)
#
# bch_serial_br_infer_indels.pl:
#	This Perl script performs bach analyses to infer (parsimonious) indels forming the input sequence alignments,
#	all of which share a species tree, stored in multiple gzipped tar archives.
#	The script was created by modifying 'bch_serial_PAML.pl'.
#	(CREATED on Nov 22 (Fri), 2013)
#
# bch_serial_PAML.pl:
#	This Perl script performs bach PAML analyses on the input sequence alignments,
#	all of which share a species tree, stored in multiple gzipped tar archives.
#	The script was created by modifying 'bchbch_refine_msas.pl',
#	and borrowing functions to perform PAML anayses from 'compare_ml_msa.bch.extJC.6t.PhyML.pl'.
#	(CREATION STARTED on Nov 3 (Sun), 2013, ENDED on Nov 3 (Sun), 2013.)
#
# bchbch_refine_msas.pl:
#	This Perl script refines MSAs stored in multiple archives.
#	The script was created by using parts of 'serial_pgene_ortho_aln_analyses.pl'.
#	(CREATION STARTED on Oct 21 (Mon), 2013, ENDED on Oct xx (yyy), 2013.)
#
# serial_pgene_ortho_aln_analyses.pl:
#	This Perl script inspects the alignments accommodating progo-genes,
#	and examines whether the orthologous region is 'missing' or not
#	in each of 3 other yeast species.
#	Once the orthologous region is found not "missing", 
#	the script goes on to calculate the alignment coverage, nucleotide similarity, and the three RFC scores.
#	The script was created by modifying 'serial_pgene_RFC_analyses.pl'.
#	(CREATION STARTED on Oct 7 (Mon), 2013, ENDED on Oct 11 (Fri), 2013)
#


use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext
	tfm_labels2indices tfm_labels2indices_gen
	);

=for the predecessor 'bchbch_refine_msas.pl'
use MyArrayManipu qw( multiple_sort );

use MyMalnIO qw(
	extract_fasta extract_clw
	pr_ODEN_input
	pr_clustal_like_lite
	);
	## pr_phylip_aln
	## pr_clustal_like_lite_aa

use MyNucSeq2 qw(
	revcom print_sequence
	);

use MyProbDistri qw(
	poisson_cum
	binomial_cum_via_betai
	);
=cut

	# IMPORTED from 'compare_ml_msa.bch.extJC.6t.PhyML.pl' #
use MyPAML_parser qw( prepare_paml_ctrl );
	# retrieve_aa_freqs 

sub peek_into_archive (\$$$$);

=for the predecessor 'bchbch_refine_msas.pl'
sub cal_pvs_sbst_indel ($$$\@\@$$);
sub pwaln_surveillance (\$\$$$$$$$$);
sub msa_surveillance (\@\@$$$$\%$\%$);
sub mask_aln_regions (\@\@$$$$$);
=cut


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;

	## Test Output Levels.
	## 0, silent; 1, very brief; 2, pretty brief; 3, rather verbose; 4, verbose; 5, very verbose, >5: extremely verbose. ##

	# (for the main part):
##our $TEST_LEVEL = 1;
our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;


	### Executable file. ###

my $main_indel_script = "Scripts/br_infer_psm_indel_events_spt_odr_dawg.hs.pl";
##my $main_indel_script = "Scripts/br_infer_psm_indel_events_spt_odr_dawg.v2.pl";

=for the predecessor 'bch_serial_br_infer_indels.pl'.
	### Input MSA type to be used. ###
	##  -- This basically determines the type of input MSAs and the label of the output directory. -- ##

##my $input_msa_type = "V20/Pv_Sbst0.01_Indel0.01";
my $input_msa_type = "V20/Pv_Sbst0.05_Indel0.05";
=cut

	##### Input/Output Files #####

##my $session_name = 'hogd_alt_lite_sbcl.1'; # For the full-analysis. #
=for 'Proof_of_Principle2'
##my $session_name = 'trio_blen0.05'; # For the full-analysis.
##my $session_name = 'trio_blen0.2'; # For the full-analysis.
=cut
my $session_name = 'TESTs/Test1a'; # For TESTING the script. #

	## Archives of tiling MSAs. ##
	# $in_archive = "$inarch_dir/$archive_pref$id_class$archive_suff".

my $inarch_dir = "../../Inputs/Processed_MSAs1/$session_name/Outputs";
=for the past sessions.
##my $inarch_dir = "../../Inputs/Processed_MSAs2_1/$session_name/Outputs";
##my $inarch_dir = "../../Inputs/Simulated_MSAs2_1/$session_name";
##my $inarch_dir = "../../Inputs/Simulated_MSAs1/hogd_alt_lite_sbcl.1";
=cut

##my $archive_pref = 'class_'; # For 'Proof_of_Principle2'.
my $archive_pref = 'hog_d';
##my $archive_pref = 'Chr_';
my $archive_suff = '.tgz';


	### Top Output Directory ###

my $topoutdir = $session_name; # For full-scale analyses.
##my $topoutdir = "TESTs/Test3c"; ## For TESTING. ##
##my $topoutdir = 'hogd_alt_lite_sbcl.1'; # For full-scale analyses.


	## Temporary Input Files ##
	# $in_msa = "$tmp_dir/$inaln_pref$lcl_msa_id$inaln_suff", and
	# $ctrl_file = "$tmp_dir/$ctrl_pref$lcl_msa_id$ctrl_suff",
	# where $tmp_dir = "$intmp_aln_dir/$archive_pref$id_class",
	# where $id_class is the ID of the class of MSAs,
	# and $lcl_msa_id is the ID of a MSA in a particular class.

my $intmp_aln_dir = "$topoutdir/Temporary";

my $inaln_pref = 'prcsd_dawg';
##my $inaln_pref = 'dawg';
##my $inaln_pref = 'e';
##my $inaln_pref = 'optionG_';
my $inaln_suff = '.aln';

my $ctrl_pref = 'dawg_inp';
my $ctrl_suff = '.txt';


=for the predecessors down to 'bch_serial_br_infer_indels.pl'
	## Template files. (ADDED in creation of 'bch_serial_PAML.pl'.) ##
	##
	## $tmpl_intree = "$tmpl_dir/$intree_name",
	## $tmpl_ctrl_file = "$tmpl_dir/$ctrl_file_name",

my $tmpl_dir = "Input_Files";
##my $tmpl_dir = "Templates";
my $intree_name = 'four_yeasts.tree';
##my $ctrl_file_name = 'baseml.ctl';

my $tmpl_intree = "$tmpl_dir/$intree_name";
##my $tmpl_ctrl_file = "$tmpl_dir/$ctrl_file_name";
=cut

	# Directory storing auxiliary Input Data Files. ##
	# (ADDED in creation of 'bch_serial_br_infer_indels_dawg.hs.pl'.) #
my $aux_indir = 'InData/PL1.6_co100';


	## Output Files. (ADDED in creation of 'bch_serial_br_infer_indels.pl'.) ##
	## (MODIFIED in creation of 'bch_serial_br_infer_indels_dawg.pl'.) 
	##
	## $outfile1 = "$out_indel_dir/$pref_outfile1$lcl_msa_id$suff_outfile1",
	## $outfile2 = "$out_indel_dir/$pref_outfile2$lcl_msa_id$suff_outfile2",
	## $logfile  = "$out_indel_dir/$pref_logfile$lcl_msa_id$suff_logfile",
	# where $out_indel_dir = "$top_indel_dir/$archive_pref$id_class/";

my $top_indel_dir = "$topoutdir/Outputs";
##my $top_indel_dir = "$topoutdir/Indel_analyses";


	# MODIFIED in creation of 'bch_serial_br_infer_indels_dawg.pl'. #
my $pref_outfile1 = 'output_detail.';
my $pref_outfile2 = 'output_brief.';
my $pref_logfile = 'log_infer_indel.';

my $suff_outfile1 = my $suff_outfile2 = my $suff_logfile = '.txt';

=for the predecessor 'bch_serial_br_infer_indels.pl'
	## $outfile1 = "$out_indel_dir/$name_outfile1",
	## $outfile2 = "$out_indel_dir/$name_outfile2",
	## $logfile  = "$out_indel_dir/$name_logfile",
	# where $out_indel_dir = "$top_indel_dir/$archive_pref$chr_no/$anchor_orf_id/";
my $name_outfile1 = 'output_detail.txt';
my $name_outfile2 = 'output_brief.txt';
my $name_logfile = 'log_br_infer_psm_indel_events.txt';
=cut


	## Check files/directories ##

	# Main script (ADDED in creation of 'bch_serial_br_infer_indels.pl').
foreach my $main_scr ($main_indel_script) {
    unless (( -f $main_scr ) and ( -r $main_scr )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable main script '$main_scr' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
}

foreach my $indir ($inarch_dir, $aux_indir) {
##foreach my $indir ($inarch_dir) {...} # For the predecessor 'bch_serial_br_infer_indels_dawg.pl'.
    unless (( -d $indir ) and ( -r $indir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input directory '$indir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
}

=for the predecessors down to 'bch_serial_br_infer_indels.pl'
	# Template files (ADDED in creation of 'bch_serial_PAML.pl'). #
foreach my $tmpl_file ($tmpl_intree) {
##foreach my $tmpl_file ($tmpl_intree, $tmpl_ctrl_file) {
    unless (( -f $tmpl_file) and ( -r $tmpl_file)) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable template file '$tmpl_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
}
=cut

my $list_id_classes = list_sessions_ext ($inarch_dir, $archive_pref, $archive_suff, '-');
##my $list_chr_nos = list_sessions_ext ($inarch_dir, $archive_pref, $archive_suff, '-');

unless (defined $list_id_classes) {
##unless (defined $list_chr_nos) {...} # For the predecessors down to 'bch_serial_br_infer_indels.pl'. #
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the archive directory '$inarch_dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
}

## (@{$list_id_classes}>2) and (@{$list_id_classes} = @{$list_id_classes}[0,1]); # For TESTING the script. #

foreach my $out_dir ($intmp_aln_dir, $top_indel_dir) {
##foreach my $out_dir ($intmp_aln_dir, $top_paml_dir) {...}
    unless ( -d $out_dir) {
	print STDOUT "[WARNING '$program_name'] CanNOT FIND an output directory '$out_dir' ! ... I will try to create one...\n\n";
	unless (0 == system (" mkdir -p $out_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the output directory '$out_dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
    unless ( -w $out_dir ) {
	print STDOUT "[WARNING '$program_name'] The output directory '$out_dir' is NOT WRITABLE ! ... I will try to change its permission mode...\n\n";
	unless (0 == system (" chmod -R u+w $out_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$out_dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<<< Program name: $program_name (path: $program_path) >>>>\n\n";


print STDOUT "Session Name: $session_name .\n\n"; # ADDED in creation of 'bch_serial_br_infer_indels_dawg.pl'. #

=for the predecessor 'bch_serial_br_infer_indels.pl'.
print STDOUT "Input MSA Type: '$input_msa_type' .\n\n";
=cut

print STDOUT '$CURR_WORK_DIR = ', $CURR_WORK_DIR, " .\n\n";
##print STDOUT '$DIS_LEN = ', $DIS_LEN, " .\n";

	# MODIFIED in creation of 'bch_serial_br_infer_indels.pl'. 
print STDOUT "Main script for the analysis: $main_indel_script .\n\n";
##print STDOUT "Main PAML executable file: $main_bin_paml .\n\n"; # ADDED in creation of 'bch_serial_PAML.pl'.

print STDOUT "Input MSA archives: $inarch_dir/$archive_pref", '{ID_class}', "$archive_suff ,\n";
print STDOUT "          where {ID_class} = ", join (', ', @{$list_id_classes}), " .\n\n";
print STDOUT "Temporary input MSAs: $intmp_aln_dir/$archive_pref", '{ID_class}', "/$inaln_pref", '{local_MSA_ID}', "$inaln_suff .\n\n";   
=for the predecessor 'bch_serial_br_infer_indels.pl'
print STDOUT "Input MSA archives: $inarch_dir/$archive_pref", '{chromosome_no.}', "$archive_suff ,\n";
print STDOUT "          where {chromosome no.} = ", join (', ', @{$list_chr_nos}), " .\n\n";
print STDOUT "Temporary input MSAs: $intmp_aln_dir/$archive_pref", '{chromosome_no.}', "/$inaln_pref", '{anchor_ORF_ID}', "$inaln_suff .\n\n";   
=cut


	## ADDED in creation of 'bch_serial_PAML.pl' ##

print STDOUT "Auxiliary (temporary) input file for indel analyses:\n";

	# MODIFIED in creation of 'bch_serial_br_infer_indels_dawg.pl'. #
print STDOUT "  (input tree and indel parameters)  $intmp_aln_dir/$archive_pref", '{ID_class}', "/$ctrl_pref", '{local_MSA_ID}', "$ctrl_suff .\n\n";   
##print STDOUT "     (input tree)  $tmpl_intree .\n\n";


			# ADDED in creation of 'bch_serial_br_infer_indels_dawg.hs.pl'. #
print STDOUT "Directory storing auxiliary input data files (recording multiplication factors):\n\n";
print STDOUT "           $aux_indir .\n\n";

print STDOUT "Final Output Directories: {out_indel_dir} = $top_indel_dir/$archive_pref", '{chromosome_no.}/', " .\n\n";
print STDOUT "Output Files:\n";
print STDOUT "     (detailed) {out_indel_dir}/$pref_outfile1", '{local_MSA_ID}', "$suff_outfile1 ,\n";
print STDOUT "     (brief)    {out_indel_dir}/$pref_outfile2", '{local_MSA_ID}', "suff_outfile2 ,\n";
print STDOUT "     (log)      {out_indel_dir}/$pref_logfile", '{local_MSA_ID}', "suff_logfile .\n\n";
	##
	## $outfile1 = "$out_indel_dir/$pref_outfile1$lcl_msa_id$suff_outfile1",
	## $outfile2 = "$out_indel_dir/$pref_outfile2$lcl_msa_id$suff_outfile2",
	## $logfile  = "$out_indel_dir/$pref_logfile$lcl_msa_id$suff_logfile",
	# where $out_indel_dir = "$top_indel_dir/$archive_pref$id_class/";

=for the predecessor 'bch_serial_br_infer_indels.pl'
	## ADDED in creation of 'bch_serial_PAML.pl' ##
print STDOUT "Final Output Directories: {out_indel_dir} = $top_indel_dir/$archive_pref", '{chromosome_no.}/{anchor_ORF_ID}/', " .\n\n";
print STDOUT "Output Files:\n";
print STDOUT "     (detailed) {out_indel_dir}/$name_outfile1 ,\n";
print STDOUT "     (brief)    {out_indel_dir}/$name_outfile2 ,\n";
print STDOUT "     (log)      {out_indel_dir}/$name_logfile .\n\n";
	##
	## $outfile1 = "$out_indel_dir/$name_outfile1",
	## $outfile2 = "$out_indel_dir/$name_outfile2",
	## $logfile  = "$out_indel_dir/$name_logfile",
	# where $out_indel_dir = "$top_indel_dir/$archive_pref$chr_no/$anchor_orf_id/";
=cut



##my $ct_chr = 
my $ct_classes = my $ct_msas = my $ct_anom = my $ct_fail = my $ct_succ = 0;

foreach my $id_class (@{$list_id_classes}) { # Outer foreach-loop. (over archives (or classes)) #
##foreach my $chr (@{$list_chr_nos}) {...} # Outer foreach-loop. (over archives (or chromosomes)) #

    ##my $chr_no = $chr;

    ## ($id_class eq '1_4') or next; # For TESTING the script ('Test3b'). #
    ## ($id_class eq '2_3') or next; # For TESTING the script ('Test3c'). #
    ## ($id_class eq '2_7') or next; # For TESTING the script ('Test2a'). #
    ## (($id_class eq '1_1') or ($id_class eq '2_2')) or next; # For TESTING the script ('Test1a' & 'Test1b' ). #
    ## ($id_class eq '0003') or next; # For TESTING the script ('Test5a'). #
    ## ($id_class eq '0005') or next; # For TESTING the script ('Test6a'). #
    ## ($id_class eq '0002') or next; # For TESTING the script ('Test6a' & 'Test7a'). #
    ## ($id_class eq '0001') or next; # For TESTING the script ('Test8a' ). #

=for the predecessor 'bch_serial_br_infer_indels.pl'.`
    ## ($chr eq '01') or next; # For TESTING the script ('Test1a' & 'Test5a,c') #
    ## ($chr eq '05') or next; # For TESTING the script ('Test2a') #
=cut

    print STDOUT "<< Class: $id_class >>\n\n";
    ##print STDOUT "<< Chromosome: $chr >>\n\n";

    my $in_archive = join ('', $inarch_dir, '/', $archive_pref, $id_class, $archive_suff);
    ##my $in_archive = join ('', $inarch_dir, '/', $archive_pref, $chr_no, $archive_suff);
		# $in_archive = "$inarch_dir/$archive_pref$chr_no$archive_suff", and 

    unless (( -f $in_archive ) and ( -r $in_archive ) ) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input archive file '$in_archive' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
    }

    my $mid_dir = join ('', $archive_pref, $id_class);
    ##my $mid_dir = join ('', $archive_pref, $chr_no);

    my $list_in_msa_ids = peek_into_archive ($in_archive, $mid_dir, $inaln_pref, $inaln_suff);
    unless (defined $list_in_msa_ids) {
	    print STDOUT "[ERROR '$program_name'] 'peek_into_archive' FAILED on the input archive '$in_archive' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
    }
    ## (@{$list_in_msa_ids}>2) and (@{$list_in_msa_ids} = @{$list_in_msa_ids}[0,1]); ## For TESTING the script. ##
    ## @{$list_in_msa_ids} = ('0004', '0122'); ## For TESTING the script ('Test2a'). ##
    ## @{$list_in_msa_ids} = ('0227'); ## For TESTING the script ('Test3b'). ##
    ## @{$list_in_msa_ids} = ('0120'); ## For TESTING the script ('Test3c'). ##
    ## @{$list_in_msa_ids} = ('0000746'); ## For TESTING the script ('Test5a'). ##
    ## @{$list_in_msa_ids} = ('0001033'); ## For TESTING the script ('Test6a'). ##
    ## @{$list_in_msa_ids} = ('0000385'); ## For TESTING the script ('Test7a'). ##
    ## @{$list_in_msa_ids} = ('0000073'); ## For TESTING the script ('Test8a'). ##

  (print STDOUT "#{Input MSAs in the archive} = ", scalar (@{$list_in_msa_ids}), " .\n\n") if ($TEST_LEVEL>0);


		# MODIFIED in creation of 'bch_serial_br_infer_indels.pl'. #

    my $parent_out_indel_dir = "$top_indel_dir/$mid_dir";
	# $out_indel_dir = "$top_indel_dir/$archive_pref$chr_no/$anchor_orf_id/";

    unless ( -d $parent_out_indel_dir) { # MODIFIED in creation of 'bch_serial_br_infer_indels.pl'. #
	unless (0 == system (" mkdir -p $parent_out_indel_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the parent output directory '$parent_out_indel_dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
    unless ( -w $parent_out_indel_dir ) { # MODIFIED in creation of 'bch_serial_br_infer_indels.pl'. #
	unless (0 == system (" chmod -R u+w $parent_out_indel_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the parent output directory '$parent_out_indel_dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }


    $ct_classes++;
    ##$ct_chr++;


    foreach my $in_msa_id (@{$list_in_msa_ids}) { # Middle foreach-loop (over MSAs in the archive) #

	$ct_msas++;

	## ($ct_msas>3) and last; # For TESTING the script. ('Test1a' & 'Test2a') for the OLD ones. #
	##($ct_msas>0) and last; # For TESTING the script. #
	##($in_msa_id eq 'YAL014C') or next; # For TESTING the script ('Test5a') #
	##($in_msa_id eq 'YAR008W') or next; # For TESTING the script ('Test5c') #

  (print STDOUT "< Input MSA ID: $in_msa_id >\n\n") if ($TEST_LEVEL>0);


		# Extract the input MSA file. #

	my $in_msa_name = join ('', $inaln_pref, $in_msa_id, $inaln_suff);
	my $extracted_file = "$mid_dir/$in_msa_name";
	my $tmpin_msa = "$intmp_aln_dir/$extracted_file";

		# $tmpin_msa = "$intmp_aln_dir/$archive_pref$chr_no/$inaln_pref$in_msa_id$inaln_suff",
		# where $chr_no is the numerical part of the chromosome ID (in the BED file(s)).

	unless (0 == system ("tar -C $intmp_aln_dir -xpzf $in_archive $extracted_file") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to extract a file '$extracted_file' into '$intmp_aln_dir' from the archive '$in_archive', for the MSA '$in_msa_id' belonging to the class '$id_class': $? !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
	unless (( -f $tmpin_msa ) and ( -r $tmpin_msa )) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable temporary input MSA file '$tmpin_msa': $! !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}

		# Extract the Dawg control file. #
		# (ADDED in creation of 'bch_serial_br_infer_indels_dawg.pl'.)

	my $ctrl_file_name = join ('', $ctrl_pref, $in_msa_id, $ctrl_suff);
	my $extracted_ctrl_file = "$mid_dir/$ctrl_file_name";
	my $tmpin_ctrl = "$intmp_aln_dir/$extracted_ctrl_file";

		# $ctrl_file = "$tmp_dir/$ctrl_pref$lcl_msa_id$ctrl_suff",
		# where $tmp_dir = "$intmp_aln_dir/$archive_pref$id_class",

	unless (0 == system ("tar -C $intmp_aln_dir -xpzf $in_archive $extracted_ctrl_file") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to extract a file '$extracted_ctrl_file' into '$intmp_aln_dir' from the archive '$in_archive', for the MSA '$in_msa_id' belonging to the class '$id_class': $? !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
	unless (( -f $tmpin_ctrl ) and ( -r $tmpin_ctrl )) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable temporary Dawg control file '$tmpin_ctrl': $! !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}


		# Create a directory for performing an Indel analysis. #
		# (ADDED in creation of 'bch_serial_br_infer_indels.pl'.)

	my $out_indel_dir = $parent_out_indel_dir;
	##my $out_indel_dir = "$parent_out_indel_dir/$in_msa_id";

	unless ( -d $out_indel_dir) {
	    unless (0 == system (" mkdir -p $out_indel_dir ") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to create the Indel output directory '$out_indel_dir': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}
	unless ( -w $out_indel_dir ) {
	    unless (0 == system (" chmod -R u+w $out_indel_dir ") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the Indel output directory '$out_indel_dir': $?  !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}


		## ADDED in creation of 'bch_serial_br_infer_indels.pl'. ##
		## => MODIFIED in creation of 'bch_serial_br_infer_indels_dawg.pl'. ##

	## $outfile1 = "$out_indel_dir/$pref_outfile1$lcl_msa_id$suff_outfile1",
	## $outfile2 = "$out_indel_dir/$pref_outfile2$lcl_msa_id$suff_outfile2",
	## $logfile  = "$out_indel_dir/$pref_logfile$lcl_msa_id$suff_logfile",
	# where $out_indel_dir = "$top_indel_dir/$archive_pref$id_class/";

	my $outfile1 = "$out_indel_dir/$pref_outfile1$in_msa_id$suff_outfile1";
	my $outfile2 = "$out_indel_dir/$pref_outfile2$in_msa_id$suff_outfile2";
	my $logfile = "$out_indel_dir/$pref_logfile$in_msa_id$suff_logfile";

		## MAJOR MODIFICATIOn in creation of 'bch_serial_br_infer_indels_dawg.hs.pl'. ##

	unless (0 == system ("(time perl $main_indel_script $tmpin_msa $tmpin_ctrl $aux_indir $outfile1 $outfile2) > $logfile 2>&1") ) {
	##unless (0 == system ("(time perl $main_indel_script $tmpin_msa $tmpin_ctrl $outfile1 $outfile2) > $logfile 2>&1") ) {...} # FOr the predecessor 'bch_serial_br_infer_indels_dawg.pl'.

	    print STDERR "[ERROR '$program_name'] FAILED 'system ('perl $main_indel_script ...')' FAILED for the input MSA ID '$in_msa_id' in the class '$id_class': $! !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
=for the predecessor 'bch_serial_br_infer_indels_dawg.pl'. 
	## $main_indel_script = "Scripts/br_infer_psm_indel_events.pl";
	##
	## $outfile1 = "$out_indel_dir/$name_outfile1",
	## $outfile2 = "$out_indel_dir/$name_outfile2",
	## $logfile  = "$out_indel_dir/$name_logfile",
	# where $out_indel_dir = "$top_indel_dir/$archive_pref$chr_no/$anchor_orf_id/";

	my $outfile1 = "$out_indel_dir/$name_outfile1";
	my $outfile2 = "$out_indel_dir/$name_outfile2";
	my $logfile = "$out_indel_dir/$name_logfile";

	unless (0 == system ("(time perl $main_indel_script $tmpin_msa $tmpl_intree $out_indel_dir) > $logfile 2>&1") ) {
	    print STDERR "[ERROR '$program_name'] FAILED 'system ('perl $main_indel_script ...')' FAILED for the input MSA ID '$in_msa_id': $! !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
=cut

	foreach my $ofile ($outfile1, $outfile2, $logfile) {
	    unless (( -f $ofile ) and ( -r $ofile )) {
		print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable output file '$ofile'... I guess that the analysis via the script '$main_indel_script' FAILED on the temporary input MSA '$tmpin_msa' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}

  (print STDOUT "\n ... Successfully extracted the input alignment '$tmpin_msa' and performed an Indel analysis on it ...\n\n") if ($TEST_LEVEL>1);

		# Remove the temporary input files. #
		# ( MODIFIED in creation of 'bch_serial_br_infer_indels_dawg.pl'.)

	foreach my $tmpin_file ($tmpin_msa, $tmpin_ctrl) {
	    unless (0 == system ("rm -f $tmpin_file") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to remove the temporary input file '$tmpin_file': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}

=for the predecessor 'bch_serial_br_infer_indels_dawg.pl'
		# Remove the temporary input MSA file. #
	unless (0 == system ("rm -f $tmpin_msa") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to remove the temporary input MSA file '$tmpin_msa': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
=cut

	$ct_succ++;

    }					# End of the middle foreach-loop (over MSAs in the archive). #

		## Archive the Middle output directory (storing masked MSAs for the same chromosome.) ##

    my $out_archive_name = "$mid_dir.tgz";

     # $parent_out_indel_dir = "$top_indel_dir/$mid_dir";
     # [ $parent_inout_paml_dir = "$top_paml_dir/$mid_dir"; ] --- for the predecessor 'bch_serial_PAML.pl'.
     # [$outaln_dir = "$topoutaln_dir/$mid_dir";]  --- for the predecessor 'bchbch_refine_msas.pl'.

    unless ( chdir $top_indel_dir ) { # MODIFIED in creation of 'bch_serial_br_infer_indels.pl'.
    ##unless ( chdir $top_paml_dir ) { # MODIFIED in creation of 'bch_serial_PAML.pl'.
    ##unless ( chdir $topoutaln_dir ) {
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to change the directory to '$top_indel_dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    unless (0 == system ("tar -cpzf $out_archive_name $mid_dir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED TO CREATE a tar-gzipped ARCHIVE '$out_archive_name out of the directory '$mid_dir' under the directory '$top_indel_dir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    unless (0 == system ("rm -rf $mid_dir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to REMOVE the directory '$mid_dir' under the directory '$top_indel_dir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
 (print STDOUT "\n... Successfully archived the masked MSAs into '$out_archive_name' under '$top_indel_dir'!!\n\n") if ($TEST_LEVEL>1);

    unless ( chdir $CURR_WORK_DIR ) {
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to RETURN TO the original place, '$CURR_WORK_DIR' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    ##($ct_chr>0) and last; ## For TESTING the script. ##

}			# END of the outer foreach-loop. (over archives (or classes)) #


print STDOUT "\n=== All Processes FINISHED !! ===\n\n";

print STDOUT "#{MSA Classes} = $ct_classes .\n";
##print STDOUT "#{Chromosomes} = $ct_chr .\n";
print STDOUT "#{Input MSAs}  = $ct_msas .\n";
print STDOUT "    #{Anomalous} = $ct_anom ,\n";
print STDOUT "    #{Failed}    = $ct_fail ,\n";
print STDOUT "    #{Succeeded} = $ct_succ .\n\n";



	### DEFINITIONS OF THE SUBROUTINES ###

# peek_into_archive:
#	This subroutine looks into an input archive and outputs a list of entries.
#	The archive is supposed to be storing files with the following paths:
#	"$dir/$pref$id$suff",
#	where $id specifies the ID of the file.
#
# INPUT: (\$in_archive, $dir, $pref, $suff), where
#	$in_archive is the path to the input archive;
#	$dir/$pref/$suff are templates for the paths to the files stored in the archive.
#
# OUTPUT: \@list_ids, where
#	@list_ids stores the IDs of the archived files with the above paths.
#
sub peek_into_archive (\$$$$) {

	my ($in_archive, $dir, $pref, $suff) = @_;
	local $/ = "\n"; # Use the default input-record-separator. #

	my $fh_tmp;
	unless ( open ($fh_tmp, "tar -tzf ${$in_archive} |" ) ) {
	    print STDOUT "[ERROR 'peek_into_archive'] CanNOT fork or look into ${$in_archive}: $!\n";
	    return undef;
	}

	my @list_ids = ();

	my $ext_pref = "$dir/$pref"; # Extended prefix. #
	my $len_epref = length $ext_pref;
	my $len_suff = length $suff;

	while (my $tmp = <$fh_tmp>) { # Main while-loop (over input lines) #

	    $tmp =~ s/\s+$//; # Remove white spaces at the end of the entry (serving as a separator (and possibly a pad)).
	    ($ext_pref =~ m/^\.\//) or ($tmp =~ s/^\.\///); # Remove './' at the beginning of the entry unless $ext_pref begins with './'. (ADDED in creation of 'bch_serial_PAML.pl'.)

	    my $tmp_suff = substr ($tmp, - $len_suff, $len_suff, '');

	    ($tmp_suff eq $suff) or next; # Skip if the suffix doesn't match.

	    my $tmp_epref = substr ($tmp, 0, $len_epref, '');

	    ($tmp_epref eq $ext_pref) or next; # Skip if the extended prefix doesn't match.

	    push @list_ids, $tmp; # The remnant of the entry must be the ID.


	}			# End of the main while-loop (over input lines) #
	close ($fh_tmp);

	return \@list_ids;

}	# End of the subroutine 'peek_into_archive' #



=for the predecessor 'bchbch_refine_msas.pl'
# cal_pvs_sbst_indel:
#	This satellite subroutine of 'pwaln_surveillance' calcualtes
#	P-value for substitutions and that for indels,
#	based on a binomial distribution for the former and a Poisson distribution for the latter.
#
# INPUT: ($wd_start, $wd_end, $ct_ngp_clms, \@window, \@list_indels, $prob_match, $rate_indel), where
#	$wd_start/end is the start/end coordinate of the window (0-based half-open);
#	$ct_ngp_clms is the number of non-gapped columns in the window;
#	@{$window[$c]} = ($coord, $char_ref, $char_sbj) is for the $c th column in the window;
#	@{$list_indels[$i]} = ($sgn, $start, $end) is for the $i th indel in the window, with
#	$sgn = 1 if gaps are in the reference, = -1 if gaps are in the subject;
#	$prob_match is the probability of matches among non-gapped columns;
#	$rate_indel is the rate of indels (per inter-nucleotide position).
#
# OUTPUT: ($pv_sbst, $pv_indel, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj), where
#	$pv_sbst/indel is the P-value for the number of substitutions (actually matches)/indels;
#	$ct_ins_ref/sbj is the number of "insertions" in the reference/subject.
#
sub cal_pvs_sbst_indel ($$$\@\@$$) {

	my ($wd_start, $wd_end, $ct_ngp_clms, $window, $list_indels, $prob_match, $rate_indel) = @_;

    (print STDOUT "    < Subroutine 'cal_pvs_sbst_indel' \; window = [$wd_start, $wd_end); #{non-gapped columns} = $ct_ngp_clms >\n\n") if ($TEST_LEVEL_SVY3>0); 

    if ($TEST_LEVEL_SVY3>1) {
	print STDOUT "    [ Input arguments ]\n\n";
	print STDOUT '    $prob_match = ', $prob_match, " .\n";
	print STDOUT '    $rate_indel = ', $rate_indel, " .\n\n";  
    }
    if ($TEST_LEVEL_SVY3>2) {
	print STDOUT '    @{$window} = ( # element = [$coord, $char_ref, $char_sbj],', "\n";
	foreach my $clm (@{$window}) { print STDOUT "\t\t\[", join (', ', @{$clm}), "\],\n"; }
	print STDOUT "\t\t) .\n\n";

	print STDOUT '    @{$list_indels} = ( # element = [reference(1)/subject(-1), $start, $end],', "\n";
	foreach my $indel (@{$list_indels}) { print STDOUT "\t\t\[", join (', ', @{$indel}), "\],\n"; }
	print STDOUT "\t\t) .\n\n";
    }

	my $ct_mismatches = 0;
	foreach my $site (@{$window}) { ($site->[1] eq $site->[2]) or ($ct_mismatches++); }

	my $ct_matches = $ct_ngp_clms - $ct_mismatches;

	my $pv_sbst =  binomial_cum_via_betai ($ct_ngp_clms, $ct_matches, $prob_match);
	unless (defined $pv_sbst) {
	    print STDOUT "[ERROR 'cal_pvs_sbst_indel'] 'binomial_cum_via_betai' FAILED to calculate the P-value for the substitutions within the window, [$wd_start, $wd_end), with size $ct_ngp_clms and $ct_matches matches under the match probability, $prob_match : $! !\n";
	    return undef;
	}

	my $ct_ins_sbj = my $ct_ins_ref = 0;
	foreach my $indel (@{$list_indels}) { ($indel->[0] == 1) ? ($ct_ins_sbj++) : ($ct_ins_ref++) ; }

	my $lambda = $rate_indel * $ct_ngp_clms;
	my $ct_indel = $ct_ins_sbj + $ct_ins_ref;

	my $cmpl_pv_indel = ($ct_indel>0) ? poisson_cum ($lambda, $ct_indel -1) : 0;
	##my $pv_indel = poisson_cum ($lambda, $ct_indel); #  A BUG fixed on Oct 23 (Wed), 2013. #
	unless (defined $cmpl_pv_indel) {
	##unless (defined $pv_indel) {
	    print STDOUT "[ERROR 'cal_pvs_sbst_indel'] 'poisson_cum' FAILED to calculate the P-value for the insertions/deletions within the window, [$wd_start, $wd_end], with size $ct_ngp_clms and $ct_indel indels under the expected number, $lambda: $! !\n";
	    return undef;
	}

	my $pv_indel = 1 - $cmpl_pv_indel; # MODIFIED to fix a bug on Oct 23 (Wed), 2013. #

    if ($TEST_LEVEL_SVY3>1) {
	print STDOUT "\n";
	print STDOUT "    [ Outputs ]\n\n";
	print STDOUT '    $ct_mismatches = ', $ct_mismatches, " .\n";
	print STDOUT '    $ct_ins_ref  = ', $ct_ins_ref, " .\n";
	print STDOUT '    $ct_ins_sbj  = ', $ct_ins_sbj, " .\n\n";
	print STDOUT '    $pv_sbst  = ', $pv_sbst, " .\n";
	print STDOUT '    $pv_indel = ', $pv_indel, " .\n\n";
    }

    (print STDOUT "    - Subroutine 'cal_pvs_sbst_indel' ENDED -\n\n") if ($TEST_LEVEL_SVY3>0);

	return ($pv_sbst, $pv_indel, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj);

}		# End of the subroutine 'cal_pvs_sbst_indel' #


# pwaln_surveillance:
#	This satellite subroutine of 'msa_surveillance' inspects a pairwise alignment between a reference sequence and a subject sequence.
#
#	The subroutine examines both substitutions and indels.
#	The P-values are calculated via a binomial distribution for substitutions,
#	and via a Poisson distribution for indels.
#
# INPUT: (\$ref_seq, \$sbj_seq, $gap, $ngp_clm_ct, $wd_offset, $prob_match, $lb_pv_sbst_ct, $indel_rate, $lb_pv_indel_ct), where
#	$ref/sbj_seq is the reference/subject sequence embedded in the alignment;
#	$gap is a character representing a gap;
#	$ngp_clm_ct is the number of non-gapped columns each window usually contains;
#	$wd_offset is the number of non-gapped columns discarded/newly acquired to update the window;
#	$lb_pv_sbst/indel_ct is lower-bound for the P-value of the number of substitutions/indels;
#	$prob_match is the probability that the two aligned bases match among the non-gapped columns;
#	$indel_rate is the total rate of insertions/deletions.
#
# OUTPUT: \@poor_aln_regions_pw, where
#	@{$poor_aln_regions[$k]} = ($start_coord, $end_coord, $ct_ngp_clms, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj, $pv_sbst, $pv_indel) is for the $k th poorly aligned window within the pairwise alignment.
#	(The coordinates are 0-based half-open.)
#
sub pwaln_surveillance (\$\$$$$$$$$) {

	my ($ref_seq, $sbj_seq, $gap, $ngp_clm_ct, $wd_offset, $prob_match, $lb_pv_sbst_ct, $indel_rate, $lb_pv_indel_ct) = @_;

   (print STDOUT " < Subroutine 'pwaln_surveillance' (", '$gap = ', $gap, ") >\n\n") if ($TEST_LEVEL_SVY2>0);

	my $cp_ref_seq = uc ${$ref_seq};
	$cp_ref_seq =~ tr/U/T/;
	my $cp_sbj_seq = uc ${$sbj_seq};
	$cp_sbj_seq =~ tr/U/T/;

	my $len_ref_seq = length $cp_ref_seq;
	my $len_sbj_seq = length $cp_sbj_seq;

	unless ($len_ref_seq == $len_sbj_seq) {
	    print STDOUT "[ERROR 'pwaln_surveillance'] ILLEGAL INPUT: The subject sequence length ($len_sbj_seq) does NOT EQUAL the reference sequence length: $! !\n";
	    return undef;
	}

	unless ($wd_offset <= $ngp_clm_ct) {
	    print STDOUT "[ERROR 'pwaln_surveillance'] ILLEGAL INPUT: The variable, ", '$wd_offset' , "(=$wd_offset) is LARGER THAN the variable, ", '$ngp_clm_ct', "(=$ngp_clm_ct): $! !\n";
	    return undef;
	}


    if ($TEST_LEVEL_SVY2>1) {
	print STDOUT "  {Length of the sequences embedded in the alignment (including gaps)} = $len_ref_seq .\n";
	print STDOUT '  $ngp_clm_ct = ', $ngp_clm_ct, " .\n";
	print STDOUT '  $wd_offset  = ', $wd_offset, " .\n\n";
	print STDOUT '  $prob_match = ', $prob_match, " .\n";
	print STDOUT '  $lb_pv_sbst_ct = ', $lb_pv_sbst_ct, " .\n\n";
	print STDOUT '  $indel_rate = ', $indel_rate, " .\n";
	print STDOUT '  $lb_pv_indel_ct = ', $lb_pv_indel_ct, " .\n\n";
    }

    if ($TEST_LEVEL_SVY2>5) {

	print STDOUT "  Reference sequence:\n\n";

	unless ( print_sequence (${$ref_seq}, 50, *STDOUT) ) {
	    print STDOUT "[ERROR 'pwaln_surveillance'] 'print_sequence' FAILED !\n\n";
	    return undef;
	}

	print STDOUT "\n\n  Subject sequence:\n\n";

	unless ( print_sequence (${$sbj_seq}, 50, *STDOUT) ) {
	    print STDOUT "[ERROR 'pwaln_surveillance'] 'print_sequence' FAILED !\n\n";
	    return undef;
	}
	print STDOUT "\n\n";
    }

    if ($TEST_LEVEL_SVY2>2) {
	my @pw_aln = (\$cp_ref_seq, \$cp_sbj_seq);
	my @labels = ('Reference', 'Subject');
	print STDOUT "Pairwise alignment of the reference and subject sequences:\n\n";
	unless ( pr_ODEN_input (@labels, @pw_aln, 50, *STDOUT) ) {
	    print STDOUT "[ERROR 'pwaln_surveillance'] 'pr_ODEN_input' FAILED !\n\n";
	    return undef;
 	}
	print STDOUT "\n\n";
    }

   (print STDOUT "\n  .. Now, examine the pairwise alignment ..\n\n") if ($TEST_LEVEL_SVY2>1);

	my @poor_aln_regions_pw = ();
		# element = ($start_coord, $end_coord, $ct_ngp_clms, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj, $pv_sbst, $pv_indel)

	my $wd_start = 0;
	my @window = (); # element = [$clm_coord, $char_ref, $char_sbj] for non-gapped columns. #
	my @list_indels = ();
		# list of indels WITHIN the window under investigation. #
		# element = [1 if gap is in the reference/ -1 if gap is in the subject, $start_coord, $end_coord] (0-based half-open).

	for (my $coord = 0; $coord<$len_ref_seq; $coord++) { # The main for-loop (over alignment columns) #

	    my $char_ref = substr ($cp_ref_seq, 0, 1, '');
	    my $char_sbj = substr ($cp_sbj_seq, 0, 1, '');

	    my $last_indel = (@list_indels>0) ? $list_indels[$#list_indels] : undef;

	    if (($char_ref eq $gap) and ($char_sbj eq $gap)) { # Skip if both reference and subject hold gaps.
		((defined $last_indel) and ($last_indel->[2] == $coord)) and ($last_indel->[2]++); # Extend the last indel. #
		next;
	    }

   (print STDOUT "  ( Coordinate: $coord; Reference: $char_ref; Subject: $char_sbj )\n\n") if ($TEST_LEVEL_SVY2>1);

	    if ($char_ref eq $gap) {

		if (@list_indels>0) {
		    if (($last_indel->[2] == $coord) and ($last_indel->[0] == 1)) {
			$last_indel->[2]++;
		    } else {
			push @list_indels, [1, $coord, $coord+1];
		    }
		
		} else {
		    push @list_indels, [1, $coord, $coord+1];
		}

		next;

	    } elsif ($char_sbj eq $gap) {

		if (@list_indels>0) {
		    if (($last_indel->[2] == $coord) and ($last_indel->[0] == -1)) {
			$last_indel->[2]++;
		    } else {
			push @list_indels, [-1, $coord, $coord+1];
		    }
		
		} else {
		    push @list_indels, [-1, $coord, $coord+1];
		}

		next;
	    }

	    (($char_ref =~ m/^[ATGC]$/) and ($char_sbj =~ m/^[ATGC]$/)) or next; # Skip if ambiguous characters are contained.

	    push @window, [$coord, $char_ref, $char_sbj];

	    (@window < $ngp_clm_ct) and next;

		## The window attained the desired size. ##

	    my $ct_ngp_clms = @window;

	    my $wd_end = $window[$#window]->[0] + 1;

    (print STDOUT "  { Window: [$wd_start, $wd_end) }\n\n") if ($TEST_LEVEL_SVY2>1);

	    my ($pv_sbst, $pv_indel, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj)
		= cal_pvs_sbst_indel ($wd_start, $wd_end, $ct_ngp_clms, @window, @list_indels, $prob_match, $indel_rate);
	    unless ((defined $pv_sbst) and (defined $pv_indel)) {
		print STDOUT "[ERROR 'pwaln_surveillance'] 'cal_pvs_sbst_indel' FAILED on the window, [$wd_start, $wd_end), with size $ct_ngp_clms: $! !\n\n";
		return undef;
	    }
   (print STDOUT "  ... Subroutine 'cal_pvs_sbst_indel' Ended on the window: [$wd_start, $wd_end) ...\n\n") if ($TEST_LEVEL_SVY2>1);
   (print STDOUT '   ($pv_sbst, $pv_indel, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj) = (', join (', ', $pv_sbst, $pv_indel, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj), "\) .\n\n") if ($TEST_LEVEL_SVY2>2);

	    if (($pv_sbst < $lb_pv_sbst_ct) or ($pv_indel < $lb_pv_indel_ct)) {

			# The window is judged as "poorly aligned".

		push @poor_aln_regions_pw, [$wd_start, $wd_end, $ct_ngp_clms, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj, $pv_sbst, $pv_indel];
	    }

		# ($ref_seq, $sbj_seq, $gap, $ngp_clm_ct, $wd_offset, $prob_match, $lb_pv_sbst_ct, $indel_rate, $lb_pv_indel_ct)

			# To update the window, discard some columns and indels on the 5'-side. #

	    $wd_start = $window[$wd_offset-1]->[0] + 1;

	    for (1 .. $wd_offset) { shift @window; }

	    while ((@list_indels>0) and ($list_indels[0]->[1] < $wd_start)) { shift @list_indels; }

	}		# End of the main for-loop (over alignment columns) #


		## Finally, process the incomplete window at the end. ##

	my $ct_ngp_clms = @window;
	my $wd_end = $len_ref_seq;

	my ($pv_sbst, $pv_indel, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj)
	    = cal_pvs_sbst_indel ($wd_start, $wd_end, $ct_ngp_clms, @window, @list_indels, $prob_match, $indel_rate);
	unless ((defined $pv_sbst) and (defined $pv_indel)) {
		print STDOUT "[ERROR 'pwaln_surveillance'] 'cal_pvs_sbst_indel' FAILED on the window, [$wd_start, $wd_end), with size $ct_ngp_clms: $! !\n\n";
		return undef;
	}

	if (($pv_sbst < $lb_pv_sbst_ct) or ($pv_indel < $lb_pv_indel_ct)) {

			# The window is judged as "poorly aligned".

	    push @poor_aln_regions_pw, [$wd_start, $wd_end, $ct_ngp_clms, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj, $pv_sbst, $pv_indel];
	}


    if ($TEST_LEVEL_SVY2>2) {
	print STDOUT "  [ Output ]\n\n";
	print STDOUT '   @poor_aln_regions_pw = ( # element = [$wd_start, $wd_end, $ct_ngp_clms, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj, $pv_sbst, $pv_indel],', "\n";
	foreach my $window (@poor_aln_regions_pw) { print STDOUT "\t\t\[", join (', ', @{$window}), "],\n"; }
	print STDOUT "\t\t) .\n\n";
    }

    (print STDOUT "  -- End of the Subroutine 'pwaln_surveillance' --\n\n") if ($TEST_LEVEL_SVY2>0);

	return \@poor_aln_regions_pw;
	    # element = ($start_coord, $end_coord, $ct_ngp_clms, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj, $pv_sbst, $pv_indel)


}		# End of the subroutine 'pwaln_surveillance' #


# msa_surveillance:
#	This subroutine receives an MSA of orthologous sequences,
#	and examines the quality of MSA via a stepping-window analysis on the pairwise alignments,
#	each between the reference sequence and a non-reference sequence.
#
#	The subroutine examines both substitutions and indels.
#	The P-values are calculated via a binomial distribution for substitutions,
#	and via a Poisson distribution for indels.
#
# INPUT: (\@in_msa, \@species, $ref_speces, $gap, $ngp_clm_ct, $wd_offset, \%spec2prob_match, $lb_pv_sbst_ct, \%spec2indel_rate, $lb_pv_indel_ct), where
#	${$in_msa[$k]} is the $k th orthologous sequence embedded in the input MSA;
#	$species[$k] is the name (or nickname) of the $k th species;
#	$ref_species is the reference species name;
#	$gap is a character representing a gap;
#	$ngp_clm_ct is the number of non-gapped columns each window usually contains;
#	$wd_offset is the number of non-gapped columns discarded/newly acquired to update the window;
#	$lb_pv_sbst/indel_ct is lower-bound for the P-value of the number of substitutions/indels;
#	%spec2prob_match = ( $species => $match_probability, ...) for substitutions;
#	%spec2indel_rate = ( $species => $indel_rate, ...) for indels.
#	
#
# OUTPUT: \@poor_aln_regions, where
#	@{$poor_aln_regions[$sp]->[$k]} = ($start_coord, $end_coord, $ct_ngp_clms, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj, $pv_sbst, $pv_indel) is for the $k th poorly aligned window within the pairwise alignment between the reference and the $sp th species.
#	[ @{$poor_aln_regions[$indx_ref]} is empty. ]
#	(The coordinates are 0-based half-open.)
#
sub msa_surveillance (\@\@$$$$\%$\%$) {

	my ($in_msa, $species, $ref_species, $gap, $ngp_clm_ct, $wd_offset, $spec2prob_match, $lb_pv_sbst_ct, $spec2indel_rate, $lb_pv_indel_ct) = @_;

   (print STDOUT "<< Subroutine 'msa_surveillance' (", '$ref_species = ', $ref_species, '; $gap = ', $gap, ") >>\n\n") if ($TEST_LEVEL_SVY>0);

	my $ct_seqs = @{$in_msa};
	unless ($ct_seqs>0) {
	    print STDOUT "[EROR 'msa_surveillance'] ILLEGAL INPUT: The input MSA contains NO SEQUENCES: $! !\n";
	    return undef;
	}
	unless (@{$species} == $ct_seqs) {
	    print STDOUT "[ERROR 'msa_surveillance'] ILLEGAL INPUT: The number of species name (=", scalar (@{$species}), ") does NOT EQUAL the number of sequences in the MSA (=$ct_seqs): $! !\n";
	    return undef;
	}

	my $len_msa = length (${$in_msa->[0]});
	unless ($len_msa > 0) {
	    print STDOUT "[ERROR 'msa_surveillance'] ILLEGAL INPUT: The MSA contains NO COLUMN: $! !\n";
	    return undef;
	}

	my $indx_ref;
	for (my $i=0; $i<$ct_seqs; $i++) {
	    ($species->[$i] eq $ref_species) or next;
	    $indx_ref = $i;
	    last;
	}
	unless (defined $indx_ref) {
	    print STDOUT "[ERROR 'msa_surveillance'] The input MSA does NOT contain the reference sequence '$ref_species': $! !\n";
	    return undef;
	}

    if ($TEST_LEVEL_SVY>1) {
	print STDOUT "{Reference species}  = '$ref_species' (Index: $indx_ref).\n";
	print STDOUT "#{Sequences in the MSA} = $ct_seqs .\n";
	print STDOUT "{Length of the MSA}     = $len_msa .\n\n";
	##print STDOUT "{Length of the region embedded in the MSA} = $len_embedded_rg .\n\n";
    }

    if ($TEST_LEVEL_SVY>=5) {
	print STDOUT "{ Input MSA }\n\n";
	unless ( pr_ODEN_input (@{$species}, @{$in_msa}, 50, *STDOUT) ) {
	    print STDOUT "[ERROR 'msa_surveillance'] 'pr_ODEN_input' FAILED !\n\n";
	    return undef;
 	}
	print STDOUT "\n\n";
    }

    (print STDOUT "\n... Now, compare the reference sequence and the orthologous sequences of the subject species ...\n\n") if ($TEST_LEVEL_SVY>1);

	my @poor_aln_regions = ();

	for (my $k=0; $k<$ct_seqs; $k++) { # Main for-loop (over species) #

	    my $spec = $species->[$k];

	    if ($k == $indx_ref) {

   (print STDOUT "(( The $k th species: '$spec' is the reference species itself, so skip examining the pairwise alignment ))\n\n") if ($TEST_LEVEL_SVY>1);

		push @poor_aln_regions, [];
		next; # Skip the reference species.
	    }

	    my $prob_match = $spec2prob_match->{$spec};
	    unless (defined $prob_match) {
		print STDOUT "[ERROR 'msa_surveillance'] NO MATCH PROBABILITY for the $k th species,'$spec': $! !\n";
		return undef;
	    }

	    my $indel_rate = $spec2indel_rate->{$spec};
	    unless (defined $indel_rate) {
		print STDOUT "[ERROR 'msa_surveillance'] NO INDEL RATE for the $k th species, '$spec': $! !\n"; 
		return undef;
	    }

    (print STDOUT "(( The $k th species: '$spec' \[match_probability = $prob_match\; indel_rate = $indel_rate\] ))\n\n") if ($TEST_LEVEL_SVY>1);

	    my $poor_aln_regions_pw = pwaln_surveillance (${$in_msa->[$indx_ref]}, ${$in_msa->[$k]}, $gap, $ngp_clm_ct, $wd_offset, $prob_match, $lb_pv_sbst_ct, $indel_rate, $lb_pv_indel_ct);
	    unless (defined $poor_aln_regions_pw) {
		print STDOUT "[ERROR 'msa_surveillance'] 'pwaln_surveillance' FAILED between the reference ($ref_species) and the $k th species '$spec': $! !\n";
		return undef;
	    }
    (print STDOUT " --- The subroutine 'pwaln_surveillance' Ended ..\n\n") if ($TEST_LEVEL_SVY>1); 

   if ($TEST_LEVEL_SVY>2) {
	print STDOUT "( Output of 'pwaln_surveillance' )\n\n";

	print STDOUT '@{$poor_aln_regions_pw} = ( # element = [$start_coord, $end_coord, $ct_ngp_clms, $ct_mismatches, $ct_ins_ref, $ct_ins_sbj, $pv_sbst, $pv_indel],' , "\n";
	foreach my $window (@{$poor_aln_regions_pw}) { print STDOUT "\[", join (', ', @{$window}), "\],\n"; }
	print STDOUT "\t\) .\n\n\n";
   }

	    push @poor_aln_regions, $poor_aln_regions_pw;


	}			# End of the main for-loop (over species) #

   (print STDOUT "--- END of the Subroutine 'msa_surveillance' ---\n\n") if ($TEST_LEVEL_SVY>0);

	return \@poor_aln_regions;

}		# End of the subroutine 'msa_surveillance' #


# mask_aln_regions:
#	This subroutine masks specified regions in the alignment.
#	Within each specified region, all the characters except a gap and a missing base will be masked.
#
#	NOTE that this subroutine gnaws up the input sequences.
#
# INPUT: (\@in_msa, \@list_regions, $indx_start, $indx_end, $gap, $missing, $masked), where
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	@{$list_regions[$k]->[$r]}[$indx_start, $indx_end] = ($start_coord, $end_coord)
#	is for the $r th region to be masked of the $k th sequence, with
#	$start_coord and $end_coord being the 0-based half-open coordinates along the alignment;
#	$gap/$missing is the character representing a gap/missing base;
#	$masked is the character representing a masked base.
#
# OUTPUT: \@masked_msa, where
#	${$masked_msa[$k]} is the product of masking the $k th input sequence.
#
sub mask_aln_regions (\@\@$$$$$) {

	my ($in_msa, $list_regions, $indx_start, $indx_end, $gap, $missing, $masked) = @_;

   (print STDOUT "  << Subroutine 'mask_aln_regions' >>\n\n") if ($TEST_LEVEL_MSK>0);

	my $ct_seqs = @{$in_msa};

	my @masked_msa = ();

	for (my $k=0; $k<$ct_seqs; $k++) { # Main outer for-loop (over sequences) #

   (print STDOUT "\n  [ The $k th sequence ]\n\n") if ($TEST_LEVEL_MSK>1);

		# Make a list of non-overlapping regions. #

	    my @nov_regions = (); # element = [$start, $end] for the non-overlapping regions.

   if ($TEST_LEVEL_MSK>2) {
	print STDOUT "  ( Raw regions to be masked: )\n\n"; 
	print STDOUT '  ', join ("\t", 'Start', 'end'), "\n\n";
   }

	    foreach my $reg (@{$list_regions->[$k]}) { # Preliminary middle foreach-loop (over raw regions to be masked) #
	    ##foreach my $reg (@{$list_regions}) { # Preliminary middle foreach-loop (over raw regions to be masked) # A BUG fixed on Oct 23 (Wed), 2013. #

		my ($start, $end) = @{$reg}[$indx_start, $indx_end];

   (print STDOUT '  ', join ("\t", $start, $end), " .\n") if ($TEST_LEVEL_MSK>2);

		my $last_region = (@nov_regions>0) ? $nov_regions[$#nov_regions] : undef;

		if ( (defined $last_region) and ($start <= $last_region->[1]) ) {
		##if ((defined $last_region) and ($last_region->[1] <= $start)) { # A BUG fixed on Oct 23 (Wed), 2013. #
			# The last region overlaps the current region. => Extend the last region. #
		    $last_region->[1] = $end;
		} else {
			# Create a new region. #
		    push @nov_regions, [$start, $end];
		}

	    }			# End of the preliminary middle foreach-loop (over raw regions to be masked) #

   if ($TEST_LEVEL_MSK>2) {
	print STDOUT "\n\n";
	print STDOUT "  ( Non-overlapping regions to be masked: )\n\n";
	print STDOUT '  ', join ("\t", 'Start', 'end'), "\n\n";
	foreach my $reg (@nov_regions) { print STDOUT '  ', join ("\t", @{$reg}), "\n"; }
	print STDOUT "\n\n";
   }

	    my $sbj_seq = $in_msa->[$k];
	    my $seq_len = length (${$sbj_seq});

	    my $masked_seq = '';
	    my $curr_coord = 0;

	    foreach my $reg (@nov_regions) { # Main middle foreach-loop (over non-overlapping regions to be masked) #
		my ($start, $end) = @{$reg};

		if ($curr_coord < $start) {
		    $masked_seq .= substr (${$sbj_seq}, 0, $start - $curr_coord, '');
		    $curr_coord = $start;
		}

		##my $len_rg = $end - $start;
		##my $region = substr (${$sbj_seq}, 0, $len_rg, '');

		for (; $curr_coord < $end; $curr_coord++) { # Inner for-loop (over sites within the region) #

		    my $char = substr (${$sbj_seq}, 0, 1, '');
		    $masked_seq .= (($char eq $gap) or ($char eq $missing)) ? $char : $masked;

		}			# End of the inner for-loop (over sites within the region) #

			# Update the "current" coordinate. #
		$curr_coord = $end;

	    }			# End of the main middle foreach-loop (over non-overlapping regions to be masked) #
	    if ($curr_coord < $seq_len) { # Append the 3'-end not to be masked.

		$masked_seq .= substr (${$sbj_seq}, 0, $seq_len - $curr_coord, '');
	    }

	    push @masked_msa, \$masked_seq;

	}			# End of the main outer for-loop (over sequences) #


   (print STDOUT "  -- End of the subroutine 'mask_aln_regions' --\n\n") if ($TEST_LEVEL_MSK>0);

	return \@masked_msa;

}		# End of the subroutine 'mask_aln_regions' #

=cut
