#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# bch_serial_preprocess_msas_dawg.pl:
#	This Perl script preprocesses multiple MSAs in a batch, via a Perl script, 'preprecess_msa_xxx.pl'.
#	It was created by modifying 'bch_serial_br_infer_indels_dawg.pl'.
#
#
#	[CREATION STARTED on May 13 (Wed), 2015, ENDED on May xx (yyy), 2015.]
#
# bch_serial_br_infer_indels_dawg.pl:
#	This modified version of 'bch_serial_br_infer_indels.pl' can analyze a batch of
#	multiple sequence alignments simulated via Dawg.
#	(CREATED on Feb 28 (Fri), 2014)
#
# bch_serial_br_infer_indels.pl:
#	This Perl script performs bach analyses to infer (parsimonious) indels forming the input sequence alignments,
#	all of which share a species tree, stored in multiple gzipped tar archives.
#	The script was created by modifying 'bch_serial_PAML.pl'.
#	(CREATED on Nov 22 (Fri), 2013)
#
# bch_serial_PAML.pl:
#	This Perl script performs bach PAML analyses on the input sequence alignments,
#	all of which share a species tree, stored in multiple gzipped tar archives.
#	The script was created by modifying 'bchbch_refine_msas.pl',
#	and borrowing functions to perform PAML anayses from 'compare_ml_msa.bch.extJC.6t.PhyML.pl'.
#	(CREATION STARTED on Nov 3 (Sun), 2013, ENDED on Nov 3 (Sun), 2013.)
#
# bchbch_refine_msas.pl:
#	This Perl script refines MSAs stored in multiple archives.
#	The script was created by using parts of 'serial_pgene_ortho_aln_analyses.pl'.
#	(CREATION STARTED on Oct 21 (Mon), 2013, ENDED on Oct xx (yyy), 2013.)
#
# serial_pgene_ortho_aln_analyses.pl:
#	This Perl script inspects the alignments accommodating progo-genes,
#	and examines whether the orthologous region is 'missing' or not
#	in each of 3 other yeast species.
#	Once the orthologous region is found not "missing", 
#	the script goes on to calculate the alignment coverage, nucleotide similarity, and the three RFC scores.
#	The script was created by modifying 'serial_pgene_RFC_analyses.pl'.
#	(CREATION STARTED on Oct 7 (Mon), 2013, ENDED on Oct 11 (Fri), 2013)
#


use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext
	tfm_labels2indices tfm_labels2indices_gen
	);

=for the predecessor 'bchbch_refine_msas.pl'
use MyArrayManipu qw( multiple_sort );

use MyMalnIO qw(
	extract_fasta extract_clw
	pr_ODEN_input
	pr_clustal_like_lite
	);
	## pr_phylip_aln
	## pr_clustal_like_lite_aa

use MyNucSeq2 qw(
	revcom print_sequence
	);

use MyProbDistri qw(
	poisson_cum
	binomial_cum_via_betai
	);
=cut

	# IMPORTED from 'compare_ml_msa.bch.extJC.6t.PhyML.pl' #
use MyPAML_parser qw( prepare_paml_ctrl );
	# retrieve_aa_freqs 

sub peek_into_archive (\$$$$);

=for the predecessor 'bchbch_refine_msas.pl'
sub cal_pvs_sbst_indel ($$$\@\@$$);
sub pwaln_surveillance (\$\$$$$$$$$);
sub msa_surveillance (\@\@$$$$\%$\%$);
sub mask_aln_regions (\@\@$$$$$);
=cut


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;

	## Test Output Levels.
	## 0, silent; 1, very brief; 2, pretty brief; 3, rather verbose; 4, verbose; 5, very verbose, >5: extremely verbose. ##

	# (for the main part):
##our $TEST_LEVEL = 1;
our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;


	### Executable file. ###

my $main_indel_script = "Scripts/preprocess_msa_dawg.pl" ;
##my $main_indel_script = "Scripts/br_infer_psm_indel_events_spt_odr_dawg.v2.pl";


	##### Input/Output Files #####

##my $session_name = 'hogd_alt_lite_sbcl.1'; # For the full-analysis.
##my $session_name = 'trio_blen0.05'; # For the full-analysis.
##my $session_name = 'trio_blen0.2'; # For the full-analysis.
my $session_name = 'TESTs/Test1a'; # For TESTING the script. #

	## Archives of tiling MSAs. ##
	# $in_archive = "$inarch_dir/$archive_pref$id_class$archive_suff".

my $inarch_dir = "../Simulated_MSAs1/$session_name";
##my $inarch_dir = "../Simulated_MSAs2_1/$session_name";
##my $inarch_dir = "../../Inputs/Simulated_MSAs2_1/$session_name";
##my $inarch_dir = "../../Inputs/Simulated_MSAs1/hogd_alt_lite_sbcl.1";

##my $archive_pref = 'class_';
my $archive_pref = 'hog_d';
##my $archive_pref = 'Chr_';
my $archive_suff = '.tgz';


	### Top Output Directory ###

my $topoutdir = $session_name; # For full-scale analyses.
##my $topoutdir = "TESTs/Test5a"; ## For TESTING. ##
##my $topoutdir = 'hogd_alt_lite_sbcl.1'; # For full-scale analyses.


	## Temporary Input Files ##
	# $in_msa = "$tmp_dir/$inaln_pref$lcl_msa_id$inaln_suff", and
	# $ctrl_file = "$tmp_dir/$ctrl_pref$lcl_msa_id$ctrl_suff",
	# where $tmp_dir = "$intmp_aln_dir/$archive_pref$id_class",
	# where $id_class is the ID of the class of MSAs,
	# and $lcl_msa_id is the ID of a MSA in a particular class.

my $intmp_aln_dir = "$topoutdir/Temporary";

my $inaln_pref = 'dawg';
##my $inaln_pref = 'e';
##my $inaln_pref = 'optionG_';
my $inaln_suff = '.aln';

my $ctrl_pref = 'dawg_inp';
my $ctrl_suff = '.txt';


	## Output Files. (ADDED in creation of 'bch_serial_br_infer_indels.pl'.) ##
	## (MODIFIED in creation of 'bch_serial_br_infer_indels_dawg.pl'.) 
	##
	## $outfile1 = "$out_indel_dir/$pref_outfile1$lcl_msa_id$suff_outfile1",
	## $outfile2 = "$out_indel_dir/$pref_outfile2$lcl_msa_id$suff_outfile2",
	## $logfile  = "$out_indel_dir/$pref_logfile$lcl_msa_id$suff_logfile",
	# where $out_indel_dir = "$top_indel_dir/$archive_pref$id_class/";

my $top_indel_dir = "$topoutdir/Outputs";
##my $top_indel_dir = "$topoutdir/Indel_analyses";

	# MODIFIED in creation of 'bch_serial_preprocess_msas_dawg.pl'. #

my $pref_outfile1 = "prcsd_$inaln_pref";
my $pref_outfile2 = $ctrl_pref;
my $pref_logfile = 'log_preprocess.';

my $suff_outfile1 = $inaln_suff;
my $suff_outfile2 = $ctrl_suff;
my $suff_logfile = '.txt';


=for the predecessor 'bch_serial_br_infer_indels_dawg.pl'.
	# MODIFIED in creation of 'bch_serial_br_infer_indels_dawg.pl'. #
my $pref_outfile1 = 'output_detail.';
my $pref_outfile2 = 'output_brief.';
my $pref_logfile = 'log_infer_indel.';

my $suff_outfile1 = my $suff_outfile2 = my $suff_logfile = '.txt';
=cut


	## Check files/directories ##

	# Main script (ADDED in creation of 'bch_serial_br_infer_indels.pl').
foreach my $main_scr ($main_indel_script) {
    unless (( -f $main_scr ) and ( -r $main_scr )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable main script '$main_scr' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
}

foreach my $indir ($inarch_dir) {
    unless (( -d $indir ) and ( -r $indir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input directory '$indir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
}

my $list_id_classes = list_sessions_ext ($inarch_dir, $archive_pref, $archive_suff, '-');

unless (defined $list_id_classes) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the archive directory '$inarch_dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
}

## (@{$list_id_classes}>1) and (@{$list_id_classes} = ($list_id_classes->[0]) ); # For TESTING the script. #
## (@{$list_id_classes}>2) and (@{$list_id_classes} = @{$list_id_classes}[0,1]); # For TESTING the script. #

foreach my $out_dir ($intmp_aln_dir, $top_indel_dir) {
    unless ( -d $out_dir) {
	print STDOUT "[WARNING '$program_name'] CanNOT FIND an output directory '$out_dir' ! ... I will try to create one...\n\n";
	unless (0 == system (" mkdir -p $out_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the output directory '$out_dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
    unless ( -w $out_dir ) {
	print STDOUT "[WARNING '$program_name'] The output directory '$out_dir' is NOT WRITABLE ! ... I will try to change its permission mode...\n\n";
	unless (0 == system (" chmod -R u+w $out_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$out_dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<<< Program name: $program_name (path: $program_path) >>>>\n\n";


print STDOUT "Session Name: $session_name .\n\n"; # ADDED in creation of 'bch_serial_br_infer_indels_dawg.pl'. #


print STDOUT '$CURR_WORK_DIR = ', $CURR_WORK_DIR, " .\n\n";
##print STDOUT '$DIS_LEN = ', $DIS_LEN, " .\n";

	# MODIFIED in creation of 'bch_serial_br_infer_indels.pl'. 
print STDOUT "Main script for the analysis: $main_indel_script .\n\n";
##print STDOUT "Main PAML executable file: $main_bin_paml .\n\n"; # ADDED in creation of 'bch_serial_PAML.pl'.

print STDOUT "Input MSA archives: $inarch_dir/$archive_pref", '{ID_class}', "$archive_suff ,\n";
print STDOUT "          where {ID_class} = ", join (', ', @{$list_id_classes}), " .\n\n";
print STDOUT "Temporary input MSAs: $intmp_aln_dir/$archive_pref", '{ID_class}', "/$inaln_pref", '{local_MSA_ID}', "$inaln_suff .\n\n";   


	## ADDED in creation of 'bch_serial_PAML.pl' ##

print STDOUT "Auxiliary (temporary) input file for indel analyses:\n";

	# MODIFIED in creation of 'bch_serial_br_infer_indels_dawg.pl'. #
print STDOUT "  (input tree and indel parameters)  $intmp_aln_dir/$archive_pref", '{ID_class}', "/$ctrl_pref", '{local_MSA_ID}', "$ctrl_suff .\n\n";   
##print STDOUT "     (input tree)  $tmpl_intree .\n\n";


print STDOUT "Final Output Directories: {out_indel_dir} = $top_indel_dir/$archive_pref", '{chromosome_no.}/', " .\n\n";
print STDOUT "Output Files:\n";
print STDOUT "     (processed MSAs) {out_indel_dir}/$pref_outfile1", '{local_MSA_ID}', "$suff_outfile1 ,\n";
print STDOUT "     (control files)  {out_indel_dir}/$pref_outfile2", '{local_MSA_ID}', "suff_outfile2 ,\n";
=for the predecessor 'bch_serial_br_infer_indels_dawg.pl'
print STDOUT "     (detailed) {out_indel_dir}/$pref_outfile1", '{local_MSA_ID}', "$suff_outfile1 ,\n";
print STDOUT "     (brief)    {out_indel_dir}/$pref_outfile2", '{local_MSA_ID}', "suff_outfile2 ,\n";
=cut
print STDOUT "     (log)      {out_indel_dir}/$pref_logfile", '{local_MSA_ID}', "suff_logfile .\n\n";
	##
	## $outfile1 = "$out_indel_dir/$pref_outfile1$lcl_msa_id$suff_outfile1",
	## $outfile2 = "$out_indel_dir/$pref_outfile2$lcl_msa_id$suff_outfile2",
	## $logfile  = "$out_indel_dir/$pref_logfile$lcl_msa_id$suff_logfile",
	# where $out_indel_dir = "$top_indel_dir/$archive_pref$id_class/";


##my $ct_chr = 
my $ct_classes = my $ct_msas = my $ct_anom = my $ct_fail = my $ct_succ = 0;

foreach my $id_class (@{$list_id_classes}) { # Outer foreach-loop. (over archives (or classes)) #

    ##my $chr_no = $chr;

    ## ($id_class eq '1_4') or next; # For TESTING the script ('Test3b'). #
    ## ($id_class eq '2_3') or next; # For TESTING the script ('Test3c'). #
    ## ($id_class eq '2_7') or next; # For TESTING the script ('Test2a'). #
    ## (($id_class eq '1_1') or ($id_class eq '2_2')) or next; # For TESTING the script ('Test1a' & 'Test1b' ). #
    ## ($id_class eq '0003') or next; # For TESTING the script ('Test5a'). #
    ## ($id_class eq '0005') or next; # For TESTING the script ('Test6a'). #
    ## ($id_class eq '0002') or next; # For TESTING the script ('Test6a'). #

=for the predecessor 'bch_serial_br_infer_indels.pl'.`
    ## ($chr eq '01') or next; # For TESTING the script ('Test1a' & 'Test5a,c') #
    ## ($chr eq '05') or next; # For TESTING the script ('Test2a') #
=cut

    print STDOUT "<< Class: $id_class >>\n\n";

    my $in_archive = join ('', $inarch_dir, '/', $archive_pref, $id_class, $archive_suff);
		# $in_archive = "$inarch_dir/$archive_pref$chr_no$archive_suff", and 

    unless (( -f $in_archive ) and ( -r $in_archive ) ) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input archive file '$in_archive' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
    }

    my $mid_dir = join ('', $archive_pref, $id_class);

    my $list_in_msa_ids = peek_into_archive ($in_archive, $mid_dir, $inaln_pref, $inaln_suff);
    unless (defined $list_in_msa_ids) {
	    print STDOUT "[ERROR '$program_name'] 'peek_into_archive' FAILED on the input archive '$in_archive' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
    }
    ## (@{$list_in_msa_ids}>1) and (@{$list_in_msa_ids} = ($list_in_msa_ids->[0]) ); ## For TESTING the script. ##
    ## (@{$list_in_msa_ids}>2) and (@{$list_in_msa_ids} = @{$list_in_msa_ids}[0,1]); ## For TESTING the script. ##
    ## @{$list_in_msa_ids} = ('0004', '0122'); ## For TESTING the script ('Test2a'). ##
    ## @{$list_in_msa_ids} = ('0227'); ## For TESTING the script ('Test3b'). ##
    ## @{$list_in_msa_ids} = ('0120'); ## For TESTING the script ('Test3c'). ##
    ## @{$list_in_msa_ids} = ('0000746'); ## For TESTING the script ('Test5a'). ##
    ## @{$list_in_msa_ids} = ('0001033'); ## For TESTING the script ('Test6a'). ##
    ## @{$list_in_msa_ids} = ('0000385'); ## For TESTING the script ('Test7a'). ##

  (print STDOUT "#{Input MSAs in the archive} = ", scalar (@{$list_in_msa_ids}), " .\n\n") if ($TEST_LEVEL>0);


		# MODIFIED in creation of 'bch_serial_br_infer_indels.pl'. #

    my $parent_out_indel_dir = "$top_indel_dir/$mid_dir";
	# $out_indel_dir = "$top_indel_dir/$archive_pref$chr_no/$anchor_orf_id/";

    unless ( -d $parent_out_indel_dir) { # MODIFIED in creation of 'bch_serial_br_infer_indels.pl'. #
	unless (0 == system (" mkdir -p $parent_out_indel_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the parent output directory '$parent_out_indel_dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
    unless ( -w $parent_out_indel_dir ) { # MODIFIED in creation of 'bch_serial_br_infer_indels.pl'. #
	unless (0 == system (" chmod -R u+w $parent_out_indel_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the parent output directory '$parent_out_indel_dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }


    $ct_classes++;


    foreach my $in_msa_id (@{$list_in_msa_ids}) { # Middle foreach-loop (over MSAs in the archive) #

	$ct_msas++;

	## ($ct_msas>3) and last; # For TESTING the script. ('Test1a' & 'Test2a') for the OLD ones. #
	##($ct_msas>0) and last; # For TESTING the script. #
	##($in_msa_id eq 'YAL014C') or next; # For TESTING the script ('Test5a') #
	##($in_msa_id eq 'YAR008W') or next; # For TESTING the script ('Test5c') #

  (print STDOUT "< Input MSA ID: $in_msa_id >\n\n") if ($TEST_LEVEL>0);


		# Extract the input MSA file. #

	my $in_msa_name = join ('', $inaln_pref, $in_msa_id, $inaln_suff);
	my $extracted_file = "$mid_dir/$in_msa_name";
	my $tmpin_msa = "$intmp_aln_dir/$extracted_file";

		# $tmpin_msa = "$intmp_aln_dir/$archive_pref$chr_no/$inaln_pref$in_msa_id$inaln_suff",
		# where $chr_no is the numerical part of the chromosome ID (in the BED file(s)).

	unless (0 == system ("tar -C $intmp_aln_dir -xpzf $in_archive $extracted_file") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to extract a file '$extracted_file' into '$intmp_aln_dir' from the archive '$in_archive', for the MSA '$in_msa_id' belonging to the class '$id_class': $? !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
	unless (( -f $tmpin_msa ) and ( -r $tmpin_msa )) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable temporary input MSA file '$tmpin_msa': $! !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}

		# Extract the Dawg control file. #
		# (ADDED in creation of 'bch_serial_br_infer_indels_dawg.pl'.)

	my $ctrl_file_name = join ('', $ctrl_pref, $in_msa_id, $ctrl_suff);
	my $extracted_ctrl_file = "$mid_dir/$ctrl_file_name";
	my $tmpin_ctrl = "$intmp_aln_dir/$extracted_ctrl_file";

		# $ctrl_file = "$tmp_dir/$ctrl_pref$lcl_msa_id$ctrl_suff",
		# where $tmp_dir = "$intmp_aln_dir/$archive_pref$id_class",

	unless (0 == system ("tar -C $intmp_aln_dir -xpzf $in_archive $extracted_ctrl_file") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to extract a file '$extracted_ctrl_file' into '$intmp_aln_dir' from the archive '$in_archive', for the MSA '$in_msa_id' belonging to the class '$id_class': $? !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
	unless (( -f $tmpin_ctrl ) and ( -r $tmpin_ctrl )) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable temporary Dawg control file '$tmpin_ctrl': $! !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}


		# Create a directory for performing an Indel analysis. #
		# (ADDED in creation of 'bch_serial_br_infer_indels.pl'.)

	my $out_indel_dir = $parent_out_indel_dir;
	##my $out_indel_dir = "$parent_out_indel_dir/$in_msa_id";

	unless ( -d $out_indel_dir) {
	    unless (0 == system (" mkdir -p $out_indel_dir ") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to create the Indel output directory '$out_indel_dir': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}
	unless ( -w $out_indel_dir ) {
	    unless (0 == system (" chmod -R u+w $out_indel_dir ") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the Indel output directory '$out_indel_dir': $?  !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}


		## ADDED in creation of 'bch_serial_br_infer_indels.pl'. ##
		## => MODIFIED in creation of 'bch_serial_br_infer_indels_dawg.pl'. ##

	## $outfile1 = "$out_indel_dir/$pref_outfile1$lcl_msa_id$suff_outfile1",
	## $outfile2 = "$out_indel_dir/$pref_outfile2$lcl_msa_id$suff_outfile2",
	## $logfile  = "$out_indel_dir/$pref_logfile$lcl_msa_id$suff_logfile",
	# where $out_indel_dir = "$top_indel_dir/$archive_pref$id_class/";

	my $outfile1 = "$out_indel_dir/$pref_outfile1$in_msa_id$suff_outfile1";
	my $outfile2 = "$out_indel_dir/$pref_outfile2$in_msa_id$suff_outfile2";
	my $logfile = "$out_indel_dir/$pref_logfile$in_msa_id$suff_logfile";


		# Further MODIFIED in creation of 'bch_serial_preprocess_msas_dawg.pl'. #

	unless (0 == system ("(time perl $main_indel_script $tmpin_msa $tmpin_ctrl $outfile1) > $logfile 2>&1") ) {
	    print STDERR "[ERROR '$program_name'] FAILED 'system ('perl $main_indel_script ...')' for the input MSA ID '$in_msa_id' in the class '$id_class': $! !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}

	unless (0 == system ("cp -p $tmpin_ctrl $outfile2") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED 'system ('cp -p $tmpin_ctrl $outfile2')' for the input MSA ID '$in_msa_id' in the class '$id_class': $! !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}

=for the predecessor 'bch_serial_br_infer_indels_dawg.v2.pl'.
	unless (0 == system ("(time perl $main_indel_script $tmpin_msa $tmpin_ctrl $outfile1 $outfile2) > $logfile 2>&1") ) {
	    print STDERR "[ERROR '$program_name'] FAILED 'system ('perl $main_indel_script ...')' FAILED for the input MSA ID '$in_msa_id' in the class '$id_class': $! !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
=cut

	foreach my $ofile ($outfile1, $outfile2, $logfile) {
	    unless (( -f $ofile ) and ( -r $ofile )) {
		print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable output file '$ofile'... I guess that the analysis via the script '$main_indel_script' FAILED on the temporary input MSA '$tmpin_msa' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}

  (print STDOUT "\n ... Successfully extracted the input alignment '$tmpin_msa' and preprocessed it ...\n\n") if ($TEST_LEVEL>1);
  ##(print STDOUT "\n ... Successfully extracted the input alignment '$tmpin_msa' and performed an Indel analysis on it ...\n\n") if ($TEST_LEVEL>1);

		# Remove the temporary input files. #
		# ( MODIFIED in creation of 'bch_serial_br_infer_indels_dawg.pl'.)

	foreach my $tmpin_file ($tmpin_msa, $tmpin_ctrl) {
	    unless (0 == system ("rm -f $tmpin_file") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to remove the temporary input file '$tmpin_file': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}

	$ct_succ++;

    }					# End of the middle foreach-loop (over MSAs in the archive). #

		## Archive the Middle output directory (storing masked MSAs for the same chromosome.) ##

    my $out_archive_name = "$mid_dir.tgz";

     # $parent_out_indel_dir = "$top_indel_dir/$mid_dir";
     # [ $parent_inout_paml_dir = "$top_paml_dir/$mid_dir"; ] --- for the predecessor 'bch_serial_PAML.pl'.
     # [$outaln_dir = "$topoutaln_dir/$mid_dir";]  --- for the predecessor 'bchbch_refine_msas.pl'.

    unless ( chdir $top_indel_dir ) { # MODIFIED in creation of 'bch_serial_br_infer_indels.pl'.
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to change the directory to '$top_indel_dir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    unless (0 == system ("tar -cpzf $out_archive_name $mid_dir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED TO CREATE a tar-gzipped ARCHIVE '$out_archive_name out of the directory '$mid_dir' under the directory '$top_indel_dir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    unless (0 == system ("rm -rf $mid_dir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to REMOVE the directory '$mid_dir' under the directory '$top_indel_dir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
 (print STDOUT "\n... Successfully archived the masked MSAs into '$out_archive_name' under '$top_indel_dir'!!\n\n") if ($TEST_LEVEL>1);

    unless ( chdir $CURR_WORK_DIR ) {
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to RETURN TO the original place, '$CURR_WORK_DIR' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    ##($ct_chr>0) and last; ## For TESTING the script. ##

}			# END of the outer foreach-loop. (over archives (or classes)) #


print STDOUT "\n=== All Processes FINISHED !! ===\n\n";

print STDOUT "#{MSA Classes} = $ct_classes .\n";
print STDOUT "#{Input MSAs}  = $ct_msas .\n";
print STDOUT "    #{Anomalous} = $ct_anom ,\n";
print STDOUT "    #{Failed}    = $ct_fail ,\n";
print STDOUT "    #{Succeeded} = $ct_succ .\n\n";



	### DEFINITIONS OF THE SUBROUTINES ###

# peek_into_archive:
#	This subroutine looks into an input archive and outputs a list of entries.
#	The archive is supposed to be storing files with the following paths:
#	"$dir/$pref$id$suff",
#	where $id specifies the ID of the file.
#
# INPUT: (\$in_archive, $dir, $pref, $suff), where
#	$in_archive is the path to the input archive;
#	$dir/$pref/$suff are templates for the paths to the files stored in the archive.
#
# OUTPUT: \@list_ids, where
#	@list_ids stores the IDs of the archived files with the above paths.
#
sub peek_into_archive (\$$$$) {

	my ($in_archive, $dir, $pref, $suff) = @_;
	local $/ = "\n"; # Use the default input-record-separator. #

	my $fh_tmp;
	unless ( open ($fh_tmp, "tar -tzf ${$in_archive} |" ) ) {
	    print STDOUT "[ERROR 'peek_into_archive'] CanNOT fork or look into ${$in_archive}: $!\n";
	    return undef;
	}

	my @list_ids = ();

	my $ext_pref = "$dir/$pref"; # Extended prefix. #
	my $len_epref = length $ext_pref;
	my $len_suff = length $suff;

	while (my $tmp = <$fh_tmp>) { # Main while-loop (over input lines) #

	    $tmp =~ s/\s+$//; # Remove white spaces at the end of the entry (serving as a separator (and possibly a pad)).
	    ($ext_pref =~ m/^\.\//) or ($tmp =~ s/^\.\///); # Remove './' at the beginning of the entry unless $ext_pref begins with './'. (ADDED in creation of 'bch_serial_PAML.pl'.)

	    my $tmp_suff = substr ($tmp, - $len_suff, $len_suff, '');

	    ($tmp_suff eq $suff) or next; # Skip if the suffix doesn't match.

	    my $tmp_epref = substr ($tmp, 0, $len_epref, '');

	    ($tmp_epref eq $ext_pref) or next; # Skip if the extended prefix doesn't match.

	    push @list_ids, $tmp; # The remnant of the entry must be the ID.


	}			# End of the main while-loop (over input lines) #
	close ($fh_tmp);

	return \@list_ids;

}	# End of the subroutine 'peek_into_archive' #


