#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# bch_serial_Dawg.v2.pl:
#	This modified version of 'bch_serial_Dawg.pl' takes advantage of the fact that differnt sub-classes
#	in the same class shares the same  control file templates,
#	and outputs the simulation results in an easier directory structure.
#	(CREATED on Feb 27 (Thu), 2014)
#
# bch_serial_Dawg.pl:
#	This Perl script performs bach Dawg simulations, based on the template control files.
#	The script was created by taking advantage of the overall framework of the script 'bch_serial_PAML.pl'.
#	(CREATION STARTED on Feb 24 (Mon), 2014, ENDED on Feb 26 (Wed), 2014)
#
# bch_serial_PAML.pl:
#	This Perl script performs bach PAML analyses on the input sequence alignments,
#	all of which share a species tree, stored in multiple gzipped tar archives.
#	The script was created by modifying 'bchbch_refine_msas.pl',
#	and borrowing functions to perform PAML anayses from 'compare_ml_msa.bch.extJC.6t.PhyML.pl'.
#	(CREATION STARTED on Nov 3 (Sun), 2013, ENDED on Nov 3 (Sun), 2013.)
#


use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext
	tfm_labels2indices tfm_labels2indices_gen
	);

use MyTreeManipu_II qw(
	input_tree_file
	);
	## extract_sgltree_from_file
	## output_tree

use MyTreeManipu_II_dawg qw(
	br_draw_subtree_dawg
	);


=for the predecessor 'bchbch_refine_msas.pl'
use MyArrayManipu qw( multiple_sort );

use MyMalnIO qw(
	extract_fasta extract_clw
	pr_ODEN_input
	pr_clustal_like_lite
	);
	## pr_phylip_aln
	## pr_clustal_like_lite_aa

use MyNucSeq2 qw(
	revcom print_sequence
	);

use MyProbDistri qw(
	poisson_cum
	binomial_cum_via_betai
	);
=cut

sub peek_into_archive (\$$$$);
sub assign_interior_node_names (\$$);
sub create_dawg_ctrl_file (\$\$\%\%);


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;

	## Test Output Levels.
	## 0, silent; 1, very brief; 2, pretty brief; 3, rather verbose; 4, verbose; 5, very verbose, >5: extremely verbose. ##

	# (for the main part):
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;

	# For each class, the alignment number, and the random number seed, begins with this number (integer). #
our $BASEBASE_ALN_NUM = 0;

	# Number of alignments per sub-class
our $CT_ALN_PER_SBCLS = 3; # For TESTING the script. #
#our $CT_ALN_PER_SBCLS = 300; # For the full-scale analysis. #
=for the predecessor 'bch_serial_Dawg.pl'
##our $CT_ALN_PER_SBCLS = 30; # For the full-scale analysis. #
=cut


	## Necessary for creating control files. ## 

our %IN2OUT_OPTNAME = ( # $in_optname => $out_optname,
	'Output.File' => 'File',
	'Sim.Reps' => 'Reps',
	'Root.Length' => 'Length',
	'Tree.Tree' => 'Tree',
	'Tree.Scale' => 'TreeScale',
	'Subst.Model' => 'Model',
	'Indel.Model' => 'GapModel',
	'Indel.Params' => 'GapParams',
	'Indel.Rate' => 'Lambda',
	);

	# Necessary for processing trees #

our $TOP_NODE  = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
our $INDX_INTNODENAME = $MyTreeManipu_II_dawg::INDX_INTNODENAME;

our $PREF_INTNODENAME = 'anc';

	### Executable file. ###

my $main_bin_dawg = "/Users/kezawa3/Bioinformatics/Dawg/Packaged/dawg1.2-release/bin/dawg";


	### Input/Output Files ###

	## Templates of Dawg control files. ##
	## $tmpl_file = "$indir/$tmpl_file_name", where
	## $indir = "$topindir/$inmiddir1_pref$id_class$inmiddir1_suff/$inmiddir2_pref$id_subclass$inmiddir2_suff";

my $topindir = "/Users/kezawa3/Ezawa_works_UH/Main_Project/2nd_Stage.DNA/Initial_Data/hogd_alt_lite_sbcl";
my $inmiddir1_pref = 'hog_d';
my $inmiddir1_suff = '';
my $inmiddir2_pref = 'hog_d';
my $inmiddir2_suff = '';
my $tmpl_file_name = 'dawg_inp.txt';


	## Output files ##
	## $out_ctrl_file = "$outdir/$ctrl_pref$aln_num$ctrl_suff", and
	## $out_aln_file = "$outdir/$aln_pref$aln_num$aln_suff", where
	## [ $outdir = "$topoutdir/$outmiddir1_pref$id_class$outmiddir1_suff/$outmiddir2_pref$id_sbclass$outmiddir2_suff"; ] --- for the predecessor 'bch_serial_Dawg.pl'
	## $outdir = "$topoutdir/$outmiddir1_pref$id_class$outmiddir1_suff";

my $topoutdir = "TESTs/Test1a"; ## For TESTING. ##
##my $topoutdir = 'hogd_alt_lite_sbcl.1'; # For full-scale analyses.

my $outmiddir1_pref = $inmiddir1_pref;
my $outmiddir1_suff = $inmiddir1_suff;

=for the predecessor 'bch_serial_Dawg.pl'
my $outmiddir2_pref = $inmiddir2_pref;
my $outmiddir2_suff = $inmiddir2_suff;
=cut

my $ctrl_pref = 'dawg_inp';
my $ctrl_suff = '.txt';
my $aln_pref = 'dawg';
my $aln_suff = '.aln';

=for the predecessor 'bch_serial_PAML.pl'

	## Temporary Input Files ##
	# $in_msa = "$intmp_aln_dir/$archive_pref$chr_no/$inaln_pref$anchor_orf_id$inaln_suff",
	# where $chr_no is the numerical part of the chromosome ID (in the BED file(s)).

my $intmp_aln_dir = "$topoutdir/Temporary";
my $inaln_pref = 'optionG_';
my $inaln_suff = '.aln';

	## Template files. (ADDED in creation of 'bch_serial_PAML.pl'.) ##
	##
	## $tmpl_intree = "$tmpl_dir/$intree_name",
	## $tmpl_ctrl_file = "$tmpl_dir/$ctrl_file_name",
my $tmpl_dir = "Templates";
my $intree_name = 'four_yeasts.tree';
my $ctrl_file_name = 'baseml.ctl';

my $tmpl_intree = "$tmpl_dir/$intree_name";
my $tmpl_ctrl_file = "$tmpl_dir/$ctrl_file_name";

=cut

	## Check files/directories ##

	# Main binary files #
foreach my $main_bin ($main_bin_dawg) {
    unless (( -f $main_bin ) and ( -x $main_bin )) {
	print STDERR "[ERROR '$program_name'] CanNOT FIND an executable PROGRAM '$main_bin' !\n";
	die "$! .\n\n";
    }
}

foreach my $indir ($topindir) {
    unless (( -d $indir ) and ( -r $indir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input directory '$indir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
}

my $list_id_classes = list_sessions_ext ($topindir, $inmiddir1_pref, $inmiddir1_suff, 'drw');

unless (defined $list_id_classes) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the top input directory '$topindir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
}

## (@{$list_id_classes}>2) and (@{$list_id_classes} = @{$list_id_classes}[0, 1]); # For TESTING the script. #

foreach my $out_dir ($topoutdir) {
    unless ( -d $out_dir) {
	print STDOUT "[WARNING '$program_name'] CanNOT FIND an output directory '$out_dir' ! ... I will try to create one...\n\n";
	unless (0 == system (" mkdir -p $out_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the output directory '$out_dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
    unless ( -w $out_dir ) {
	print STDOUT "[WARNING '$program_name'] The output directory '$out_dir' is NOT WRITABLE ! ... I will try to change its permission mode...\n\n";
	unless (0 == system (" chmod -R u+w $out_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$out_dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
}



	### Declare the Beginning of the Script ###

print STDOUT "\n<<<< Program name: $program_name (path: $program_path) >>>>\n\n";

print STDOUT '$CURR_WORK_DIR = ', $CURR_WORK_DIR, " .\n\n";

print STDOUT "Main Dawg executable file: $main_bin_dawg .\n\n";

print STDOUT "{Initial value of the alignment numbers} = $BASEBASE_ALN_NUM .\n\n";

print STDOUT "{Number of alignments per sub-class} = $CT_ALN_PER_SBCLS .\n\n";

print STDOUT "{Shared prefix of the interior node names} = '$PREF_INTNODENAME' .\n\n";

print STDOUT "Input Files:\n\n";
print STDOUT "   (Templates of Dawg control files) {in_dir}/$tmpl_file_name ,\n";
print STDOUT "         where {in_dir} = {top_in_dir}/$inmiddir1_pref", '{id_class}', "$inmiddir1_suff/$inmiddir2_pref", '{id_subclass}', "$inmiddir2_suff ,\n";
print STDOUT "          with {top_in_dir} = $topindir ,\n";
print STDOUT "           and {id_class} = ", join (', ', @{$list_id_classes}), " \n";

	# MODIFIED in creation of 'bch_serial_Dawg.v2.pl'. #
print STDOUT "                (only the first value of {id_subclass} will be used) .\n\n";
##print STDOUT "                ({id_subclass} will be given later) .\n\n";

	## Templates of Dawg control files. ##
	## $tmpl_file = "$indir/$tmpl_file_name", where
	## $indir = "$topindir/$inmiddir1_pref$id_class$inmiddir1_suff/$inmiddir2_pref$id_subclass$inmiddir2_suff";

print STDOUT "Output Files:\n\n";

print STDOUT "    (Dawg control files) {out_dir}/$ctrl_pref", '{aln_no}', "$ctrl_suff ,\n";
print STDOUT "    (Dawg alignments)    {out_dir}/$aln_pref", '{aln_no}', "$aln_suff ,\n";

	# MODIFIED in creation of 'bch_serial_Dawg.v2.pl'. #
print STDOUT "         where {out_dir} = {top_out_dir}/$outmiddir1_pref", '{id_class}', "$outmiddir1_suff ,\n";
##print STDOUT "         where {out_dir} = {top_out_dir}/$outmiddir1_pref", '{id_class}', "$outmiddir1_suff/$outmiddir2_pref", '{id_subclass}', "$outmiddir2_suff ,\n";

print STDOUT "         with {top_out_dir} = $topoutdir ,\n";

	# MODIFIED in creation of 'bch_serial_Dawg.v2.pl'. #
print STDOUT "         and {id_class} is the same as those for the input files ,\n";
##print STDOUT "         and {id_class}, {id_subclass} are the same as those for the input files ,\n";

printf STDOUT "          and {aln_num}  = %02d, .., %0d2 .\n\n", 1, $CT_ALN_PER_SBCLS; 

	## Output files ##
	## $out_ctrl_file = "$outdir/$ctrl_pref$aln_num$ctrl_suff", and
	## $out_aln_file = "$outdir/$aln_pref$aln_num$aln_suff", where
	## $outdir = "$topoutdir/$outmiddir1_pref$id_class$outmiddir1_suff/$outmiddir2_pref$id_sbclass$outmiddir2_suff";

print STDOUT "\n\n<<< Now, the main process begins ... >>>\n\n";

my $ct_classes = my $totct_sbclasses = my $totct_malns = 0;
my $totct_anom = my $totct_fail = my $totct_succ = 0;

foreach my $id_class (@{$list_id_classes}) { # Main outer foreach-loop (over classes). #

    my $inmiddir = "$topindir/$inmiddir1_pref$id_class$inmiddir1_suff";
    my $outmiddir_name = "$outmiddir1_pref$id_class$outmiddir1_suff";
    my $outmiddir = "$topoutdir/$outmiddir_name";

    print STDOUT "<< Class: $id_class >>\n\n";

    unless (( -d $inmiddir ) and ( -r $inmiddir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input directory '$inmiddir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    my $list_id_sbclasses = list_sessions_ext ($inmiddir, $inmiddir2_pref, $inmiddir2_suff, 'drw');

    unless (defined $list_id_sbclasses) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the middle input directory '$inmiddir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

=for the predecessor 'bch_serial_Dawg.pl'
    ## (@{$list_id_sbclasses}>2) and (@{$list_id_sbclasses} = @{$list_id_sbclasses}[0, 1]); # For TESTING the script. #
=cut

  if ($TEST_LEVEL>1) {
    print STDOUT "Middle input directory:  $inmiddir .\n";
    print STDOUT "Middle output directory: $outmiddir .\n\n";

    print STDOUT "{id_subclasses} = ", join (', ', @{$list_id_sbclasses}), " ,\n";
    print STDOUT " out of which only the 1st subclass will be used ...\n\n";
  }

	## ADDED in creation of 'bch_serial_Dawg.v2.pl'. (Used whether it's the FULL-SCALE analysis or TESTING the script). #
    (@{$list_id_sbclasses}>1) and (@{$list_id_sbclasses} = ($list_id_sbclasses->[0]) );

    my $ct_sbclasses = my $ct_malns = 0;
    my $ct_anom = my $ct_fail = my $ct_succ = 0;

    my $log_name = "log_dawg.txt";

    my $base_aln_num = $BASEBASE_ALN_NUM;


    foreach my $id_sbclass (@{$list_id_sbclasses}) { # Main middle foreach-loop (over sub-classes). #

	print STDOUT "< Sub-class: $id_sbclass >\n\n";

	my $indir = "$inmiddir/$inmiddir2_pref$id_sbclass$inmiddir2_suff";

		## MODIFIED in creation of 'bch_serial_Dawg.v2.pl'. ##
	my $outdir = $outmiddir;
	##my $outdir = "$outmiddir/$outmiddir2_pref$id_sbclass$outmiddir2_suff";

	unless (( -d $indir ) and ( -r $indir )) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input directory '$indir' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}

	unless ( -d $outdir) {
	    unless (0 == system (" mkdir -p $outdir ") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to create the output directory '$outdir': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}
	unless ( -w $outdir ) {
	    print STDOUT "[WARNING '$program_name'] The output directory '$outdir' is NOT WRITABLE ! ... I will try to change its permission mode...\n\n";
	    unless (0 == system (" chmod -R u+w $outdir ") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$outdir': $?  !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}

	my $tmpl_file = "$indir/$tmpl_file_name";

   if ($TEST_LEVEL>1) {
	print STDOUT "Input directory: $indir .\n";
	print STDOUT "Output directory: $outdir .\n\n";
   }

	$ct_sbclasses++;


	    ## Templates of Dawg control files. ##
	    ## $tmpl_file = "$indir/$tmpl_file_name", where
	    ## $indir = "$topindir/$inmiddir1_pref$id_class$inmiddir1_suff/$inmiddir2_pref$id_subclass$inmiddir2_suff";

	unless (( -f $tmpl_file ) and ( -r $tmpl_file )) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable control file template '$tmpl_file' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ; 
	}

    (print STDOUT "Template of the Dawg control file: $tmpl_file .\n\n") if ($TEST_LEVEL>2);

	for (my $n=1; $n <= $CT_ALN_PER_SBCLS; $n++) { # Main inner for-loop (over simulated alignments). #

	    my $number = $n + $base_aln_num;
	    my $aln_num = sprintf ("%04d", $number);
	    ##my $aln_num = sprintf ("%03d", $n);
	    ####my $aln_num = sprintf ("%02d", $n);

   (print STDOUT "( Alignment number: $aln_num )\n\n") if ($TEST_LEVEL>2);

	    my $ctrl_file_name = "$ctrl_pref$aln_num$ctrl_suff";
	    my $out_ctrl_file = "$outdir/$ctrl_file_name";
	    my $aln_file_name = "$aln_pref$aln_num$aln_suff";
	    my $out_aln_file = "$outdir/$aln_file_name";

		## Output files ##
		## $out_ctrl_file = "$outdir/$ctrl_pref$aln_num$ctrl_suff", and
		## $out_aln_file = "$outdir/$aln_pref$aln_num$aln_suff", where

	    my %specified_options = (
		'File' => join ('', '"' , $aln_file_name, '"'),
		'Reps' => 1,
		'Model' => '"JC"',
		'GapModel' => '"PL"',
		'Format' => '"Clustal"',
		'Seed' => $number,
		##'Seed' => $n,
		## 'Output.Split' => 'CommentOut', # Comment out this option. #
		);

	    unless ( create_dawg_ctrl_file ($tmpl_file, $out_ctrl_file, %IN2OUT_OPTNAME, %specified_options) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'create_dawg_ctrl_file' FAILED on the output file '$out_ctrl_file' with the template '$tmpl_file' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	    }

	    ### next; # For TESTING the subroutine 'create_dawg_ctrl_file'. #


			### RESTART FROM HERE ###

		# Perform the Dawg simulation in the output directory. #

	    unless ( chdir $outdir ) {
		print STDERR "[ERROR '$program_name'] FAILED the command 'chdir' to change the directory to '$outdir' to perform the Dawg simulation of the number '$aln_num' in the sub-class '$id_sbclass' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }

	    unless (0 == system ("echo 'Aln_no.: $aln_num' >> $log_name") ) {
		print STDERR "[ERROR '$program_name'] 'system ('echo ...')' FAILED before the Dawg simulation of number '$aln_num' in the sub-class '$id_sbclass' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }

	    unless (0 == system ("$main_bin_dawg $ctrl_file_name >> $log_name 2>&1") ) {
		print STDERR "[ERROR '$program_name'] 'system ('$main_bin_dawg ...')' FAILED for the alignment number '$aln_num' in the sub-class '$id_sbclass' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
  (print STDOUT "\n ... Successfully performed a Dawg simulation for the alignment number '$aln_num' in the sub-class '$id_sbclass' ...\n\n") if ($TEST_LEVEL>2);

	    unless (0 == system ("echo '' >> $log_name") ) {
		print STDERR "[ERROR '$program_name'] 'system ('echo ...')' FAILED after the Dawg simulation of number '$aln_num' in the sub-class '$id_sbclass' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }

		# Get back to the original directory. #

	    unless ( chdir $CURR_WORK_DIR ) {
		print STDERR "[ERROR '$program_name'] FAILED the subroutine 'chdir' to MOVE BACK to the original working directory '$CURR_WORK_DIR' after performing the Dawg simulation of the number '$aln_num' in the sub-class '$id_sbclass' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }

	    $ct_succ++;


	}					# End of the main inner for-loop (over simulated alignments). #

	$base_aln_num += $CT_ALN_PER_SBCLS;

    }					# End of the main middle foreach-loop (over sub-classes). #

    print STDOUT "#{sub-classes} = $ct_sbclasses .\n";
    print STDOUT "#{MSAs}  = $ct_malns .\n";
    print STDOUT "    #{Anomalous} = $ct_anom ,\n";
    print STDOUT "    #{Failed}    = $ct_fail ,\n";
    print STDOUT "    #{Succeeded} = $ct_succ .\n\n";

    $totct_sbclasses += $ct_sbclasses;
    $totct_malns += $ct_malns;
    $totct_anom +=  $ct_anom;
    $totct_fail += $ct_fail;
    $totct_succ += $ct_succ;


	## Archive the Middle output directory (storing the outputs of Dawg simulations in the class '$id_class'). ##

    my $out_archive_name = "$outmiddir_name.tgz";
    ##my $out_archive_name = "$mid_dir.tgz";

       ## $outmiddir = "$topoutdir/$outmiddir_name";

    unless ( chdir $topoutdir ) {
	print STDOUT "[ERROR '$program_name'] FAILED the command 'chdir' to change the directory to '$topoutdir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    unless (0 == system ("tar -cpzf $out_archive_name $outmiddir_name") ) {
	print STDOUT "[ERROR '$program_name'] FAILED TO CREATE a tar-gzipped ARCHIVE '$out_archive_name' out of the directory '$outmiddir_name' under the directory '$topoutdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    unless (0 == system ("rm -rf $outmiddir_name") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to REMOVE the directory '$outmiddir_name' under the directory '$topoutdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
 (print STDOUT "\n... Successfully archived the simulation results of the class '$id_class' into '$out_archive_name' under '$topoutdir'!!\n\n") if ($TEST_LEVEL>1);

    unless ( chdir $CURR_WORK_DIR ) {
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to RETURN TO the original place, '$CURR_WORK_DIR' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

}			# End of the main outer foreach-loop (over classes). #


print STDOUT "\n=== All Processes FINISHED !! ===\n\n";

print STDOUT "#{Classes} = $ct_classes .\n";
print STDOUT "Total#{sub-classes} = $totct_sbclasses .\n";
print STDOUT "Total#{MSAs}  = $totct_malns .\n";
print STDOUT "   Total#{Anomalous} = $totct_anom ,\n";
print STDOUT "   Total#{Failed}    = $totct_fail ,\n";
print STDOUT "   Total#{Succeeded} = $totct_succ .\n\n";



	### DEFINITIONS OF THE SUBROUTINES ###

# peek_into_archive:
#	This subroutine looks into an input archive and outputs a list of entries.
#	The archive is supposed to be storing files with the following paths:
#	"$dir/$pref$id$suff",
#	where $id specifies the ID of the file.
#
# INPUT: (\$in_archive, $dir, $pref, $suff), where
#	$in_archive is the path to the input archive;
#	$dir/$pref/$suff are templates for the paths to the files stored in the archive.
#
# OUTPUT: \@list_ids, where
#	@list_ids stores the IDs of the archived files with the above paths.
#
sub peek_into_archive (\$$$$) {

	my ($in_archive, $dir, $pref, $suff) = @_;
	local $/ = "\n"; # Use the default input-record-separator. #

	my $fh_tmp;
	unless ( open ($fh_tmp, "tar -tzf ${$in_archive} |" ) ) {
	    print STDOUT "[ERROR 'peek_into_archive'] CanNOT fork or look into ${$in_archive}: $!\n";
	    return undef;
	}

	my @list_ids = ();

	my $ext_pref = "$dir/$pref"; # Extended prefix. #
	my $len_epref = length $ext_pref;
	my $len_suff = length $suff;

	while (my $tmp = <$fh_tmp>) { # Main while-loop (over input lines) #

	    $tmp =~ s/\s+$//; # Remove white spaces at the end of the entry (serving as a separator (and possibly a pad)).
	    ($ext_pref =~ m/^\.\//) or ($tmp =~ s/^\.\///); # Remove './' at the beginning of the entry unless $ext_pref begins with './'. (ADDED in creation of 'bch_serial_PAML.pl'.)

	    my $tmp_suff = substr ($tmp, - $len_suff, $len_suff, '');

	    ($tmp_suff eq $suff) or next; # Skip if the suffix doesn't match.

	    my $tmp_epref = substr ($tmp, 0, $len_epref, '');

	    ($tmp_epref eq $ext_pref) or next; # Skip if the extended prefix doesn't match.

	    push @list_ids, $tmp; # The remnant of the entry must be the ID.


	}			# End of the main while-loop (over input lines) #
	close ($fh_tmp);

	return \@list_ids;

}	# End of the subroutine 'peek_into_archive' #



# assign_interior_node_names:
#	This subroutine receives a tree in the Newick format,
#	assigns the names to the interior nodes in a systematic way,
#	and returns the resulting tree.
#
# INPUT: (\$intree, $pref_intnodename), where
#	$intree is a tree in the Newick format;
#	$pref_intnodename is the common prefix for the interior node name.
#
# OUTPUT: \$outree, where
#	$outtree is the same tree as $intree, but with names of the interior nodes.
#
sub assign_interior_node_names (\$$) {

	my ($intree, $pref_intnodename) = @_;

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_attr;
	unless ( input_tree_file (${$intree}, %node2pa, %node2ch, %name2id, %id2name, %br_attr) ) {
	    print STDOUT "[ERROR 'assign_interior_node_names'] 'input_tree_file' FAILED: $! !\n";
	    return undef;
	}

		## If branches are found to have negative lengths, the lengths will be forcefully set at zero. ##

	foreach my $br (keys %br_attr) { # Preliminary foreach-loop (over branches). #
	    my $attr = $br_attr{$br};
	    ((defined $attr) and (defined $attr->[$INDX_LENGTH])) or next;
	    ($attr->[$INDX_LENGTH] < 0) and ($attr->[$INDX_LENGTH] = 0);
	}			# End of the preliminary foreach-loop (over branches). #

		## Systematically assign names to the interior nodes. ##

	my $topnode_id = $name2id{$TOP_NODE};

	my @nodes = ($topnode_id);
	my $intnode_num = 1;
	my $topnode_name = sprintf ("${pref_intnodename}%04d", $intnode_num);

	while (@nodes>0) { # Main outer while-loop (until the interior nodes run out). #

	    my @new_nodes = ();
	    foreach my $node (@nodes) { # Main middle foreach-loop (over nodes of this generation). #

		my $children = $node2ch{$node};
		((defined $children) and (@{$children}>0)) or next; # Skip exterior nodes. #

		unless ($node == $topnode_id) {
		    $intnode_num++;
		    my $nodename = sprintf ("${pref_intnodename}%04d", $intnode_num);
		    my $attr = $br_attr{$node};
		    (defined $attr) or ($attr = $br_attr{$node} = []);
		    $attr->[$INDX_INTNODENAME] = $nodename;
		}

		push @new_nodes, @{$children};

	    }			# End of the main middle foreach-loop (over nodes of this generation). #
		# Update the node set to the next generation. #
	    @nodes = @new_nodes;

	}	# End of the main outer while-loop (until the interior nodes run out). #


	my $option_bootlabel = 0; # = 0 if you display bootstrap values at the nodes.
	my $option_nodeid = 0; # = 0 if you don't want the node ID displayed in the tree.

	my $outtree = br_draw_subtree_dawg ($topnode_id, %node2ch, %id2name, %br_attr, $option_bootlabel, $option_nodeid);
	unless (defined $outtree) {
	    print STDOUT "[ERROR 'assign_interior_node_names'] 'br_draw_subtree_dawg' FAILED (topnode: $topnode_id): $! !\n";
	    return undef;
	}

		# A BUG fixed on Feb 25 (Tue), 2014. #
	##$outtree .= "$topnode_name\;";
	##return \$outtree;

	${$outtree} .= "$topnode_name\;";

	return $outtree;

}	# End of the subroutine 'assign_interior_node_names' #


# create_dawg_ctrl_file:
#	This subroutine creates	a Dawg control file
#	based on a template, and options specified by the input.
#
# INPUT: (\$in_tmpl_file, \$out_ctrl_file, \%in2out_optname, \%specified_options), where
#	$in_tmpl_file is the path to the input template of the control file;
#	$out_ctrl_file is the path to the output control file;
#	%in2out_optname = ($in_optname => $out_optname, ...) with
#	  $in_optname being the option name used in the input template, and
#	  $out_optname being the option name to be used in the output file;
#	%specified_options = ($option_name => $specified_value, ...) for the options whose values must be fixed as specified. (These values will override the values given previously.);
#
# OUTPUT: 1 if the subroutine successfully ended,
#	  0 if it failed.
#
sub create_dawg_ctrl_file (\$\$\%\%) {

	my ($in_tmpl_file, $out_ctrl_file, $in2out_optname, $specified_options) = @_;
	$/ = "\n"; # Use a newline character as an input record separator.

	my $fh_tmpl = open_input_file2 (${$in_tmpl_file});
	unless (defined $fh_tmpl) {
	    print STDOUT "[ERROR 'create_dawg_ctrl_file'] 'open_input_file2' FAILED on the input template file '${$in_tmpl_file}': $! !\n";
	    return 0;
	}

	my $fh_ctrl = open_output_file (${$out_ctrl_file}, 0);
	unless (defined $fh_ctrl) {
	    print STDOUT "[ERROR 'create_dawg_ctrl_file'] 'open_output_file' FAILED on the output control file '${$out_ctrl_file}': $! !\n";
	    return 0;
	}

	my %already; # = ($out_optname => 1, ...) for options already printed out. #

	my $target;

	while (my $line = <$fh_tmpl>) { # Main while-loop (over lines in the template file). #

	   if ($line =~ m/^\#\s+Target\:/) {
		$target = $line;
		next;

	    } elsif ($line =~ m/^\#/) { # Output comment lines as they are. #
		print $fh_ctrl $line;
		next;
	    }

	    $line =~ s/\s+$//;

	    my ($in_optname, $in_value) = split /\s*\=\s*/, $line;
	    unless (defined $in_value) {
		print STDOUT "[CAUTION 'create_dawg_ctrl'] Found an ANOMALOUS line: $! !\n";
		print STDOUT "'$line' .\n\n";
		next;
	    }

	    my $out_optname = $in2out_optname->{$in_optname};
	    unless (defined $out_optname) { # Comment out the option if it has no corresponding output option name.
		print $fh_ctrl '#', $line, "\n";
		next;
	    }

	    $already{$out_optname} = 1;

	    if ($out_optname eq 'Tree') {

		my $out_tree = assign_interior_node_names ($in_value, $PREF_INTNODENAME);
		unless (defined $out_tree) {
		    print STDOUT "[ERROR 'create_dawg_ctrl'] 'assign_interior_node_names' FAILED on the tree in the template '${$in_tmpl_file}': $! !\n";
		    return 0;
		}
		print $fh_ctrl join (' = ', $out_optname, ${$out_tree}), "\n";

	    } elsif ($out_optname eq 'GapParams') {

		my $out_value = "\{$in_value\}"; # Slightly reformat the value. #
		print $fh_ctrl join (' = ', $out_optname, $out_value ), "\n";

	    } else {
		my $out_value = $specified_options->{$out_optname};
		(defined $out_value) or ($out_value = $in_value);
		print $fh_ctrl join (' = ', $out_optname, $out_value), "\n";
	    }

	}		# End of the main while-loop (over lines in the template file). #
	close ($fh_tmpl);

	foreach my $optname (sort keys %{$specified_options}) { # Trailing foreach-loop (over specified options). #

	    (defined $already{$optname}) and next; # Skip if the option is already printed out. #

	    my $value = $specified_options->{$optname};

	    print $fh_ctrl join (' = ', $optname, $value), "\n";

	}						# End of the trailing foreach-loop (over specified options). #

	(defined $target) and (print $fh_ctrl $target);

	close ($fh_ctrl);

	return 1;

}	# End of the subroutine 'create_dawg_ctrl_file'. #


