#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# br_infer_psm_indel_events_spt_odr_dawg.hs.pl:
#	This modified version of 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'
#	attempts to calculate the occurrence probabilities of the homology structures of local PWAs,
#	which are ingredients of local MSAs.
#	[ For the description of the "homology structure," see Lunter et al. (2005) BMC Bioinformatics 6:83.]
#
#	It mainly uses the perl module, "MyTreeMap_indels_ML_hs.pm," instead of "MyTreeMap_indels_ML.pm."
#
#	[CREATION STARTED on May 16 (Sat), 2015, ENDED on May xx (yyy), 2015.]
#
# br_infer_psm_indel_events_spt_odr_dawg.v2.pl:
#	This modified version of 'br_infer_psm_indel_events_dawg.v2.pl' infers parsimonious indel scenarios
#	while respecting the spatial order of gap-blocks in each cluster,
#	using the new module 'MyTreeMap_indels_spt_odr.pm' instead of is predecessor, 'MyTreeMap_indels.pm'.
#	(CREATED on Mar 21 (Fri), 2014)
#
# br_infer_psm_indel_events_dawg.v2.pl:
#	This modified version of 'br_infer_psm_indel_events_dawg.pl' prints out the start and end coordiates of each gap-block cluster.
#	(CREATED on Mar 9 (Sun), 2014)
#
#	REVISED on Mar 10 (Mon), 2014:
#	The script now copes with the new revision of 'MyTreeMap_indels.pm', where a new subroutine 'scan_gbcls_for_long_gaps' was created.
#
# br_infer_psm_indel_events_dawg.pl:
#	This modified version of 'br_infer_psm_indel_events.pl' can handle a control file and an output MSA of Dawg.
#	(CREATION STARTED on Feb 27 (Thu), 2014, ENDED on Feb 28 (Fri), 2014) 
#
#	REVISED on Mar 2 (Sun), 2014:
#	The script now continues the analysis even if it finds some gap-block clusters containing potential indel events
#	longer than the cut-off length.
#
#	REVISED on Mar 6 (Thu), 2014:
#	The script now continues the analysis even if encounters situations where the satellite subroutine,
#	'cal_lnlk1_indel_scn_cls', of the subroutine 'apprx1_tot_lnlk_gpattern_msa' failed for some reasons
#	(most likely because it encounters 3 or more indel events that overlap each other).
#
#
# br_infer_psm_indel_events.pl:
#	This Perl script infers parsimonious indel events resulting in a given MSA under a simple phylogeny,
#	as well as their relative probabilities.
#	The script was created by using some essential parts of 'compare_ref_vs_rec_abs_scoring.PhyML.pl'.
#	(CREATION STARTED on Nov 6 (Wed), 2013, ENDED on Nov 22 (Fri), 2013)
#

use strict;
use warnings;

use MyArrayManipu_num qw( log_sum_exp );
use MyArrayManipu qw( binsearch_string );
use MyFile_manipu qw( open_input_file2 open_output_file list_sessions_ext );

use MyMalnIO qw(
	extract_fasta
	extract_alignment
	pr_ODEN_input
	pr_phylip_aln
	);
	## pr_clustal_like_lite_aa
use MyMalnManipu qw( rm_gapcolumn2 );

use MyTreeManipu_II qw(
	extract_sgltree_from_file
	input_tree_file output_tree
	br_output_tree
	);

use MyTreeManipu_II_dawg qw(
	extract_sgltree_from_file_dawg
	input_tree_file_dawg
	br_output_tree_dawg
		) ;
	## read_tree_file_dawg input_subtree_dawg
	## output_tree_dawg draw_subtree_dawg
 	## br_draw_subtree_dawg
	## br_output_tree_sep_dawg br_draw_subtree_sep_dawg

use MyTreeManipu_IV qw(
	bi2multifurcated_top
	construct_neat_subtree
	construct_skeleton_subtree
	);

=for the predecessor 'compare_ref_vs_rec_abs_scoring.PhyML.pl'
use MyPhyML_parser qw(
	est_resfreqs_frm_msa print_custom_sbst_model
	extract_phyml_statistics2
	fetch_clmwise_likelihoods_phyml
	);
	## extract_phyml_statistics
=cut


	### For enumerating parsimonious indel scenarios. ###

	# A MAJOR MODIFICATION in creation of 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'. #
use MyTreeMap_indels_spt_odr qw(
	enum_psm_indel_scns_for_msa
	);
=for the predecessor 'br_infer_psm_indel_events_dawg.v2.pl'
use MyTreeMap_indels qw(
	enum_psm_indel_scns_for_msa
	);
	## extract_cls_gblocks msa_rep_gpatterns
=cut


	### For log-likelihoods of indel scenarios. ###

		## ADDED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. ##
use MyReadAlnProbIngredients qw(
	readin_delta_del
	assign_br2log_mfacs
	);


		## MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. ##
use MyTreeMap_indels_ML_hs qw(
	apprx1_tot_lnlk_gpattern_msa2
	);
	## cal_delta_exp_del_5p2
	## cal_log_normal_facs_indel_distri
	## cal_lnlk1_indel_scn_cls2

=for the predecessor 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'.
use MyTreeMap_indels_ML qw(
	cal_log_normal_facs_indel_distri
	cal_delta_exp_del_5p
	cal_lnlk1_indel_scn_cls
	apprx1_tot_lnlk_gpattern_msa
	);
=cut


sub partition_msa_via_usable_patterns (\@\%);
sub inspect_indel_events (\@\@$$$);
sub read_dawg_ctrl_file (\$);

=for the predecessor 'compare_ref_vs_rec_abs_scoring.PhyML.pl'
sub readin_list_segs (\$\%);
sub readin_simul_params (\$);

sub readin_dawg_tree (\$); # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011 
sub ln_ml_sgl_seq2 (\$\%); # ADDED in creation of 'compare_apbayes_msa.bch.extJC.6.pl' on Mar 7 (Mon), 2011 #
sub add_dummy_seq (\@\@\%\%\%\%\%\@); # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 12 (Tue), 2011 #
=cut

$| = 1; # Flush the buffer of standard output everytime data flows into it.

my $program_path = $0;
my ($program_name) = ($program_path =~ m/([^\/]+)$/); 

	# INPORTED from 'ml_scoring_msa.bch.pl' on Mar 15 (Tue), 2011 #
(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;


	## File names storing important parameters. ##
	## (ADDED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'.) ##

our $FILENAME_DELTA_DEL = 'delta_del.txt'; # The name of the file storing '$delta_del', which is important for calculating the exit rate of the entire sequence. #

our @FILENAMES_LOG_MFACS = ( # Names of the files storing logarithmic multiplication factors, which are in $indir_mfacs. #

		# In the bulk #
	'lmfacs_asites_vs_dgaps.txt', # For the cases where there are some ancestral sites and no descendant sites in between preserved ancestral sites (PASs). #
	'lmfacs_agaps_vs_dsites.txt', # For the cases where there are no ancestral sites and some descendant sites in between PASs.
	'lmfacs_asites_nhom_dsites.txt', # For the cases where there are some ancestral sites and some descendant sites in between PASs.

		# On either boundary (i.e., either PWA end). "bd_$filename" corresponds to $filename above. #
	'bd_lmfacs_asites_vs_dgaps.txt',
	'bd_lmfacs_agaps_vs_dsites.txt',
	'bd_lmfacs_asites_nhom_dsites.txt'
	);



	### Fixed Parameters ###

	# (MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'.) #
	# In this analysis of an MSA simulated by Dawg,
	# null columns will play important roles.
$MyTreeMap_indels_spt_odr::FL_IGNORE_NULL_COLUMNS = 0; # Updated in creation of 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl' (on Mar 21 (Fri), 2014). #
##$MyTreeMap_indels::FL_IGNORE_NULL_COLUMNS = 0;

	# (ADDED increation of 'br_infer_psm_indel_events_dawg.v2.pl'.) #
	# Null-clusters (=clusters consisting only of a null-block each) will be analysed.
$MyTreeMap_indels_spt_odr::FL_HANDLE_NULL_CLS = 1; # Updated in creation of 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl' (on Mar 21 (Fri), 2014). #
##$MyTreeMap_indels::FL_HANDLE_NULL_CLS = 1;

	# (ADDED in revision on Mar 10 (Mon), 2014.)
	# This variable sets an upper-bound for the gap-length.
	# IF a gap-block cluster contains a run of gaps longer than this variable,
	# the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses on the cluster.
	# The variable will be used in the new subroutine, 'scan_gbcls_for_long_gaps'.
	#
our $UPPER_BOUND_ANAL_GAPLEN = 120;

$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN; # Updated in creation of 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl' (on Mar 21 (Fri), 2014). #
##$MyTreeMap_indels::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN;

	# MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. #
$MyTreeMap_indels_ML_hs::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN;
##$MyTreeMap_indels_ML::UPPER_BOUND_ANAL_GAPLEN = $UPPER_BOUND_ANAL_GAPLEN; # For the predecessor 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'. #

	# (ADDED in revision on Mar 11 (Tue), 2014.)
	# This variable sets an upper-bound for the length of the gap-block cluster itself.
	#
our $UPPER_BOUND_ANAL_LEN_GBCLS = 150;

$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS; # Updated in creation of 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl' (on Mar 21 (Fri), 2014). #
##$MyTreeMap_indels::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS;

	# MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. #
$MyTreeMap_indels_ML_hs::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS;
##$MyTreeMap_indels_ML::UPPER_BOUND_ANAL_LEN_GBCLS = $UPPER_BOUND_ANAL_LEN_GBCLS; # For the predecessor 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. #

=for the predecessor 'br_infer_psm_indel_events.pl'
	# We just ignore null columns encountered,
	# because we know that the null columns in this analysis
	# resulted in indel events outside of the tree of analyzed sequences.
	# (ADDED on Nov 16 (Sat), 2013)
	#
$MyTreeMap_indels::FL_IGNORE_NULL_COLUMNS = 1;
=cut

our $DIS_LEN = 60; # Maximum length of the sequences to be displayed in a line

##our $GAP = '-'; # Character representing a gap
our $CHAR_ABS =  our $GAP  = '-';
our $CHAR_PRES = our $NGAP = 'X';

	# Characters indicating that the site is missing or that its orthology is questionable. 
our %CHAR_MISSING = ('?' => 1, 'N' => 1, 'n' => 1);

	# ID of the target sequence. # ADDED on Nov 19 (Tue), 2013. #
our $TGT_SEQ = 'seq0001'; # This is just a formality. #
##our $TGT_SEQ = 'seq0000'; # This is just a formality. #
##our $TGT_SEQ = 'Scer';

	# (ADDED on Nov 29 (Mon), 2010) #
our $TOP_NODE  = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
our $INDX_INTNODENAME = $MyTreeManipu_II_dawg::INDX_INTNODENAME;

	# (ADDED on Mar 7 (Mon), 2011) #
our $TOT_BOOT = -1; # Because bootstrap values are NOT necessary for this analysis. 
our @INDICES_ATTR = ($INDX_LENGTH, $INDX_BOOT, $TOT_BOOT);
our $COMMENT = '#';  # Indicator of the comment line in the tree file

our $INDX_STRING = undef;
	# Specifies the index for the reference to a string in the element array of a set of non-redundant columns.
	# = undef if the element itself is the reference to a string.

	### Set Parameters for the relative probabilities of parsimonious Indel Scenarios. ###

our %INDEL_MODEL2DISTRI = ( # = (indel model => distribution type, ...)
	'zipf' => 'P',
	'"PL"' => 'P',
	);

=for the predecessor 'br_infer_psm_indel_events.pl'.
our $INDEL_MODEL = 'zipf';
our $DISTRI = $INDEL_MODEL2DISTRI{$INDEL_MODEL};

our $TOTRATE_INS = 0.05; # Total rate of insertions at a site relative to the substitution rate.
our $TOTRATE_DEL = 0.05; # Total rate of deletions at a site relative to the substitution rate.
our $PARAM_INS = 1.5; # Exponent for the power-law distribution of insertion length (when $INDEL_MODEL = 'zipf').
our $PARAM_DEL = 1.5; # Exponent for the power-law distribution of deletion length (when $INDEL_MODEL = 'zipf').
our $CUTOFF_3P = 10000; # Cut-off indel length at the 3'-end.
our $CUTOFF_5P = 10000; # Cut-off indel length at the 5'-end.

our @PARAMS_INS = ($TOTRATE_INS, $DISTRI, $PARAM_INS);
our @PARAMS_DEL = ($TOTRATE_DEL, $DISTRI, $PARAM_DEL);
=cut


	### Parameters to be adjusted ###

	# The level of tests to examine the behavior of the script (ADDED on Sep 1 (Sun), 2012):
	# 0, nothing; 1, very brief; 2, a bit brief; 3, a bit verbose; 5, very verbose.

	# For the main part. #
##our $TEST_LEVEL = 1;
our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;

	# For the subroutine 'inspect_indel_events'. #
our $TEST_LEVEL_INSP = 0;
##our $TEST_LEVEL_INSP = 1;
##our $TEST_LEVEL_INSP = 3;
##our $TEST_LEVEL_INSP = 5;

our $MIN_FREQ = 1E-06; # If the frequency of a residue is zero in the anchor MSA, this nominal frequency is used instead.

	# Essential for calculating the log-likelihoods of indel scenarios. #
	# $OPT_BD_INDELS = 0, fixed length; = 1, sum of rates; = 2, exact calculation.
#our $OPT_BD_INDELS = 0;
#our $OPT_BD_INDELS = 1;
our $OPT_BD_INDELS = 2;

		### ADJUST THE FOLLOWING GLOBAL VARIABLES ONLY WHEN DEBUGGING ###

	# Control the level of verboseness of the test message for 'MyTreeMap_indels.pm'.
	# (Basically, believe in the behaviours of the main subroutine 'enum_psm_indel_scns_for_msa',
	#  as well as of its satellite subroutines.)
	# [NOTE:  'MyTreeMap_indels' was silently REPLACED with 'MyTreeMap_indels_spt_odr'
	#   in creation of 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl' (on Mar 21 (Fri), 2014). # ]
#$MyTreeMap_indels_spt_odr::TEST_LEVEL = 3;
$MyTreeMap_indels_spt_odr::TEST_LEVEL = 1;

$MyTreeMap_indels_spt_odr::TEST_LEVEL1 = 0;
#$MyTreeMap_indels_spt_odr::TEST_LEVEL1 = 5;

#$MyTreeMap_indels_spt_odr::TEST_LEVEL1hf = 5; # For revising the subroutine, 'recede_absence', on Mar 10 (Mon) & 11 (Tue), 2014.
$MyTreeMap_indels_spt_odr::TEST_LEVEL1hf = 0;

$MyTreeMap_indels_spt_odr::TEST_LEVEL2 = 0;
$MyTreeMap_indels_spt_odr::TEST_LEVEL2i = 0; # For testing 'create_initial_set_indels_spt_odr'.
$MyTreeMap_indels_spt_odr::TEST_LEVEL2c = 0; # For testing 'create_bestset_mvd_events'.

#$MyTreeMap_indels_spt_odr::TEST_LEVEL2inb = 5; # For DEBUGGING the subroutine, 'incorporate_null_block', on Mar 21 (Fri), 2014. #
$MyTreeMap_indels_spt_odr::TEST_LEVEL2inb = 0;

#$MyTreeMap_indels_spt_odr::TEST_LEVEL3 = 2; # For revising the subroutine, 'incorporate_new_event', on Mar 10 (Mon) & 11 (Tue), 2014.
$MyTreeMap_indels_spt_odr::TEST_LEVEL3 = 0;

$MyTreeMap_indels_spt_odr::TEST_LEVEL3d = 0; # For testing 'recede_spec_deletion'.


$MyTreeMap_indels_spt_odr::TEST_LEVEL3i = 0;
#$MyTreeMap_indels_spt_odr::TEST_LEVEL3i = 5; # For testing the subroutine 'incorporate_new_event_spt_odr', to  DEBUG 'MyTreeMap_indels_spt_odr.pm' (on Mar 24 (Mon), 2014). #

	# ADDED in revision on Mar 10 (Mon), 2014. #
$MyTreeMap_indels_spt_odr::TEST_LEVEL_GEN_NULL_BLK = 0;
$MyTreeMap_indels_spt_odr::TEST_LEVEL_SCN = 0;
#$MyTreeMap_indels_spt_odr::TEST_LEVEL_GEN_NULL_BLK = 1;
#$MyTreeMap_indels_spt_odr::TEST_LEVEL_SCN = 1;
##$MyTreeMap_indels_spt_odr::TEST_LEVEL_GEN_NULL_BLK = 5;
##$MyTreeMap_indels_spt_odr::TEST_LEVEL_SCN = 5;

=for DEBUGGING
$MyTreeMap_indels_spt_odr::TEST_LEVEL = 5;
$MyTreeMap_indels_spt_odr::TEST_LEVEL1 = 5;
$MyTreeMap_indels_spt_odr::TEST_LEVEL1hf = 5;
$MyTreeMap_indels_spt_odr::TEST_LEVEL2 = 5;
$MyTreeMap_indels_spt_odr::TEST_LEVEL2i = 5;
=cut

	# [ Control the level of verboseness of the test message for 'MyTreeMap_indels_ML.pm'. ] -- for the predecessor 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'. #
	# Control the level of verboseness of the test message for 'MyTreeMap_indels_ML_hs.pm'.
	# (Basically, believe in the behaviours of the main subroutine 'apprx1_tot_lnlk_gpattern_msa'
	#  as well as of its satellite subroutines.)
$MyTreeMap_indels_ML_hs::TEST_LEVEL = 1;
$MyTreeMap_indels_ML_hs::TEST_LEVEL1 = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL1a = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL2a = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL2b = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL2c = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL2s = 0;
$MyTreeMap_indels_ML_hs::TEST_LEVEL3 = 0;

		### END of "ADJUST THE FOLLOWING GLOBAL VARIABLES ONLY WHEN DEBUGGING" ###

	### Take in Command-Line Arguments. ###
	# (MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'.) #

unless (@ARGV > 4) {
##unless (@ARGV > 3) {
	print STDERR "\nUsage:  [perl] $program_name {in_msa_file} {dawg_ctrl_file} {dir_mfac_data} {outfile_detail} {outfile_brief}\n";
	die "$!\n\n";
}
my ($in_msa_file, $dawg_ctrl_file, $indir_mfacs, $outfile_detail, $outfile_brief) = @ARGV[0 .. 4];
##my ($in_msa_file, $dawg_ctrl_file, $indir_mfacs, $outfile_detail, $outfile_brief) = @ARGV[0 .. 3]; # A BUG fixed on May 18 (Mon), 2015.
##my ($in_msa_file, $dawg_ctrl_file, $outfile_detail, $outfile_brief) = @ARGV[0 .. 3];
my $intree_file = $dawg_ctrl_file;

my $infile_delta_del = "$indir_mfacs/$FILENAME_DELTA_DEL"; # This file stores '$delta_del', which is important for calculating the exit rate of the entire sequence. #


	# Output files. #
	# (MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'.) #

my $outfile1 = $outfile_detail;
my $outfile2 = $outfile_brief;

(my $outdir1 = $outfile1) =~ s!/?[^/]+$!!;
(my $outdir2 = $outfile2) =~ s!/?[^/]+$!!;

($outdir1 =~ m/\S/) or ($outdir1 = '.'); 
($outdir2 =~ m/\S/) or ($outdir2 = '.'); 

=for the predecessor 'br_infer_psm_indel_events.pl'
my $out_name1 = 'output_detail.txt';
my $out_name2 = 'output_brief.txt';

my $outfile1 = "$outdir/$out_name1";
my $outfile2 = "$outdir/$out_name2";
=cut

	### Check input files/directories. ###

unless (( -d $indir_mfacs ) and ( -r $indir_mfacs )) { # ADDED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'.
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable DIRECTORY, '$indir_mfacs', that stores data files for the multiplication factors !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

foreach my $file ($in_msa_file, $intree_file, $infile_delta_del) {
##foreach my $file ($in_msa_file, $intree_file) {...}
    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

foreach my $filename (@FILENAMES_LOG_MFACS) { # ADDED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. #
    my $file = "$indir_mfacs/$filename";
    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}

=DISABLED in revision of 'br_infer_psm_indel_events.pl'
unless (( -d $intree_dir ) and ( -r $intree_dir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable DIRECTORY, '$intree_dir', that stores input tree files !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
=cut

foreach my $dir ($outdir1, $outdir2) { # Preliminary foreach-loop (over output directories). #
##foreach my $dir ($outdir) {...} # Preliminary foreach-loop (over output directories). # For the predecessor 'br_infer_psm_indel_events.pl'.
    unless ( -d $dir ) {
	unless (0 == system ("mkdir -p $dir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the directory '$dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
    unless ( -w $dir ) {
	unless (0 == system (" chmod -R u+w $dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
} 	# End of the preliminary foreach-loop (over output directories). #


	### Declare the Beginning of the Script ###

print STDOUT "\n<<<<<<< Program name: $program_name (path: $program_path) >>>>>>>>\n\n";

print STDOUT "Current working directory: $CURR_WORK_DIR .\n\n";

print STDOUT "Input MSA file: $in_msa_file .\n\n";

	## MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. ##
print STDOUT "Input tree and indel parameters: $intree_file .\n\n";
=for the predecessor 'br_infer_psm_indel_events.pl'
print STDOUT "Input tree file: $intree_file .\n\n";
##print STDOUT "Directory storing input tree files: $intree_dir .\n\n";
=cut

	## ADDED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. ##
print STDOUT "Input file recording ", '$delta_del', ": $infile_delta_del .\n\n";
print STDOUT "Input directory storing (log-)multiplication factors: $indir_mfacs ,\n\n";
print STDOUT "   which holds the following input files:\n\n";
print STDOUT '     ', join (', ', @FILENAMES_LOG_MFACS), " .\n\n\n";

print STDOUT "Output files:\n";
print STDOUT "   (Details)     $outfile1 ,\n";
print STDOUT "  (Brief record) $outfile2 .\n\n\n";
##print STDOUT "Output file: $outfile .\n\n\n";

print STDOUT "\n<< Options and Parameters  >>\n\n";

print STDOUT "Option for the calculation of log-likelihoods of indel scenarios:\n\n";
print STDOUT '   $OPT_BD_INDELS  = ', "'$OPT_BD_INDELS' .\n\n";

print STDOUT "Levels of tests for debugging or behavior tests:\n\n";

print STDOUT '  (for the main part)  $TEST_LEVEL        =  ', $TEST_LEVEL, " ,\n";
print STDOUT "  \(for 'inspect_indel_events'\)  ", '$TEST_LEVEL_INSP = ', $TEST_LEVEL_INSP, " .\n\n";

print STDOUT "Characters indicating that the site is missing or that its orthology is questionable:\n\n";

print STDOUT ' ' x 5, join (', ', sort keys %CHAR_MISSING), " .\n\n";

	## our %CHAR_MISSING = ('?' => 1, 'N' => 1, 'n' => 1);

print STDOUT '$MyTreeMap_indels_spt_odr::FL_IGNORE_NULL_COLUMNS =' , $MyTreeMap_indels_spt_odr::FL_IGNORE_NULL_COLUMNS, " ,\n"; # ADDED on Nov 16 (Sat), 2013. # => silently MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl' (on Mar 21 (Fri), 2014). #

	# ADDED in revision on Mar 10 (Mon), 2014. #
	# [NOTE: 'MyTreeMap_indels' was silently REPLACED with 'MyTreeMap_indels_spt_odr'
	# in creation of 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl' (on Mar 21 (Fri), 2014).] #

print STDOUT '$MyTreeMap_indels_spt_odr::FL_HANDLE_NULL_CLS = ', $MyTreeMap_indels_spt_odr::FL_HANDLE_NULL_CLS, " ,\n";
print STDOUT '$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN = $MyTreeMap_indels_ML::UPPER_BOUND_ANAL_GAPLEN = ', $MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_GAPLEN, " .\n\n";
print STDOUT '$MyTreeMap_indels_spt_odr::UPPER_BOUND_ANAL_LEN_GBCLS = $MyTreeMap_indels_ML::UPPER_BOUND_ANAL_LEN_GBCLS = ', $UPPER_BOUND_ANAL_LEN_GBCLS, " .\n\n"; # ADDED in revision on Mar 11 (Tue), 2014. #
print STDOUT '$MyTreeMap_indels_spt_odr::$TEST_LEVEL_GEN_NULL_BLK = ', $MyTreeMap_indels_spt_odr::TEST_LEVEL_GEN_NULL_BLK, " ,\n";
print STDOUT '$MyTreeMap_indels_spt_odr::$TEST_LEVEL_SCN  = ', $MyTreeMap_indels_spt_odr::TEST_LEVEL_SCN , " .\n\n";

print STDOUT "Name (ID) of the target sequence: ", '$TGT_SEQ = ', $TGT_SEQ, " .\n\n";

	### RETRIEVE the Parameters necessary for Indel Probabilities, as well as the guide tree. ###
	### (ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'.) ###

print STDOUT "\n<<<< As a preliminary process, read the Dawg control file '$intree_file' >>>\n\n\n";

my ($tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate)
	= read_dawg_ctrl_file ($intree_file);
	##= read_dawg_ctrl_file ($in_msa_file); # A BUG fixed on Feb 28 (Fri), 2014. #`
unless (defined $tree_string) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'read_dawg_ctrl-file' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

our $INDEL_MODEL = $gapmodel;
our $DISTRI = $INDEL_MODEL2DISTRI{$INDEL_MODEL};

	# Total rate of insertions at a site relative to the substitution rate.
our $TOTRATE_INS = $tot_indel_rate /2;
	# Total rate of deletions at a site relative to the substitution rate.
our $TOTRATE_DEL = $tot_indel_rate /2;

 	# Exponent for the power-law distribution of insertion length (when $INDEL_MODEL = 'zipf').
our $PARAM_INS = $gap_param;
	# Exponent for the power-law distribution of deletion length (when $INDEL_MODEL = 'zipf').
our $PARAM_DEL = $gap_param;


	# MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'.

	# Cut-off insertion length.
our $CUTOFF_INS_LEN = $gaplen_cutoff ;
	# Cut-off deletion length.
our $CUTOFF_DEL_LEN = $gaplen_cutoff ;

=for the predecessor 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'.
	# Cut-off indel length at the 3'-end.
our $CUTOFF_3P = $gaplen_cutoff;
	# Cut-off indel length at the 5'-end.
our $CUTOFF_5P = $gaplen_cutoff;

=cut

our @PARAMS_INS = ($TOTRATE_INS, $DISTRI, $PARAM_INS);
our @PARAMS_DEL = ($TOTRATE_DEL, $DISTRI, $PARAM_DEL);

	### Print out the Parameters for Indel Probabilities ###

print STDOUT "\n<<<< Input Parameters for Indel Probabilities >>>>\n\n";

print STDOUT '@PARAMS_INS = ($TOTRATE_INS, $DISTRI, $PARAM_INS) = (', join (', ', @PARAMS_INS), ") ,\n";
print STDOUT '@PARAMS_DEL = ($TOTRATE_DEL, $DISTRI, $PARAM_DEL) = (', join (', ', @PARAMS_DEL), ") .\n\n";


	# MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. #

print STDOUT '$CUTOFF_INS_LEN = ', "$CUTOFF_INS_LEN ,\n";
print STDOUT '$CUTOFF_DEL_LEN = ', "$CUTOFF_DEL_LEN .\n\n\n";

=for the predecessor 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'.
	# ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
print STDOUT '$CUTOFF_3P = ', $CUTOFF_3P, " ,\n";
print STDOUT '$CUTOFF_5P = ', $CUTOFF_5P, " .\n\n\n";
=cut

	## MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. ##

print STDOUT "\n<<< Now, as another preliminary process, read in the multiplication factor '\$delta_del' for a 'universal' term of log-lilekihood ... >>>\n\n";

my $delta_del = readin_delta_del ($infile_delta_del);
unless (defined $delta_del) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'readin_delta_del' FAILED !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die;
}


=for the predecessor 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'.

print STDOUT "\n<<< Now, as another preliminary process, calculate the normalization factors for the indel distributions >>>\n\n";

my $log_nfacs = cal_log_normal_facs_indel_distri (@PARAMS_INS, @PARAMS_DEL, $CUTOFF_5P, $CUTOFF_3P);
unless (defined $log_nfacs) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'cal_log_normal_facs_indel_distri' FAILED !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die;
}

our @LABELS_NFACS = ("Insertions(5p)", "Deletions(5p)", "Insertions(3p)", "Deletions(3p)");

print STDOUT "\n<< Outputs of 'cal_log_normal_facs_indel_distri' >>\n\n";
print STDOUT join ("\t", 'Category', 'log(norm_fac)', 'norm_fac'), "\n";

for (my $i=0; $i<4; $i++) { # Preliminary for-loop #
    my $label = $LABELS_NFACS[$i];

    my $log_nfac;
    my $nfac = (defined $log_nfac) ? exp ($log_nfac) : 'undef';
    (defined $log_nfac) or ($log_nfac = 'undef');
    print STDOUT join ("\t", $label, $log_nfac, $nfac), "\n";
}		# End of the preliminary for-loop #

print STDOUT "\n\n";


my $log_nfac_del_5p = $log_nfacs->[1];
print STDOUT "\n<<< Now, calculate the multiplicative factor '\$delta_del_5p' for a common term of log-likelihood (with \$log_nfac_del_5p = $log_nfac_del_5p). >>>\n\n";

my $delta_del_5p = cal_delta_exp_del_5p (@PARAMS_DEL, $CUTOFF_5P, $log_nfac_del_5p); 
unless (defined $delta_del_5p) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'cal_delta_exp_del_5p' FAILED !\n\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
print STDOUT "\n<< Output of 'cal_delta_exp_del_5p' >>\n\n";
print STDOUT '$delta_del_5p = ', $delta_del_5p, " .\n\n";

=cut

	### Read in the input tree. ###

print STDOUT "\n<<<<< Now, read in the Input Tree >>>>>\n\n";

my %node2pa; # = ($node_id => $parent_node_id, ...) for all nodes except the top node. 
my %node2ch; # = ($node_id => \@child_node_ids, ...) for all interior nodes.
my %name2id; # = ($sequence_name => $node_id, ...) for all exterior nodes and the top-node.
my %id2name; # = ($node_id => $sequence_name, ...) for all exterior nodes.
# [my %br_attr; # = ($node_id => [$length, $bootstrap, ...], ...) for all nodes at the lowe-ends of the branches. ] --- for the predecessor 'br_infer_psm_indel_events.pl'. #
my %br_attr; # = ($node_id => [$length, $int_node_name, ...], ...) for all nodes at the lowe-ends of the branches.

	# ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
my %name2id_intnode; # = ($node_name => $node_id, ...) for all interior nodes. #
my $topnode_name; # The name of the top-node (root).

	# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #

unless ( input_tree_file_dawg (${$tree_string}, %node2pa, %node2ch, %name2id, %id2name, %br_attr, %name2id_intnode, $topnode_name) ) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input tree from the control file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

=for the predecessor 'br_infer_psm_indel_events.pl'.
unless ( extract_sgltree_from_file ($intree_file, %node2pa, %node2ch, %name2id, %id2name, %br_attr, $COMMENT) ) {

	print STDOUT "[ERROR '$program_name'] FAILED to extract the input tree from the file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
print STDOUT " --- 'extract_sgltree_from_file' performed on the file '$intree_file' ...\n\n";
=cut

my $topnode_id = $name2id{$TOP_NODE};

print STDOUT "[ Input Tree (with interior node names shown) ]:\n\n"; # MODIFIED on Mar 11 (Tue), 2014.
##print STDOUT "[ Input Tree ]:\n\n";

	# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #

unless ( br_output_tree_dawg ($topnode_id, %node2ch, %id2name, %br_attr, $topnode_name, *STDOUT, 0, 0) ) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'br_output_tree_dawg' FAILED to output the input tree extracted from the control file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
##=for the predecessor 'br_infer_psm_indel_events.pl'.

	# ADDED for DEBUGGING the module 'MyTreeMap_indel.pm' (on Mar 11 (Tue), 2014). #
print "\n\n[ Input Tree (with node IDs shown) ]:\n\n";
my %br2blength;
foreach my $br (keys %br_attr) {
	my @attr = ();
	$attr[$INDX_LENGTH] = $br_attr{$br}->[$INDX_LENGTH];
	$br2blength{$br} = \@attr;
}

unless ( br_output_tree ($topnode_id, %node2ch, %id2name, %br2blength, *STDOUT, 1, 1) ) {
##unless ( output_tree ($topnode_id, %node2ch, %id2name, %br_attr, *STDOUT, 1, 1) ) {...} # --- For the predecessor 'br_infer_psm_indel_events.pl'.
	print STDOUT "[ERROR '$program_name'] Subroutine 'br_output_tree' FAILED to output the input tree extracted from the file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
##=cut
print STDOUT "\n\n";

	### Read in the logarithmic multiplication factors. ###
	### (ADDED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'.) ###

print STDOUT "\n<<<<< Now, assign the logarithmic multiplication factors to each branche, based on the files in '$indir_mfacs' ... >>>>>\n\n";

my $br2log_mfacs = assign_br2log_mfacs (%br_attr, $TOTRATE_INS + $TOTRATE_DEL, $indir_mfacs, @FILENAMES_LOG_MFACS);
unless (defined $br2log_mfacs) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'assign_br2log_mfacs' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

## @FILENAMES_LOG_MFACS


	### Read in the input MSA. ###

print STDOUT "\n<<<<< Now, read in the Input MSA >>>>>\n\n";

my ($in_msa, $seqids, $seq_lens) = extract_alignment ($in_msa_file);
unless (defined $in_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input MSA from the file '$in_msa_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_sequences = @{$seqids};

	# REMOVE the Ancestral Sequences from the Input MSA. #
	# (ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'.) #

my @msa_extseqs = my @ext_seqids = ();
for (my $i=0; $i<$ct_sequences; $i++) { # Preliminary for-loop (over sequences in the input MSA). #

	my $seqid = $seqids->[$i];
	($seqid  =~ m/^anc/) and next;

	push @ext_seqids, $seqid;
	push @msa_extseqs, $in_msa->[$i];

}			# End of the preliminary for-loop (over sequences in the input MSA). #

@{$in_msa} = @msa_extseqs;
@{$seqids} = @ext_seqids;
$ct_sequences = @{$seqids};

unless ($ct_sequences > 1) {
	print STDOUT "[ERROR '$program_name'] The number of sequences ($ct_sequences) is TOO SMALL ! I will stop here ...\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $totlen_aln = length ${$in_msa->[0]};

	## ADDED on Nov 19 (Tue), 2013. ##
my $indx_tgt;
for (my $k=0; $k<$ct_sequences; $k++) { ($seqids->[$k] eq $TGT_SEQ) and ($indx_tgt = $k); }
unless (defined $indx_tgt) {
	print STDOUT "[ERROR '$program_name'] Couldn't find the target sequence name '$TGT_SEQ' in the list of sequence names !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


print STDOUT "#{sequences in MSA} =  $ct_sequences .\n";
print STDOUT "#{columns in MSA}   = $totlen_aln .\n";

print STDOUT "{Index for the target sequence '$TGT_SEQ'} =  $indx_tgt .\n\n"; # ADDED on Nov 19 (Tue), 2013. #

if ($TEST_LEVEL>2) {
    print STDOUT "\n[ Input MSA ]:\n\n";

    unless ( pr_ODEN_input (@{$seqids}, @{$in_msa}, 50, *STDOUT) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    print STDOUT "\n\n";
}


	### Partition the input MSA into segments, each with a particular pattern of 'usable' sequences. ###

print STDOUT "\n<<<<< Now, partition the Input MSA via usable site patterns >>>>>\n\n";

# @{$in_msa_segments->[$s]} = ($seg_start, $seg_end, \@seg_pattern, \@seg_msa) is for the $s th segment, where
#	$seg_start/end is the start/end coordinate of the segment (in the 0-based half-open convention);
#	$seg_pattern[$k] = 1 if the site in the $k th sequence is usable for the analysis,
#			 = 0 if the site is missing or its orthology is questionable (i.e., $CHAR_MISSING{$char} = 1);
#	${$seg_msa[$k]} is the $k th sequence embedded in the segment MSA.
#
my $in_msa_segments = partition_msa_via_usable_patterns (@{$in_msa}, %CHAR_MISSING);
unless (defined $in_msa_segments) {
	print STDOUT "[ERROR '$program_name'] 'partition_msa_via_usable_patterns' FAILED on the input MSA !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
print STDOUT " --- 'partition_msa_via_usable_patterns' performed on the input MSA ...\n\n";

my $ct_segments = @{$in_msa_segments};

print STDOUT "#{Segments of particular missing site patterns} = $ct_segments .\n\n";

if ($TEST_LEVEL>3) {
    print STDOUT "[ MSA Segments ]:\n\n";

    for (my $s=0; $s<$ct_segments; $s++) { # Preliminary for-loop (over MSA segments) #

	my ($seg_start, $seg_end, $seg_pattern, $seg_msa) = @{$in_msa_segments->[$s]};

	my $cnct_seg_pattern = join ('', @{$seg_pattern});
	print STDOUT "  { The $s th segment\; Segion = \[$seg_start\, $seg_end\) }\n\n";
	print STDOUT "  Pattern (1:usable/0:missing) = '$cnct_seg_pattern' .\n\n";
	print STDOUT "  Alignment:\n\n";

	unless ( pr_ODEN_input (@{$seqids}, @{$seg_msa}, 50, *STDOUT) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the $s th segment !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	}
	print STDOUT "\n\n";

    }				# End of the preliminary for-loop (over MSA segments) #
    print STDOUT "\n";
}

	### Open the Output File and Record the Input Parameters, etc. ### 
	### (MOVED from below, and ADDED on Nov 18 (Mon), 2013) ###

my $fh_out = open_output_file ($outfile1, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile1' !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $fh_out2 = open_output_file ($outfile2, 0);
unless (defined $fh_out2) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile2' !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

foreach my $fh ($fh_out, $fh_out2) { print $fh "<<<< Input Parameters >>>>\n\n"; }


our @LABELS_NFACS = ("Insertions(5p)", "Deletions(5p)", "Insertions(3p)", "Deletions(3p)"); # Just a formality, which is a legacy of the predecessor 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'.

print $fh_out "\n[[ Normalization factors for the indel probability calculation ]]\n\n";
print $fh_out join ("\t", 'Category', 'log(norm_fac)', 'norm_fac'), "\n";

for (my $i=0; $i<4; $i++) { # Preliminary for-loop #
    my $label = $LABELS_NFACS[$i];

	# MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. #
    my $log_nfac;
    ##my $log_nfac = $log_nfacs->[$i];

    my $nfac = (defined $log_nfac) ? exp ($log_nfac) : 'undef';
    (defined $log_nfac) or ($log_nfac = 'undef');
    print $fh_out join ("\t", $label, $log_nfac, $nfac), "\n";
}		# End of the preliminary for-loop #

print $fh_out "\n\n";

print $fh_out "\n[[ Exponent for the Universal Multiplication Factor ]]\n\n";

	# MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. #
print $fh_out '$delta_del = ', $delta_del, " .\n\n\n";
##print $fh_out '$delta_del_5p = ', $delta_del_5p, " .\n\n\n";

foreach my $fh ($fh_out, $fh_out2) {

    print $fh "[[ Input Tree (top_node_name = $topnode_name) ]]\n\n";

    unless ( br_output_tree_dawg ($topnode_id, %node2ch, %id2name, %br_attr, $topnode_name, *${$fh}, 0, 0) ) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'br_output_tree_dawg' FAILED to output the input tree extracted from the control file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
=for the predecessor 'br_infer_psm_indel_events.pl'
    print $fh "[[ Input Tree (top_node_ID = $topnode_id) ]]\n\n";
    unless ( output_tree ($topnode_id, %node2ch, %id2name, %br_attr, *${$fh}, 1, 1) ) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'output_tree' FAILED to output the input tree extracted from the file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
=cut
    print $fh "\n\n";

    print $fh "[[ Input MSAs (#{sequences} = $ct_sequences) ]]\n\n";

    print $fh 'Sequences: ', join (', ', @{$seqids}), " .\n\n";

    print $fh "#{columns in MSA} = $totlen_aln .\n\n";

    print $fh "{Index for the target sequence '$TGT_SEQ'} =  $indx_tgt .\n\n"; # ADDED on Nov 19 (Tue), 2013. #
}

print $fh_out2 "#{MSA Segments} = $ct_segments .\n\n\n";

print $fh_out "[[ MSA Segments with particular sequence availability (#{segments} = $ct_segments) ]]\n\n";

print $fh_out join ("\t", 'Seg_No.', 'Start', 'End', 'Length', 'pattern(1,available;0,missing)'), "\n\n";

for (my $s=0; $s<$ct_segments; $s++) { # Preliminary for-loop (over MSA segments) #
	my ($seg_start, $seg_end, $seg_pattern, $seg_msa) = @{$in_msa_segments->[$s]};
	my $seg_len = $seg_end - $seg_start;
	my $cnct_seg_pattern = join ('', @{$seg_pattern});
	print $fh_out join ("\t", $s, $seg_start, $seg_end, $seg_len, $cnct_seg_pattern), "\n";
}				# End of the preliminary for-loop (over MSA segments) #
print $fh_out "\n\n";


	### MAIN ANALYSES:
	## Estimate and Compare the Relative Probabilities
	## of the Candidate (parsimonious) Indel Scenarios for the MSA segments. ###

print STDOUT "\n<<<<<<< The MAIN ANALYSES >>>>>>>\n\n";

my $ct_anom_segs = my $ct_in_segs = my $ct_succ_segs = my $ct_fail_segs = my $ct_skipped_segs = 0;

## my $ref_coord_beg = my $rec_coord_beg = 0; # The 'beginning' coordinates of the segment in the reference MSA and reconstructed MSA, respectively. #

for (my $s = 0; $s<$ct_segments; $s++) { # Main outer for-loop (over MSA segments) #

	# @{$in_msa_segments->[$s]} = ($seg_start, $seg_end, \@seg_pattern, \@seg_msa) is for the $s th segment, where
	#	$seg_start/end is the start/end coordinate of the segment (in the 0-based half-open convention);
	#	$seg_pattern[$k] = 1 if the site in the $k th sequence is usable for the analysis,
	#			 = 0 if the site is missing or its orthology is questionable (i.e., $CHAR_MISSING{$char} = 1);
	#	${$seg_msa[$k]} is the $k th sequence embedded in the segment MSA.

	my ($seg_start, $seg_end, $seg_pattern, $seg_msa) = @{$in_msa_segments->[$s]};

	my $seg_len = $seg_end - $seg_start;
	my $cnct_seg_pattern = join ('', @{$seg_pattern});

	print STDOUT "<<<<<< The $s th Segment \; Region = \[$seg_start, $seg_end\) \; Length = $seg_len\; Usability pattern = '$cnct_seg_pattern'  >>>>>>\n\n";

	print $fh_out "<<<< The $s th Segment \; Region = \[$seg_start, $seg_end\) \; Length = $seg_len\; Usability pattern = '$cnct_seg_pattern'  >>>>\n\n";
	print $fh_out2 "<<<< The $s th Segment \; Region = \[$seg_start, $seg_end\) \; Length = $seg_len\; Usability pattern = '$cnct_seg_pattern'  >>>>\n\n";

    ##if ($TEST_LEVEL>1) { # MODIFIED on Nov 22 (Fri), 2013. # ...}
    if ($TEST_LEVEL>2) { # RE-Switched on Feb 28 (Fri), 2014. #
	print STDOUT "Original Alignment for the Segment:\n\n";
	unless ( pr_ODEN_input (@{$seqids}, @{$seg_msa}, 50, *STDOUT) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reference MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n\n";
    }

		## (i) Pre-processing: Create a reduced MSA & a reduced tree, using just 'usable' sequences.

	print STDOUT "<<<< Step (i) Pre-processing: Creating a reduced segment MSA and a reduced tree >>>>\n\n";

			# Prepare a 'reduced' alignment by keeping only 'usable' sequences. #

	my @red_seqnames = (); # element = the name of a 'usable' sequence
	my @red_smsa = (); # element = \${a 'usable' sequence embedded in the segmental MSA}

	for (my $k=0; $k<$ct_sequences; $k++) { # Preliminary middle for-loop #

	    ($seg_pattern->[$k] eq 1) or next; # Skip questionable sequences.

	    push @red_seqnames, $seqids->[$k];
	    push @red_smsa, $seg_msa->[$k];

	}			# End of the preliminary middle for-loop #

	my $sbct_seqs = @red_seqnames;

		## MODIFIED on Nov 19 (Tue), 2013. ##
    ##if ($TEST_LEVEL>2) {
	print $fh_out "[[ Reduced MSA Segment (#{involved sequences} = $sbct_seqs, #{columns} = $seg_len) ]]\n\n";
	##print STDOUT "<< Reduced MSA Segment (#{involved sequences} = $sbct_seqs, #{columns} = $seg_len) >>\n\n";

	print $fh_out "... MSA was OMITTED to save the disk space ...\n\n\n"; # TEMPORARILY ADDED in order to reduce the output size. (on Feb 28 (Fri), 2014). #

=TEMPORARILY DISABLED in order to reduce the output size. (on Feb 28 (Fri), 2014). #
	if ($sbct_seqs>0) { ## ADDED (on Nov 22 (Fri), 2013) to work around the problem that 'pr_ODEN_input' cannot output an MSA with no sequence. #

	   unless ( pr_ODEN_input (@red_seqnames, @red_smsa, 50, *${$fh_out}) ) {
	   ##unless ( pr_ODEN_input (@red_seqnames, @red_smsa, 50, *STDOUT) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the reduced MSA for the $s th segment !\n";
	 	print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	}
	print $fh_out "\n\n";
     ##}
=cut

		## Skip if the reduced alignment is inadequate. ## (ADDED on Nov 19 (Tue), 2013.) ##

	my $avail_tgt = $seg_pattern->[$indx_tgt];

	unless (($avail_tgt eq '1')  and ($sbct_seqs>1)) {

	    print STDOUT "\n ... Because of INADEQUATE INFORMATION to infer ancestral states (target availability = $avail_tgt; #{available sequences} = $sbct_seqs), I will skip the segment ...\n\n\n";
	    print $fh_out "\n[[ INADEQUATE INFORMATION to infer ancestral states (target availability = $avail_tgt; #{available sequences} = $sbct_seqs) ]]\n\n\n";

	    print $fh_out2 "SKIPPED INFERRING Ancestral Indel States, because of the inadequate reduced MSA (target availability = $avail_tgt, #{available sequences} = $sbct_seqs).\n\n\n";

	    $ct_skipped_segs++;
	    next;
		## our $TGT_SEQ = 'Scer';
		## my $indx_tgt;
	}


			# Prepare a 'reduced' phylogenetic tree by excluding questionable sequences ##

	my %sb_name2id;
	my %sb_id2name;
	my ($sb_topnode_id, $sb_node2ch, $sb_node2pa, $sb_br_attr);

		# ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
	my %sb_name2id_intnode;
	my $sb_topnode_name;

	##my %node2pa; # = ($node_id => $parent_node_id, ...) for all nodes except the top node. 
	##my %node2ch; # = ($node_id => \@child_node_ids, ...) for all interior nodes.
	##my %name2id; # = ($sequence_name => $node_id, ...) for all exterior nodes and the top-node.
	##my %id2name; # = ($node_id => $sequence_name, ...) for all exterior nodes.
	##my %br_attr; # = ($node_id => [$length, $bootstrap, ...], ...) for all nodes at the lowe-ends of the branches.

	if ($sbct_seqs == $ct_sequences) {
		$sb_topnode_id = $topnode_id;
		$sb_node2ch = \%node2ch;
		$sb_node2pa = \%node2pa;
		$sb_br_attr = \%br_attr;
		%sb_name2id = %name2id;
		%sb_id2name = %id2name;

		# ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		%sb_name2id_intnode = %name2id_intnode;
		$sb_topnode_name = $topnode_name;

	} else {

	  my @sub_nodes = ();
	  foreach my $seqname (@red_seqnames) { # Preliminary inner foreach-loop (over non-gapped sequences) #

	    my $nodeid = $name2id{$seqname};
	    unless (defined $nodeid) {
		print STDOUT "[ERROR '$program_name'] For the $s th segment, the sequence name '$seqname' has NO node ID in the input tree !\n\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
	    push @sub_nodes, $nodeid;

	    $sb_name2id{$seqname} = $nodeid;
	    $sb_id2name{$nodeid} = $seqname;

	  }			# End of the preliminary inner foreach-loop (over non-gapped sequences) #

	  (my $sb_topnode_ids, $sb_node2ch, $sb_node2pa, $sb_br_attr)
		= construct_skeleton_subtree (%node2pa, @sub_nodes, %br_attr);
		##= construct_neat_subtree (%{$node2pa}, @sub_nodes, %{$br_attr}); # For the predecessor 'compare_ref_vs_rec_abs_scoring.PhyML.pl'. #

	  unless (defined $sb_topnode_ids) {
	    print STDOUT "[ERROR '$program_name'] FAILED to reduce the input tree for the $s th segment !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	  }
	  unless (1 == @{$sb_topnode_ids}) {
	    print STDOUT "[ERROR '$program_name'] Found MULTIPLE TOP NODES for the reduced tree of the input tree for the $s th segment: ", join (', ', @{$sb_topnode_ids}), " !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	  }
	  $sb_topnode_id = $sb_topnode_ids->[0];
	  $sb_name2id{$TOP_NODE} = $sb_topnode_id;

		# ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. #

	  $sb_topnode_name = ($sb_topnode_id == $topnode_id) ? $topnode_name : $br_attr{$sb_topnode_id}->[$INDX_INTNODENAME];
	  (defined $sb_topnode_name) or ($sb_topnode_name = $id2name{$sb_topnode_id});
	  unless (defined $sb_topnode_name) {
		print STDOUT "[ERROR '$program_name'] FAILED to assign the name to the topnode (ID: $sb_topnode_id) of the reduced tree !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	  }

	  foreach my $name (keys %name2id_intnode) { # 2nd preliminary inner for-loop (over interior nodes). #
		my $nodeid = $name2id_intnode{$name};
		my $ch = $sb_node2ch->{$nodeid};
		((defined $ch) and (@{$ch}>0)) and ($sb_name2id_intnode{$name} = $nodeid);
	  }				# End of the 2nd preliminary inner for-loop (over interior nodes). #

	}
		# Handle Exceptional Case -- Part I + 1/2 -- #


		# Normal output #

	foreach my $fh ($fh_out, $fh_out2) {

		# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #

	    print $fh "\n[[ Reduced Tree (top_node_name = $sb_topnode_name) ]]\n\n";

	    my $fl_bootlabel = 0; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	    my $fl_nodeid = 0; # Flag to determine whether node IDs are displayed (=1) or not (=0).
	    unless ( br_output_tree_dawg ($sb_topnode_id, %{$sb_node2ch}, %sb_id2name, %{$sb_br_attr}, $sb_topnode_name, *${$fh}, $fl_bootlabel, $fl_nodeid) ) {
		print STDOUT "[ERROR '$program_name'] 'br_output_tree_dawg' FAILED to output the reduced tree of the input tree for the $s th segment: $! !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }

=for the predecessor 'br_infer_psm_indel_events.pl'.
	    print $fh "\n[[ Reduced Tree (top_node_ID = $sb_topnode_id) ]]\n\n";
	    my $fl_bootlabel = 1; # Flag to determine whether bootstrap values are displayed at the branch (=1) or node (=0).
	    my $fl_nodeid = 1; # Flag to determine whether node IDs are displayed (=1) or not (=0).
	    unless ( output_tree ($sb_topnode_id, %{$sb_node2ch}, %sb_id2name, %{$sb_br_attr}, *${$fh}, $fl_bootlabel, $fl_nodeid) ) {
		print STDOUT "[ERROR '$program_name'] 'output_tree' FAILED to output the reduced tree of the input tree for the $s th segment: $! !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	    }
=cut
	    print $fh "\n\n";

	}


		## (ii) Main Process:
		## Calculate the relative probabilities of the parsimonious indel events 
		## in each gap-block cluster in the segment,
		##  under the input tree.

	print STDOUT "<<<< Step (ii) Main Process A: Calculating the relative probabilities of the Parsimonious INDEL Events in the segment >>>>\n\n";


		# Copy the reduced MSA #
	my @cp_red_smsa = ();
	for (my $k=0; $k< $sbct_seqs; $k++) { my $cp_seg = ${$red_smsa[$k]}; push @cp_red_smsa, \$cp_seg; }

	##my %sb_name2id;
	##my %sb_id2name;
	##my ($sb_topnode_id, $sb_node2ch, $sb_node2pa, $sb_br_attr);

	##my @red_seqnames = (); # element = the name of a 'usable' sequence
	##my @red_smsa = (); # element = \${a 'usable' sequence embedded in the segmental MSA}

	print STDOUT "\n\n<< (i) Enumerate the parsimonious indel scenarios, each of which realizes the reduced segment MSA >>\n\n";

	my ($set_cls_gblocks, $min_totcts_indel_for_cls, $sets_psm_scns_for_cls, $set_gpatterns, $set_node2dparsstat, $set_br2change)
	   = enum_psm_indel_scns_for_msa (%{$sb_node2ch}, %{$sb_node2pa}, %sb_name2id, @red_seqnames, @cp_red_smsa, $CHAR_ABS, $CHAR_PRES);
	unless (defined $set_cls_gblocks) {
	    print STDOUT "[ERROR '$program_name'] The preliminary subroutine 'enum_psm_indel_scns_for_msa' FAILED !\n";
	    print STDERR "[ERROR '$program_name'] $! !\n\n";
	    die ;
	}
    (print STDOUT "\n--- The preliminary subroutine 'enum_psm_indel_scns_for_msa' ENDED ...\n\n") if ($TEST_LEVEL>1);

	print $fh_out "\n[[ Secondary Inputs (i): Outputs of 'enum_psm_indel_scns_for_msa' ]]\n\n";

	my $ct_gp = @{$set_gpatterns};

    if ($TEST_LEVEL>2) {

	print STDOUT "\n[[ Secondary Inputs (i): Outputs of 'enum_psm_indel_scns_for_msa' ]]\n\n";

	print STDOUT "[1] List of gap patterns encountered (#{patterns} = $ct_gp):\n\n";
	print STDOUT join ("\t", 'Index', "'gap_patterns(horizontal_representation)'"), "\n\n";
	for (my $indx_gp1=0; $indx_gp1<$ct_gp; $indx_gp1++) { # 1st interim outer for-loop (over the gap patterns) #

	    my $gpattern = ${$set_gpatterns->[$indx_gp1]};
	    print STDOUT join ("\t", $indx_gp1, "'$gpattern'"), "\n";

	}	# End of the 1st interim outer for-loop (over the gap patterns) #
	print STDOUT "\n\n";
    } ##  END of "if ($TEST_LEVEL>2)" 

    if ($TEST_LEVEL>3) {

	print STDOUT "[2] Dollo parsimonious scenarios realizing the gap patterns:\n\n";
	for (my $indx_gp1=0; $indx_gp1<$ct_gp; $indx_gp1++) { # 2nd interim outer for-loop (over the gap patterns) #
    
	    my $gpattern = ${$set_gpatterns->[$indx_gp1]};
	    my $node2dparsstat = $set_node2dparsstat->[$indx_gp1];
	    my $br2change = $set_br2change->[$indx_gp1];

	    print STDOUT "\n{ The $indx_gp1 th gap pattern }\n\n";

	    print STDOUT "Pattern:  '$gpattern'\n\n";
	    print STDOUT "( Map of Dollo parsimonious states onto the nodes )\n\n";

		# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #

	    print STDOUT join ("\t", 'Node_ID', 'Node_Name', 'state', '#{child nodes with #{exterior descendants with \'absence\' state}>0}'), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2dparsstat}) {
		my $node_name = ($node == $sb_topnode_id) ? $sb_topnode_name : ((defined $sb_node2ch->{$node})? $sb_br_attr->{$node}[$INDX_INTNODENAME] : $sb_id2name{$node}) ;
		print STDOUT join ("\t", $node, $node_name, @{$node2dparsstat->{$node}}), "\n";
	    }
=for the predecessor 'br_infer_psm_indel_events.pl'
	    print STDOUT join ("\t", 'Node_ID', 'state', '#{child nodes with #{exterior descendants with \'absence\' state}>0}'), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2dparsstat}) { print STDOUT join ("\t", $node, @{$node2dparsstat->{$node}}), "\n"; }
=cut

	    print STDOUT "\n\n";
	    print STDOUT "( Map of Dollo parsimonious indel events onto the branches )\n\n";

	    print STDOUT join ("\t", 'Branch_ID', 'Lower_End_Name', 'state(upper-end)', 'state(lower-end)'), "\n\n";
	    foreach my $br (sort {$a <=> $b} keys %{$br2change}) {
		my $le_name = (defined $sb_node2ch->{$br}) ? $sb_br_attr->{$br}[$INDX_INTNODENAME] : $sb_id2name{$br} ;
		print STDOUT join ("\t", $br, $le_name,  @{$br2change->{$br}}), "\n";
	    }
=for the predecessor 'br_infer_psm_indel_events.pl'
	    print STDOUT join ("\t", 'Branch_ID', 'state(upper-end)', 'state(lower-end)'), "\n\n";
	    foreach my $br (sort {$a <=> $b} keys %{$br2change}) { print STDOUT join ("\t", $br, @{$br2change->{$br}}), "\n"; }
=cut
	    print STDOUT "\n\n";
    
	}			# End of the 2nd interim outer for-loop (over the gap patterns) #

    } ##  END of "if ($TEST_LEVEL>3)"


	my $ct_cls = @{$set_cls_gblocks};
	my $min_totct_indels = 0;

	print $fh_out "\n[3] Clusters of gap-pattern-blocks (#{clusters} = $ct_cls) and Genuinely Parsimonious Scenarios for each of the clusters:\n\n\n";

	for (my $cl = 0; $cl<$ct_cls; $cl++) { # 3rd interim outer for-loop (over the clusters) #

	    my $cls_gblks = $set_cls_gblocks->[$cl];
	    my $ct_gblks = @{$cls_gblks};

		# COPIED from below, around the output to $fh_out2, and MODIFIED in creation of 'br_infer_psm_indel_events_dawg.v2.pl' on Mar 9 (Sun), 2014. #
	    my $cl_start = $cls_gblks->[0][2] + $seg_start;
	    my $cl_end = $cls_gblks->[$ct_gblks-1][3] + $seg_start + 1; # Convert from full-closed to half-open. #

		# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.v2.pl' on Mar 9 (Sun), 2014. #
	    print $fh_out "< The $cl th cluster \; region = \[$cl_start\, $cl_end\) \; #{gap-blocks} = $ct_gblks >\n\n";
	    ##print $fh_out "< The $cl th cluster (#{gap-blocks} = $ct_gblks) >\n\n";

	    print $fh_out "[ Gap-blocks in the cluster ]\n\n";
	    print $fh_out join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal_representation)'"), "\n\n";

	    for (my $bl=0; $bl<$ct_gblks; $bl++) { # 1st middle for-loop (over the gap-blocks) #

	    	## Shift the start/end coordinates to the left by $coord_beg. ##
	    	## Actually, this operation is very important when calculating the log-likelihoods of indel scenarios. ##

		my $info_gblock = $cls_gblks->[$bl];
		$info_gblock->[2] += $seg_start;
		$info_gblock->[3] += $seg_start; 

		my ($indx_gp, $len_gb, $start, $end) = @{$info_gblock};

		$end++; # Convert from the full-closed to the half-open convention. (ADDED on Nov 18 (Mon), 2013). #
	
		my $gpattern = ${$set_gpatterns->[$indx_gp]};
		print $fh_out join ("\t", $bl, $len_gb, $start, $end, $indx_gp, "'$gpattern'"), "\n";

	    }			# End of the 1st middle for-loop (over the gap-blocks) #
	    print $fh_out "\n\n";

	    my $min_sbtotct_indels = $min_totcts_indel_for_cls->[$cl];

		# MODIFIED in revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014. #
	    unless ($min_totct_indels eq 'N/A') {
	       ($min_sbtotct_indels eq 'N/A') ? ($min_totct_indels = 'N/A') : ($min_totct_indels += $min_sbtotct_indels);
	    }
	    ##$min_totct_indels += $min_sbtotct_indels;

	    my $set_psm_scns = $sets_psm_scns_for_cls->[$cl];

		# MODIFIED in revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014. #
	    my $ct_psm_scns = ($min_sbtotct_indels eq 'N/A') ? 'N/A' : @{$set_psm_scns};
	    ##my $ct_psm_scns = @{$set_psm_scns};

	    print $fh_out "[ Parsimonious Scenarios (Min(Total#{indel events}) = $min_sbtotct_indels\; #{scenarios} = $ct_psm_scns) ]\n\n";

		# ADDED in revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014. #
	    if ($min_sbtotct_indels eq 'N/A') {
		print $fh_out "[INFORMATION] 'N/A' was assigned as the Minimun Total Number Of Indel Events, likely because the cluster contains (a) long gap(s)...\n\n\n";
		next;
	    }

	    for (my $ps=0; $ps<$ct_psm_scns; $ps++) { # 2nd middle for-loop (over the parsimonious scenarios) #

		print $fh_out ' ' x 3, "{ The $ps th scenario }\n\n";

		my ($br2indels, $node2bw_gstats) = @{$set_psm_scns->[$ps]};

    ##if ($TEST_LEVEL>2) {
		print $fh_out ' ' x 3, "( Map of block-wise gap states onto the nodes )\n\n";

			# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		print $fh_out ' ' x 3, join ("\t", 'Node_Name', "'block_wise_gap_states(concatenated)'"), "\n\n";
		##print $fh_out ' ' x 3, join ("\t", 'Node_ID', "'block_wise_gap_states(concatenated)'"), "\n\n";

		foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats}) { # 1st inner foreach-loop #
		    my $cnct_bw_gstats = join ('', @{$node2bw_gstats->{$node}});

			# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		    my $node_name = ($node == $sb_topnode_id) ? $sb_topnode_name : ((defined $sb_node2ch->{$node}) ? $sb_br_attr->{$node}[$INDX_INTNODENAME] : $sb_id2name{$node} ) ;
	  	    print $fh_out ' ' x 3, join ("\t", $node_name, "'$cnct_bw_gstats'"), "\n";
	  	    ##print $fh_out ' ' x 3, join ("\t", $node, "'$cnct_bw_gstats'"), "\n";
		}				# End of the 1st inner foreach-loop #
		print $fh_out "\n\n";
    ##} ## END of "if ($TEST_LEVEL>2)"

		## ($seg_start, $seg_end, $seg_pattern, $seg_msa) = @{$in_msa_segments->[$s]};

		print $fh_out ' ' x 3, "( Map of indel events onto the branches )\n\n";

			# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		print $fh_out ' ' x 3, join ("\t", 'Branch_le_Name', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";
		##print $fh_out ' ' x 3, join ("\t", 'Branch_ID', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";

		foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # 2nd inner foreach-loop #
		    my @out_indel_info = ();
		    foreach my $indel_info (@{$br2indels->{$br}}) { # Innermost foreach-loop #
			my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
			(defined $stat_ue) or ($stat_ue = 'undef');
			(defined $stat_le) or ($stat_le = 'undef');
			my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef';
			push @out_indel_info, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
		    }			# End of the innermost foreach-loop #

			# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		    my $br_le_name = (defined $sb_node2ch->{$br}) ? $sb_br_attr->{$br}[$INDX_INTNODENAME] : $sb_id2name{$br} ;
		    print $fh_out ' ' x 3, join ("\t", $br_le_name, @out_indel_info), "\n";
		    ##print $fh_out ' ' x 3, join ("\t", $br, @out_indel_info), "\n";

		}				# End of the 2nd inner foreach-loop #
		print $fh_out "\n\n";

	    }			# End of the 2nd middle for-loop (over the parsimonious scenarios) #

	}		# End of the 3rd interim outer for-loop (over the clusters) #

    (print STDOUT "\n\n.....Finished displaying the Secondary Inputs (i).....\n\n") if ($TEST_LEVEL>1);

	print $fh_out "\n< Minimum Total #{indels} = $min_totct_indels >\n\n";


		## Important Output (i) ##

	## $set_ct_indels->[$t] = $min_totct_indels;

	## ($msa_type, $seg_msa, $coord_beg, $seg_len, $totlen_aln, $set_seg_scores_gp, $set_ct_indels, $fh_log_g) = @{$set_info};

		## Before beginning to calculate the log-likelihoods of the parsimonious indel scenarios,
		## check whether all indels are shorter than or equal to the specified cut-off length.
		## (ADDED on Sep 29 (Sat), 2012)

		# MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'. #
	my $min_cutoff_len_ins = $CUTOFF_INS_LEN;
	my $min_cutoff_len_del = $CUTOFF_DEL_LEN;
	##my $min_cutoff_len_ins = my $min_cutoff_len_del = ($CUTOFF_5P < $CUTOFF_3P) ? $CUTOFF_5P : $CUTOFF_3P; 

	print STDOUT "\n\n<< Examine whether indel events are shorter than or equal to the cut-off length (= $min_cutoff_len_ins for insertions, = $min_cutoff_len_del for deletions) >>\n\n";

	my $illegal_indels = inspect_indel_events (@{$set_cls_gblocks}, @{$sets_psm_scns_for_cls}, $min_cutoff_len_ins, $min_cutoff_len_del, $CHAR_ABS);
	unless (defined $illegal_indels) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'inspect_indel_events' FAILED !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ;
	}

	my @interim_in_sets_psm_scns_for_cls = @{$sets_psm_scns_for_cls};

	if (@{$illegal_indels}>0) {
	    my $ct_illegal = @{$illegal_indels};

	# @{$illegal_indels[$il]} = ($cl, $ps, $br_id, $id, $length) is for the $il th illegal indel event, with

	    print STDOUT "[WARNING '$program_name'] The reduced MSA for the $s th segment has $ct_illegal illegal indels:\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'No.', 'ID_cls', 'ID_psm_scn', 'Br_ID(Lower_End_Name)', 'ID_indel', 'Length'), "\n\n";
	    ##print STDOUT ' ' x 3, join ("\t", 'No.', 'ID_cls', 'ID_psm_scn', 'Br_ID', 'ID_indel', 'Length'), "\n\n";
	    ##print $fh_out ' ' x 3, join ("\t", 'No.', 'ID_cls', 'ID_psm_scn', 'Br_ID', 'ID_indel', 'Length'), "\n\n";

	    for (my $il=0; $il<$ct_illegal; $il++) {

			# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		my @out = @{$illegal_indels->[$il]};
		my $br = $out[2];
		my $br_le_name = (defined $sb_node2ch->{$br}) ? $sb_br_attr->{$br}[$INDX_INTNODENAME] : $sb_id2name{$br} ;
	        print STDOUT ' ' x 3, join ("\t", $il, @out[0,1], "$br($br_le_name)", @out[3,4]), "\n";
	        ##print STDOUT ' ' x 3, join ("\t", $il, @{$illegal_indels->[$il]}), "\n";

			# ADDED in revision of 'br_infer_psm_indel_events_dawg.pl' (on Mar 2 (Sun), 2014). #
		my $cl = $out[0];
		my $ct_psns = @{$sets_psm_scns_for_cls->[$cl]};
		my @new_interim_in_set = ();
		for (1 .. $ct_psns) { push @new_interim_in_set, []; }
		$interim_in_sets_psm_scns_for_cls[$cl] = \@new_interim_in_set;
	    }

			# MODIFIED in revision of 'br_infer_psm_indel_events_dawg.pl' (on Mar 2 (Sun), 2014). #
	    print STDOUT "\n\n", ' ' x 3, "... I will refrain from calculating the likelihoods of indel scenarios for those clusters. Instead, I will just set them to be 'N/A' ...\n\n";
	    print STDOUT '               $tot_lnlk_gp = N/A', " .\n\n\n";
=For the OLD VERSION of 'br_infer_psm_indel_events_dawg.pl' (DISABLED on Mar 2 (Sun), 2014). #
	    print STDOUT "\n\n", ' ' x 3, "... I will refrain from calculating the likelihood of indel scenarios. Instead, I will just set it to be 'undef' ...\n\n";
	    print STDOUT '               $tot_lnlk_gp = undef', " .\n\n\n";

	        ## Important Output (ii) (Exceptional) ##
	        ##$set_seg_scores_gp->[$t] = undef;

	    print $fh_out "\nFOUND ILLEGAL INDEL EVENTS (#{events} = $ct_illegal) ...\n\n";
	    print $fh_out "\n\\\\\n\n";

	    next;
=cut

	    ## ($msa_type, $seg_msa, $coord_beg, $seg_len, $totlen_aln, $set_seg_scores_gp, $set_ct_indels, $fh_log_g) = @{$set_info};
	}


	print STDOUT "\n\n<< (ii) Calculate the log-likelihoods of the parsimonious indel scenarios, and therefore the log-likelihood of the gap-pattern of this MSA segment >>\n\n";

	print $fh_out "\n\n<< (ii) Calculate the log-likelihoods of the parsimonious indel scenarios, and therefore the log-likelihood of the gap-pattern of this MSA segment >>\n\n";

	##my %sb_name2id;
	##my %sb_id2name;
	##my ($sb_topnode_id, $sb_node2ch, $sb_node2pa, $sb_br_attr);

	my ($tot_lnlk_gp0, $shared_fac1, $shared_fac2, $set_clwise_tot_lnlk_gp, $tot_div_tree, $totlen_ng, $setset_lnlk_gp_cls_scn)
	 = apprx1_tot_lnlk_gpattern_msa2 (%{$sb_br_attr}, %{$sb_node2pa}, $totlen_aln, $CHAR_ABS, @{$set_cls_gblocks}, @interim_in_sets_psm_scns_for_cls, $TOTRATE_INS, $TOTRATE_DEL, $CUTOFF_INS_LEN, $CUTOFF_DEL_LEN, $delta_del, %{$br2log_mfacs}); # MODIFIED in creation of 'br_infer_psm_indel_events_spt_odr_dawg.hs.pl'.

=for the predecessor 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'.
	 = apprx1_tot_lnlk_gpattern_msa (%{$sb_br_attr}, %{$sb_node2pa}, $totlen_aln, $CHAR_ABS, @{$set_cls_gblocks}, @interim_in_sets_psm_scns_for_cls, @PARAMS_INS, @PARAMS_DEL, $CUTOFF_5P, $CUTOFF_3P, @{$log_nfacs}, $delta_del_5p, $OPT_BD_INDELS); # MODIFIED in revision of 'br_infer_psm_indel_events_dawg.pl' (on Mar 2 (Sun), 2014). #
	 ##= apprx1_tot_lnlk_gpattern_msa (%{$sb_br_attr}, %{$sb_node2pa}, $totlen_aln, $CHAR_ABS, @{$set_cls_gblocks}, @{$sets_psm_scns_for_cls}, @PARAMS_INS, @PARAMS_DEL, $CUTOFF_5P, $CUTOFF_3P, @{$log_nfacs}, $delta_del_5p, $OPT_BD_INDELS); # For the OLD VERSION (that had been effective until Mar 2 (Sun), 2014). #
=cut

	unless (defined $tot_lnlk_gp0) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'apprx1_tot_lnlk_gpattern_msa2' FAILED: $! !\n\n";
	    print STDERR "[ERROR '$program_name'] $! !\n\n";
	    die ;
	}
	unless (@{$set_clwise_tot_lnlk_gp} == $ct_cls) {
	    print STDOUT "[ERROR '$program_name'] The output ", '@{$set_clwise_tot_lnlk_gp} ', "has the size ", scalar (@{$set_clwise_tot_lnlk_gp}), ", which CLASHES with $ct_cls, the number of gap-block clusters !\n";
	    print STDOUT "[ERROR '$program_name'] $!\n";
	    print STDERR "[ERROR '$program_name'] $! !\n\n";
	    die ;
	}
	unless (@{$setset_lnlk_gp_cls_scn} == $ct_cls) {
	    print STDOUT "[ERROR '$program_name'] The output ", '@{$setset_lnlk_gp_cls_scn} ', "has the size ", scalar (@{$setset_lnlk_gp_cls_scn}), ", which CLASHES with $ct_cls, the number of gap-block clusters !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ;
	}
    (print STDOUT "\n--- The main subroutine 'apprx1_tot_lnlk_gpattern_msa2' ENDED ...\n\n") if ($TEST_LEVEL>1);
    ##(print STDOUT "\n--- The main subroutine 'apprx1_tot_lnlk_gpattern_msa' ENDED ...\n\n") if ($TEST_LEVEL>1);


		## An Important Adjustment, from 'MSA-wide' log-likelihoods to segment-wise log-likelihoods. ##

		# FURTHER MODIFIED in revision on Mar 6 (Thu), 2014. #
	my $tot_lnlk_gp = ($tot_lnlk_gp0 eq 'N/A') ? 'N/A' : (($tot_lnlk_gp0 eq 'undef') ? 'undef' : $tot_lnlk_gp0 - $shared_fac1 + ($TOTRATE_INS + $TOTRATE_DEL) * $tot_div_tree * ($totlen_aln - $seg_len)) ; 
	##	# MODIFIED in revision on Mar 2 (Sun), 2014. #
	##my $tot_lnlk_gp = ($tot_lnlk_gp0 eq 'N/A') ? 'N/A' : $tot_lnlk_gp0 - $shared_fac1 + ($TOTRATE_INS + $TOTRATE_DEL) * $tot_div_tree * ($totlen_aln - $seg_len); 
	####my $tot_lnlk_gp = $tot_lnlk_gp0 - $shared_fac1 + ($TOTRATE_INS + $TOTRATE_DEL) * $tot_div_tree * ($totlen_aln - $seg_len); 


	print $fh_out "\n[[ Outputs of 'apprx1_tot_lnlk_gpattern_msa2' ]]\n\n";
	##print $fh_out "\n[[ Outputs of 'apprx1_tot_lnlk_gpattern_msa' ]]\n\n";

	print $fh_out "[1] Scalar Outputs:\n\n";
	print $fh_out '($tot_div_tree, $totlen_ng) = ', "($tot_div_tree, $totlen_ng) ,\n";
	print $fh_out '($shared_fac1, $shared_fac2) = ', "($shared_fac1, $shared_fac2) .\n\n";
	print $fh_out 'RAW $tot_lnlk_gp = ', $tot_lnlk_gp0, " .\n\n";

	print $fh_out "\n[2] Cluster-wise log-likelihoods (for the $ct_cls gap-block clusters):\n\n";

	for (my $cl = 0; $cl<$ct_cls; $cl++) { # Main outer for-loop (over the gap-block clusters) #
    
	    my $cw_tot_lnlk_gp = $set_clwise_tot_lnlk_gp->[$cl];
	    my $ct_psm_scn = @{$sets_psm_scns_for_cls->[$cl]};
	    my $set_lnlk_gp_scn = $setset_lnlk_gp_cls_scn->[$cl];

	    unless (@{$set_lnlk_gp_scn} == $ct_psm_scn) {
		print STDOUT "\n[CAUTION '$program_name'] In the $cl th cluster, the number of log-likelihoods of parsimonious scenarios (=", scalar (@{$set_lnlk_gp_scn}), ") CLASHES with the number of scenarios themselves (=$ct_psm_scn) ... !!\n\n";
	    }

		# MODIFIED in revision on Mar 11 (Tue), 2014. #
	    my $out_ct_psm_scn = ($min_totcts_indel_for_cls->[$cl] eq 'N/A') ? 'N/A' : $ct_psm_scn;
	     print $fh_out "< The $cl th cluster (#{parsimonious scenarios} = $out_ct_psm_scn\; Total{log-likelihood} = $cw_tot_lnlk_gp)  >\n\n";
	    ## print $fh_out "< The $cl th cluster (#{parsimonious scenarios} = $ct_psm_scn\; Total{log-likelihood} = $cw_tot_lnlk_gp)  >\n\n";

	    ## print $fh_out "[ Parsimonious Scenarios (Min(Total#{indel events}) = $min_sbtotct_indels\; #{scenarios} = $ct_psm_scns) ]\n\n";
		## print $fh_out ' ' x 3, "{ The $ps th scenario }\n\n";

	    print $fh_out join ("\t", 'Scn_no.', 'log-likelihood', 'Relative_probability'), "\n\n";

	    for (my $psm = 0; $psm<$ct_psm_scn; $psm++) { # Inner for-loop (over the parsimonious scenarios). #

		my $lnlk_gp_scn = $set_lnlk_gp_scn->[$psm];

			# FURTHER MODIFIED in revision on Mar 6 (Thu), 2014. #
		my $rel_prob = (($cw_tot_lnlk_gp eq 'N/A') or ($lnlk_gp_scn eq 'N/A')) ? 'N/A' : ( (($cw_tot_lnlk_gp eq 'undef') or ($lnlk_gp_scn eq 'undef')) ?  'undef' : exp ($lnlk_gp_scn - $cw_tot_lnlk_gp) ) ;
		##	# MODIFIED in revision on Mar 2 (Sun), 2014. #
		##my $rel_prob = (($cw_tot_lnlk_gp eq 'N/A') or ($lnlk_gp_scn eq 'N/A')) ? 'N/A' : exp ($lnlk_gp_scn - $cw_tot_lnlk_gp);
		####my $rel_prob = exp ($lnlk_gp_scn - $cw_tot_lnlk_gp);

		print $fh_out join ("\t", $psm, $lnlk_gp_scn, $rel_prob), "\n";

	    }				# End of the inner for-loop (over the parsimonious scenarios). #

	    print $fh_out "\n";

	}		# End of the main outer for-loop (over the clusters) #
	
    (print STDOUT "\n\n.....Finished displaying the 'Final' Outputs (ii) ....\n\n") if ($TEST_LEVEL>1);

	print $fh_out "\n", '[3] (Main!!) FINAL $tot_lnlk_gp (= $tot_lnlk_gp0 - $shared_fac1 + ($TOTRATE_INS + $TOTRATE_DEL) * $tot_div_tree * ($totlen_aln - $seg_len)):', "\n";
	print $fh_out '               $tot_lnlk_gp = ', $tot_lnlk_gp, " .\n\n\n";


		### Output Parsimonious indel scenarios and ancestral gap states to $outfile2. ###
		### (ADDED on Nov 19 (Tue), 2013) ###

	print $fh_out2 "[[ Parsimonious indel scenarios and resulting ancestral gap states ]]\n\n";

	my $curr_coord = $seg_start;

	for (my $cl = 0; $cl<$ct_cls; $cl++) { # 3rd interim outer for-loop (over the clusters) #

	    my $cls_gblks = $set_cls_gblocks->[$cl];
	    my $ct_gblks = @{$cls_gblks};

	    my $cl_start = $cls_gblks->[0][2];
	    my $cl_end = $cls_gblks->[$ct_gblks-1][3] + 1; # Convert fro full-closed to half-open. #
	    my $len_glbl = $cl_start - $curr_coord;
	    my $cl_len = $cl_end - $cl_start;

	    print $fh_out2  "< Gapless block left-flanking the $cl th cluster \; region = \[$curr_coord, $cl_start\) \; length = $len_glbl >\n\n"; 

	    my $cw_tot_lnlk_gp = $set_clwise_tot_lnlk_gp->[$cl]; # Total log-likelihood of the gap-pattern of the cluster.
	    my $set_lnlk_gp_scn = $setset_lnlk_gp_cls_scn->[$cl]; # This array is necessary for calculating the relative probability. #

	    print $fh_out2 "< The $cl th cluster \; region = \[$cl_start, $cl_end\) \; length = $cl_len \; #{gap-blocks} = $ct_gblks >\n\n";

	    print $fh_out2 "[ Gap-blocks in the cluster ]\n\n";
	    print $fh_out2 join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', 'gap_pattern'), "\n\n";

	    for (my $bl=0; $bl<$ct_gblks; $bl++) { # 1st middle for-loop (over the gap-blocks) #

	    	## Shift the start/end coordinates to the left by $coord_beg. ##
	    	## Actually, this operation is very important when calculating the log-likelihoods of indel scenarios. ##

		my $info_gblock = $cls_gblks->[$bl];
		##$info_gblock->[2] += $seg_start;
		##$info_gblock->[3] += $seg_start; 

		my ($indx_gp, $len_gb, $start, $end) = @{$info_gblock};

		$end++; # Convert from the full-closed to the half-open convention. (ADDED on Nov 18 (Mon), 2013). #
	
		my $gpattern = ${$set_gpatterns->[$indx_gp]};
		print $fh_out2 join ("\t", $bl, $len_gb, $start, $end, $indx_gp, $gpattern), "\n";

	    }			# End of the 1st middle for-loop (over the gap-blocks) #
	    print $fh_out2 "\n\n";

	    my $min_sbtotct_indels = $min_totcts_indel_for_cls->[$cl];
	    ##$min_totct_indels += $min_sbtotct_indels;

	    my $set_psm_scns = $sets_psm_scns_for_cls->[$cl];

		# MODIFIED in revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014. #
	    my $ct_psm_scns = ($min_sbtotct_indels eq 'N/A') ? 'N/A' : @{$set_psm_scns};
	    ##my $ct_psm_scns = @{$set_psm_scns};

	    print $fh_out2 "[ Parsimonious Scenarios (Min_Total#{indel_events} = $min_sbtotct_indels\; #{scenarios} = $ct_psm_scns) ]\n\n";

			## Handle an exceptinal case, which involves (a) long gap(s). ##
			## (ADDED accompanying the revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014.) ##

	    if ($min_sbtotct_indels eq 'N/A') {

		print $fh_out2 "[INFORMATION] 'N/A' was assigned as the Minimun Total Number Of Indel Events, likely because the cluster contains (a) long gap(s)...\n\n\n";

		$curr_coord = $cl_end; # Update the current coordinate. #
		next;
	    }

	    for (my $ps=0; $ps<$ct_psm_scns; $ps++) { # 2nd middle for-loop (over the parsimonious scenarios) #

			## Calculate the relative probability of the scenario. ##
		my $lnlk_gp_scn = $set_lnlk_gp_scn->[$ps];

			# FURTHER MODIFIED in revision on Mar 6 (Thu), 2014. #
		my $rel_prob = (($cw_tot_lnlk_gp eq 'N/A') or ($lnlk_gp_scn eq 'N/A')) ? 'N/A' : ( (($cw_tot_lnlk_gp eq 'undef') or ($lnlk_gp_scn eq 'undef')) ?  'undef' : exp ($lnlk_gp_scn - $cw_tot_lnlk_gp) ) ;
		##	# MODIFIED in revision on Mar 2 (Sun), 2014. #
		## my $rel_prob = (($cw_tot_lnlk_gp eq 'N/A') or ($lnlk_gp_scn eq 'N/A')) ? 'N/A' : exp ($lnlk_gp_scn - $cw_tot_lnlk_gp);
		#### my $rel_prob = exp ($lnlk_gp_scn - $cw_tot_lnlk_gp);

		print $fh_out2 "{ Scenario no. $ps \; rel_prob. = $rel_prob }\n\n";

		my ($br2indels, $node2bw_gstats) = @{$set_psm_scns->[$ps]};

		print $fh_out2 "( Block-wise Gap States mapped onto the nodes )\n\n";

			# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		print $fh_out2 join ("\t", 'Node_Name', 'block_wise_gap_states(concatenated)'), "\n\n";
		##print $fh_out2 join ("\t", 'Node_ID', 'block_wise_gap_states(concatenated)'), "\n\n";

		foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats}) { # 1st inner foreach-loop #
	 	    my $cnct_bw_gstats = join ('', @{$node2bw_gstats->{$node}});

			# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		    my $node_name = ($node == $sb_topnode_id) ? $sb_topnode_name : ((defined $sb_node2ch->{$node}) ? $sb_br_attr->{$node}[$INDX_INTNODENAME] : $sb_id2name{$node} ) ;
	 	    print $fh_out2 join ("\t", $node_name, $cnct_bw_gstats), "\n";
	 	    ##print $fh_out2 join ("\t", $node, $cnct_bw_gstats), "\n";
		}				# End of the 1st inner foreach-loop #
		print $fh_out2 "\n\n";

		## ($seg_start, $seg_end, $seg_pattern, $seg_msa) = @{$in_msa_segments->[$s]};

		print $fh_out2 "( Indel Events mapped onto the branches )\n\n";

			# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		print $fh_out2 join ("\t", 'Branch_le_Name', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";
		##print $fh_out2 join ("\t", 'Branch_ID', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";

		foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # 2nd inner foreach-loop #

		    my @out_indel_info = ();
		    foreach my $indel_info (@{$br2indels->{$br}}) { # Innermost foreach-loop #
			my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
			(defined $stat_ue) or ($stat_ue = 'undef');
			(defined $stat_le) or ($stat_le = 'undef');
			my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef';
			push @out_indel_info, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
		    }			# End of the innermost foreach-loop #

			# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		    my $br_le_name = (defined $sb_node2ch->{$br}) ? $sb_br_attr->{$br}[$INDX_INTNODENAME] : $sb_id2name{$br} ;
		    print $fh_out2 join ("\t", $br_le_name, @out_indel_info), "\n";
		    ##print $fh_out2 join ("\t", $br, @out_indel_info), "\n";

		}				# End of the 2nd inner foreach-loop #
		print $fh_out2 "\n\n";

	    }			# End of the 2nd middle for-loop (over the parsimonious scenarios) #

	    $curr_coord = $cl_end; # Update the current coordinate. #

	}		# End of the 3rd interim outer for-loop (over the clusters) #

	my $len_glbl = $seg_end - $curr_coord;
	print $fh_out2  "< Gapless block on the right-end of the segment \; region = \[$curr_coord, $seg_end\) \; length = $len_glbl >\n\n\n"; 

	$ct_succ_segs++;

}			# END of the main outer for-loop (over MSA segments) #

close ($fh_out);
close ($fh_out2);


print STDOUT "\n\n === All the Analyses FINISHED !! ===\n\n";

print STDOUT "#{Segments skipped} = $ct_skipped_segs .\n";
print STDOUT "#{Anomalous segments} = $ct_anom_segs .\n";
print STDOUT "#{Segments used as inputs} = $ct_in_segs .\n";
print STDOUT "#{Segments successfully processed} = $ct_succ_segs .\n";
print STDOUT "#{Segments failed to be processed} = $ct_fail_segs .\n\n";

unless (0 == $ct_fail_segs) {
	print STDOUT "[ERROR '$program_name'] FAILED to process $ct_fail_segs segments !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}


	### DEFINITIONS of the Subroutines. ###


# partition_msa_via_usable_patterns:
#	This subroutine partitions an input MSA according to the
#	pattern of usability of sequences for the indel inference afterwards.
#
# INPUT: (\@in_msa, \%char_missing), where
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	%char_missing = ($character => {some defined value}, ...) for the characters
#	indicating that the site is missing or that the site's orthology is questionable.
#
# OUTPUT: \@msa_segments, where
#	@{$msa_segments[$s]} = ($seg_start, $seg_end, \@seg_pattern, \@seg_msa) is for the $s th segment, where
#	$seg_start/end is the start/end coordinate of the segment (in the 0-based half-open convention);
#	$seg_pattern[$k] = 1 if the site (including a gap) in the $k th sequence is usable for the analysis,
#			 = 0 if $char_missing{$char} is defined for the $k th sequence at the site in question.
#	${$seg_msa[$k]} is the $k th sequence embedded in the segment MSA.
#
sub partition_msa_via_usable_patterns (\@\%) {

	my ($in_msa, $char_missing) = @_;

	my $ct_seqs = @{$in_msa};
	unless ($ct_seqs>0) {
	    print STDOUT "[ERROR 'partition_msa_via_usable_patterns'] ILLEGAL INPUT: The input MSA contains NO SEQUENCE: $! !\n";
	    return undef;
	}

	my $msa_len = length ${$in_msa->[0]};

	my @msa_segments = ();

	my $last_cnct_pattern = '';

	for (my $l=0; $l<$msa_len; $l++) { # Main for-loop (over sites) #

	    my @chars = my @seg_pattern = ();

	    for (my $k=0; $k<$ct_seqs; $k++) { # Inner for-loop (over sequences) #

		my $char = substr (${$in_msa->[$k]}, 0, 1, '');
		push @chars, $char;
		push @seg_pattern, (defined $char_missing->{$char}) ? 0 : 1; 

	    }				# End of the inner for-loop (over sequences) #

	    my $cnct_pattern = join ('', @seg_pattern);

	    if ($cnct_pattern eq $last_cnct_pattern) {

			# Elongate the last segment. #

		my $last_segment = $msa_segments[$#msa_segments];

		$last_segment->[1]++;

		for (my $k=0; $k<$ct_seqs; $k++) { ${$last_segment->[3][$k]} .= $chars[$k]; }

	    } else {
			# Create a new segment. #

		my @seg_msa = ();
		foreach my $char (@chars) { push @seg_msa, \$char; }

		push @msa_segments, [$l, $l+1, \@seg_pattern, \@seg_msa];

		$last_cnct_pattern = $cnct_pattern;
	    }

	}			# End of the main for-loop (over sites) #

	return \@msa_segments;

}	# End of the subroutine 'partition_msa_via_usable_patterns' #


# inspect_indel_events:
#	This subroutine examines whether all the indel events are shorter than or equal to the cutoff-lengths.
#
# INPUT: (\@set_cls_gblocks, \@sets_psm_scns_for_cls, $cutoff_len_ins, $cutoff_len_del, $char_abs), where
#
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)
#	is for the $gb th block in the $cl th cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordiate of the block (in the 0-based convention);
#
#	@{$set_psm_scns_in_cls[$cl]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th of the parsimonious scenarios realizing the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) if the map of gap states onto the interior nodes,
#	with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id;
#	
#	$cutoff_len_ins/del is the cutoff length for the insertion/deletion events;
#
#	$char_abs is the character representing the 'absence' state.
#
# OUTPUT: \@illegal_indels, where
#	@{$illegal_indels[$il]} = ($cl, $ps, $br_id, $id, $length) is for the $il th illegal indel event, with
#	$cl specifying a cluster,
#	$ps specifying a parsimonious scenario in the cluster,
#	$br_id specifying a branch,
#	$id specifying an illegal event in the node,
#	and $length is the length of the event.
#
#	The subroutine returns undef if it fails.
#
sub inspect_indel_events (\@\@$$$) {

	my ($set_cls_gblocks, $sets_psm_scns_for_cls, $cutoff_len_ins, $cutoff_len_del, $char_abs) = @_;

    (print STDOUT ' ' x 3, "< Subroutine 'inspect_indel_events' >\n\n") if ($TEST_LEVEL_INSP>1);

	my $ct_cls = @{$set_cls_gblocks};
	unless (@{$sets_psm_scns_for_cls} == $ct_cls) {
	    print STDOUT "[ERROR 'inspect_indel_events'] ILLEGAL INPUT: The size of ", '@{$sets_psm_scns_for_cls} = (', scalar (@{$sets_psm_scns_for_cls}), ") CLASHES with the number of gap-block clusters (=$ct_cls): $! !\n";
	    return undef;
	}

	my @illegal_indels = ();

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Outer for-loop (over clusters) #

	    my $cls_gblock = $set_cls_gblocks->[$cl];
	    my $psm_scns_for_cls = $sets_psm_scns_for_cls->[$cl];
	    my $ct_psm_scns = @{$psm_scns_for_cls};

    (print STDOUT "\n", ' ' x 3, "[ The $cl th cluster has $ct_psm_scns parsimonious scenarios of indel events ]\n\n") if ($TEST_LEVEL_INSP>2);

	    for (my $ps=0; $ps<$ct_psm_scns; $ps++) { # Middle for-loop (over parsiminous scenarios) #

    (print STDOUT "\n", ' ' x 5, "\( The $ps th parsimonious scenario \)\n\n") if ($TEST_LEVEL_INSP>3);

		my ($br2indels, $node2bw_gstat) = @{$psm_scns_for_cls->[$ps]};

		foreach my $br_id (sort {$a <=> $b} keys %{$br2indels}) { # Inner-middle foreach-loop (over branches) #
		    my $indel_events = $br2indels->{$br_id};
		    my $ct_events = @{$indel_events};

    (print STDOUT ' ' x 7, "( The branch '$br_id' has $ct_events nominal indel events )\n") if ($TEST_LEVEL_INSP>3);

		    for (my $id=0; $id<$ct_events; $id++) { # Inner for-loop (over indel events) #
			my ($stat_pa, $stat_node, $indices_blk) = (defined $indel_events->[$id]) ? @{$indel_events->[$id]} : ();
			unless (defined $stat_pa) {
			    print STDOUT "[WARNING 'inspect_indel_events'] The $id th event on the branch '$br_id' is EMPTY  for the $ps th indel scenario for the $cl th cluster: $! !\n";
			    next;
			}
			my $cutoff_len = ($stat_pa eq $char_abs) ? $cutoff_len_ins : $cutoff_len_del;
			my $indel_len = 0;
			foreach my $indx_blk (@{$indices_blk}) { $indel_len += $cls_gblock->[$indx_blk][1] ; }

			unless ($indel_len <= $cutoff_len) {
			    print ' ' x 7, "\n[WARNING 'inspect_indel_events'] The $id th indel event along the $br_id th branch is EXCEEDINGLY LONG (length = $indel_len)  for the $ps th scenario in the $cl th gap-block cluster: $! !\n\n";
			    push @illegal_indels, [$cl, $ps, $br_id, $id, $indel_len];
			}

		    }				# End of the inner for-loop (over indel events) #

		} 		# End of the inner-middle foreach-loop (over branches) #

	    }			# End of the middle for-loop (over parsimonious scenarios) #

	}			# End of the outer for-loop (over clusters) #

    (print STDOUT "\n\n", ' ' x 3, "--- End of the Subroutine 'inspect_indel_events'  ---\n\n\n") if ($TEST_LEVEL_INSP>1);

	return \@illegal_indels;

}	# End of the subroutine 'inspect_indel_events' #


	## ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. ##

# read_dawg_ctrl_file:
#	This subroutine reads a Dawg control file,
#	and retrieves indel parameters and a guide tree.
#
# INPUT: \$ctrl_file, where
#	$infile is the path to a Dawg control file.
#
# OUTPUT: (\$tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate), where
#	$tree_string is a string representing the guide tree;
#	$gapmodel specifies the gap model;
#	$gap_param is the power exponent if $gapmodel = '"PL"',
#			the negative binomial parameter if $gapmodel = '"NB"';
#	$gaplen_cutoff is the cut-off for the gap-length;
#	$tot_indel_rate is the total insertion/deletion rate (per site).
#
sub read_dawg_ctrl_file (\$) {

	my ($ctrl_file) = @_;
	local $/ = "\n";

	my $fh_ctrl = open_input_file2 (${$ctrl_file});
	unless (defined $fh_ctrl) {
	    print STDOUT "[ERROR 'read_dawg_ctrl_file'] 'open_input_file2' FAILED on the control file '${$ctrl_file}': $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

	my ($tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate);

	while (my $line = <$fh_ctrl>) { # Main while-loop (over lines). #

	    ($line =~ m/^\#/) and next; # Skip a comment line. #

	    $line =~ s/\s+$//;
	    $line =~ s/\;$//; # Added to prevent an error message (concerning $tot_indel_rate) on Mar 10 (Mon), 2014. #

	    my ($option_name, $value) = split /\s*\=\s*/, $line;

	    if ($option_name eq 'Tree') {
		$tree_string = $value;

	    } elsif ($option_name eq 'GapModel') {
		$gapmodel = $value;

	    } elsif ($option_name eq 'GapParams') {

		$value =~ s/^\{//;
		$value =~ s/\}$//;
		($gap_param, $gaplen_cutoff) = split /\s*\,\s*/, $value;

	    } elsif ($option_name eq 'Lambda') {

		$tot_indel_rate = $value;
	    }

	}			# End of the main while-loop (over lines). #

	close ($fh_ctrl);

	foreach my $var ($tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate) { # Trailing foreach-loop. #
	    (defined $var) and next;
	    return (undef, undef, undef, undef, undef);
	}		# End of the trailing foreach-loop. #

	return (\$tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate);

}	# End of the subroutine 'read_dawg_ctrl_file'. # 

