#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# bchbch_retrieve_abs_prob_gpatterns.pl:
#	This Perl script retrieves the absolute (log-)probabilities and gap-patterns of the gap-block clusters
#	from the 'output_detailxxx.txt' file output by the Perl script 'br_infer_psm_indel_events_dawg.v2.pl'.
#	It was created by taking advantage of the overall framework of the script 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
#	(CREATION STARATED on Mar 12 (Wed), 2014, ENDED on Mar xx (yyy), 2014)
#
# bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl:
#	This script was created by stripping most commented out parts off the script
#	'bchbch_cmp_true_vs_psm_inferred_ancstates.v2.pl'.
#	(CREATED on Mar 12 (Wed), 2014)
#
# bchbch_cmp_true_vs_psm_inferred_ancstates.v2.pl:
#	This revised version of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'
#	copes with some changes in the 'output_briefxxxx.txt' file introduced 
#	in the creation (and revision) of 'br_infer_psm_indel_events_dawg.v2.pl'.
#	(CREATED on Mar 11 (Tue), 2014)
#
# bchbch_cmp_true_vs_psm_inferred_ancstates.pl:
#	This Perl script compares true ancestral states (in each Dawg simulated MSA)
#	with inferred ones, and judges how accurately my parsimony method can infer ancestral states.
#	(CREATION STARTED on Mar 3 (Mon), 2014, ENDED on Mar 6 (Thu), 2014)
#
#	REVISED on Mar 6 (Thu), 2014:
#	The script now handles the gap-block cluster (, which likely consists ONLY of a NULL BLOCK),
#	that was assigned an EMPTY SCENARIO DISTRIBUTION by the subroutine 'parse_brief_output_infer_psm_indel_events_dawg'.
#
# bchbch_anal_muts_on_sbjcts_in_bestMSA.pl:
#	This Perl script analyzes mutations, from the parent node to the reference sequence,
#	on each subject region embedded in its best MSA.
#	It was created based on the template 'template_anal_anc_sbjcts_in_bestMSA.pl'.
#	(CREATION STARTED on Dec 24 (Tue), 2013, ENDED on Jan 1 (Wed), 2014)
#
# template_anal_anc_sbjcts_in_bestMSA.pl:
#	This template uses the list of best MSAs embedding the subjects.
#	It was created by modifying 'template_anal_relv_region_in_MSA.pl'.
#	(CREATED on Dec 24 (Tue), 2013)
#
# template_anal_relv_region_in_MSA.pl:
#	This is a template for the Perl scripts that analyzes the ancestral states of 
#	the subject regions that are embedded in the MSA.
#	It was created by modifying the script 'serial_pgene_RFC_analyses.pl'.
#	(CREATION STARTED on Dec 2 (Mon), 2013, ENDED on Dec xx (yyy), 2013)
#
# serial_pgene_RFC_analyses.pl:
#	This Perl script calculates three Reading Frame Conservation (RFC) scores of each proto-gene
#	in a set of proto-genes each fully covered by at least one tiling MSA.
#	The script was created by using some parts in 'check_ORF_completeness.pl' and in 'extract_proto_gene_MSAs.pl'.
#	(CREATION STARTED on Sep 3 (Tue), 2013, ENDED on Sep 14 (Sat), 2013)
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2 open_output_file list_sessions_ext
	tfm_labels2indices tfm_labels2indices_gen
	);

use MyArrayManipu qw(
	diff_array
	multiple_sort
	);

use MyMalnIO qw(
	extract_fasta extract_clw
	pr_ODEN_input
	);
	## pr_phylip_aln
	## pr_clustal_like_lite_aa

use MyInferAncStates_parser qw(
	linear_order_indel_state_distributions
	);

use MyInferAncStates_parser_dawg qw(
	parse_brief_output_infer_psm_indel_events_dawg
	);

use MyTreeManipu_II_dawg qw(
	input_tree_file_dawg read_tree_file_dawg input_subtree_dawg
	br_output_tree_dawg
	);

=for the predecessor 'bchbch_anal_muts_on_sbjcts_in_bestMSA.pl'.
use MyNucSeq2 qw(
	revcom print_sequence
	);

	## ADDED in creation of 'template_anal_anc_sbjcts_in_MSA.pl'. ##

use MyArrayManipu_num qw( binsearch_num );

use MyInferAncStates_parser qw(
	map_treenode2treenode
	extract_sitewise_anc_state_distri_baseml
	from_1based_to_0based_paml_hash
	revcom_anc_base_state_distr
	parse_brief_output_infer_psm_indel_events
	linear_order_indel_state_distributions
	revcom_distr_psm_indel_events
	);

use MyTreeManipu_II qw( output_tree );
=cut

sub peek_into_archive (\$$$$);
sub br_parse_detail_output_infer_psm_indel_events_dawg (\$);

$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);

(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;

our $TOP_NODE = $MyTreeManipu_II_dawg::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II_dawg::INDX_LENGTH;
##our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
our $INDX_INTNODENAME = $MyTreeManipu_II_dawg::INDX_INTNODENAME;



	### User-specified parameters ###

	## Test Output Levels.
	## 0, silent; 1, very brief; 2, pretty brief; 3, rather verbose; 4, verbose; 5, very verbose. ##

	# Test output level (for the main part):
##our $TEST_LEVEL = 1;
our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;
##our $TEST_LEVEL = 6; # For "Pre/Test2a". #

our $GAP = '-'; # This string specifies a gap in the MSA.
our $NGAP = 'X'; # This string specifies a non-gap character (i.e., a residue) in the MSA.

	# Flag to specify how extensively extract the gap states of the nodes.
	# (0, top node only; 1, all interior nodes; 2, all nodes) 
our $FL_FULL_NODES = 2;

our $PREF_ANC = 'anc'; # Prefix shared by the ancestral sequences. #
our $REGX_PREF_ANC = qr/^$PREF_ANC/;


	### Input/Output Files ###

##my $session_name = 'trio_blen0.05';
my $session_name = 'trio_blen0.2';
##my $session_name = 'TESTs/Test1a'; # For TESTING the script ('Test1a'). #
##my $session_name = 'TESTs/Test5a'; # For TESTING the script ('Test5a'). #
##my $session_name = 'TESTs/Test6a'; # For TESTING the script ('Test6a'). #
##my $session_name = 'TESTs/Test6a.DETAILED'; # For TESTING the script ('Test6a'). #
##my $session_name = 'TESTs/Test7a'; # For TESTING the script ('Test7a'). #

####my $session_name = 'hogd_alt_lite_sbcl.1'; # --- For the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'. & For TESTING the script ('Test1_3c').

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
	## Archives of simulated MSAs. ##
	# $in_archive_msa = "$inarch_dir_msa/$archive_pref_msa$id_class$archive_suff_msa".

##=For THIS SCRIPT!!
my $inarch_dir_msa = "../../Inputs/Simulated_MSAs2_1/$session_name";
##my $inarch_dir_msa = "../../Inputs/Simulated_MSAs2_1/trio_blen0.2"; # For TESTING the script ('Test6a').
my $archive_pref_msa = 'class_';
my $archive_suff_msa = '.tgz';
##=cut
#=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl' & For TESTING the script ('Test1_3c').
##my $inarch_dir_msa = "../../Inputs/Simulated_MSAs1/$session_name";
my $inarch_dir_msa = "../../../Proof_of_Principle1/Inputs/Simulated_MSAs1/hogd_alt_lite_sbcl.1";
my $archive_pref_msa = 'hog_d';
my $archive_suff_msa = '.tgz';
#=cut

=cut

	# Archives of ancestral gap states. #
	# $in_archive_g = "$inarch_dir_g/$archive_pref_g$id_class$archive_suff_g".

##=For THIS SCRIPT!!
my $inarch_dir_g = "../../Infer_INDELS/Processed_MSAs2_1/$session_name/Outputs";
##my $inarch_dir_g = "../../Infer_INDELS/Simulated_MSAs2_1/$session_name/Outputs";
my $archive_pref_g = 'class_';
my $archive_suff_g = '.tgz';
##=cut
=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl' & For TESTING the script ('Test1_3c'). 
##my $inarch_dir_g = "../../Infer_INDELS/Simulated_MSAs1/$session_name/Outputs";
##my $inarch_dir_g = "../../Infer_INDELS/Simulated_MSAs1/TESTs/Test3b/Outputs"; # For TESTING the script 'Test3b'.
##my $inarch_dir_g = "../../Infer_INDELS/Simulated_MSAs1/TESTs/Test3c/Outputs"; # For TESTING the script 'Test3c'.
my $inarch_dir_g = "../../../Proof_of_Principle1/Infer_INDELS/Simulated_MSAs1/TESTs/Test3c/Outputs"; # For TESTING the script ('Test1_3c').
my $archive_pref_g = 'hog_d';
my $archive_suff_g = '.tgz';
=cut

	## Output files ##

	# $outfile = "$outdir/$out_pref$lcl_msa_id$out_suff", where
	# $outdir = "$topoutdir/$middir_pref$id_class$middir_suff".

my $toptopoutdir = $session_name; # For the full-scale analysis. #
##my $toptopoutdir = "TESTs/Pre/Test3a"; ## For Pre-TESTING. ##
##my $toptopoutdir = "TESTs/Test1a"; ## For TESTING. ##

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
my $topoutdir = "$toptopoutdir/Outputs";

my $middir_pref = 'class_';
##my $middir_pref = 'hog_d'; # --- For the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'.
my $middir_suff = '';

my $out_pref = 'results_cmp.'; 
my $out_suff = '.txt'; 
=cut

##my $out_list = "$toptopoutdir/tbl_results_cmp_ancstates.txt"; # --- For the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
my $out_list = "$toptopoutdir/tbl_results_abs_prob_gpatterns.txt";

	## Temporary Input Files ##

my $intmp_dir = "$toptopoutdir/Temporary";

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
	# $in_msa = "$intmp_dir/$archive_pref_msa$id_class/$inaln_pref$lcl_msa_id$inaln_suff",
	# where $id_class specifies the class of simulated MSAs.

my $inaln_pref = 'dawg';
my $inaln_suff = '.aln';
=cut

	# $in_anc_gaps = "$intmp_dir/$archive_pref_g$id_class/$in_anc_gaps_pref$lcl_msa_id$in_anc_gaps_suff",
	# where $id_class specifies the class of simulated MSAs.

###my $in_anc_gaps_pref = 'output_brief.'; # --- For the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
my $in_anc_gaps_pref = 'output_detail.';
my $in_anc_gaps_suff = '.txt';


	### Check files/directories ###

my @dummy = ();
foreach my $file (@dummy) {
    unless (( -f $file ) and ( -r $file )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input file '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
}

foreach my $indir ($inarch_dir_g) {
##foreach my $indir ($inarch_dir_msa, $inarch_dir_g) {...} # --- For the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.

    unless (( -d $indir ) and ( -r $indir )) {
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input directory '$indir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
}

my $list_id_classes = list_sessions_ext ($inarch_dir_g, $archive_pref_g, $archive_suff_g, '-');
unless (defined $list_id_classes) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the archive directory '$inarch_dir_g' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
}


=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.

my $list_id_classes1 = list_sessions_ext ($inarch_dir_msa, $archive_pref_msa, $archive_suff_msa, '-');
unless (defined $list_id_classes1) {
	print STDOUT "[ERROR '$program_name'] 'list_sessions_ext' FAILED on the archive directory '$inarch_dir_msa' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
}
my ($common0, $ext0, $ext1) = diff_array (@{$list_id_classes}, @{$list_id_classes1});
unless (@{$ext0} ==0) {
##unless ((@{$ext0} ==0) and (@{$ext1}==0)) {...} # --- For the predecessor 'bchbch_anal_muts_on_sbjcts_in_bestMSA.pl'.
	print STDOUT "[ERROR '$program_name'] The results of 'list_sessions_ext' on the two different file systems CLASH each other !\n";
	print STDOUT "Entries with the ancestral gap states only: ", join (', ', @{$ext0}), " \;\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
}
unless (@{$ext1} == 0) {
	print STDOUT "[WARNING '$program_name'] The following class IDs are found in the MSA archives while not in the archives for Ancestral gap states:\n";
	print STDOUT "[WARNING '$program_name']  ", join (', ', @{$ext1}), " .\n\n";
}
=cut

## (@{$list_id_classes}>2) and (@{$list_id_classes} = @{$list_id_classes}[0,1]); ## For TESTING the script. ("Pre/Testxx", "Test1a,b") ##


foreach my $out_dir ($intmp_dir) {
##foreach my $out_dir ($intmp_dir, $topoutdir) {...} # --- For the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
    unless ( -d $out_dir) {
	print STDOUT "[WARNING '$program_name'] CanNOT FIND an output directory '$out_dir' ! ... I will try to create one...\n\n";
	unless (0 == system (" mkdir -p $out_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the output directory '$out_dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
    unless ( -w $out_dir ) {
	print STDOUT "[WARNING '$program_name'] The output directory '$out_dir' is NOT WRITABLE ! ... I will try to change its permission mode...\n\n";
	unless (0 == system (" chmod -R u+w $out_dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$out_dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
}


	### Declare the Beginning of the Script ###

print STDOUT "\n<<<<< Program name: $program_name (path: $program_path) >>>>>\n\n";

print STDOUT "Session Name: $session_name .\n\n";

print STDOUT '$FL_FULL_NODES = ', $FL_FULL_NODES, " .\n\n"; 

print STDOUT '{Character for a gap} = ', $GAP, "  ,\n";
print STDOUT '{Character for a residue} = ', $NGAP, " .\n\n";

print STDOUT '{Prefix for the ancestral sequences} = ', $PREF_ANC , " .\n\n";

print STDOUT '$CURR_WORK_DIR = ', $CURR_WORK_DIR, " .\n\n";

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
print STDOUT "Input archives for MSAs, including ancestral gap states:     $inarch_dir_msa/$archive_pref_msa", '{ID_class}', "$archive_suff_msa ,\n";
	# $in_msa = "$intmp_dir/$archive_pref_msa$chr_no/$inaln_pref$anchor_orf_id$inaln_suff",
=cut

	# MODIFIED in creation of 'bchbch_retrieve_abs_prob_gpatterns.pl'.
print STDOUT "Input archives for the absolute probabilities of gap-block-clusters:  $inarch_dir_g/$archive_pref_g", '{ID_class}', "$archive_suff_g ,\n";
=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
print STDOUT "Input archives for inferred ancestral gap states:  $inarch_dir_g/$archive_pref_g", '{ID_class}', "$archive_suff_g ,\n";
=cut
print STDOUT "          where {ID_class} = ", join (', ', @{$list_id_classes}), " .\n\n";


=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
print STDOUT "Temporary input MSAs: $intmp_dir/$archive_pref_msa", '{ID_class}', "/$inaln_pref", '{local_MSA_ID}', "$inaln_suff .\n\n";   
=cut

print STDOUT "Temporary input for gap states:  $intmp_dir/$archive_pref_g", '{ID_class}', "/$in_anc_gaps_pref", '{local_MSA_ID}', "$in_anc_gaps_suff .\n\n";

print STDOUT "Output files:\n";

	# MODIFIED in creation of 'bchbch_retrieve_abs_prob_gpatterns.pl'. #
print STDOUT "    (Summary table of the analyses)  $out_list .\n\n";

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
print STDOUT "    (Summary table of the analyses)  $out_list ,\n\n";
print STDOUT "    (Records of the results) $topoutdir/$middir_pref", '{ID_class}', "$middir_suff/$out_pref", '{local_MSA_ID}', "$out_suff .\n\n";
=cut


	### Main Process: Read MSA from an input file, 
	###           and extract ancestral gap states from another input file.

print STDOUT "\n<<<< Now, the main process begins ... >>>>\n\n\n";

my $fh_olist = open_output_file ($out_list, 1);
unless (defined $fh_olist) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$out_list' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

print $fh_olist join ("\t", 'ID_Class', 'Lcl_MSA_ID', '#{Ext_Seqs}', '#{Anc_Seqs}', 'MSA_Length', '#{gap_block_clusters}', 'delta_del', 'total_branch_length'), "\n"; # MODIFIED to adapt to the current convention on May 22 (Fri), 2015. #
##print $fh_olist join ("\t", 'ID_Class', 'Lcl_MSA_ID', '#{Ext_Seqs}', '#{Anc_Seqs}', 'MSA_Length', '#{gap_block_clusters}', 'delta_del_5p', 'total_branch_length'), "\n";

print $fh_olist join ("\t", 'Cl_no.(exclusive)', 'Seg_no.', 'lcl_cl_no.', 'Cl_Start', 'Cl_End', 'Cl_Length', '#{gap_blocks}', 'Min[Total#{indels}]', '#{parsimonious_scenarios}', 'Log(Abs_Prob[gap-pattern])', 'Gap_Pattern_Code', '[Notes(optional)]'), "\n\n";

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
print $fh_olist join ("\t", 'ID_Class', 'Lcl_MSA_ID', '#{Ext_Seqs}', '#{Anc_Seqs}', 'MSA_Length', '#{gap_block_clusters}', '#{All_clusters&other_blocks}'), "\n";
print $fh_olist join ("\t", 'Cl_no.(exclusive)', 'Cl_no.(overall)', 'Cl_Start', 'Cl_End', 'Cl_Length', '#{gap_blocks}', 'Min[Total#{indels}]', '#{parsimonious_scenarios}', '#{inferred_ancestral_gap_patterns}', 'Rel_Prob[correct_inf_anc_gp]', 'Total/relative/probabilities/of/inferred/ancestral/gap/patterns', '[Notes(optional)]'), "\n\n";
=cut


my $ct_classes = 0;
my $ct_missing_archs = 0;
my $totct_msas = my $totct_anom = my $totct_fail = my $totct_succ = 0;

OUTER:
foreach my $id_class (@{$list_id_classes}) { # Main outermost foreach-loop (over classes of MSAs). #

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
   (($id_class eq '2_4') or ($id_class eq '3_4')) and next;  # Skip these classes because they turned out to be redundant. #
    ## ($id_class eq '2_7') or next; ### For TESTING the script ("Test2a"). ###
    ## ($id_class eq '1_4') or next; ### For TESTING the script ("Test3a"). ###
=cut

    print STDOUT "<<< MSA class: $id_class >>>\n\n";

    $ct_classes++;

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
    my $in_archive_msa = join ('', $inarch_dir_msa, '/', $archive_pref_msa, $id_class, $archive_suff_msa);
=cut
    my $in_archive_g = join ('', $inarch_dir_g, '/', $archive_pref_g, $id_class, $archive_suff_g);

    foreach my $in_archive ($in_archive_g) {
    ##foreach my $in_archive ($in_archive_msa, $in_archive_g) {...} # --- For the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
	unless (( -f $in_archive ) and ( -r $in_archive ) ) {
	    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input archive file '$in_archive': $! !\n\n";
	    $ct_missing_archs++;
	    next OUTER;
	    ##print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable input archive file '$in_archive' !\n";
	    ##print STDERR "[ERROR '$program_name'] $!\n";
	    ##die;
	}
    }

   if ($TEST_LEVEL>1) {
	print STDOUT "[ Input Archives ]\n\n";
=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
	print STDOUT "  (Storing MSAs, including true ancestral gap states) $in_archive_msa ,\n";
	print STDOUT "  (Storing inferred ancestral gap states)  $in_archive_g .\n\n";
=cut
	print STDOUT "  (Storing inferred ancestral gap states)  $in_archive_g .\n\n";
   }

    ##my $mid_dir_msa = join ('', $archive_pref_msa, $id_class); # --- For the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
    my $mid_dir_g = join ('', $archive_pref_g, $id_class);

    my $list_lcl_msa_ids = peek_into_archive ($in_archive_g, $mid_dir_g, $in_anc_gaps_pref, $in_anc_gaps_suff);
    unless (defined $list_lcl_msa_ids) {
	print STDOUT "[ERROR '$program_name'] 'peek_into_archive' FAILED on the input archive '$in_archive_g' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
    my $list_lcl_msa_ids1 = peek_into_archive ($in_archive_msa, $mid_dir_msa, $inaln_pref, $inaln_suff);
    unless (defined $list_lcl_msa_ids1) {
	print STDOUT "[ERROR '$program_name'] 'peek_into_archive' FAILED on the input archive '$in_archive_msa' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
    my ($sb_common0, $sb_ext0, $sb_ext1) = diff_array (@{$list_lcl_msa_ids}, @{$list_lcl_msa_ids1});
    unless (@{$sb_ext0} ==0) {
    	print STDOUT "[ERROR '$program_name'] The results of 'list_sessions_ext' on the two different file archives CLASH each other !\n";
    	print STDOUT "Entries with the ancestral gap states only: ", join (', ', @{$sb_ext0}), " \;\n";
    	print STDERR "[ERROR '$program_name'] $!\n";
    	die;
    }
    unless (@{$sb_ext1} == 0) {
    	print STDOUT "[WARNING '$program_name'] The following MSA IDs are found in the MSA archive while not in the archive for Ancestral gap states:\n";
    	print STDOUT "[WARNING '$program_name']  ", join (', ', @{$sb_ext1}), " .\n\n";
    }
=cut

     ## (@{$list_lcl_msa_ids}>2) and (@{$list_lcl_msa_ids} = @{$list_lcl_msa_ids}[0,1]); ## For TESTING the script. ##
     ## (@{$list_lcl_msa_ids}>1) and (@{$list_lcl_msa_ids} = ($list_lcl_msa_ids->[0])); ## For TESTING the script. ("Pre/Test2a" & 'Test1a') ##


=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
		# Create a middle-level output directory. #

    my $midoutdir = join ('', $middir_pref, $id_class, $middir_suff);

    my $outdir = "$topoutdir/$midoutdir";
    unless ( -d $outdir) {
	unless (0 == system (" mkdir -p $outdir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the output directory '$outdir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
    unless ( -w $outdir ) {
	unless (0 == system (" chmod -R u+w $outdir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$outdir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
    }
=cut


    my $ct_msas = my $ct_anom = my $ct_fail = my $ct_succ = 0;

    foreach my $lcl_msa_id (@{$list_lcl_msa_ids}) { # Main outer foreach-loop (over MSAs in a class). #

	 ## (($lcl_msa_id eq '0004') or ($lcl_msa_id eq '0122')) or next; ### For TESTING the script ('Test2a'). ###
	## ($lcl_msa_id eq '0227') or next; # For TESTING the script ('Test3a'). ###

  (print STDOUT "<< Local MSA ID: $lcl_msa_id >>\n\n") if ($TEST_LEVEL>0);

	$ct_msas++;


=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
		## Extract the simulated MSA file and the record of inferred ancestral gap states. ##

	my $in_msa_name = join ('', $inaln_pref, $lcl_msa_id, $inaln_suff);
	my $extracted_file_msa = "$mid_dir_msa/$in_msa_name";
	my $tmpin_msa = "$intmp_dir/$extracted_file_msa";
=cut

		# $intmp_msa = "$intmp_dir/$archive_pref_msa$id_class/$inaln_pref$lcl_msa_id$inaln_suff",

	my $in_anc_gaps_name = join ('', $in_anc_gaps_pref, $lcl_msa_id, $in_anc_gaps_suff);
	my $extracted_file_g = "$mid_dir_g/$in_anc_gaps_name";
	my $tmpin_anc_gaps  = "$intmp_dir/$extracted_file_g";

		# $tmpin_anc_gaps = "$intmp_dir/$archive_pref_g$chr_no/$anchor_orf_id/$name_in_anc_gaps",

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
	foreach my $set_files (
		[$in_archive_msa, $extracted_file_msa, $tmpin_msa],
		[$in_archive_g, $extracted_file_g, $tmpin_anc_gaps]
			) {...}
=cut
	foreach my $set_files (
		[$in_archive_g, $extracted_file_g, $tmpin_anc_gaps]
			) {

		my ($in_archive, $extracted_file, $tmpin_file) = @{$set_files};

		unless (0 == system ("tar -C $intmp_dir -xpzf $in_archive $extracted_file") ) {
		    print STDOUT "[ERROR '$program_name'] FAILED to extract a file '$extracted_file' into '$intmp_dir' from the archive '$in_archive', for the local MSA '$lcl_msa_id' in the class '$id_class': $? !\n\n";
		    print STDERR "[ERROR '$program_name'] $!\n";
		    die;
		    ##$ct_fail++;
		    ##print $fh_olist "FAILED\n\n";
		    ##next OUTER;
		}
		unless (( -f $tmpin_file ) and ( -r $tmpin_file )) {
		    print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable temporary input file '$tmpin_file' : $! !\n\n";
		    print STDERR "[ERROR '$program_name'] $!\n";
		    die;
		    ##$ct_fail++;
		    ##print $fh_olist "FAILED\n\n";
		    ##next OUTER;
		}
	} # End of "foreach my $set_files (...) {...}"


=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
			# Retrieve the simulated MSA from the extracted input MSA file. #

    (print STDOUT "\n{ First, retrieve the simulated MSA ... }\n\n") if ($TEST_LEVEL>1);

	my ($msa_aln, $seq_names, $seqlens, $labels) = extract_clw ($tmpin_msa, 1); # The alignment is in the CLUSTAL format. #
	unless ((defined $msa_aln) and (@{$msa_aln}>0)) {
		print STDOUT "[CAUTION '$program_name'] 'extract_clw' FAILED to extract a simulated alignment from the file '$tmpin_msa': $! !\n\n";
		$ct_fail++;
		##print $fh_olist "FAILED\n\n";
		next;
	}

    (print STDOUT "\n ... Successfully retrieved the simulated alignment '$tmpin_msa' ...\n\n") if ($TEST_LEVEL>1);

	my $ct_seqs = @{$msa_aln};
	my $msa_len = length (${$msa_aln->[0]});


    (print STDOUT "\n#{sequences} = $ct_seqs .\n") if ($TEST_LEVEL>2);
    (print STDOUT "{MSA length}  = $msa_len .\n\n") if ($TEST_LEVEL>2);

    if ($TEST_LEVEL>5) {
	print STDOUT "\[\[ Input MSA \]\]\n\n";
	unless ( pr_ODEN_input (@{$seq_names}, @{$msa_aln}, 50, *STDOUT) ) {
	    print STDOUT "[ERROR '$program_name'] 'pr_ODEN_input' FAILED to print out the entire simulated MSA ...\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
	print STDOUT "\n\n";
    }

	my @ancseq_aln = my @ancseq_names = ();
	my @extseq_aln = my @extseq_names = ();
	for (my $k=0; $k<$ct_seqs; $k++) { # Middle preliminary for-loop (over aligned sequences). #
	    my $seqname = $seq_names->[$k];
	    if ($seqname =~ m/$REGX_PREF_ANC/) {
		push @ancseq_names, $seqname;
		push @ancseq_aln, $msa_aln->[$k];
	    } else {
		push @extseq_names, $seqname;
		push @extseq_aln, $msa_aln->[$k];
	    }
	}			# End of the middle preliminary for-loop (over aligned sequences). #

	my $ct_ancseqs = @ancseq_aln;
	my $ct_extseqs = @extseq_aln;

    if ($TEST_LEVEL>2) {
	print STDOUT "\n#{Ancestral sequences} = $ct_ancseqs .\n";
	print STDOUT "  (Names: ", join (', ', @ancseq_names), ")\n\n";
    }

    if ($TEST_LEVEL>3) {

	print STDOUT "#{Exterior sequences} = $ct_extseqs .\n";
	print STDOUT "  (Names: ", join (', ', @extseq_names), ")\n\n";

	print STDOUT "\n\[\[ Ancestral sequences embedded in the Input MSA \]\]\n\n";
	unless ( pr_ODEN_input (@ancseq_names, @ancseq_aln, 50, *STDOUT) ) {
	    print STDOUT "[ERROR '$program_name'] 'pr_ODEN_input' FAILED to print out the ancestral part of the simulated MSA ...\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
	print STDOUT "\n\n";
    }
=cut

		# Retrieve the absolute probabilities of gap-block clusters. #

    (print STDOUT "\n{ Retrieve the absolute probabilities of gap-block clusters  ... }\n\n") if ($TEST_LEVEL>1);

		# [ @parameters = ($delta_del_5p, \@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with ] -- OBSOLETE as of May 22 (Fri), 2015. #
		# @parameters = ($delta_del, \@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with
		# @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr, \%fname2id_intnode, $ftopnode_name) for the full MSA,
		# @seqids storing the sequence IDs,
		# $ct_segments = scalar (@info_segments) = scalar (@indel_state_distributions);
		#
		# @{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_name), with 
		# $seg_start/end in the 0-based half-open convention,
		# $red_topnode_name is the name of the top-node (root) of the tree reduced according to the usability pattern
		#  ($red_topnode_name = undef for the segment skipped to infer ancestral states) ;
		#
		# @info_cls = @{$setset_info_cls[$s]->[$cl]} 
		# stores information on the $cl th cluster in the $s th segment. Here,
		#
		# @info_cls = ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, $ct_scenarios, $log_abs_prob), where
		#
		# @{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block,
		#
		# $ct_scenarios is the number of (parsimoinious) indel scenarios that result in the gap-block cluster, and
		# $log_abs_prob is the (logarithmic) absolute probability of the gap-block cluster.
		# 
	my ($parameters, $info_segments, $setset_info_cls)
	    	= br_parse_detail_output_infer_psm_indel_events_dawg ($tmpin_anc_gaps);
	unless (defined $info_segments) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'br_parse_detail_output_infer_psm_indel_events' FAILED on the input file '$tmpin_anc_gaps' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	my ($delta_del, $full_tree, $ext_seq_names2, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments) = @{$parameters};
	##my ($delta_del_5p, $full_tree, $ext_seq_names2, $msa_len2, $tgt_seqid, $indx_tgt, $ct_segments) = @{$parameters};
	my ($topnode_id, $node2pa, $node2ch, $name2id, $id2name, $br_attr, $name2id_intnode, $topnode_name) = @{$full_tree};

	my $ct_extseqs = @{$ext_seq_names2};
	my $ct_ancseqs = scalar (keys %{$node2pa}) + 1 - $ct_extseqs;

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
		# Retrieve the inferred gap states. #

    (print STDOUT "\n{ Third, retrieve the inferred ancestral gap states ... }\n\n") if ($TEST_LEVEL>1);

		# @{$parameters} = (\@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with
		# @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr, \%fname2id_intnode, $topnode_name) for the full MSA,
		# @seqids storing the sequence IDs;
		#
		# @{$info_segments->[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_id), with 
		# $seg_start/end in the 0-based half-open convention,
		# $red_topnode_id is the ID of the top-node (root) of the tree reduced according to the usability pattern
		#  ($red_topnode_id = undef for the segment skipped to infer ancestral states) ;
		#
		# @{$sw_indel_state_distributions[2*$cl]} = ($cl_start, $cl_end, $state)
		# for the gapless block left-flanking the $cl th gap-block cluster
		# (or on the right-end of the segment (with $cl = #{clusters})),
		# or (always with $cl = 0) for the skipped segment due to inadquate information.
		# $state	= 'X' for the gapless block,
		# 	= '?' for the skipped segment due to inadequate information.
		#
		# @{$sw_indel_state_distributions[2*$cl + 1]}
		# = ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution) 
		# is for the $cl th gap-block cluster, where
		# @{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block, and
		# @{$scenario_distribution[$ps]} = ($rel_prob, \%node2bw_gap_states, \%br2indel_events)
		#  is for the $ps th parsimonious indel scenario, with
		# %node2bw_gap_states = ($node_name => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and
		# %br2indel_events = ($branch_le_name => \@events_on_branch, ...) only for branches experiencing indel events, with
		# @{$events_on_branch[$e]} = ($state_upper_end, $state_lower_end, \@involved_blocks) for the $e th event along the branch $branch_id.

	my ($parameters, $info_segments, $indel_state_distrs)
	    	= parse_brief_output_infer_psm_indel_events_dawg ($tmpin_anc_gaps, $FL_FULL_NODES) ;
	unless (defined $info_segments) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'parse_brief_output_infer_psm_indel_events' FAILED on the input file '$tmpin_anc_gaps' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	my ($full_tree, $ext_seq_names2, $msa_len2, $tgt_seqid, $indx_tgt, $ct_segments) = @{$parameters};
	my ($topnode_id, $node2pa, $node2ch, $name2id, $id2name, $br_attr, $name2id_intnode, $topnode_name) = @{$full_tree};
	    ##my $ct_seqs = @{$seqids};
	unless (@{$ext_seq_names2} == $ct_extseqs) {
		my $ct_extseqs2 = @{$ext_seq_names2};
		print STDOUT "[ERROR '$program_name'] The number (=$ct_extseqs2) of exterior sequence IDs, ", join (', ', @{$ext_seq_names2}), ", extracted from the file '$tmpin_anc_gaps' does NOT MATCH the number of exterior sequences (=$ct_extseqs) in the input MSA: $!\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}
	unless ($msa_len == $msa_len2) {
		print STDOUT "[ERROR '$program_name'] The number (=$msa_len2) of MSA columns extracted from the file '$tmpin_anc_gaps' does NOT MATCH the actual number (=$msa_len) of columns in the input MSA: $!\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

    (print STDOUT "\n{ Then, convert the extracted gap state distribution from the dual-hierarchical to the linear structure... }\n\n") if ($TEST_LEVEL>1);

	my $linear_indel_state_distr = linear_order_indel_state_distributions (@{$indel_state_distrs});
	unless (defined $linear_indel_state_distr) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'linear_order_indel_state_distributions' FAILED on the MSA '$lcl_msa_id' in the class '$id_class' !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}
	my $totct_cls = @{$linear_indel_state_distr};

    (print STDOUT "#{clusters (including gapless blocks)} = $totct_cls .\n\n") if ($TEST_LEVEL>2);
=cut

    if ($TEST_LEVEL>3) {

	print STDOUT "\[\[ Extracted information concerning the absolute probabilities of gap-block clusters \]\]\n\n";
	##print STDOUT "\[\[ Extracted information concerning the ancestral gap states \]\]\n\n"; # --- For the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.

	print STDOUT '$delta_del = ', $delta_del, " .\n\n"; # MODIFIED to adapt to the current convention on May 22 (Fri), 2015. #
	##print STDOUT '$delta_del_5p = ', $delta_del_5p, " .\n\n";
	
	print STDOUT "\[ Input Tree (top_node_name = $topnode_name) \]\n\n";
	
	unless ( br_output_tree_dawg ($topnode_id, %{$node2ch}, %{$id2name},  %{$br_attr}, $topnode_name, *STDOUT, 0, 0) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'br_output_tree_dawg' FAILED !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}
	print STDOUT "\n\n";
	
	print STDOUT 'Exterior sequences in MSA (#{seqs} = ', $ct_extseqs, ') : ', join (', ', @{$ext_seq_names2}), " .\n\n";
	
	print STDOUT "#{columns in MSA} = $msa_len .\n\n";
	##print STDOUT "#{columns in MSA} = $msa_len2 .\n\n";
	
	print STDOUT "\{Index for the target sequence '$tgt_seqid'\} = $indx_tgt .\n\n";
	
	print STDOUT "\{ Segment Information (#{MSA Segments} = $ct_segments) \}\n\n";

	print STDOUT join ("\t", 'Indx', 'Start', 'End', 'Length', 'Usability_pattern', 'top_node_(reduced_tree)'), "\n";

	for (my $s=0; $s<$ct_segments; $s++) { # Test outer for-loop (over MSA segments). #
	
		my ($seg_start, $seg_end, $us_pattern, $red_topnode_id) = @{$info_segments->[$s]};
		my $seg_len = $seg_end - $seg_start;
	
		    # @{$info_segments->[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_id), with 
		    # $seg_start/end in the 0-based half-open convention,
		    # $red_topnode_id is the ID of the top-node (root) of the tree reduced according to the usability pattern
		    #  ($red_topnode_id = undef for the segment skipped to infer ancestral states) ;

		print STDOUT join ("\t", $s, $seg_start, $seg_end, $seg_len, $us_pattern, $red_topnode_id), "\n";

	}			# End of the test outer for-loop (over MSA segments). #

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
	print STDOUT "\n\n\{ Back-bone of the linear-order distribution #{clusters (including gapless blocks)} = $totct_cls) \}\n\n";
	
	print STDOUT join ("\t", 'Indx_cls', 'Indx_seg', 'Cl_start', 'Cl_end', 'State'), "\n\n";
	
	for (my $cl=0; $cl<$totct_cls; $cl++) { # 2nd test for-loop (over gap-block clusters and gapless blocks) #
	    my $cw_indel_state_distr = $linear_indel_state_distr->[$cl];
	    print STDOUT join ("\t", $cl, @{$cw_indel_state_distr}[0..3]), "\n";
	}			# End of the 2nd test for-loop (over gap-block clusters and gapless blocks) #
=cut

	print STDOUT "\n\n";

    }  # END of "if ($TEST_LEVEL>3) {...}".

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
		# Open an output file and record some input parameters. #

	my $outfile = "$outdir/$out_pref$lcl_msa_id$out_suff";
	my $fh_out = open_output_file ($outfile, 0);
	unless (defined $fh_out) {
	    print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile' !\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die ;
	}

	print $fh_out "<<<< Input Parameters >>>>\n\n";
	print $fh_out "\[\[ Input Tree (top_node_name = $topnode_name) \]\]\n\n";
	
	unless ( br_output_tree_dawg ($topnode_id, %{$node2ch}, %{$id2name},  %{$br_attr}, $topnode_name, *${$fh_out}, 0, 0) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'br_output_tree_dawg' FAILED !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}
	print $fh_out "\n\n";
	print $fh_out "\[\[ Input MSA (#{exterior sequences} = $ct_extseqs, #{ancestral sequences} = $ct_ancseqs) \]\]\n\n";
	print $fh_out "Exterior sequences:  ", join (', ', @extseq_names), " .\n\n";
	print $fh_out "Ancestral sequences: ", join (', ', @ancseq_names), " .\n\n";

	print $fh_out "#{columns in MSA} = $msa_len .\n\n";

	print $fh_out "#{MSA Segments} = $ct_segments .\n\n";

	my $net_ct_cls = 0;
	for (my $cl=0; $cl<$totct_cls; $cl++) { ($linear_indel_state_distr->[$cl][3] eq 'X/-') and ($net_ct_cls++); }

	my $ct_others = $totct_cls - $net_ct_cls;

	print $fh_out "#{Gap-block clusters} = $net_ct_cls .\n\n";
	print $fh_out "#{Gapless blocks and unusable MSA blocks} = $ct_others .\n\n\n";
=cut

			# Calculate the total branch length. #
	my $tot_blen = 0;
	foreach my $br (keys %{$br_attr}) { $tot_blen += $br_attr->{$br}[$INDX_LENGTH]; }

			# Calculate the total number of gap-block clusters. #
	my $net_totct_cls = 0;
	for (my $s=0; $s<$ct_segments; $s++) { $net_totct_cls += @{$setset_info_cls->[$s]}; }

    if ($TEST_LEVEL>2) {
	print STDOUT "[ Derived properties ]\n\n";
	print STDOUT "\n\n{Total Branch Length} = $tot_blen ,\n";
	print STDOUT "Total#{gap-block clusters} = $net_totct_cls .\n\n\n";
    }

			## Record summary information of the MSA into the output list. ##

		# MODIFIED in creation of 'bchbch_retrieve_abs_prob_gpatterns.pl'. #
		# => Further MODIFIED to adapt to the current convention (on May 22 (Fri), 2015). #
	print $fh_olist join ("\t", $id_class, $lcl_msa_id, $ct_extseqs, $ct_ancseqs, $msa_len, $net_totct_cls, $delta_del, $tot_blen), "\n";
	##print $fh_olist join ("\t", $id_class, $lcl_msa_id, $ct_extseqs, $ct_ancseqs, $msa_len, $net_totct_cls, $delta_del_5p, $tot_blen), "\n";
		# print $fh_olist join ("\t", 'ID_Class', 'Lcl_MSA_ID', '#{Ext_Seqs}', '#{Anc_Seqs}', 'MSA_Length', '#{gap_block_clusters}', 'delta_del_5p', 'total_branch_length'), "\n";

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
	print $fh_olist join ("\t", $id_class, $lcl_msa_id, $ct_extseqs, $ct_ancseqs, $msa_len, $net_ct_cls, $totct_cls), "\n";
		# print $fh_olist join ("\t", 'ID_Class', 'Lcl_MSA_ID', '#{Ext_Seqs}', '#{Anc_Seqs}', 'MSA_Length', '#{gap_block_clusters}', '#{All_clusters&other_blocks}'), "\n";
=cut

			## Record absolute probabilities and other basic attributes of the gap-block clusters. ##
			# (ADDED in creation of 'bchbch_retrieve_abs_prob_gpatterns.pl'.) #

    (print STDOUT "\n{{ MAIN PROCESS: Record absolute probabilities and other baic attributes of Gap-BLock Clusters... }}\n\n") if ($TEST_LEVEL>1);

	my $null_gblock = $GAP x $ct_extseqs;

	my $serial_cl_no = 0;
	for (my $s=0; $s<$ct_segments; $s++) { # Main middle for-loop (over MSA segments). #

	    my $set_cls_in_seg = $setset_info_cls->[$s];
	    my $ct_cls = @{$set_cls_in_seg};

	    for (my $cl = 0; $cl<$ct_cls; $cl++) { # Main inner for-loop (over gap-block clusters). #

		my ($cl_start, $cl_end, $state, $gblocks, $min_totct_indels, $ct_psm_scns, $log_abs_prob) = @{$set_cls_in_seg->[$cl]};
		#
		# @info_cls = ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, $ct_scenarios, $log_abs_prob), where
		#

		my $ct_gblks = @{$gblocks};
		my $gp_code = '';
		for (my $gb=0; $gb<$ct_gblks; $gb++) { # Inner-inner for-loop (over gap-blocks). #

		    my ($bl_start, $bl_end, $indx_gp, $gpattern) = @{$gblocks->[$gb]};
			#
			# @{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block,

		    $gp_code .= sprintf (",%d%s", $bl_end - $bl_start, $gpattern);

		}			# End of the inner-inner for-loop (over gap-blocks). #
		$gp_code =~ s/^\,//;


			## Assign a note, if necessary. ## 

		my $note;

		if (!(defined $min_totct_indels) ) { # ADDED in creation of 'bchbch_retrieve_abs_prob_gpatterns.pl' on Mar 13 (Thu), 2014. #
		    $min_totct_indels = 'undef';

		    if (!(defined $ct_psm_scns) ) {

			$ct_psm_scns = 'undef';

    (print STDOUT "\n", '$min_totct_indels AND $ct_psm_scns are UNDEFINED', " for the cluster $serial_cl_no (segment ID = $s, local cluster no. = $cl) !\n\n") if ($TEST_LEVEL>1);

			$note = '[Unexpected Errors 1&2]';

		    } else {

    (print STDOUT "\n", '$min_totct_indels is UNDEFINED', " for the cluster $serial_cl_no (segment ID = $s, local cluster no. = $cl) !\n\n") if ($TEST_LEVEL>1);

			$note = '[Unexpected Error 1]';
		    }

		} elsif (!(defined $ct_psm_scns) ) { # ADDED in creation of 'bchbch_retrieve_abs_prob_gpatterns.pl' on Mar 13 (Thu), 2014. #
		    $ct_psm_scns = 'undef';

    (print STDOUT "\n", '$ct_psm_scns is UNDEFINED', " for the cluster $serial_cl_no (segment ID = $s, local cluster no. = $cl) !\n\n") if ($TEST_LEVEL>1);

		    $note = '[Unexpected Error 2]';


		} elsif ($min_totct_indels eq 'N/A') {

			# Successor of "Preliminary 0: If $min_totct_indels = 'N/A', skip the further analysis of the cluster." #
	    
    (print STDOUT "\nMin_Total#{indels} = 'N/A' for the cluster $serial_cl_no (segment ID = $s, local cluster no. = $cl) !\n\n") if ($TEST_LEVEL>1);

		    $note = '[Cluster itself or some gaps too long]';

		} elsif ($ct_psm_scns eq 'N/A') { # ADDED in creation of 'bchbch_retrieve_abs_prob_gpatterns.pl'. 

    (print STDOUT "\n#{Parsimonious Scenarios} = 'N/A' for the cluster $serial_cl_no (segment ID = $s, local cluster no. = $cl) !\n\n") if ($TEST_LEVEL>1);

		    $note = '[Cluster itself or some gaps too long (II)]';

		} elsif ($log_abs_prob eq 'N/A') {

			# Successor of "Preliminary 1: Briefly examine if any of the scenario has the relative probability = 'N/A'." #
    (print STDOUT "\nLog(Abs_Prob[gap-pattern]) = 'N/A' for the cluster $serial_cl_no (segment ID = $s, local cluster no. = $cl) !\n\n") if ($TEST_LEVEL>1);

		    $note = '[Absolute Probability Not Available]';

		} elsif ($log_abs_prob eq 'undef') {

			# Successor of "Preliminary 1': Briefly examine if any of the scenario has the relative probability = 'undef'." #

    (print STDOUT "\nLog(Abs_Prob[gap-pattern]) = 'undef' for the cluster $serial_cl_no (segment ID = $s, local cluster no. = $cl) !\n\n") if ($TEST_LEVEL>1);

		    $note = "[FAILURE of the subroutine 'cal_lnlk1_indel_scn_cls']";

=Unfortunately, this cannot be done in this script, 'bchbch_retrieve_abs_prob_gpatterns.pl'.

		# Preliminary 2: Examine if the sites flanking the gap-block cluster contains gaps or nor. #
		# (ADDED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'.) #

	    my $ct_ancgaps_lflank = my $ct_ancgaps_rflank  = 0;
	    if (length (${$ancmsa_lflank[0]}) > 0) {
		for (my $k=0; $k<$ct_ancseqs; $k++) { ($GAP eq substr (${$ancmsa_lflank[$k]}, -1, 1) ) and ($ct_ancgaps_lflank++); }
	    }
	    if ($cl_end < $msa_len) {
		for (my $k=0; $k<$ct_ancseqs; $k++) { ($GAP eq substr (${$ancseq_aln[$k]}, 0, 1) ) and ($ct_ancgaps_rflank++) ; }
	    }

	    if (0 < $ct_ancgaps_lflank + $ct_ancgaps_rflank) {

    (print STDOUT "\n#{ancestral gaps in flanking sites} = $ct_ancgaps_lflank (left), $ct_ancgaps_rflank (right).\n\n") if ($TEST_LEVEL>1);

		print $fh_out "\n[INFORMATION] GAPS WERE FOUND IN THE FLANKING SITES of the ancestral sequences (#{flanking gaps} = $ct_ancgaps_lflank (left), $ct_ancgaps_rflank (right)) !!\n\n\n";

	    print $fh_olist join ("\t", $num_cls, $cl, $cl_start, $cl_end, $cl_len, $ct_gblks, $min_totct_indels, $ct_psm_scns, 'N/A', 'N/A', 'N/A', "[Inconsistent Ancestral Flanking Sites]"), "\n";

		next;
	    }

=cut # END of "Unfortunately, this cannot be done in this script, 'bchbch_retrieve_abs_prob_gpatterns.pl'."

		} elsif ($ct_psm_scns == 0) {

		# Successor of "Preliminary 3: Examine if the scenario distribution is empty." #

    (print STDOUT "\nThe Scenario Distribution is EMPTY for the cluster $serial_cl_no (segment ID = $s, local cluster no. = $cl) !\n\n") if ($TEST_LEVEL>1);

		     $note = '[EMPTY Scenario Distribution]';

		} elsif (($ct_gblks == 1) and ($gblocks->[0][3] eq $null_gblock)) { # ADDED in creation of 'bchbch_retrieve_abs_prob_gpatterns.pl'. #

    (print STDOUT "\nThe cluster is NULL (i.e., it consists only of gaps) for the cluster $serial_cl_no (segment ID = $s, local cluster no. = $cl) !\n\n") if ($TEST_LEVEL>1);

		    $note = '[Null Cluster]';

		} # END of "if (...) {...} elsif (...) {...} ... elsif (...) {...}"

			# Finally, output the summary description of the clustrer. #

		my @output = ($serial_cl_no, $s, $cl, $cl_start, $cl_end, $cl_end - $cl_start, $ct_gblks, $min_totct_indels, $ct_psm_scns, $log_abs_prob, $gp_code);
			#
			# print $fh_olist join ("\t", 'Cl_no.(exclusive)', 'Seg_no.', 'lcl_cl_no.', 'Cl_Start', 'Cl_End', 'Cl_Length', '#{gap_blocks}', 'Min[Total#{indels}]', '#{parsimonious_scenarios}', 'Log(Abs_Prob[gap-pattern])', 'Gap_Pattern_Code', '[Notes(optional)]'), "\n\n";

		(defined $note) and (push @output, $note);

		print $fh_olist join ("\t", @output), "\n";

		$serial_cl_no++;
	    }				# End of the main inner for-loop (over gap-block clusters). #

	}			# End of the main middle for-loop (over MSA segments). #
	print $fh_olist "\n";

			### RESTART FROM HERE ###


=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
		###  MAIN PROCESS: Compare the True ancestral gap states with the Inferred ancestral gap states. ###

	print $fh_out "<<<< Comparing the True vs Inferred ancestral gap states in each of the $net_ct_cls gap-block clusters >>>>\n\n";
	
   (print STDOUT "\n[[ Now, Compare the True & Inferred ancestral gap states in each of the gap-block clusters (#{clusters} = $net_ct_cls) ... ]]\n\n") if ($TEST_LEVEL>1);


	##print STDOUT "\{\{ Details on the ancestral gap states \}\}\n\n"; # For the predecessor 'bchbch_anal_muts_on_sbjcts_in_bestMSA.pl'. #


	my $last_cl_end = 0; # End coordinate (0-based half-open) of the last cluster. #

	my $num_cls = -1;

	for (my $cl=0; $cl<$totct_cls; $cl++) { # Main middle for-loop (over gap-block clusters and gapless blocks) #
	    my $cw_indel_state_distr = $linear_indel_state_distr->[$cl];

	    my ($indx_seg, $cl_start, $cl_end, $state) = @{$cw_indel_state_distr}[0..3];
	    ($state eq 'X/-') or next;

	    $num_cls++;


		# Chop off MSAs of ancestral sequences in the gap-block cluster and in its left-flanking gapless block. #
		# (ADDED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'.)

	    my @ancmsa_lflank = ();
	    my @ancmsa_gbcluster = ();
	    for (my $k=0; $k<$ct_ancseqs; $k++) { # 1st inner for-loop (over ancestral sequences). #

		my $leftflank = ($last_cl_end < $cl_start) ? substr (${$ancseq_aln[$k]}, 0, $cl_start - $last_cl_end, '') : '';
		##my $leftflank = ($last_cl_end < $cl_start) ? substr (${$ancseq_aln[$k]}, 0, $cl_start - $cl_end, '') : ''; # A BUG fixed on Mar 5 (Wed), 2014. #

		my $rlv_seq = ($cl_start < $cl_end) ? substr (${$ancseq_aln[$k]}, 0, $cl_end - $cl_start, '') : '';
		push @ancmsa_lflank, \$leftflank;
		push @ancmsa_gbcluster, $rlv_seq;
	    }				# End of the 1st inner for-loop (over ancestral sequences). #

	    $last_cl_end = $cl_end;


	    my ($gap_blocks, $min_totct_indels, $scenario_distribution)
		= @{$cw_indel_state_distr}[4..6];

		# @{$sw_indel_state_distributions[2*$cl + 1]}
		# = ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution) 
		# is for the $cl th gap-block cluster, where

	    my $cl_len = $cl_end - $cl_start;
	    my $ct_gblks = @{$gap_blocks};

		# MODIFIED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2.pl' on Mar 11 (Tue), 2014. #
	    my $ct_psm_scns = ($min_totct_indels eq 'N/A') ? 'N/A' : @{$scenario_distribution};
	    ##my $ct_psm_scns = @{$scenario_distribution}; # MOVED from below. #

    (print STDOUT "\< The $num_cls th cluster (with overall-number=$cl) (, which belongs to the $indx_seg th segment); region = \[$cl_start\, $cl_end\) ; length = $cl_len ; #{gap-blocks} = $ct_gblks \>\n\n") if ($TEST_LEVEL>2);

	    print $fh_out "\< The $num_cls th cluster (overall-number=$cl) \; region = \[$cl_start\, $cl_end\) \; length = $cl_len \; #{gap-blocks} = $ct_gblks \>\n\n";
	    print $fh_out '[ Gap-blocks in the cluster ]', "\n\n";
	    print $fh_out join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', 'gap_pattern'), "\n\n";

    if ($TEST_LEVEL>=5) {
	print STDOUT '[ Gap-blocks in the cluster ]', "\n\n";
	print STDOUT join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', 'gap_pattern'), "\n\n";
    }

	    my @gblk_lengths = (); # ADDED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'. #

	    for (my $g=0; $g<$ct_gblks; $g++) { # 2nd inner for-loop (over gap-blocks) #

		my ($bl_start, $bl_end, $indx_gp, $gpattern) = @{$gap_blocks->[$g]};
		my $bl_len = $bl_end - $bl_start;

		$gblk_lengths[$g] = $bl_len; # ADDED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'. #

		    # @{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block, and

    		print $fh_out join ("\t", $g, $bl_len, $bl_start, $bl_end, $indx_gp, $gpattern), "\n";

    (print STDOUT join ("\t", $g, $bl_len, $bl_start, $bl_end, $indx_gp, $gpattern), "\n") if ($TEST_LEVEL>=5);

	    }				# End of the 2nd inner for-loop (over gap-blocks) #


		## ADDED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'. ##

    (print STDOUT "\n{Gap-block lengths} = (", join (', ', @gblk_lengths), ") .\n\n") if ($TEST_LEVEL>3);

    if ($TEST_LEVEL>5) {
	print STDOUT "[ Ancestral Sequences in the Left-flanking Region ]\n\n";
	unless ( pr_ODEN_input (@ancseq_names, @ancmsa_lflank, 50, *STDOUT) ) {
	    print STDOUT "[ERROR '$program_name'] 'pr_ODEN_input' FAILED to print out the ancestral MSA in the region left-flanking the $cl th cluster ...\n";
	    print STDERR "[ERROR '$program_name'] $!\n";
	    die;
	}
	print STDOUT "\n\n";
    }
    if ($TEST_LEVEL>=5) {
	print STDOUT "[ Ancestral Sequences in the Gap-Block Cluster in question ]\n\n";
	for (my $k=0; $k<$ct_ancseqs; $k++) { print STDOUT sprintf ("%-20s", $ancseq_names[$k]), $ancmsa_gbcluster[$k], "\n"; }
	print STDOUT "\n\n";
    }

		# Preliminary 0: If $min_totct_indels = 'N/A', skip the further analysis of the cluster. #
	    if ($min_totct_indels eq 'N/A') {

    (print STDOUT "\nMin_Total#{indels} = 'N/A' !\n\n") if ($TEST_LEVEL>2);

		print $fh_out "\n[INFORMATION] 'N/A's are assigned as the minimum total number of indels !!\n\n\n";

		print $fh_olist join ("\t", $num_cls, $cl, $cl_start, $cl_end, $cl_len, $ct_gblks, $min_totct_indels, $ct_psm_scns, 'N/A', 'N/A', 'N/A', "[Cluster itself or some gaps too long]"), "\n";

		next;
	    }

		# Preliminary 1: Briefly examine if any of the scenario has the relative probability = 'N/A'. #
		# (ADDED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'.) #

	    my $ct_na_relprobs = 0;
	    foreach my $scn_info (@{$scenario_distribution}) {($scn_info->[0] eq 'N/A') and ($ct_na_relprobs++); }

	    if ($ct_na_relprobs>0) {

    (print STDOUT "\n#{Scenarios with relative probabilities being 'N/A'} = $ct_na_relprobs !\n\n") if ($TEST_LEVEL>2);

		print $fh_out "\n[INFORMATION] 'N/A's are assigned as the relative probabilities of $ct_na_relprobs out of $ct_psm_scns parsimonious scenarios !!\n\n\n";

			# MODIFIED on March 6 (Thu), 2014. #
		print $fh_olist join ("\t", $num_cls, $cl, $cl_start, $cl_end, $cl_len, $ct_gblks, $min_totct_indels, $ct_psm_scns, 'N/A', 'N/A', 'N/A', "[Some Indel Events Are Too Long]"), "\n";
		##print $fh_olist join ("\t", $num_cls, $cl, $cl_start, $cl_end, $cl_len, $ct_gblks, $min_totct_indels, $ct_psm_scns, 'N/A', 'N/A', 'N/A', "[Relative Probabilities Not Available]"), "\n";

		next; 
	    }

		# Preliminary 1': Briefly examine if any of the scenario has the relative probability = 'undef'. #
		# (ADDED in revision of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl' on Mar 6 (Thu), 2014.) #
	    my $ct_undef_relprobs = 0;
	    foreach my $scn_info (@{$scenario_distribution}) {($scn_info->[0] eq 'undef') and ($ct_undef_relprobs++); }

	    if ($ct_undef_relprobs>0) {
	    ##if ($ct_na_relprobs>0) {...} # A BUG fixed on Mar 6 (Thu), 2014.#

    (print STDOUT "\n#{Scenarios with relative probabilities being 'undef'} = $ct_undef_relprobs !\n\n") if ($TEST_LEVEL>2);

		print $fh_out "\n[INFORMATION] 'undef's are assigned as the relative probabilities of $ct_undef_relprobs out of $ct_psm_scns parsimonious scenarios !!\n\n\n";

		print $fh_olist join ("\t", $num_cls, $cl, $cl_start, $cl_end, $cl_len, $ct_gblks, $min_totct_indels, $ct_psm_scns, 'undef', 'undef', 'undef', "[FAILURE of the subroutine 'cal_lnlk1_indel_scn_cls']"), "\n";

		next; 
	    }
	    

		# Preliminary 2: Examine if the sites flanking the gap-block cluster contains gaps or nor. #
		# (ADDED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'.) #

	    my $ct_ancgaps_lflank = my $ct_ancgaps_rflank  = 0;
	    if (length (${$ancmsa_lflank[0]}) > 0) {
		for (my $k=0; $k<$ct_ancseqs; $k++) { ($GAP eq substr (${$ancmsa_lflank[$k]}, -1, 1) ) and ($ct_ancgaps_lflank++); }
	    }
	    if ($cl_end < $msa_len) {
		for (my $k=0; $k<$ct_ancseqs; $k++) { ($GAP eq substr (${$ancseq_aln[$k]}, 0, 1) ) and ($ct_ancgaps_rflank++) ; }
	    }

	    if (0 < $ct_ancgaps_lflank + $ct_ancgaps_rflank) {

    (print STDOUT "\n#{ancestral gaps in flanking sites} = $ct_ancgaps_lflank (left), $ct_ancgaps_rflank (right).\n\n") if ($TEST_LEVEL>2);

		print $fh_out "\n[INFORMATION] GAPS WERE FOUND IN THE FLANKING SITES of the ancestral sequences (#{flanking gaps} = $ct_ancgaps_lflank (left), $ct_ancgaps_rflank (right)) !!\n\n\n";

	    print $fh_olist join ("\t", $num_cls, $cl, $cl_start, $cl_end, $cl_len, $ct_gblks, $min_totct_indels, $ct_psm_scns, 'N/A', 'N/A', 'N/A', "[Inconsistent Ancestral Flanking Sites]"), "\n";

		next;
	    }

		# Preliminary 3: Examine if the scenario distribution is empty. #
		# (ADDED in revision of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl' on Mar 6 (Thu), 2014.)

	    if ($ct_psm_scns == 0) {

    (print STDOUT "\nThe Scenario Distribution is EMPTY !\n\n") if ($TEST_LEVEL>2);

		print $fh_out "\n[INFORMATION] The Scenario Distribution is EMPTY !!\n\n\n";

		print $fh_olist join ("\t", $num_cls, $cl, $cl_start, $cl_end, $cl_len, $ct_gblks, $min_totct_indels, $ct_psm_scns, 'undef', 'undef', 'undef', "[EMPTY Scenario Distribution]"), "\n";

		next;

	    }


		# Main 1: Prepare the true ancestral gap-states in this gap-block cluster. #
		# (ADDED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'.) #

	    my @true_anc_gp = ();
	    for (my $k=0; $k<$ct_ancseqs; $k++) { # 3rd inner for-loop (over ancestral sequences). #

		my $ancseq = $ancmsa_gbcluster[$k];
		##my $ancseq = ${$ancmsa_gbcluster[$k]}; # A BUG fixed on Mar 5 (Wed), 2014.

		my $len = length $ancseq;
		my $gpattern = '';
		for (my $l=0; $l<$len; $l++) { $gpattern .= ($GAP eq substr ($ancseq, 0, 1, '')) ? $GAP : $NGAP; }
		push @true_anc_gp, $gpattern;
	    }				# End of the 3rd inner for-loop (over ancestral sequences). #


	    print $fh_out "\n\n[ TRUE Ancestral Gap-Patterns ]\n\n";
	    for (my $k=0; $k<$ct_ancseqs; $k++) { print $fh_out sprintf ("%-20s", $ancseq_names[$k]), $true_anc_gp[$k], "\n"; }
	    print $fh_out "\n\n";

    if ($TEST_LEVEL>3) {
	print STDOUT "[ TRUE Ancestral Gap-Patterns ]\n\n";
	for (my $k=0; $k<$ct_ancseqs; $k++) { print STDOUT sprintf ("%-20s", $ancseq_names[$k]), $true_anc_gp[$k], "\n"; }
	print STDOUT "\n\n";
    }

	    my $cnct_true_anc_gp = join (':', @true_anc_gp);

		### RESTART FROM HERE ###

		# Main 2: Examine the Inferred Ancestral Gap States (in the parsimonious indel scenarios). #
		# (MODIFIED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'.) #

	    my %inf_anc_gp2info; # ($cnct_inf_anc_gp => [$tot_rel_prob, \@list_psm_scns], ...)

	    ##my $ct_psm_scns = @{$scenario_distribution}; # MOVED to above (on Mar 5 (Wed), 2014). #

	    print $fh_out "\n\[ Parsimonious Scenarios: \(Min_Total#{indel_events} = $min_totct_indels; #{scenarios} = $ct_psm_scns\) \]\n\n";

    (print STDOUT "\n\n\[ Parsimonious Scenarios \(Min_Total#{indel_events} = $min_totct_indels; #{scenarios} = $ct_psm_scns\) \]\n\n") if ($TEST_LEVEL>3);

	    for (my $ps = 0; $ps < $ct_psm_scns; $ps++) { # 5th inner for-loop (over parsimonious scenarios). #

		my ($rel_prob, $node2bw_gap_states, $br2indel_events) = @{$scenario_distribution->[$ps]}; 

		    # @{$scenario_distribution[$ps]} = ($rel_prob, \%node2bw_gap_states, \%br2indel_events)
		    #  is for the $ps th parsimonious indel scenario, with


    (print STDOUT "\{ Scenario no. $ps ; rel_prob. = $rel_prob }\n\n") if ($TEST_LEVEL>3);

    if ($TEST_LEVEL>=5) {
	print STDOUT '( Block-wise Gap States mapped onto the nodes )', "\n\n";

		# MODIFIED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'. #
	print STDOUT join ("\t", 'Node_Name', 'block_wise_gap_states(concatenated)'), "\n\n";
	##print STDOUT join ("\t", 'Node_ID', 'block_wise_gap_states(concatenated)'), "\n\n";

			# MODIFIED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'. #
	foreach my $node (sort keys %{$node2bw_gap_states}) { print STDOUT join ("\t", $node, $node2bw_gap_states->{$node}), "\n"; }
	##foreach my $node (sort {$a <=> $b} keys %{$node2bw_gap_states}) { print STDOUT join ("\t", $node, $node2bw_gap_states->{$node}), "\n"; }
		# %node2bw_gap_states = ($node_id => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and

	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL>=5) {...}

    if ($TEST_LEVEL>5) {
	print STDOUT '( Indel Events mapped onto the branches )', "\n\n";

			# MODIFIED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'. #
	print STDOUT join ("\t", 'Branch_le_Name', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";
	##print STDOUT join ("\t", 'Branch_ID', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";

			# MODIFIED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'. #
	foreach my $br (sort keys %{$br2indel_events}) { # 2nd inner-inner foreach-loop (over indel-undergoing branches). #
	##foreach my $br (sort {$a <=> $b} keys %{$br2indel_events}) {...} # 2nd inner-inner foreach-loop (over indel-undergoing branches). #

	    my @out_info_events = ();

	    foreach my $event_info (@{$br2indel_events->{$br}}) { # Inner-most foreach-loop (over events along the branch). #
		my ($state_ue, $state_le, $involved) = @{$event_info};
		my $cnct_involved = join (',', @{$involved});
		push @out_info_events, "$state_ue/$state_le/$cnct_involved";

		# %br2indel_events = ($branch_id => \@events_on_branch, ...) only for branches experiencing indel events, with
		# @{$events_on_branch[$e]} = ($state_upper_end, $state_lower_end, \@involved_blocks) for the $e th event along the branch $branch_id.
	    }				# End of the inner-most foreach-loop (over events along the branch). #

	    print STDOUT join ("\t", $br, @out_info_events), "\n";

	}		# End of the 2nd inner-inner foreach-loop (over indel-undergoing branches). #

	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL>5) {...}"

			# Construct inferred ancestral gap states. #
			# (ADDED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'.) #

		my @inf_anc_gp = ();
		for (my $k=0; $k<$ct_ancseqs; $k++) { # 3rd inner-inner for-loop (over ancestral sequences). #
		    my $anc_name = $ancseq_names[$k];
		    my $cnct_bw_gpstat = $node2bw_gap_states->{$anc_name};
		    unless (defined $cnct_bw_gpstat) {
			print STDOUT "[ERROR '$program_name'] The $ps th scenario in the $cl th cluster does NOT have the BLOCK-WISE GAP STATE of the ancestral sequence '$anc_name' !\n";
			print STDERR "[ERROR '$program_name'] $!\n";
			die ;
		    }

		    my @bw_gpstats = split //, $cnct_bw_gpstat;
		    unless ($ct_gblks == @bw_gpstats) {
			print STDOUT "[ERROR '$program_name'] The number of block-wise gap-states (=", scalar (@bw_gpstats), ") for the $ps th parsimonious scenario CLASHES with the number of gap-blocks (=$ct_gblks) in the $cl th gap-block cluster !\n";
			print STDERR "[ERROR '$program_name'] $!\n";
			die ;
		    }
		    my $full_gpstat = '';
		    for (my $g=0; $g<$ct_gblks; $g++) { # Innermost for-loop (over gap-blocks). #
			my $bw_gpstat = $bw_gpstats[$g];
			my $bl_len = $gblk_lengths[$g];
			$full_gpstat .= $bw_gpstat x $bl_len;
		    }				# End of the innermost for-loop (over gap-blocks). #
		    $inf_anc_gp[$k] = $full_gpstat;

		}   				# End of the 3rd inner-inner for-loop (over ancestral sequences). #

    if ($TEST_LEVEL>5) {
	print STDOUT "[ INFERRED Ancestral Gap-Patterns in the Gap-Block Cluster in question ]\n\n";
	for (my $k=0; $k<$ct_ancseqs; $k++) { print STDOUT sprintf ("%-20s", $ancseq_names[$k]), $inf_anc_gp[$k], "\n"; }
	print STDOUT "\n\n";
    }

		my $cnct_inf_anc_gp = join (':', @inf_anc_gp);
		my $info = $inf_anc_gp2info{$cnct_inf_anc_gp};

		if (defined $info) {
		    my ($tot_rel_prob, $list_psm_scns) = @{$info};
		    $tot_rel_prob = (($tot_rel_prob eq 'N/A') or ($rel_prob eq 'N/A')) ? 'N/A' : $tot_rel_prob + $rel_prob;
		    $info->[0] = $tot_rel_prob;
		    push @{$list_psm_scns}, $ps;
		} else {
		    $inf_anc_gp2info{$cnct_inf_anc_gp} = [$rel_prob, [$ps]];
		}

	    	# %inf_anc_gp2info; # ($cnct_inf_anc_gp => [$tot_rel_prob, \@list_psm_scns], ...)

	    }				# End of the 5th inner for-loop (over parsimonious scenarios). #

		# Enumerate the inferred ancestral gap-patterns and their relative-probabilities. #
		# (ADDED in creation of 'bchbch_cmp_true_vs_psm_inferred_ancstates.pl'.)

	    print $fh_out "\n[ INFERRED Ancestral Gap-Patterns, their relative probabilities, and scenarios giving the pattern ]\n\n";

    (print STDOUT "\n[[ INFERRED Ancestral Gap-Patterns in the Gap-Block Cluster in question, their relative probabilities, and scenarios giving the pattern ]]\n\n") if ($TEST_LEVEL>= 5);

	    my @list_totrel_probs = ();

	    foreach my $cnct_inf_anc_gp (sort { $inf_anc_gp2info{$b}->[0] <=> $inf_anc_gp2info{$a}->[0] } keys %inf_anc_gp2info) { # Testing 6th inner foreach-loop (over ancestral gap patterns). #

		my ($totrel_prob, $list_psm_scns) = @{$inf_anc_gp2info{$cnct_inf_anc_gp}};
		push @list_totrel_probs, $totrel_prob;

		my @inf_anc_gp = split /\:/, $cnct_inf_anc_gp;

		print $fh_out "\n\{ Resulting from Scenarios: ", join (', ', @{$list_psm_scns}), " \; Rel_prob = $totrel_prob \}\n\n"; 
		for (my $k=0; $k<$ct_ancseqs; $k++) { print $fh_out sprintf ("%-20s", $ancseq_names[$k]), $inf_anc_gp[$k], "\n"; }
		print $fh_out "\n\n";

    if ($TEST_LEVEL>=5) {
	    print STDOUT "\n\[ The gap-pattern resulting from the scenarios: ", join (', ', @{$list_psm_scns}), " \; Rel_prob = $totrel_prob \]\n\n"; 
	    for (my $k=0; $k<$ct_ancseqs; $k++) { print STDOUT sprintf ("%-20s", $ancseq_names[$k]), $inf_anc_gp[$k], "\n"; }
	    print STDOUT "\n\n";
    }

	    }   # End of the 6 th inner foreach-loop (over ancestral gap patterns). #

			## RECORD THE FINAL RESULT on the cluster ##
			### RESTART FROM HERE ###

	    my $ct_inf_anc_gps = @list_totrel_probs;

	    print $fh_out "\n[ Summary ]\n\n";
	    print $fh_out "#{Inferred ancestral gap-patterns} = $ct_inf_anc_gps .\n";
	    print $fh_out "List of total-relative-probabilities: ", join (', ', @list_totrel_probs), " .\n\n";

    if ($TEST_LEVEL >3) {
	print STDOUT "[ Summary: ]\n\n";
	print STDOUT "#{Inferred ancestral gap-patterns} = $ct_inf_anc_gps .\n";
	print STDOUT "List of total-relative-probabilities: ", join (', ', @list_totrel_probs), " .\n\n";
    }

			# Examine information on the correct inference. #

	    my $info_correct_inf = $inf_anc_gp2info{$cnct_true_anc_gp};
	    my $relprob_correct_inf = (defined $info_correct_inf) ? $info_correct_inf->[0] : 'NonParsimonious';

	    print $fh_out "{Relative-probability of the correct inference} = $relprob_correct_inf .\n\n\n";

    if ($TEST_LEVEL>3) {
	print STDOUT "{Relative-probability of the correct inference} = $relprob_correct_inf .\n\n\n";
    }


			## Record the summary information of the cluster into the output list. ##

			# Examine if the subject cluster is null or not. #
	    my $cnct_totrel_probs = join ('/', @list_totrel_probs);
	    my $null_block = $GAP x $ct_extseqs;
	    if (($ct_gblks == 1) and ($gap_blocks->[0][3] eq $null_block)) {
			# The cluster is a null cluster. #

     (print STDOUT "[Final INFORMATION] The cluster consists ONLY of a single NULL gap-block .\n\n\n") if ($TEST_LEVEL>3);

		print $fh_out "[NOTE]: NULL CLUSTER .\n\n\n";

		print $fh_olist join ("\t", $num_cls, $cl, $cl_start, $cl_end, $cl_len, $ct_gblks, $min_totct_indels, $ct_psm_scns, $ct_inf_anc_gps, $relprob_correct_inf, $cnct_totrel_probs, '[Null Cluster]'), "\n";
	    } else {
			# Otherwise. #
		print $fh_olist join ("\t", $num_cls, $cl, $cl_start, $cl_end, $cl_len, $ct_gblks, $min_totct_indels, $ct_psm_scns, $ct_inf_anc_gps, $relprob_correct_inf, $cnct_totrel_probs), "\n";
	    }


	}			# End of the main middle for-loop (over gap-block clusters and gapless blocks) #

	print $fh_olist "\n";
=cut


		# Remove the temporary input files. #

	foreach my $tmpin_file ($tmpin_anc_gaps) { # Xth middle foreach-loop (over temporary files). #
	##foreach my $tmpin_file ($tmpin_msa, $tmpin_anc_gaps) {...} # Xth middle foreach-loop (over temporary files). # --- For the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.

=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
    unless (0 == system ("cp -f $tmpin_file $outdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to copy the temporary input file '$tmpin_file' into the directory '$outdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
=cut

	    ## next; # For TESTING the framework ("Pre/Test2a")

	    unless (0 == system ("rm -f $tmpin_file") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to remove the temporary input file '$tmpin_file': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die;
	    }
	}					# End of the Xth middle foreach-loop (over temporary files). #

	$ct_succ++;

    }				# End of the main outer foreach-loop (over MSAs in a class). #

    print STDOUT "\n  --- Finished Processing the class '$id_class' ---\n\n";

    print STDOUT "\n  #{Input MSAs}  = $ct_msas .\n";
    print STDOUT "      #{Anomalous} = $ct_anom ,\n";
    print STDOUT "      #{Failed}    = $ct_fail ,\n";
    print STDOUT "      #{Succeeded} = $ct_succ .\n\n\n";

    $totct_msas += $ct_msas;
    $totct_anom += $ct_anom;
    $totct_fail += $ct_fail;
    $totct_succ += $ct_succ;


=for the predecessor 'bchbch_cmp_true_vs_psm_inferred_ancstates.v2_lite.pl'.
		## Archive the Middle output directory (storing masked MSAs for the same chromosome.) ##

    my $out_archive_name = "$midoutdir.tgz";

     # $parent_out_indel_dir = "$top_indel_dir/$mid_dir";
     # [ $parent_inout_paml_dir = "$top_paml_dir/$mid_dir"; ] --- for the predecessor 'bch_serial_PAML.pl'.
     # [$outaln_dir = "$topoutaln_dir/$mid_dir";]  --- for the predecessor 'bchbch_refine_msas.pl'.

    unless ( chdir $topoutdir ) { # MODIFIED in creation of 'bch_serial_br_infer_indels.pl'.
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to change the directory to '$topoutdir' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    unless (0 == system ("tar -cpzf $out_archive_name $midoutdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED TO CREATE a tar-gzipped ARCHIVE '$out_archive_name' out of the directory '$midoutdir' under the directory '$topoutdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }

    unless (0 == system ("rm -rf $midoutdir") ) {
	print STDOUT "[ERROR '$program_name'] FAILED to REMOVE the directory '$midoutdir' under the directory '$topoutdir': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
 (print STDOUT "\n... Successfully archived the masked MSAs into '$out_archive_name' under '$topoutdir'!!\n\n") if ($TEST_LEVEL>1);

    unless ( chdir $CURR_WORK_DIR ) {
	print STDOUT "[ERROR '$program_name'] FAILED the subroutine 'chdir' to RETURN TO the original place, '$CURR_WORK_DIR' !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die;
    }
=cut

}				# End of the main outermost foreach-loop (over classes of MSAs). #

close ($fh_olist);


print STDOUT "\n\n=== All Processes FINISHED !! ===\n\n";


print STDOUT "#{MSA Classes} = $ct_classes .\n";
print STDOUT "Total#{Input MSAs}  = $totct_msas .\n";
print STDOUT "    Total#{Anomalous} = $totct_anom ,\n";
print STDOUT "    Total#{Failed}    = $totct_fail ,\n";
print STDOUT "    Total#{Succeeded} = $totct_succ .\n\n\n";



	### DEFINITIONS OF THE SUBROUTINES ###

# peek_into_archive:
#	This subroutine looks into an input archive and outputs a list of entries.
#	The archive is supposed to be storing files with the following paths:
#	"$dir/$pref$id$suff",
#	where $id specifies the ID of the file.
#
# INPUT: (\$in_archive, $dir, $pref, $suff), where
#	$in_archive is the path to the input archive;
#	$dir/$pref/$suff are templates for the paths to the files stored in the archive.
#
# OUTPUT: \@list_ids, where
#	@list_ids stores the IDs of the archived files with the above paths.
#
sub peek_into_archive (\$$$$) {

	my ($in_archive, $dir, $pref, $suff) = @_;
	local $/ = "\n"; # Use the default input-record-separator. #

	my $fh_tmp;
	unless ( open ($fh_tmp, "tar -tzf ${$in_archive} |" ) ) {
	    print STDOUT "[ERROR 'peek_into_archive'] CanNOT fork or look into ${$in_archive}: $!\n";
	    return undef;
	}

	my @list_ids = ();

	my $ext_pref = "$dir/$pref"; # Extended prefix. #
	my $len_epref = length $ext_pref;
	my $len_suff = length $suff;

	while (my $tmp = <$fh_tmp>) { # Main while-loop (over input lines) #

	    $tmp =~ s/\s+$//; # Remove white spaces at the end of the entry (serving as a separator (and possibly a pad)).
	    ($ext_pref =~ m/^\.\//) or ($tmp =~ s/^\.\///); # Remove './' at the beginning of the entry unless $ext_pref begins with './'. (ADDED in creation of 'bch_serial_PAML.pl'.)

	    my $tmp_suff = substr ($tmp, - $len_suff, $len_suff, '');

	    ($tmp_suff eq $suff) or next; # Skip if the suffix doesn't match.

	    my $tmp_epref = substr ($tmp, 0, $len_epref, '');

	    ($tmp_epref eq $ext_pref) or next; # Skip if the extended prefix doesn't match.

	    push @list_ids, $tmp; # The remnant of the entry must be the ID.


	}			# End of the main while-loop (over input lines) #
	close ($fh_tmp);

	return \@list_ids;

}	# End of the subroutine 'peek_into_archive' #



# br_parse_detail_output_infer_psm_indel_events_dawg:
#	This subroutine parses the brief output of the Perl script 'br_infer_psm_indel_events_dawg.pl',
#	which calculates the (logarithmic) 'absolute' probabilities of the gap-block clusters,
#	based on the parsimonious indel scenarios.
#	The subroutine was created by taking advantage of the overall framework of the subroutine,
# 	'parse_brief_output_infer_psm_indel_events_dawg'.
#	(CREATED on Mar 12 (Wed), 2014)
#
# parse_brief_output_infer_psm_indel_events_dawg:
#	This subroutine was created by modifying 'parse_brief_output_infer_psm_indel_events'
#	so that it will be cut out for the output of the Perl script 'br_infer_psm_indel_events_dawg.pl'.
#	(CREATED on Feb 28 (Fri), 2014)
#
#	TEMPORARILY REVISED on Mar 6 (Thu), 2014:
#	The subroutine now copes with a gap-block cluster with "Min_Total#{indel_events} = 0; #{scenarios} = 0",
#	which is likely because the cluster consists ONLY of a NULL BLOCK,
#	by assigning an EMPTY SCENARIO DISTRIBUTION.
#
# parse_brief_output_infer_psm_indel_events:
#	This subroutine parses the brief output of the Perl script 'br_infer_psm_indel_events.pl',
#	which calculates the 'probability distribution' of ancestral gap states (and indel scenarios),
#	based on the parsimonious indel scenarios.
#
# INPUT: (\$infile), where
#	$infile is the path to the input file, which is an brief output of 'br_infer_psm_indel_events.pl';
#
#
# OUTPUT: (\@parameters, \@info_segments, \@setset_info_cls), where
#	[ @parameters = ($delta_del_5p, \@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with ] --- OBSOLETE as of May 22 (Fri), 2015.
#	@parameters = ($delta_del, \@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with
#	@full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr, \%fname2id_intnode, $ftopnode_name) for the full MSA,
#	@seqids storing the sequence IDs,
#	$ct_segments = scalar (@info_segments) = scalar (@indel_state_distributions);
#
#	@{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_name), with 
#	$seg_start/end in the 0-based half-open convention,
#	$red_topnode_name is the name of the top-node (root) of the tree reduced according to the usability pattern
#	 ($red_topnode_name = undef for the segment skipped to infer ancestral states) ;
#
#	@info_cls = @{$setset_info_cls[$s]->[$cl]} 
#	stores information on the $cl th cluster in the $s th segment. Here,
#
#	@info_cls = ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, $ct_scenarios, $log_abs_prob), where
#
#	@{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block,
#
#	$ct_scenarios is the number of (parsimoinious) indel scenarios that result in the gap-block cluster, and
#	$log_abs_prob is the (logarithmic) absolute probability of the gap-block cluster.
#	
sub br_parse_detail_output_infer_psm_indel_events_dawg (\$) {

	my ($infile) = @_;
	local $/ = ''; # Use blank lines as input-record-separators. #

	my $fh_in = open_input_file2 (${$infile});
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] 'open_input_file2' FAILED on the input file '${$infile}': $! !\n";
	    return (undef, undef, undef); 
	}

		## Read the phase factor. ##
		# (ADDED in creation of 'br_parse_detail_output_infer_psm_indel_events_dawg' on Mar 12 (Wed), 2014.) #
		# (=> MODIFIED to adapt to the current convention on May 22 (Fri), 2015.) #

	my $delta_del;
	##my $delta_del_5p;
	while (my $in = <$fh_in>) {
	    my ($tmp) = ($in =~ m/^\$delta_del\s*\=\s*([^\s]+)/);
	    ##my ($tmp) = ($in =~ m/^\$delta_del_5p\s*\=\s*([^\s]+)/);
	    (defined $tmp) or next;
	    ($delta_del = $tmp) =~ s/\s*\.$//;
	    ##($delta_del_5p = $tmp) =~ s/\s*\.$//;
	    last;
	}
	unless (defined $delta_del) {
	##unless (defined $delta_del_5p) {...}
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT FIND the Phase factor, '", '$delta_del', "': $! !\n";
	    ##print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT FIND the Phase factor, '", '$delta_del_5p', "': $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}


		## Read the input sequence tree. ##

	my $top_node_name0;
	##my $top_node_id0;

	while (my $in = <$fh_in>) {
	    ($in =~ m/^\[\[\s*Input\s+Tree/) or next;
	    ##($in =~ m/^\[\[\s*Input\s+tree/) or next; # A BUG fixed on Nov 26 (Tue), 2013. #

	    ($top_node_name0) = ($in =~ m/top_node_name\s*\=\s*([^\s\(\)]+)/);
	    ##($top_node_id0) = ($in =~ m/top_node_ID\s*\=\s*([^\s\(\)]+)/);
	    last;
	}
	unless (defined $top_node_name0) {
	##unless (defined $top_node_id0) {...}
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT FIND the top-node name: $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $full_tree = <$fh_in>;

	my %fnode2pa; my %fnode2ch; my %fname2id; my %fid2name; my %fbr_attr;
	my %fname2id_intnode; my $ftopnode_name;

	unless ( input_tree_file_dawg ($full_tree, %fnode2pa, %fnode2ch, %fname2id, %fid2name, %fbr_attr, %fname2id_intnode, $ftopnode_name) ) {
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] The subroutine 'input_tree_file_dawg' FAILED to extract the full tree: $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}
=for the predecessor 'parse_brief_output_infer_psm_indel_events'.
	unless ( input_ext_tree_file ($full_tree, %fnode2pa, %fnode2ch, %fname2id, %fid2name, %fbr_attr, 1) ) {
	    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The subroutine 'input_ext_tree_file' FAILED to extract the full tree: $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}
=cut

	my $ftopnode_id = $fname2id{$TOP_NODE};
	unless ($ftopnode_name eq $top_node_name0) {
	##unless ($ftopnode_id eq $top_node_id0) {...} #For the predecessor 'parse_brief_output_infer_psm_indel_events'.
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The top-node name in the label (=$top_node_name0) CONFLICTS with the name displayed in the tree (=$ftopnode_name): $! !\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

		# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #

	my @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr, \%fname2id_intnode, $ftopnode_name);

		# @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr, \%fname2id_intnode, $ftopnode_name) for the full MSA,

=for the predecessor 'parse_brief_output_infer_psm_indel_events'.
	my @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr);
		# @full_tree = ($ftopnode_id, \%fnode2pa, \%fnode2ch, \%fname2id, \%fid2name, \%fbr_attr) for the full MSA,
=cut

		## Read the MSA information ##

	my $label_msa = <$fh_in>;

	my ($ct_seqs0) = ($label_msa =~ m/\(\#\{sequences\}\s*\=\s*(\d+)\)/);
	unless (defined $ct_seqs0) {
	    $label_msa =~ s/\s+$//;
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The number of sequences canNOT be extracted from the label: $!\n";
	    print STDOUT "'$label_msa' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	(my $list_seqids = <$fh_in>) =~ s/\s*\.\s+$//; 
	##(my $list_seqids = <$fh_in>) =~ s/\.\s+$//; # A BUG fixed on Dec 19 (Thu), 2013. #
	$list_seqids =~ s/^Sequences\:\s+//;
	my @seqids = split /\,\s*/, $list_seqids;

	unless (@seqids == $ct_seqs0) {
	    my $ct_seqids = @seqids;
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The number of sequences in the label (=$ct_seqs0) CONFLICTS with $ct_seqids, which is the number of IDs in the list: $! !\n";
	    print STDOUT ' ' x 3, join (', ', @seqids), " .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $descr_msa_len = <$fh_in>;
	my ($msa_len) = ($descr_msa_len =~ m/^\#\{columns\s+in\s+MSA\}\s*\=\s*(\d+)/);
	unless (defined $msa_len) {
	    $descr_msa_len =~ s/\s+$//;
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The length of the MSA canNOT be extracted from the line: $!\n";
	    print STDOUT "'$descr_msa_len' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $descr_tgt_indx = <$fh_in>;
	my ($tgt_seqid, $indx_tgt) = ($descr_tgt_indx =~ m/^\{Index\s+for\s+the\s+target\s+sequence\s+\'([^\s\']+)\'\}\s*\=\s*(\S+)/ );
	unless (defined $indx_tgt) {
	    $descr_tgt_indx =~ s/\s+$//; 
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The index of the target sequence canNOT be extracted from the line: $!\n";
	    print STDOUT "'$descr_tgt_indx' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

	my $descr_ct_segs = <$fh_in>;

		# MODIFIED in creation of 'br_parse_detail_output_infer_psm_indel_events_dawg'. #
	my ($ct_segments) = ($descr_ct_segs =~ m/\#\{segments\}\s*\=\s*(\d+)/);
	##my ($ct_segments) = ($descr_ct_segs =~ m/^\#\{segments\}\s*\=\s*(\d+)/); # A BUG fixed on Mar 13 (Thu), 2014. #
	##my ($ct_segments) = ($descr_ct_segs =~ m/^\#\{MSA\s+Segments\}\s*\=\s*(\d+)/);
	unless (defined $ct_segments) {
	    $descr_ct_segs =~ s/\s+$//; 
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The number of MSA segments canNOT be extracted from the line: $! !\n";
	    print STDOUT "'$descr_ct_segs' .\n\n";
	    close ($fh_in);
	    return (undef, undef, undef);
	}

		# MODIFIED in creation of 'br_parse_detail_output_infer_psm_indel_events_dawg'. #
		# => Further MODIFIED to adapt to the current convention on May 22 (Fri), 2015. #
	my @parameters = ($delta_del, \@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments);
	##my @parameters = ($delta_del_5p, \@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments);
	##my @parameters = (\@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments);

		# [ @parameters = ($delta_del_5p, \@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with ] --- OBSOLETE as of May 22 (Fri), 2015. #
		# @parameters = ($delta_del, \@full_tree, \@seqids, $msa_len, $tgt_seqid, $indx_tgt, $ct_segments), with

=for the predecessor 'parse_brief_output_infer_psm_indel_events_dawg'.
		### MAIN PROCESS: Read the cluster-wise distributions of the indel scenarios
		### (and the associated ancestral gap states), which were bundled into segments.

	my @info_segments = my @indel_state_distributions = ();
=cut

		### MAIN PROCESS: retrieve the (logarithmic) 'absolute' probabilities, as well as other attributes,
		## of the gap-block clusters. 
		## (MODIFIED in creation of 'br_parse_detail_output_infer_psm_indel_events_dawg'.) #

	my @info_segments = ();
	my @setset_info_cls = ();
	my ($seg_no, $cls_no);

	while (my $in_chunk = <$fh_in>) { # Main outer while-loop (over input chunks). #

=for the predecessor 'parse_brief_output_infer_psm_indel_events_dawg'.
	while (my $head_seg = <$fh_in>) {...} # Main outer while-loop (over segments) #

	    unless (defined $seg_no) {
		$head_seg =~ s/\s+$//;
		print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT Extract the segment no. from the segment heading: $! !\n";
		print STDOUT "'$head_seg' .\n\n";
		close ($fh_in);
		return (undef, undef, undef);
	    }
=cut

	    my ($tmp_seg_no) = ($in_chunk =~ m/^\<\<\<\<\s+The\s+(\d+)\s+th\s+Segment/) ;

	    if (defined $tmp_seg_no) {

		$seg_no = $tmp_seg_no;

		my $head_seg = $in_chunk;
		my ($seg_start, $seg_end) = ($head_seg =~ m/Region\s+\=\s+\[(\d+)\,\s*(\d+)\)/);
		unless (defined $seg_end) {
		    $head_seg =~ s/\s+$//;
		    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT Extract the segment end coordinate from the segment heading: $! !\n";
		    print STDOUT "'$head_seg' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my ($us_pattern) = ($head_seg =~ m/Usability\s+pattern\s+\=\s*\'([^\s\']+)\'/);
		unless (defined $us_pattern) {
		    $head_seg =~ s/\s+$//;
		    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT Extract the segment end coordinate from the segment heading: $! !\n";
		    print STDOUT "'$head_seg' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my @seg_info = ($seg_start, $seg_end, $us_pattern);
		$info_segments[$seg_no] = \@seg_info;

		$setset_info_cls[$seg_no] = []; # Prepare an empty array for the cluster. #

		next;

	    }

		# @{$info_segments[$s]} = ($seg_start, $seg_end, $usability_pattern, $red_topnode_id), with 
		# $seg_start/end in the 0-based half-open convention,
		# $red_topnode_id is the ID of the top-node (root) of the tree reduced according to the usability pattern
		#  ($red_topnode_id = undef for the segment skipped to infer ancestral states) ;


		# MODIFIED in creation of 'br_parse_detail_output_infer_psm_indel_events_dawg'. #
	    if ($in_chunk =~ s/^\[\[\s+Reduced\s+Tree\s+//) {
		my $second_chunk = $in_chunk;
=for the predecessor 'parse_brief_output_infer_psm_indel_events_dawg'.
	    my $second_chunk = <$fh_in>;

	    if ($second_chunk =~ s/^\[\[\s+Reduced\s+Tree\s+//) {
=cut

			# Normal case. #

			# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
		my ($red_topnode_name) = ($second_chunk =~ m/\(top_node_name\s*\=\s*([^\s\(\)]+)\)/);
		unless (defined $red_topnode_name) {
		##my ($red_topnode_id) = ($second_chunk =~ m/\(top_node_ID\s*\=\s*([^\s\(\)]+)\)/);
		##unless (defined $red_topnode_id) {...}
		    $second_chunk =~ s/\s+$//;
		    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT Extract the top-node name for the reduced tree from the tree-header: $! !\n";
		    print STDOUT "'$second_chunk' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		    # MODIFIED in creation of 'br_parse_detail_output_infer_psm_indel_events_dawg'. #
		push @{$info_segments[$seg_no]}, $red_topnode_name;
		next;
	    }

=for the predecessor 'parse_brief_output_infer_psm_indel_events_dawg'.
		push @seg_info, $red_topnode_name;
		##push @seg_info, $red_topnode_id;

			# Go ahead with reading the indel scenario distribution.

	    } elsif ($second_chunk =~ m/^SKIPPED\s+INFERRING/) {
	    ##} elsif ($second_chunk =~ m/^SKIPPED\s+INDERRING/) { # A BUG fixed on Nov 26 (Tue), 2013. #

			# The ancestral state inference was skipped for this segment. #

		my @sw_indel_state_distributions = ( [$seg_start, $seg_end, '?'] );
		$indel_state_distributions[$seg_no] = \@sw_indel_state_distributions;

		next;	# Go ahead to the next segment.
		
	    } else {
			# An Unexpected Case. #

		$second_chunk =~ s/\s+$//;
		print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: an ANOMALOUS SECOND CHUNK for the $seg_no th segment: $! !\n";
		print STDOUT "'$second_chunk' .\n\n";
		close ($fh_in);
		return (undef, undef, undef);
	    }
=cut

=for the predecessor 'parse_brief_output_infer_psm_indel_events_dawg'.
	    my $red_topnode_name = $seg_info[3];
	    ##my $red_topnode_id = $seg_info[3];

	    my $red_tree = <$fh_in>; # Emit the chunk for the reduced tree.
	    my $heading_main = <$fh_in>; # Emit the heading that tells that the main description begins. 

	    my @sw_indel_state_distributions = ();
=cut

	    my ($tmp_cls_no) = ($in_chunk =~ m/^\<\s+The\s+(\d+)\s+th\s+cluster/);
	    if (defined $tmp_cls_no) {

		$cls_no = $tmp_cls_no;

		my ($cl_start, $cl_end) = ($in_chunk =~ m/region\s*\=\s*\[(\d+)\,\s*(\d+)\)/);

		unless (defined $cl_end) {

			# Then, the chunk (actially line) must record the (logarithmic) 'absolute' probability. #

		    my ($log_abs_prob) = ($in_chunk =~ m/Total\{log\-likelihood\}\s*\=\s*([^\s\(\)]+)/ );
		    unless (defined $log_abs_prob) {
			$in_chunk =~ s/\s+$//;
			print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the (logarithmic) absolute probability from the following heading of the $cls_no th gap-block cluster in the $seg_no th segment: $! !\n";
			print STDOUT "'$in_chunk' .\n\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }

		    $setset_info_cls[$seg_no]->[$cls_no][6] = $log_abs_prob;

		    next;
		}

		my ($ct_gblks0) = ($in_chunk =~ m/\#\{gap\-blocks\}\s*\=\s*(\d+)/);
		unless (defined $ct_gblks0) {
		    $in_chunk =~ s/\s+$//;
		    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the number of gap-blocks from the following heading of the $cls_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$in_chunk' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}
		my $head_blk = <$fh_in>;
		unless ($head_blk =~ m/^\[\s+Gap\-blocks/) {
		    $head_blk =~ s/\s+$//;
		    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: Found an ANOMALOUS HEADER of the gap-blocks in the $cls_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$head_blk' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}
		my $fld_tags = <$fh_in>;
		unless ($fld_tags =~ m/^Block_no\./) {
		    $fld_tags =~ s/\s+$//;
		    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: Found an ANOMALOUS FIELD TAGS of the gap-blocks in the $cls_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$fld_tags' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my $info_gblocks = <$fh_in>;
		my @lines_gblocks = split /\n+/, $info_gblocks;
		my @gap_blocks = ();
		foreach my $line_gb (@lines_gblocks) { # Inner foreach-loop (over gap-blocks). #
		    my ($blk_no, $blen, $bl_start, $bl_end, $indx_gp, $gpattern) = split /\s+/, $line_gb;
		    $gap_blocks[$blk_no] = [$bl_start, $bl_end, $indx_gp, $gpattern];
		}			# End of the inner foreach-loop (over gap-blocks). #

		unless (@gap_blocks == $ct_gblks0) {
		    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: The size of the array (=", scalar (@gap_blocks), ") describing gap-blocks does NOT MATCH the number (=$ct_gblks0) extracted from the header of the $cls_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$fld_tags' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		$setset_info_cls[$seg_no]->[$cls_no] = [$cl_start, $cl_end, 'X/-', \@gap_blocks];
		
		next;
	    }

	    if ($in_chunk =~ m/^\[\s+Parsimonious\s+Scenarios/) {

		my ($min_totct_indels) = ($in_chunk =~ m/Min\(Total\#{indel\s+events\}\)\s*\=\s*([^\s\;]+)/ );
		my ($ct_scenarios) = ($in_chunk =~ m/\#\{scenarios\}\s*\=\s*([^\s\(\)]+)/);

		unless ((defined $min_totct_indels) and (defined $ct_scenarios)) {
		    $in_chunk =~ s/\s+$//;
		    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: Found an ANOMALOUS HEADER for the parsimonious scenarios in the $cls_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$in_chunk' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		@{$setset_info_cls[$seg_no]->[$cls_no]}[4,5] = ($min_totct_indels, $ct_scenarios);

		next;
	    }

		# @info_cls = @{$setset_info_cls[$s]->[$cl]} 
		# stores information on the $cl th cluster in the $s th segment. Here,
		#
		# @info_cls = ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, $ct_scenarios, $log_abs_prob), where
		#
		# @{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block,
		#
		# $ct_scenarios is the number of (parsimoinious) indel scenarios that result in the gap-block cluster, and
		# $log_abs_prob is the (logarithmic) absolute probability of the gap-block cluster.
		# 

		### RESTART FROM HERE ###

=for the predecessor 'parse_brief_output_infer_psm_indel_events_dawg'.
	    while (my $head_cls = <$fh_in>) { # Main middle while-loop (over clusters) #

		if ($head_cls =~ s/^\<\s+Gapless\s+block\s+// ) {

			# Gapless block. #

		    my ($bl_start, $bl_end) = ($head_cls =~ m/region\s+\=\s+\[(\d+)\,\s*(\d+)\)/ );
		    unless (defined $bl_end) {
			$head_cls =~ s/\s+$//;
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the block-end coordinate from the following heading of the gapless block in the $seg_no th segment: $! !\n";
			print STDOUT "'< Gapless block $head_cls' .\n\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }

		    my ($rfl_cl_no) = ($head_cls =~ m/^left\-flanking\s+the\s+(\d+)/) ;
		    ##my ($rfl_cl_no) = ($head_cls =~ m/^left\-flanking\s+the\s+(\d)/) ; # A BUG fixed on Nov 26 (Tue), 2013. #

		    unless (defined $rfl_cl_no) {
			if ($head_cls =~ m/^on\s+the\s+right\-end/) {

				# The right-end of the segment. #

			    push @sw_indel_state_distributions, [$bl_start, $bl_end, 'X'];

			    last; # End the middle while-loop.

			} else {
				# An unexpected case. #

			    $head_cls =~ s/\s+$//;
			    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the right-flanking cluster number from the following heading of the gapless block in the $seg_no th segment: $! !\n";
			    print STDOUT "'< Gapless block $head_cls' .\n\n";
			    close ($fh_in);
			    return (undef, undef, undef);
			}

		    } # END of "unless (defined $right_cl_no) {...}"

		    unless (@sw_indel_state_distributions == 2* $rfl_cl_no) {
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The size of the array, ", '@sw_indel_state_distributions (=', scalar (@sw_indel_state_distributions), ") is NOT EQUAL to twice the number of the right-flanking cluster (=$rfl_cl_no): $! !\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }

		    $sw_indel_state_distributions[2*$rfl_cl_no] = [$bl_start, $bl_end, 'X'];

		    next; # Go ahead to the gap-block cluster on the right. 

			# @{$sw_indel_state_distributions[2*$cl]} = ($cl_start, $cl_end, $state)
			# for the gapless block left-flanking the $cl th gap-block cluster
			# (or on the right-end of the segment (with $cl = #{clusters})),
			# or (always with $cl = 0) for the skipped segment due to inadquate information.
			# $state	= 'X' for the gapless block,
			# 	= '?' for the skipped segment due to inadequate information.

		} # END of "if ($head_cls =~ s/^\<\s+Gapless\s+block\s+/) {...}"


			## Read information on the gap-block cluster. ##

		my ($cl_no) = ($head_cls =~ m/^\<\s+The\s+(\d+)\s+th\s+cluster/ );
		unless (defined $cl_no) {
		    $head_cls =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the cluster number from the following heading of a cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$head_cls' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		unless (@sw_indel_state_distributions == 2 * $cl_no + 1) {
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The temporary size of the array, ", '@sw_indel_state_distributions (=', scalar (@sw_indel_state_distributions), ') is NOT EQUAL to 2 * $cl_no + 1, where $cl_no = ', "$cl_no, for the $seg_no th segment: $! !\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my ($cl_start, $cl_end) = ($head_cls =~ m/region\s+\=\s+\[(\d+)\,\s*(\d+)\)/ );
		unless (defined $cl_end) {
		    $head_cls =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the end-coordinate of the cluster from the following heading of a cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$head_cls' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

		my ($ct_gblks) = ($head_cls =~ m/\#\{gap\-blocks\}\s+\=\s+(\d+)/ );
		unless (defined $ct_gblks) {
		    $head_cls =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the number of gap-blocks from the following heading of a cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$head_cls' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

			# Read information on the gap-blocks. #

		my $head_gblocks = <$fh_in>;
		my $fld_tags_gblocks = <$fh_in>;
		my $info_gblocks = <$fh_in>;

		my @gap_blocks = ();

		foreach my $gblock (split /\n+/, $info_gblocks) { # 1st inner foreach-loop (over gap-blocks) #
		     my ($bl_no, $bl_len, $bl_start, $bl_end, $indx_gp, $gpattern) = split /\s+/, $gblock;
		     push @gap_blocks, [$bl_start, $bl_end, $indx_gp, $gpattern];

			# @{$gap_blocks[$g]} = ($bl_start, $bl_end, $indx_gp, $gpattern) is for the $g th gap-block, and

		}					# End of the 1st inner foreach-loop (over gap-blocks) #

		unless (@gap_blocks == $ct_gblks) {
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The size of ", '@gap_blocks (=', scalar (@gap_blocks), ") CONFLICTS with the number of gap-blocks (=$ct_gblks) described in the heading: $! !\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

			# Read the distribution of parsimonious scenarios for a gap-block cluster. #

		my $heading_list_psm_scns = <$fh_in>;

			# MODIFIED accompanying the revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014.) #
		my ($min_totct_indels) = ($heading_list_psm_scns =~ m/Min_Total\#\{indel_events\}\s+\=\s+([^\;\s]+)/ );
		##my ($min_totct_indels) = ($heading_list_psm_scns =~ m/Min_Total\#\{indel_events\}\s+\=\s+(\d+)/ );
		unless (defined $min_totct_indels) {
		    $heading_list_psm_scns =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the minimum total count of indel events from the following heading for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$heading_list_psm_scns' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

			# MODIFIED accompanying the revision of 'MyTreeMap_indels.pm' on Mar 11 (Tue), 2014.) #
		my ($ct_psm_scns) = ($heading_list_psm_scns =~ m/\#\{scenarios\}\s+\=\s+([^\)\s]+)/ );
		##my ($ct_psm_scns) = ($heading_list_psm_scns =~ m/\#\{scenarios\}\s+\=\s+(\d+)/ );
		unless (defined $ct_psm_scns) {
		    $heading_list_psm_scns =~ s/\s+$//;
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the count of parsimonious indel scenarios from the following heading for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
		    print STDOUT "'$heading_list_psm_scns' .\n\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}

			## When encountering ($min_totct_indels eq 'N/A'),
			## assign an empty scenario distribution to the cluster.
			## (TEMPORARILY ADDED accompanying the revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014.) ##

		if ($min_totct_indels eq 'N/A') { # Further Modified on Mar 11 (Tue), 2014. #
		##if (($min_totct_indels eq 'N/A') and ($ct_psm_scns == 0)) {...} 
		    print STDOUT "[CAUTION 'parse_brief_output_infer_psm_indel_events_dawg'] The $cl_no th gap-block cluster in the $seg_no th segment has Min_Total#{indel_events} = $min_totct_indels and #{scenarios} = $ct_psm_scns, probably because it contains (a) long gap(s): $! !\n\n";
		    my @scenario_distribution = (); # An empty scenario distribution !!
		    $sw_indel_state_distributions[2 * $cl_no + 1] = [$cl_start, $cl_end, 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution];
		    my $info = <$fh_in>;
		    unless ($info =~ m/^\[INFORMATION\]/) {
			$info  =~ s/\s+$//;
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] Found an UNEXPECTED CHUNK: $! !\n";
			print STDOUT "'$info' .\n\n";
		 	close ($fh_in);
			return (undef, undef, undef);
		    }
		    next;
		}

			## When encountering ($min_totct_indels == 0) and ($ct_psm_scns == 0),
			## assign an empty scenario distribution to the cluster.
			## (TEMPORARILY ADDED (as a first-aid) on Mar 6 (Thu), 2014.) ##

		if (($min_totct_indels == 0) and ($ct_psm_scns == 0)) {
		    if (($ct_gblks == 1) and ($gap_blocks[0]->[3] eq '-' x $ct_seqs0)) {
			print STDOUT "[CAUTION 'parse_brief_output_infer_psm_indel_events_dawg'] The $cl_no th gap-block cluster in the $seg_no th segment consists ONLY of a NULL BLOCK: $! !\n";
		    } else {
			print STDOUT "[CAUTION 'parse_brief_output_infer_psm_indel_events_dawg'] Somehow, the $cl_no th gap-block cluster in the $seg_no th segment has Min_Total#{indel_events} = 0 and #{scenarios} = 0: $! !\n";
		    }
		    my @scenario_distribution = (); # An empty scenario distribution !!
		    $sw_indel_state_distributions[2 * $cl_no + 1] = [$cl_start, $cl_end, 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution];
		    next;
		}

		my @scenario_distribution = ();

		while (my $head_scn = <$fh_in>) { # 2nd inner whlie-loop (over parsimonious scenarios). #

		    my ($scn_no, $rel_prob) = ($head_scn =~ m/^\{\s+Scenario\s+no\.\s+(\d+)\s*\;\s+rel_prob\.\s*\=\s*([^\s\{\}]+)/ );
		    unless (defined $rel_prob) {
			$head_scn =~ s/\s+$//;
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] ILLEGAL INPUT: CanNOT extract the relative probability from the following heading of a parsimonious scenario for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
			print STDOUT "'$head_scn' .\n\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }
		    unless (@scenario_distribution == $scn_no) {
			print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The temporary size of ", '@scenario_distribution (=', scalar (@scenario_distribution), ") CLASHES with the parsimonious scenario no. (=$scn_no) for the $cl_no th gap-block cluster in the $seg_no th segment: $! !\n";
			close ($fh_in);
			return (undef, undef, undef);
		    }

			# Read block-wise gap states mapped onto the nodes. #

		    my $head_bw_gstates = <$fh_in>;
		    my $fld_tags_bw_gstates = <$fh_in>;
		    my $info_bw_gstates = <$fh_in>;

		    my %node2bw_gap_states;
		    my $ct_rlv_nodes = my $ct_gaponly = 0; # A measure to avoid a buggy behavior. (introduced on Nov 26 (Tue), 2013). #

		    foreach my $line (split /\n+/, $info_bw_gstates) { # 1st inner-inner foreach-loop (over nodes). #
				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			my ($node_name, $cnct_bw_gstats) = split /\s+/, $line;
			##my ($node_id, $cnct_bw_gstats) = split /\s+/, $line;

				# Introduced (on Nov 26 (Tue), 2013) in order to avoid a buggy behavior when encountering gap-only columns. #
			$ct_rlv_nodes++;
			($cnct_bw_gstats =~ m/X/) or ($ct_gaponly++);

			if ($fl_full_nodes <= 0) {
				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			    ($node_name eq $red_topnode_name) or next; # Skip except the top node.
			    ##($node_id eq $red_topnode_id) or next; # Skip except the top node.
			} elsif ($fl_full_nodes < 2) {
				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			    (defined $fname2id_intnode{$node_name}) or next; # Skip an exterior node. #
			    ##my $children = $fnode2ch{$node_id};
			    ##((defined $children) and (@{$children}>0)) or next; # Skip an exterior node.
			}
				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			$node2bw_gap_states{$node_name} = $cnct_bw_gstats;
			##$node2bw_gap_states{$node_id} = $cnct_bw_gstats;

				# %node2bw_gap_states = ($node_name => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and
				# [ %node2bw_gap_states = ($node_id => $concatenated_block_wise_gap_states, ...) for nodes specified by $fl_full_nodes, and ] --- for the predecessor 'parse_brief_output_infer_psm_indel_events'.

		    }						# End of the 1st inner-inner foreach-loop (over nodes). #

			# Read indel events mapped onto the branches. #

		    my $head_indels = <$fh_in>;
		    my $fld_tags_indels = <$fh_in>;

				# FURTHER MODIFIED (on Mar 9 (Sun), 2014) to cope with the new revision of 'MyTreeMap_indels.pm' that handles gap-only columns. #
		    my $info_indels = ($min_totct_indels > 0) ? <$fh_in> : '';
		    ##		#  MODIFIED (on Nov 26 (Tue), 2013) in order to avoid a buggy behavior when encountering gap-only columns. #
		    ##my $info_indels = ($ct_gaponly < $ct_rlv_nodes) ? <$fh_in> : '';
		    ####my $info_indels = <$fh_in>;

		    my %br2indel_events;

		    foreach my $line (split /\n+/, $info_indels) { # 2nd inner-inner foreach-loop (over indel-undergoing branches). #

			my @fields = split /\s+/, $line;
				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			my $br_le_name = shift @fields;
			##my $br_id = shift @fields;

			my @events_on_branch = ();

			foreach my $fld (@fields) { # Inner-most foreach-loop (over indel events on the branch). # 
			    my ($stat_ue, $stat_le, $cnct_blks) = split /\//, $fld;
			    my @involved_blocks = split /\,/, $cnct_blks;
			    push @events_on_branch, [$stat_ue, $stat_le, \@involved_blocks];

			}		# End of the inner-most foreach-loop (over indel events on the branch). #

				# MODIFIED in creation of 'parse_brief_output_infer_psm_indel_events_dawg'. #
			$br2indel_events{$br_le_name} = \@events_on_branch;
			##$br2indel_events{$br_id} = \@events_on_branch;

			    # [ %br2indel_events = ($branch_id => \@events_on_branch, ...) only for branches experiencing indel events, with ] --- for the predecessor 'parse_brief_output_infer_psm_indel_events'.
			    # %br2indel_events = ($branch_lower_end_name => \@events_on_branch, ...) only for branches experiencing indel events, with
			    # @{$events_on_branch[$e]} = ($state_upper_end, $state_lower_end, \@involved_blocks) for the $e th event

		    }				# End of the 2nd inner-inner foreach-loop (over indel-undergoing branches). #

		    $scenario_distribution[$scn_no] = [$rel_prob, \%node2bw_gap_states, \%br2indel_events];


			# @{$scenario_distribution[$ps]} = ($rel_prob, \%node2bw_gap_states, \%br2indel_events)
			#  is for the $ps th parsimonious indel scenario, with

		    ($scn_no == $ct_psm_scns - 1) and last; # We know that this is the last scenario. #


		}			# End of the 2nd inner while-loop (over parsimonious scenarios). #

		unless (@scenario_distribution == $ct_psm_scns) {
		    print STDOUT "[ERROR 'parse_brief_output_infer_psm_indel_events_dawg'] The size of ", '@scenario_distribution (=', scalar (@scenario_distribution), ") CLASHES with the number of parsimonious scenarios (=$ct_psm_scns) recorded in the heading: $! !\n";
		    close ($fh_in);
		    return (undef, undef, undef);
		}


		$sw_indel_state_distributions[2 * $cl_no + 1] = [$cl_start, $cl_end, 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution];

			# @{$sw_indel_state_distributions[2*$cl + 1]}
			# = ($cl_start, $cl_end, $state = 'X/-', \@gap_blocks, $min_totct_indels, \@scenario_distribution) 
			# is for the $cl th gap-block cluster, where

	    }			# End of the main middle while-loop (over clusters) #

	    $indel_state_distributions[$seg_no] = \@sw_indel_state_distributions;

		# @sw_indel_state_distributons = @{$indel_state_distributons[$s]}
		# stores the indel state distributions for the $s th segment, and
=cut

	}		# End of the main outer while-loop (over segments) #

	close ($fh_in);

	unless (@info_segments == $ct_segments) {
	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] The size of ", '@info_segments', " (=", scalar (@info_segments), ") CLASHES with the number of MSA segments (=$ct_segments): $! !\n";
	    return (undef, undef, undef);
	}

		# MODIFIED in creation of 'br_parse_detail_output_infer_psm_indel_events_dawg' on Mar 13 (Thu), 2014. #
	unless (@setset_info_cls == $ct_segments) {
	##unless (@indel_state_distributions == $ct_segments) {...} # For the predecessor 'parse_brief_output_infer_psm_indel_events_dawg'.

	    print STDOUT "[ERROR 'br_parse_detail_output_infer_psm_indel_events_dawg'] The size of ", '@setset_info_cls', " (=", scalar (@setset_info_cls), ") CLASHES with the number of MSA segments (=$ct_segments): $! !\n";
	    return (undef, undef, undef);
	}

		# MODIFIED in creation of 'br_parse_detail_output_infer_psm_indel_events_dawg' on Mar 13 (Thu), 2014. #
	return (\@parameters, \@info_segments, \@setset_info_cls);
	##return (\@parameters, \@info_segments, \@indel_state_distributions);

}	# End of the subroutine 'br_parse_detail_output_infer_psm_indel_events_dawg'


