package MyReadAlnProbIngredients;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyReadAlnProbIngredients.pm:
#	This package is a collection of subroutines
#	that read in sets of pre-computed ingredients, including the multiplication factors,
#	for the probability of an alignment,
#	and those that assign appropriate ingredients according to the pre-computed ones.
#
#	[CREATION STARTED on May 17 (Sun), 2015, ENDED on May xx (yyy), 2015.]
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	construct_label2indx const_indices_from_labels tfm_labels2indices tfm_labels2indices_gen
	);


use MyArrayManipu_num qw( diff_nmarray log_sum_exp binsearch_num );
	## union_nmarray

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite



use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	readin_delta_del
	assign_br2log_mfacs
	);


	# This specifies the level of tests on the main subroutines,
	# 'xxx', ...
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;

	# This specifies the level of tests on the primary satellite subroutines,
	# 'xxx', ...
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
our $TEST_LEVEL1 = 1;
##our $TEST_LEVEL1 = 2;
##our $TEST_LEVEL1 = 3;
##our $TEST_LEVEL1 = 5;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;


	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###

# readin_delta_del:
#	This subroutine reads a file recording the factor, '$delta_del',
#	which is essential for the calculation of the exit rate of an entire sequence.
#
# INPUT: \$infile, where
#	$infile is the path to the input file.
#
# OUTPUT: $delta_del.
#
sub readin_delta_del (\$) {

	my ($infile) = @_;
	local $/ = "\n"; # Use the newline as an input-record-separator. #

	my $fh_in = open_input_file2 (${$infile});
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_delta_del'] Subroutine 'open_input_file2' FAILED: $! !\n";
	    return undef;
	}

	my $delta_del;

	while (my $in = <$fh_in>) { # Main while-loop #

	    ($delta_del) = ($in =~ m/^Delta_del\s*\=\s*(\S+)/);
	    (defined $delta_del) and last;

	}	# End of the main while-loop. #
	close ($fh_in);

	unless (defined $delta_del) {
	    print STDOUT "[ERROR 'readin_delta_del'] I could NOT FIND the delta_del in the file '${$infile}': $! !\n";
	    return undef;
	}

	return $delta_del;

}	# End of the subroutine 'readin_delta_del'. #



# readin_sgl_sitect_vs_flds:
#	This satelite subroutine of 'assign_br2log_mfacs' reads in a file in the following format:
#
#	* [Possible comments]
#	*
#
#	[Corner_label]	[label_field]_1	[label_field]_2	...
#	
#	[Site_count]_1	{cells ...}
#	[Site_count]_2	{cells ...}
#	...
#
#	The, it stores the results in a matrix.
#
# INPUT: \$infile, where
#	$infile is the path to the input file.
#
# OUTPUT: (\@labels_fld, \@matrix), where
#
#	$labels_fld[$indx_fld] is the label for the $indx_fld th field. ($indx_fld is 0-based);
#
#	$matrix[$indx_fld]->[$site_count] is the value in the cell, which is at the $indx_fld th field
#	in the column for $site_count subject sites.
#
sub readin_sgl_sitect_vs_flds (\$) {

	my ($infile) = @_;
	local $/ = "\n"; # Use newline characters as input-record-separators.

	my $fh_in = open_input_file2 (${$infile}) ;
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_sgl_sitect_vs_flds'] 'open_input_file2' FAILED on '${$infile}': $! !\n\n";
	    return (undef, undef);
	}

	my $header ;

	while (my $in = <$fh_in>) { # Preliminary while-loop. #
	    ($in  =~ m/^\S/) or next;
	    ($in =~ m/^\*/) and next;
	    ($header = $in) =~ s/\s+$//;
	    last;
	}			# End of the preliminary while-loop. #

	my @labels_fld = split /\t/, $header;
	my $corner = shift @labels_fld;

	my $ct_flds = @labels_fld;

		# Initialize the main output. #
	my @matrix = ();
	for (1..$ct_flds) { push @matrix, []; }


		# Read and store the main content. # 

	while (my $in = <$fh_in>) { # Main while-loop (over input lines). # 

	   ($in =~ m/^\S/) or next;
	   $in =~ s/\s+$//;

	    my @flds = split /\t/, $in;
	    (@flds > $ct_flds) or next;

	    my $sitect = shift @flds;
	    unless (($sitect =~ m/^\d+$/) and ($sitect>0)) {
		print STDOUT "[ERROR 'reading_sgl_sitect_vs_flds'] Encountered an ANOMALOUS site-count (=$sitect) in the input line:\n";
		print STDOUT " '$in' : $! !\n\n";
		return (undef, undef);

	    }

	    for (my $i=0; $i<$ct_flds; $i++) { # Main inner for-loop (over fields). #

		my $fld = $flds[$i];
		($fld eq 'log(zero)') and ($fld = undef);

		$matrix[$i]->[$sitect] = $fld ;

	    }				# End of the main inner for-loop (over fields). #

	}			# End of the main while-loop (over input lines). #

	close ($fh_in);


	return (\@labels_fld, \@matrix);

}	# End of the subroutine 'readin_sgl_sitect_vs_flds'. #


# readin_dbl_sitects_vs_flds:
#	This satelite subroutine of 'assign_br2log_mfacs' reads in a file in the following format:
#
#	* [Possible comments]
#	*
#
#	[Corner_label]	[label_field]_1	[label_field]_2	...
#	
#	[Count_siteAs:count_siteBs]_1	{cells ...}
#	[Count_siteAs:count_siteBs]_2	{cells ...}
#	...
#
#	The, it stores the results in a matrix.
#
# INPUT: \$infile, where
#	$infile is the path to the input file.
#
# OUTPUT: (\@labels_fld, @threeD_array), where
#
#	$labels_fld[$indx_fld] is the label for the $indx_fld th field. ($indx_fld is 0-based);
#
#	$threeD_array[$indx_fld]->[$ct_siteAs][$ct_siteBs] is the value in the cell, which is at the $indx_fld th field
#	in the column for the pair of cite-counts, "$ct_siteAs:$ct_siteBs".
#
sub readin_dbl_sitects_vs_flds (\$) {

	my ($infile) = @_;
	local $/ = "\n"; # Use newline characters as input-record-separators.

	my $fh_in = open_input_file2 (${$infile}) ;
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_dbl_sitects_vs_flds'] 'open_input_file2' FAILED on '${$infile}': $! !\n\n";
	    return (undef, undef);
	}

	my $header ;

	while (my $in = <$fh_in>) { # Preliminary while-loop. #
	    ($in  =~ m/^\S/) or next;
	    ($in =~ m/^\*/) and next;
	    ($header = $in) =~ s/\s+$//;
	    last;
	}			# End of the preliminary while-loop. #

	my @labels_fld = split /\t/, $header;
	my $corner = shift @labels_fld;

	my $ct_flds = @labels_fld;

		# Initialize the main output. #
	my @threeD_array = ();
	for (1..$ct_flds) { push @threeD_array, []; }


		# Read and store the main content. # 

	while (my $in = <$fh_in>) { # Main while-loop (over input lines). # 

	   ($in =~ m/^\S/) or next;
	   $in =~ s/\s+$//;

	    my @flds = split /\t/, $in;
	    (@flds > $ct_flds) or next;

	    my $pair_cts = shift @flds;
	    my ($ct_siteAs, $ct_siteBs) = ($pair_cts =~ m/^(\d+)\:(\d+)$/);

	    unless ((defined $ct_siteAs) and (defined $ct_siteBs)) {
		print STDOUT "[ERROR 'reading_dbl_sitects_vs_flds'] Encountered an ANOMALOUS site-count pair (=$pair_cts) in the input line:\n";
		print STDOUT " '$in' : $! !\n\n";
		return (undef, undef);

	    }

	    for (my $i=0; $i<$ct_flds; $i++) { # Main inner for-loop (over fields). #

		my $fld = $flds[$i];
		($fld eq 'log(zero)') and ($fld = undef);

		my $array = $threeD_array[$i]->[$ct_siteAs];
		(defined $array) or ($array = $threeD_array[$i]->[$ct_siteAs] = []);
 		$array->[$ct_siteBs] = $fld;

	    }				# End of the main inner for-loop (over fields). #

	}			# End of the main while-loop (over input lines). #

	close ($fh_in);


	return (\@labels_fld, \@threeD_array);

}	# End of the subroutine 'readin_dbl_sitects_vs_flds'. #


# mk_num1Darray_linear:
#	This satellite subroutine of 'assign_br2log_mfacs'
#	creates a 1D array of numbers out of a set of 1D arrays	via linear interpolation.
#
#	[CREATED on May 18 (Mon), 2015.]
#
# INPUT: ($sbjct_fld, \@set_flds, \@set_1Darrays), where 
#
#	$sbjct_fld is the value of the subject field;
#
#	@set_flds stores the field values at which 1D arrays are available.
#	It is ASSUMED that the values in @set_flds are sorted in ASCENDING ORDER;
#
#	@{$set_1Darrays[$indx_fld]} is the 1D array associated with the field value $set_flds[$indx_fld].
#
# OUTPUT: \@out_1Darray, where
#	@out_1Darray is the 1D array created by the linear interpolation.
#
sub mk_num1Darray_linear ($\@\@) {

	my ($sbjct_fld, $set_flds, $set_1Darrays) = @_;

   (print STDOUT "  < Subroutine 'mk_num1Darray_linear' >\n\n") if ($TEST_LEVEL1>0) ;

	unless (@{$set_flds}>0) {
	    print STDOUT "[ERROR 'mk_num1Darray_linear'] ILLEGAL INPUT: EMPTY SET of FIELDS: $! !\n";
	    return undef;
	}
	my $ct_flds = @{$set_flds};
	unless ($ct_flds == @{$set_1Darrays}) {
	    my $ct_arrays = @{$set_1Darrays};
	    print STDOUT "[ERROR 'mk_num1Darray_linear'] ILLEGAL INPUT: The NUMBER OF AVAILABLE ARRAYS (=$ct_arrays) CLASHES with the NUMBER of FIELDS (=$ct_flds): $! !\n";
	    return undef;
	}

   if ($TEST_LEVEL1>1) {
	print STDOUT "  [ Inputs ]\n\n";
	print STDOUT '   $sbjct_fld = ', $sbjct_fld, " ;\n";
	print STDOUT '   @{$set_flds} = (', join (', ', @{$set_flds}), ") .\n\n";
   }

	if ($sbjct_fld <= 0) { # In this case, set all multiplication factors zero (i.e., their logarithms 'undef') !!
	    my $ct_els = @{$set_1Darrays->[0]};
	    my @out_1Darray = ();
	    for (my $k=0; $k<$ct_els; $k++) { $out_1Darray[$k] = undef; }
	    return \@out_1Darray ;
	}

	my $indx_num;
	my @indices_rlv = (0 .. $ct_flds -1);
	my ($if_exist, $indx2_qu, $indx_qu) = binsearch_num ($sbjct_fld, @{$set_flds}, $indx_num, @indices_rlv);
	unless (defined $if_exist) {
	    print STDOUT "[ERROR 'mk_num1Darray_linear'] Subroutine 'binsearch_num' FAILED: $! !\n";
	    return undef;
	}

   if ($TEST_LEVEL1>1) {
	print STDOUT "   [ Output of 'binsearch_num' ]\n\n";
	print STDOUT '   ($if_exist, $indx2_qu, $indx_qu) = ', "($if_exist, $indx2_qu, $indx_qu) .\n\n"; 
   }

	if ($if_exist) { # If the $sbjct_fld is in @{$set_flds}.

	    my $out_1Darray = $set_1Darrays->[$indx_qu];

	    return $out_1Darray ;
	}

	if ($indx_qu == $ct_flds -1) { # The $sbjct_fld is LARGER than ANY fields in @{$set_flds}.

		# We do NOT expect this situation!! #

	    print STDOUT "[ERROR 'mk_num1Darray_linear'] UNEXPECTED!! ... The subject (=$sbjct_fld) is LARGER THAN ANY FIELDS in the set: $! !\n\n";
	    return undef;

	}

	if ($indx_qu < 0) { # The $sbjct_fld is SMALLER than ANY fields in @{$set_flds}.

		# In this case, assume that the (original, non-logarithm)  multiplication factor is proportional to the smallest one. #

	    my $fld_smallest = $set_flds->[0];
	    my $smallest = $set_1Darrays->[0];
	    my $ct_els = @{$smallest};

	    my @out_1Darray = ();

	    my $diff = log ($sbjct_fld) - log ($fld_smallest) ;

	    for (my $i=0; $i<$ct_els; $i++) { # 1st main for-loop (over elements of the array). #
		my $element = $smallest->[$i];
		if (defined $element) {
		    $out_1Darray[$i] = $element + $diff ;
		} else {
		    $out_1Darray[$i] = undef;
		}
	    }				# End of the 1st main for-loop (over elements of the array). #

	    return \@out_1Darray ;
	    
	}

		# The normal situation we consider. #

	my $fld_lb = $set_flds->[$indx_qu]; # Lower-bound.
	my $fld_ub = $set_flds->[$indx_qu+1]; # Upper-bound.
	##my $fld_ub = $set_flds->[$indx_qu]; # Upper-bound. # A BUG fixed on May 18 (Mon), 2015.

	my $array_lb = $set_1Darrays->[$indx_qu]; # Lower-bound.
	my $array_ub = $set_1Darrays->[$indx_qu+1]; # Upper-bound.

	my $wgt_ub = ($sbjct_fld - $fld_lb) /($fld_ub - $fld_lb) ;
	##my $wgt_ub = (log ($sbjct_fld) - log ($fld_lb)) /(log($fld_ub) - log($fld_lb)) ;

	my $wgt_lb = 1 - $wgt_ub ;

	my $ct_els_lb = @{$array_lb};
	my $ct_els_ub = @{$array_ub};
	my $ct_els = ($ct_els_lb < $ct_els_ub) ? $ct_els_lb : $ct_els_ub ;

	my @out_1Darray = ();

	for (my $i=0; $i<$ct_els; $i++) { # 2nd main for-loop (over elements of the output array).

	    my $el_lb = $array_lb->[$i];
	    my $el_ub = $array_ub->[$i];

	    my $element = ((defined $el_lb) and (defined $el_ub)) ? $wgt_lb * $el_lb + $wgt_ub * $el_ub : undef ;

	    $out_1Darray[$i] = $element;

	}			# End of the 2nd may for-loop (over elements of the output array). #

   (print STDOUT "  < Subroutine 'mk_num1Darray_linear' ENDED />\n\n") if ($TEST_LEVEL1>0) ;

	return \@out_1Darray ;

}	# End of the subroutine 'mk_num1Darray_linear'. #


# mk_num2Darray_linear:
#	This satellite subroutine of 'assign_br2log_mfacs'
#	creates a 2D array of numbers out of a set of 2D arrays via linear interpolation.
#
#	[CREATED on May 18 (Mon), 2015.]
#
# INPUT: ($sbjct_fld, \@set_flds, \@set_2Darrays), where 
#
#	$sbjct_fld is the value of the subject field;
#
#	@set_flds stores the field values at which 1D arrays are available.
#	It is ASSUMED that the values in @set_flds are sorted in ASCENDING ORDER;
#
#	@{$set_2Darrays[$indx_fld]} is the 2D array associated with the field value $set_flds[$indx_fld].
#
# OUTPUT: \@out_2Darray, where
#	@out_2Darray is the 2D array created by the linear interpolation.
#
sub mk_num2Darray_linear ($\@\@) {

	my ($sbjct_fld, $set_flds, $set_2Darrays) = @_;

   (print STDOUT "  < Subroutine 'mk_num2Darray_linear' >\n\n") if ($TEST_LEVEL1>0);

	unless (@{$set_flds}>0) {
	    print STDOUT "[ERROR 'mk_num2Darray_linear'] ILLEGAL INPUT: EMPTY SET of FIELDS: $! !\n";
	    return undef;
	}
	my $ct_flds = @{$set_flds};
	unless ($ct_flds == @{$set_2Darrays}) {
	    my $ct_arrays = @{$set_2Darrays};
	    print STDOUT "[ERROR 'mk_num1Darray_linear'] ILLEGAL INPUT: The NUMBER OF AVAILABLE ARRAYS (=$ct_arrays) CLASHES with the NUMBER of FIELDS (=$ct_flds): $! !\n";
	    return undef;
	}

   if ($TEST_LEVEL1>1) {
	print STDOUT "  [ Inputs ]\n\n";
	print STDOUT '   $sbjct_fld = ', $sbjct_fld, " ;\n";
	print STDOUT '   @{$set_flds} = (', join (', ', @{$set_flds}), ") .\n\n";
   }

	if ($sbjct_fld <= 0) { # In this case, set all multiplication factors zero (i.e., their logarithms 'undef') !!

	    my $ct_rows = @{$set_2Darrays->[0]};

	    ##my $ct_clms = @{$set_2Darrays->[0][0]}; # A BUG fixed on May 22 (Fri), 2015.

	    my @out_2Darray = ();
	    for (my $i=0; $i<$ct_rows; $i++) {

			# ADDED to FIX the BUG above on May 22 (Fri), 2015. #
		my $in_row = $set_2Darrays->[0][$i];
		(defined $in_row) or next; # Skip in this case (most likely when $i=0). #
		my $ct_clms = @{$in_row};

		my @out_row = ();
		for (my $j=0; $j <$ct_clms; $j++) { $out_row[$j] = undef; }

		$out_2Darray[$i] = \@out_row;
	    }

	    return \@out_2Darray ;
	}

	my $indx_num;
	my @indices_rlv = (0 .. $ct_flds -1);
	my ($if_exist, $indx2_qu, $indx_qu) = binsearch_num ($sbjct_fld, @{$set_flds}, $indx_num, @indices_rlv);
	unless (defined $if_exist) {
	    print STDOUT "[ERROR 'mk_num2Darray_linear'] Subroutine 'binsearch_num' FAILED: $! !\n";
	    return undef;
	}

   if ($TEST_LEVEL1>1) {
	print STDOUT "   [ Output of 'binsearch_num' ]\n\n";
	print STDOUT '   ($if_exist, $indx2_qu, $indx_qu) = ', "($if_exist, $indx2_qu, $indx_qu) .\n\n"; 
   }

	if ($if_exist) { # If the $sbjct_fld is in @{$set_flds}.

	    my $out_2Darray = $set_2Darrays->[$indx_qu];

	    return $out_2Darray ;
	}

	if ($indx_qu == $ct_flds -1) { # The $sbjct_fld is LARGER than ANY fields in @{$set_flds}.

		# We do NOT expect this situation!! #

	    print STDOUT "[ERROR 'mk_num2Darray_linear'] UNEXPECTED!! ... The subject (=$sbjct_fld) is LARGER THAN ANY FIELDS in the set: $! !\n\n";
	    return undef;

	}

	if ($indx_qu < 0) { # The $sbjct_fld is SMALLER than ANY fields in @{$set_flds}.

		# In this case, assume that the (original, non-logarithm)  multiplication factor is proportional to the smallest one. #

	    my $fld_smallest = $set_flds->[0];
	    my $smallest = $set_2Darrays->[0];
	    my $ct_rows = @{$smallest};

	    my @out_2Darray = ();

	    my $diff = log ($sbjct_fld) - log ($fld_smallest) ;

	    for (my $i=0; $i<$ct_rows; $i++) { # 1st main for-loop (over rows of the 2D array). #

		my $row = $smallest->[$i];

		(defined $row) or next; # Skip in this case (most likely when $i =0). # A DEBUGGING MEASURE on May 22 (Fri), 2015. #

		my $ct_clms = @{$row};

		my @out_row = ();

		for (my $j=0; $j<$ct_clms; $j++) { # 1st inner for-loop (over columns). #

		    my $element = $row->[$j];

		    if (defined $element) {
			$out_row[$j] = $element + $diff;
		    } else {
			$out_row[$j] = undef;
		    }
		}			# End of the 1st inner for-loop (over columns). #

		$out_2Darray[$i] = \@out_row ;

	    }				# End of the 1st main for-loop (over rows of the 2D array). #

	    return \@out_2Darray ;
	    
	}

		# The normal situation we consider. #

	my $fld_lb = $set_flds->[$indx_qu]; # Lower-bound.
	my $fld_ub = $set_flds->[$indx_qu+1]; # Upper-bound.
	##my $fld_ub = $set_flds->[$indx_qu]; # Upper-bound. # A BUG fixed on May 18 (Mon), 2015.

	my $array2D_lb = $set_2Darrays->[$indx_qu]; # Lower-bound.
	my $array2D_ub = $set_2Darrays->[$indx_qu+1]; # Upper-bound.

	my $wgt_ub = ($sbjct_fld - $fld_lb) /($fld_ub - $fld_lb) ;
	##my $wgt_ub = (log ($sbjct_fld) - log ($fld_lb)) /(log($fld_ub) - log($fld_lb)) ;

	my $wgt_lb = 1 - $wgt_ub ;

	my $ct_rows_lb = @{$array2D_lb};
	my $ct_rows_ub = @{$array2D_ub};
	my $ct_rows = ($ct_rows_lb < $ct_rows_ub) ? $ct_rows_lb : $ct_rows_ub ;

	my @out_2Darray = ();

	for (my $i=0; $i<$ct_rows; $i++) { # 2nd main for-loop (over rows of the output 2D array).

	    my $row_lb = $array2D_lb->[$i];
	    my $row_ub = $array2D_ub->[$i];

	    ((defined $row_lb) and (defined $row_ub)) or next; # Skip in this case (most likely when $i=0). #

	    my $ct_clms_lb = @{$row_lb};
	    my $ct_clms_ub = @{$row_ub};
	    my $ct_clms = ($ct_clms_lb < $ct_clms_ub) ? $ct_clms_lb : $ct_clms_ub ; 

	    my @out_row = ();

	    for (my $j=0; $j<$ct_clms; $j++) { # 2nd inner for-loop (over columns of the output 2D array). #

		my $el_lb = $row_lb->[$j];
		my $el_ub = $row_ub->[$j];
		my $element = ((defined $el_lb) and (defined $el_ub)) ? $wgt_lb * $el_lb + $wgt_ub * $el_ub : undef ;

		$out_row[$j] = $element ;

	    }				# End of the 2nd inner for-loop (over columns of the output 2D array). #

	    $out_2Darray[$i] = \@out_row ;

	}			# End of the 2nd may for-loop (over rows of the output 2D array). #

   (print STDOUT "  < Subroutine 'mk_num2Darray_linear' ENDED />\n\n") if ($TEST_LEVEL1>0);

	return \@out_2Darray ;

}	# End of the subroutine 'mk_num2Darray_linear'. #


# assign_br2log_mfacs:
#	This subroutine first reads in pre-computed logarithmic multiplication factors
#	that are recorded in a set of input files.
#	Then it calculates the multiplication factors relevant for each of the branches.
#
#	[CREATION STARTED on May 17 (Sun), 2015, ENDED on May 18 (Mon), 2015.]
#	
#
# INPUT: (\%br2attr, $totrate_indels, \$indir, \@infilenames), where
#
#	%br2attr = ($branch_ID => \@attr, ...) for all branches,
#	with $attr[$INDX_LENGTH] being the branch length;
#
#	$totrate_indels is the total rate of indels (per site per unit time);
#
#	$indir is the path to the input directory;
#
#	@infilenames stores the names of the input files storing multiplication factors at grid points.
#	More precisely,
#	@infilenames = (
#		# Factors for the region in the bulk of the PWA. #
#	$filename_as_dg, # For cases with some ancestral sites and no descendant sites in between a pair of PASs. 
#	$filename_ag_ds, # For cases with no ancestral sites and some descendant sites in between a pair of PASs.
#	$filename_as_ds, # For cases with some ancestral sites and some descendant sites in between a pair of PASs.
#
#		# Factors for the region on either boundary of the PWA. #
#	$filename_bd_as_dg, # Some ancestral sites and no descendant sites.
#	$filename_bd_ag_ds, # No ancestral sites and some descendant sites.
#	$filename_bd_as_as, # Some ancestral sites and some descendant sites.
#	);
#
# OUTPUT: \%br2log_mfacs, where
#
#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,
#
#	$log_mfacs_ins/del_bd[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertions/deletions of $l sites on either boundary of the MSA,
#
#	$log_mfacs_pair[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites in the bulk of the MSA,
#
#	$log_mfacs_pair_bd[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites on either boundary of the MSA.
#
#	The subroutine returns undef if it encounters an error.
#
sub assign_br2log_mfacs (\%$\$\@) {
##sub asign_br2log_mfacs (\%$\$\@) {...} # A BUG fixed on May 18 (Mon), 2015. #

	my ($br2attr, $totrate_indels, $indir, $infilenames) = @_;

		# First, read in the pre-computed multiplication factors. #

	my ($filename_as_dg, $filename_ag_ds, $filename_as_ds,
	   $filename_bd_as_dg, $filename_bd_ag_ds, $filename_bd_as_ds) = @{$infilenames};

   (print STDOUT "<< Subroutine 'assign_br2log_mfacs' >>\n\n") if ($TEST_LEVEL>0);

   if ($TEST_LEVEL>2) {
	print STDOUT "[ Inputs ]\n\n";
	print STDOUT '$totrate_indels = ', $totrate_indels , " .\n\n";
	print STDOUT '%{$br2attr} = ( # $br_ID => [attributes] ', "\n";

	foreach my $br (sort {$a <=> $b} keys %{$br2attr}) {
	    print STDOUT '              ', $br, ' => [', join (', ', @{$br2attr->{$br}}), "],\n";
	}
	print STDOUT '              )', " .\n\n";
   }

	# @{$set_e_indelcts_xxxx} stores the expected numbers of indels used for the computation of the multiplication factors recorded in "$filename_xxxx". #
	#
	# $mtrx_lmfacs_xxxx->[$indx_fld][$nsites] is the multiplication factor for the number of sites = $nsites, recorded in the field, $set_e_indelcts_xxxx->[$indx_fld].
	#
	# $threeD_lmfacs_(bd_)as_as->[$indx_fld][$n_A][$n_D] is the multiplication factor for $n_A ancestral sites and $n_D descendant sites, recorded in the fields, $set_e_indelcts_(bd_)as_as->[$indx_fld].
	#

	my $file_as_dg = "${$indir}/$filename_as_dg";
	my ($set_e_indelcts_as_dg, $mtrx_lmfacs_as_dg) = readin_sgl_sitect_vs_flds ($file_as_dg);
	unless (defined $mtrx_lmfacs_as_dg) {
	    print STDOUT "[ERROR 'assign_br2log_mfacs'] 'readin_sgl_sites_vs_flds' FAILED on the input file '$file_as_dg': $! !\n";
	    return undef;
	}

	my $file_ag_ds = "${$indir}/$filename_ag_ds";
	my ($set_e_indelcts_ag_ds, $mtrx_lmfacs_ag_ds) = readin_sgl_sitect_vs_flds ($file_ag_ds);
	unless (defined $mtrx_lmfacs_ag_ds) {
	    print STDOUT "[ERROR 'assign_br2log_mfacs'] 'readin_sgl_sites_vs_flds' FAILED on the input file '$file_ag_ds': $! !\n";
	    return undef;
	}

	my $file_bd_as_dg = "${$indir}/$filename_bd_as_dg";
	my ($set_e_indelcts_bd_as_dg, $mtrx_lmfacs_bd_as_dg) = readin_sgl_sitect_vs_flds ($file_bd_as_dg);
	unless (defined $mtrx_lmfacs_bd_as_dg) {
	    print STDOUT "[ERROR 'assign_br2log_mfacs'] 'readin_sgl_sites_vs_flds' FAILED on the input file '$file_bd_as_dg': $! !\n";
	    return undef;
	}

	my $file_bd_ag_ds = "${$indir}/$filename_bd_ag_ds";
	my ($set_e_indelcts_bd_ag_ds, $mtrx_lmfacs_bd_ag_ds) = readin_sgl_sitect_vs_flds ($file_bd_ag_ds);
	unless (defined $mtrx_lmfacs_bd_ag_ds) {
	    print STDOUT "[ERROR 'assign_br2log_mfacs'] 'readin_sgl_sites_vs_flds' FAILED on the input file '$file_bd_ag_ds': $! !\n";
	    return undef;
	}

	my $file_as_ds = "${$indir}/$filename_as_ds";
	my ($set_e_indelcts_as_ds, $threeD_lmfacs_as_ds) = readin_dbl_sitects_vs_flds ($file_as_ds);
	unless (defined $threeD_lmfacs_as_ds) {
	    print STDOUT "[ERROR 'assign_br2log_mfacs'] 'readin_dbl_sitects_vs_flds' FAILED on the input file '$file_as_ds': $! !\n";
	    return undef;
	}

	my $file_bd_as_ds = "${$indir}/$filename_bd_as_ds";
	my ($set_e_indelcts_bd_as_ds, $threeD_lmfacs_bd_as_ds) = readin_dbl_sitects_vs_flds ($file_bd_as_ds);
	unless (defined $threeD_lmfacs_bd_as_ds) {
	    print STDOUT "[ERROR 'assign_br2log_mfacs'] 'readin_dbl_sitects_vs_flds' FAILED on the input file '$file_bd_as_ds': $! !\n";
	    return undef;
	}


		# Now, assign log-multiplication factors to each branch according to the retrieved ones. #

	my %ect2log_mfacs; # = ($expected_count => \@setset_log_mfacs, ...) for the ones used thus far.
	my %br2log_mfacs; # = ($branch_ID => \@setset_log_mfacs).

	foreach my $br_id (sort keys %{$br2attr}) { # Main foreach-loop (over branches). #

	    my $attr = $br2attr->{$br_id};
	    my $blen = $attr->[$INDX_LENGTH];

	    my $e_indelct = $totrate_indels * $blen ; # Expected number of indels. #

   (print STDOUT "  [ Branch: $br_id (length = $blen; Exp#{indels} = $e_indelct) ]\n\n") if ($TEST_LEVEL>1);

	    if (defined $ect2log_mfacs{$e_indelct}) {# If the desired set already exists... #
		$br2log_mfacs{$br_id} = $ect2log_mfacs{$e_indelct};
		next;
	    }

			# Create the arrays of multiplication factors. #

	    my $log_mfacs_ins = mk_num1Darray_linear ($e_indelct, @{$set_e_indelcts_ag_ds}, @{$mtrx_lmfacs_ag_ds});
	    unless (defined $log_mfacs_ins) {
		print STDOUT "[ERROR 'assign_br2log_mfacs'] 'mk_num1Darray_linear' FAILED for the effective insertions in the bulk: $! !\n\n";
		return undef;
	    }

	    my $log_mfacs_del = mk_num1Darray_linear ($e_indelct, @{$set_e_indelcts_as_dg}, @{$mtrx_lmfacs_as_dg});
	    unless (defined $log_mfacs_ins) {
		print STDOUT "[ERROR 'assign_br2log_mfacs'] 'mk_num1Darray_linear' FAILED for the effective deletions in the bulk: $! !\n\n";
		return undef;
	    }

	    my $log_mfacs_ins_bd = mk_num1Darray_linear ($e_indelct, @{$set_e_indelcts_bd_ag_ds}, @{$mtrx_lmfacs_bd_ag_ds});
	    unless (defined $log_mfacs_ins) {
		print STDOUT "[ERROR 'assign_br2log_mfacs'] 'mk_num1Darray_linear' FAILED for the effective insertions on either PWA boundary: $! !\n\n";
		return undef;
	    }

	    my $log_mfacs_del_bd = mk_num1Darray_linear ($e_indelct, @{$set_e_indelcts_bd_as_dg}, @{$mtrx_lmfacs_bd_as_dg});
	    unless (defined $log_mfacs_ins) {
		print STDOUT "[ERROR 'assign_br2log_mfacs'] 'mk_num1Darray_linear' FAILED for the effective deletions on either PWA boundary: $! !\n\n";
		return undef;
	    }

	    my $log_mfacs_pair = mk_num2Darray_linear ($e_indelct, @{$set_e_indelcts_as_ds}, @{$threeD_lmfacs_as_ds});
	    unless (defined $log_mfacs_pair) {
		print STDOUT "[ERROR 'assign_br2log_mfacs'] 'mk_num2Darray_linear' FAILED for the pairs of effective insertions and effective deletions in the bulk: $! !\n\n";
		return undef;
	    }

	    my $log_mfacs_pair_bd = mk_num2Darray_linear ($e_indelct, @{$set_e_indelcts_bd_as_ds}, @{$threeD_lmfacs_bd_as_ds});
	    unless (defined $log_mfacs_pair) {
		print STDOUT "[ERROR 'assign_br2log_mfacs'] 'mk_num2Darray_linear' FAILED for the pairs of effective insertions and effective deletions in the bulk: $! !\n\n";
		return undef;
	    }

			### RESTART FROM HERE ### 

	# @setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
	#
	# $log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,
	#
	# $log_mfacs_ins/del_bd[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertions/deletions of $l sites on either boundary of the MSA,
	#
	# $log_mfacs_pair[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites in the bulk of the MSA,
	#
	# $log_mfacs_pair_bd[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites on either boundary of the MSA.

	    my @setset_log_mfacs = ($log_mfacs_ins, $log_mfacs_del, $log_mfacs_pair, $log_mfacs_ins_bd, $log_mfacs_del_bd, $log_mfacs_pair_bd);

	    $br2log_mfacs{$br_id} = $ect2log_mfacs{$e_indelct} = \@setset_log_mfacs ;

	}				# End of the main foreach-loop (over branches). #

   (print STDOUT "\n<< Subroutine 'assign_br2log_mfacs' ENDED />>\n\n") if ($TEST_LEVEL>0);

	return \%br2log_mfacs ;

}	# End of the subroutine 'assign_br2log_mfacs'. #



1;	# a "spell" which makes the module "use"-able and/or "require"-able #


