package MyPopgenGenealogy;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyPopgenGenealogy.pm:
#	This package contains modules to perform population-genetics analyses
#	using genealogies of sampled chromosomes.
#	CREATION STARTED on Nov 12 (Sat))
#	

use strict;
use warnings;

use MyTaxonomy qw(
	fetch_common_ancestors fetch_ext_offsprings_nr
	);
	## fetch_ancestry 
	## fetch_last_common_ancestor_ov2
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_II qw(
	br_output_tree move_topnode
	);
	## input_tree_file read_tree_file
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

use MyTreeManipu_IV qw(
	br_contract_branch
	);
	## rm_external_node 

use MyTreeManipu_VI qw( leastsq_brlen_nn );
use MyTreeManipu_VI2 qw( neighbor_joining_tree );

use MyTreeMap_II qw(
	map_parsim_states_sankoff_1clm locate_parsim_changes
	);

	# For TESTING the package #
use MyMalnIO qw( pr_ODEN_input );

=set aside for later use
use MyArrayManipu qw( diff_array union_array nr_array);
use MyArrayManipu_num qw( diff_nmarray );

use MyTreeManipu_III qw(
	make_node2abs_depth
	);
=cut

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	list_nonred_partitions mk_pw_div_mtrx
	locate_midpoint enumerate_parsim_changes_1site
	incompl_neut_genealogy_via_nj
	);
	## min_penalties_sankoff_bo enum_node2parsstat_sankoff_td
	## map_parsim_states_sankoff_clm
	## map_parsim_states_sankoff_msa map_parsim_states_sankoff_1clm
	## map_parsim_gapstates_sankoff locate_parsim_changes
	## map_parsim_states_sankoff_clm

our $LEVEL_TEST = 1;
##our $LEVEL_TEST = 2;
	# = 0 if no monitoring outputs,
	# = 1 if you want brief monitoring outputs,
	# = 2 if you want a bit detailed monitoring outputs.

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.

	## These two indices specify where the information is stored
	## in %nodee2pars_stats = ($node_id => \@stat_info, ...) #
our $INDX_STAT = 0; # The state (or the character) mapped onto the node.
our $INDX_SBTOT_PEN = 1; # The subtotal of penalties on the subtree under $node_id. 

	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###

# list_nonred_partitions:
#	This subroutine receives a set of haplotypes, and
#	makes a list of non-redundant partitions.
#	NOTE that this subroutine eats up the input haplotypes.
#
# INPUT: (\@haplotypes, \@positions), where
#	${$haplotypes[$k]} is the haplotype of the $k th chromosome in the sample,
#	which is made by concatenating its states at the neutral segregating sites;
#	$positions[$l] is the position (i.e. chromosomal coordinate) of the $l th segregating site;
#	
# OUTPUT: \%partition2positions,  where
#	%partitoin2positions = ($partition => [\@rlv_indices, \@rlv_positions], ...)
#	with @rlv_indices listing the indices (in @positions) of the partition,
#	and with @rlv_positions listing the positions of the partition.
#
sub list_nonred_partitions (\@\@) {

	my ($haplotypes, $positions) = @_;

	my $ct_chrm = @{$haplotypes};
	unless ($ct_chrm>0) {
	    print STDOUT "[ERROR 'list_nonred_partitions'] INNEGAL INPUT: THe set of haplotypes is EMPTY : $! !\n";
	    return (undef, undef);
	}
	my $ct_segsites = length ${$haplotypes->[0]};
	unless ($ct_segsites>0) {
	    print STDOUT "[ERROR 'list_nonred_partitions'] ILLEGAL INPUT: The length of haplotype is ZERO: $! !\n";
	    return (undef, undef);
	}
	unless ($ct_segsites == @{$positions}) {
	    print STDOUT "[ERROR 'list_nonred_parritions'] ILLEGAL INPUT: The number of segregating site positions (", scalar (@{$positions}), ") CLASHES with the length of haplotypes ($ct_segsites): $! !\n";
	    return (undef, undef);
	}

	##	# Make a copy of haplotypes #
	##my @cp_haplotypes = ();
	##foreach my $haplo (@{$haplotypes}) { my $cp = ${$haplo}; push @cp_haplotypes, \$cp; }

	my %partition2positions;
	##my %partition2indices; # DISABLED on a BUG-Fixation (on Nov 13 (Sun), 2011)
	for (my $l=0; $l<$ct_segsites; $l++) { # Main outer for-loop #

	    my $partition = '';
	    for (my $k=0; $k<$ct_chrm; $k++) {  # Main inner for-loop #
		my $char = substr (${$haplotypes->[$k]}, 0, 1, '');
		##my $char = substr (${$cp_haplotypes[$k]}, 0, 1, '');
		$partition .= $char;
	    }				# End of the main inner for-loop #
	    my $pos = $positions->[$l];

	    if (defined $partition2positions{$partition}) {
		    # A BUG Fixed (on Nov 13 (Sun), 2011)
		push @{$partition2positions{$partition}->[0]}, $l;
		##push @{$partition2indices{$partition}->[0]}, $l;

		push @{$partition2positions{$partition}->[1]}, $pos;
	    } else {
		$partition2positions{$partition} = [[$l], [$pos]];
	    }

	}			# End of the main outer for-loop #

	return \%partition2positions;

}		# End of the subroutine 'list_nonred_partitions' #


# mk_pw_div_mtrx:
#	This subroutine receives a list of non-redundant partitions,
#	and construct a matrix of pairwise chromosome divergences.
#	It assumes that all the characters in the input haplotypes are comparable.
#
# INPUT: ($ct_chrm, \%partition2ct_info, $indx_ct, $indx_pos),  where
#	$ct_chrm is the number of chromosomes to be compared;
#	%partition2ct_info = ($partition => $ct_info, ...) with
#	$ct_info = #{partitions} if $indx_ct = undef,
#	$ct_info->[$indx_ct] = #{partitoins} if $indx_ct >= 0, or
#	$ct_info = \@positions if $indx_ct < 0, with
#	@{$positions[$indx_pos]} (or @positions itself if $indx_pos = undef) storing the the positions that show the partition.
#
# OUTPUT: \@pw_div_mtrx, where
#	$pw_div_mtrx[$i]->[$j] (0 <= $j < $i <$ct_chrm) is the pairwise divergence, namely the number of partitions differentiating, between the $i th and $j th chromosomes.
#
sub mk_pw_div_mtrx ($\%$$) {

	my ($ct_chrm, $partition2ct_info, $indx_ct, $indx_pos) = @_;

		# Initialize the output matrix #
	my @pw_div_mtrx = ();
	for (my $i=0; $i<$ct_chrm; $i++) { # Preliminary inner for-loop #
	    my @row = ();
	    for (my $j=0; $j<$i; $j++) { $row[$j] = 0; }
	    $pw_div_mtrx[$i] = \@row;
	}			# End of the preliminary inner for-loop #

	foreach my $partition (sort keys %{$partition2ct_info}) { # Main outer foreach-loop (over partitions) #

	    my @chars = split //, $partition;
	    unless ($ct_chrm == @chars) {
		print STDOUT "[ERROR 'mk_pw_div_mtrx'] The partition '$partition' contains ", scalar (@chars), " characters, which CLASHES with the number of chromosomes to be compared ($ct_chrm): $! !\n";
		return (undef, undef);
	    }
	    my $ct_info = $partition2ct_info->{$partition};
	    my $ct_ptit = $ct_info;
	    if (defined $indx_ct) {
		if ($indx_ct>=0) {
		    $ct_ptit = $ct_info->[$indx_ct];
		} else {
		    my $positions = (defined $indx_pos) ? $ct_info->[$indx_pos] : $ct_info ;
		    $ct_ptit = @{$positions};
		}
	    }

	    for (my $i=0; $i<$ct_chrm; $i++) { # Main middle for-loop (over chromosomes) #
		my $char1 = $chars[$i];
		my $row = $pw_div_mtrx[$i];
		for (my $j=0; $j<$i; $j++) { ($char1 ne $chars[$j]) and ($row->[$j] += $ct_ptit); }
	    }			# End of the main middle for-loop (over chromosomes) #
	    
	}			# End of the main outer foreach-loop (over partitions) #

	return \@pw_div_mtrx;

}	# End of the subroutine 'mk_pw_div_mtrx' #


# locate_midpoint:
#	This subroutine locates a mid-point between a pair of nodes in a tree.
#
# INPUT: ($node1, $node2, \%node2pa, \%br_attr, $indx_len), where
#	$node1 and $node2 are the pair of nodes to be examined;
#	%node2pa = ($node_id => $parent, ...);
#	%br_attr = ($br_id => $attr, ...) with
#	$attr->[$indx_len], or $attr itself if $indx_len = undef, giving the branch length.
#
# OUTPUT: ($midpoint1, $midpoint2), where
#	$midpoint1 and $midpoint2 are node IDs that specify where the mid-point should be located.
#	$midpoint2 is undef if the mid-point is exactly at $midpoint1.
#	Otherwise, $midpoint2 = $node2pa->{$midpoint1}.
#
sub locate_midpoint ($$\%\%$) {

	my ($node1, $node2, $node2pa, $br_attr, $indx_len) = @_;

	# @{$common_ancs} = ($top_node, ..., $node_MRCA);
	# @{$lineage_chr1/2} = ($first_node_after_divergence, ..., $node1/2)
	#
	my ($common_ancs, $lineage_chr1, $lineage_chr2)
		= fetch_common_ancestors ($node1, $node2, %{$node2pa});
	unless (defined $common_ancs) {
	   print STDOUT "[ERROR 'locate_midpoint'] 'fetch_common_ancestors' FAILED: $! !\n";
	   return (undef, undef);
	}

	my $len_lin1 = my $len_lin2 = 0;
	foreach my $pair ([\$len_lin1, $lineage_chr1], [\$len_lin2, $lineage_chr2]) { # 1st outer foreach-loop #
	    my ($rf_len_lin, $lineage_chr) = @{$pair};
	    foreach my $br (@{$lineage_chr}) { # 1st inner foreach-loop #
		my $attr = $br_attr->{$br};
		my $brlen = (defined $indx_len) ? $attr->[$indx_len] : $attr ;
		${$rf_len_lin} += $brlen;
	    }			# End of the 1st inner foreach-loop #
	}			# End of the 1st outer foreach-loop #

	if ($len_lin1 == $len_lin2) { # The mid-point is exactly at the MRCA of the two nodes #
	    my $mrca = pop @{$common_ancs};
	    return ($mrca, undef);
	}

	my $mid_len = ($len_lin1 + $len_lin2)/2;

		# Explore the longer lineage #

	my ($to_be_searched, $len_lin) = ($len_lin1 > $len_lin2) ?
		($lineage_chr1, $len_lin1) : ($lineage_chr2, $len_lin2) ;

	while (@{$to_be_searched}>0) { # 2nd while-loop # 
	    my $mp1 = shift @{$to_be_searched}; 
	    my $attr = $br_attr->{$mp1};
	    my $brlen = (defined $indx_len) ? $attr->[$indx_len] : $attr ;
	    $len_lin -= $brlen;

	    if ($len_lin == $mid_len) { # The node is the mid-point #
		return ($mp1, undef);
	    } elsif ($len_lin < $mid_len) { # The node comes below the mid-point for the first time #
		my $mp2 = $node2pa->{$mp1};
		return ($mp1, $mp2);
	    }
	}			# End of the 2nd while-loop #

		# An ANOMALOUS END #

	print STDOUT "[ERROR 'locate_midpoint'] Somehow, the mid-point between '$node1' and '$node2' was NOT FOUND, I will leave here: $! !\n";
	return (undef, undef);

}		# End of the subroutine 'locate_midpoint' #


# enumerate_parsim_changes_1site:
#	This subroutine receives a genealogy of sequences and a set of charactersthat the sequences take at a site (or locus),
#	and returns ALL the sets of changes that most parsimoniously explain the current state.
#
# INPUT: ($name_topnode, \@genealogy, \@seqnames, \@site_state, \@charlist, \@pen_mtrx), where
#	$name2id{$name_topnode} = $id_topnode is the ID of the top node;
#	@genealogy = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr),
#	whose elements follow the format specified in 'MyTreeManipu_II.pm';
#	$seqnames[$k] is the name of the $k th sequence in an entry;
#	$site_state[$k] is the character (trait) of the $k th sequence;
#	@charlist list the possible characters;
#	$pen_mtrx[$i]->[$j] gives the penalty for the change from $charlist[$i] to $charlist[$j].
#
# OUTPUT: (\@anc_states, \@sets_parsim_chgs, \@cts_parsim_chgs, \@parsim_penalties), where
#	$anc_states[$s] is the ancestral state for the $s th scenario of the most parsimonious changes at the site;
#	%{$sets_parsim_chgs[$s]} = ($br_id => [$stat_parent, $stat_node], ...), where $br_id is actually the ID of the node immediately below the branch in question,
#	gives the $s th scenario of the most parsimonious changes at the site;
#	$cts_parsim_chgs[$s] gives the number of changes in the $s th scenario of the most parsimonious changes;
#	$parsim_penalties[$s] gives the total penalty on the $s th scenario of the most parsimonious changes;
#
#
#
sub enumerate_parsim_changes_1site ($\@\@\@\@\@) {

	my ($name_topnode, $genealogy, $seqnames, $site_state, $charlist, $pen_mtrx) = @_;

	my ($node2pa, $node2ch, $name2id, $id2name, $br_attr) = @{$genealogy};
	my $id_topnode = $name2id->{$name_topnode};
	unless (defined $id_topnode) {
	    print STDOUT "[ERROR 'enumerate_parsim_changes_1site'] CanNOT FIND an ID for the top node (name: $name_topnode): $! !\n";
	    return (undef, undef, undef, undef);
	}

	my $fl_full_mode = 1; # This has to be 1, because information on ALL nodes is needed.
	# %{$set_node2parsstat->[$s]} = ($node_id => [$char, $sum_penal], ...) for the $s th scenario of maximum parsimonious changes;
	#
	my $set_node2parsstat = map_parsim_states_sankoff_1clm ($id_topnode, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$site_state}, @{$charlist}, @{$pen_mtrx}, $fl_full_mode);
	unless (defined $set_node2parsstat) {
	    print STDOUT "[ERROR 'enumerate_parsim_changes_1site'] The subroutine 'map_parsim_states_sankoff_1clm' FAILED: $! !\n";
	    return (undef, undef, undef);
	}
	my $ct_pars_scenarios = @{$set_node2parsstat};

	my $indx_stat = 0; # The index specifying where the state is stored in @{$set_node2parsstat->[$i]{$node_id}}.
	my $indx_penal = 1; # The index specifying where the sub-total penalty is stored in @{$set_node2parsstat->[$i]{$node_id}}.

	my @anc_states = ();
	my @sets_parsim_chgs = ();
	my @counts_parsim_chgs = ();
	my @parsim_penalties = ();

	for (my $i = 0; $i<$ct_pars_scenarios; $i++) { # The main for-loop #

	    my $node2parsstat = $set_node2parsstat->[$i];

	    my $topnode_parsstat = $node2parsstat->{$id_topnode};
	    my $anc_state = $topnode_parsstat->[$indx_stat]; # ADDED afterwards on Nov 6 (Sun), 2011 #
	    my $total_penalty = $topnode_parsstat->[$indx_penal];

	    my ($br2change, $ct_changes) = locate_parsim_changes (%{$node2pa}, %{$node2parsstat}, $indx_stat);
	    unless (defined $br2change) {
		print STDOUT "[ERROR 'enumerate_parsim_changes_1site'] The subroutine 'locate_parsim_changes' FAILED on the $i th out of the $ct_pars_scenarios scenario of the most parsimonious changes: $! !\n";
		return (undef, undef, undef, undef);
	    }

	    push @anc_states, $anc_state;
	    push @sets_parsim_chgs, $br2change;
	    push @counts_parsim_chgs, $ct_changes;
	    push @parsim_penalties, $total_penalty;

	}			# End of the main for-loop #

	return (\@anc_states, \@sets_parsim_chgs, \@counts_parsim_chgs, \@parsim_penalties);

}	# End of the subroutine 'enumerate_parsim_changes_1site' #


# incompl_neut_genealogy_via_nj:
#	This subroutine infers a genealogy of sampled chromosomes
#	from a set of putatively neutral segregating sites by
#	(i) first reconstructing a neighbor-joining (NJ) tree (Saitou N and Nei M. 1987. 4:406-425)
#	from a set of haplotypes
#	[and removing branches with negative lengths (--- it's unnecessary, come to think of it)], then
#	(ii) rooting the NJ tree at the mid-point of the most divergent pair 
#		(if there are more than one such pair, choose one arbitrarily), then
#	(iii) removing the branches not supported by any partitions (i.e. polymorphism patterns),
#	 while pointing out partitions conflicting with the NJ tree, and
#	(iv) finally rescaling the NJ tree branches by the factor:
#	    (1+1/2 +...+1/(#{chromosomes}-1)) /#{segregating sites consistent with the NJ tree} .
#
#	NOTE that this subroutine eats up the input haplotypes.
#	(CREATED on Nov 12 (Sat), 2011)
#
# INPUT: (\@haplotypes, \@chr_names, \@positions, $putat_anc_state, \@in_charlist, \@in_pen_mtrx), where
#	${$haplotypes[$k]} is the haplotype of the $k th chromosome in the sample,
#	which is made by concatenating its states at the neutral segregating sites;
#	$chr_names[$k] is the name of the $k th chromosome;
#	$positions[$l] is the position (i.e. chromosomal coordinate) of the $l th segregating site;
#	$putat_anc_state is the character representing the putative ancestral state (= undef if no particular single state is specified);
#	@in_charlist is a list of characters to appear,
#	it is empty if it is not known in advance what characters will appear;
#	$in_pen_mtrx[$i]->[$j] (0 <= $i, $j < $#in_charlist) is the penalty
#	on a change from $in_charlist[$i] to $in_charlist[$j],
#	@in_pen_mtrx is empty if the list of characters is unknown yet.
#
# OUTPUT: (\@genealogy, \@consist_partitions, \@inconsist_partitions, \@anc_inconsist_partitions, \@pw_div_mtrx, \@unsupported_branches), where
#	@genealogy = (\%node2pa, \%node2ch, \%name2id, \%id2name, \%br_attr) represents
#	the inferred genealogy, with each component following the format in 'MyTreeManipu_II.pm'; 
#	@{$consist_partitions[$c]} = (\$partition, \@rlv_positions, \@pars_scenarios)
#	is for the $c th partition that is consistent with the inferred genealogy;
#	@{$inconsist_partitions[$i]} = (\$partition, \@rlv_positions, \@pars_scenarios)
#	is for the $i th partition that is NOT consistent with the inferred genealogy (namely that cannot be represented by a single change along the genealogy), with
#	$partition being a string representing the partition, with
#	@rlv_positions storing all the positions showing the partition, and with
#	%{$pars_scenarios[$s]} = ($br_id => [$stat_upper, $stat_lower]), where $stat_upper/lower gives the state at the upper/lower end of the branch, describing the $s th parsimonious scenario;
#	@{$anc_inconsist_partitions[$i]} = (\$partition, \@rlv_positions, \@anc_states, \@pars_scenarios)
#	is  for the $i th partition that is consistent with the inferred genealogy, but none of whose inferred ancestral states is $putat_anc_state, with
#	$partition being a string representing the partiton, with
#	@rlv_positions storing all the positions with the partition, and with
#	@anc_states storing all the possible ancestral states;
#	$pw_div_mtrx[$i]->[$j] (0 <= $j < $i < #chromosomes) gives the pairwise divergence (, which is simply the number of state differences) between the $i th and $j th chromosomes;
#	@{$unsupported_branches[$u]} = ($br_id, \@br_attr_nj, \@ext_descendants)
#	is for the $u th branch removed either because it had non-positive branch length
#	or because it was NOT supported by the partitions (and therefore removed),
#	with @br_attr_nj describing the branch attribute of the NJ tree,
#	and with @ext_descendants listing the branch's descendants that are exterior nodes.
#	
sub incompl_neut_genealogy_via_nj (\@\@\@$\@\@) {

	my ($haplotypes, $chr_names, $positions, $putat_anc_state, $in_charlist, $in_pen_mtrx) = @_;

   (print STDOUT "\n< Subroutine 'incompl_neut_genealogy_via_nj' >\n\n") if ($LEVEL_TEST>0);

	my $ct_chrm = @{$haplotypes};
	unless ($ct_chrm>0) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] INNEGAL INPUT: THe set of haplotypes is EMPTY : $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
	unless ($ct_chrm == @{$chr_names}) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] ILLEGAL INPUT: The number of chromosome names (", scalar (@{$chr_names}), ") CLASHES with the number of chromosomes ($ct_chrm): $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
	my $ct_segsites = length ${$haplotypes->[0]};
	unless ($ct_segsites>0) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] ILLEGAL INPUT: The length of haplotype is ZERO: $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
	unless ($ct_segsites == @{$positions}) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] ILLEGAL INPUT: The number of segregating site positions (", scalar (@{$positions}), ") CLASHES with the length of haplotypes ($ct_segsites): $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

	my $size_in_charlist = @{$in_charlist};
	unless ($size_in_charlist == @{$in_pen_mtrx}) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] ILLEGAL INPUT: The size of input penalty matrix (", scalar (@{$in_pen_mtrx}), ") CLASHES with that of input character list ($size_in_charlist): $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

   (print STDOUT "#{Chromosomes} = $ct_chrm\n\n") if ($LEVEL_TEST>0);
   if ($LEVEL_TEST>1) {
	print STDOUT "Haplotypes:\n\n";
	unless ( pr_ODEN_input (@{$chr_names}, @{$haplotypes}, 50, *STDOUT) ) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] 'pr_ODEN_input' FAILED to print out haplotypes : $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
	print STDOUT "\n";
   } elsif ($LEVEL_TEST>0) {
	print STDOUT "Chromosome names:  ", join (', ', @{$chr_names}), ".\n\n";
   }
   (print STDOUT "#{Segregating sites} = $ct_segsites\n\n") if ($LEVEL_TEST>0);
   (print STDOUT "Positions of segregating sites: ", join (', ', @{$positions}), ".\n\n") if ($LEVEL_TEST>1);
   (print STDOUT "Putative ancestral state:  ", ((defined $putat_anc_state) ? $putat_anc_state : 'undef'), "\n\n") if ($LEVEL_TEST>0);

	# %{$partition2positions} = ($partition => [\@rlv_indices, \@rlv_positions], ...)
	# with @rlv_indices listing the indices (in @positions) of the partition,
	# and with @rlv_positions listing the positions of the partition.
	#
	my $partition2positions = list_nonred_partitions (@{$haplotypes}, @{$positions});
	unless (defined $partition2positions) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] The subroutine 'list_nonred_partitions' FAILED : $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

   (print STDOUT " --- 'list_nonred_partitions' ENDED ---\n\n") if ($LEVEL_TEST>0);
   if ($LEVEL_TEST>1) {
	print STDOUT "{ Output of 'list_nonred_partitions' }\n\n";
	foreach my $partition (sort keys %{$partition2positions}) {
	    my $pos_info = $partition2positions->{$partition};
	    print STDOUT "Partition: '$partition'\n";
	    print STDOUT "Indices: ", join (', ', @{$pos_info->[0]}), "\n";
	    print STDOUT "Positions: ", join (', ', @{$pos_info->[1]}), "\n";
	    print STDOUT "\n";
	}
   }

	# $pw_div_mtrx->[$i][$j] (0 <= $j<$i<$ct_chrm) is the pairwise divergence (i.e. the number of state differences) between the $i th and the $j th chromosomes.
	#
	my ($indx_ct, $indx_pos) = (-1, 0); # These indices specify how to get the number of partitions fro %{$partition2positions}.
	my $pw_div_mtrx = mk_pw_div_mtrx ($ct_chrm, %{$partition2positions}, $indx_ct, $indx_pos);
	unless (defined $pw_div_mtrx) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] The subroutine 'mk_pw_div_mtrx' FAILED: $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
   (print STDOUT " --- 'mk_pw_div_mtrx' ENDED ---\n\n") if ($LEVEL_TEST>0);
   if ($LEVEL_TEST>1) {
	print STDOUT "{ Pairwise divergences output by 'mk_pw_div_mtrx' }\n\n";
	print STDOUT join ("'\t'", "'Chr1'\'Chr2", @{$chr_names}), "'\n\n";
	for (my $k=0; $k<$ct_chrm; $k++) { print STDOUT join ("\t", "'$chr_names->[$k]'", @{$pw_div_mtrx->[$k]}), "\n"; } 
	print STDOUT "\n";
   }

	my ($name2id, $id2name, $node2pa, $node2ch, $br_attr_nj)
		= neighbor_joining_tree (@{$chr_names}, @{$pw_div_mtrx});
	unless (defined $name2id) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] The subroutine 'neighbor_joining_tree' FAILED: $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
   (print STDOUT " --- 'neighbor_joining_tree' ENDED ---\n\n") if ($LEVEL_TEST>0);
   if ($LEVEL_TEST>1) {
	my $topnode = $name2id->{$TOP_NODE};
	my $option_bl = 1; # =1 if you display bootstrap at the branch, = 0 at the node.
	my $option_ni = 1; # = 1 if you display node ID, = 0 otherwise.
	print STDOUT "{ Raw NJ tree output by 'neighbor_joining_tree' (top_node: $topnode) }\n\n";
	unless ( br_output_tree ($topnode, %{$node2ch}, %{$id2name}, %{$br_attr_nj}, *STDOUT, $option_bl, $option_ni) ) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] 'br_output_tree' FAILED while testing the subroutine : $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
	print STDOUT "\n";
   }

=This may be useless in the following processes.
		# Re-estimate the branch lengths using a least-square method #
	my ($power, $buffer) = (1, 1); # This parameter choice is equivalent to that for the quasi-ordinary LS framework with $alpha = 1.
	my $br_attr_ls = leastsq_brlen_nn (@{$chr_names}, @{$pw_div_mtrx}, %{$name2id}, %{$node2pa}, $power, $buffer);
	unless (defined $br_attr_ls) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] The subroutine 'leastsq_brlen_nn' FAILED: $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
=cut

	    ## Root the NJ tree at the mid-point between the most separated pair of chromosomes ##

	my ($max_div, $max_i, $max_j) ;

	for (my $i=0; $i<$ct_chrm; $i++) { # 1st outer for-loop (over 1st chromosomes) #
	    my $row = $pw_div_mtrx->[$i];
	    for (my $j=0; $j<$i; $j++) { # 1st inner for-loop (over 2nd chromosomes) #
			# A BUG-Fixed on Nov 13 (Sun), 2011 
		my $div = $row->[$j];
		##my $div = $row->[$i];
		unless ((defined $max_div) and ($max_div >= $div)) {
		    ($max_div, $max_i, $max_j) = ($div, $i, $j);
		}
	    }			# End of the 1st inner for-loop (over 2nd chromosomes) #
	}		# End of the 1st outer for-loop (over 1st chromosomes) #


		# Names and IDs of the most separated pair of chromosomes #
	my $name_mx_chr1 = $chr_names->[$max_i];
	my $name_mx_chr2 = $chr_names->[$max_j];
	my $nodeid_mx_chr1 = $name2id->{$name_mx_chr1};
	my $nodeid_mx_chr2 = $name2id->{$name_mx_chr2};

    (print STDOUT "\n{Maximum divergence} = $max_div (between the $max_i th and $max_j th chromosomes)\n\n") if ($LEVEL_TEST>0);
   if ($LEVEL_TEST>1) {
	print STDOUT "The $max_i th chromosome: (name: $name_mx_chr1, node_ID: $nodeid_mx_chr1)\n";
	print STDOUT "The $max_j th chromosome: (name: $name_mx_chr2, node_ID: $nodeid_mx_chr2)\n";
	print STDOUT "\n\n";
   }

	# $midpoint2 = undef if the mid-point is exactly at $midpoint1,
	#            = $node2pa->{$midpoint1} otherwise. 
	#
	my ($midpoint1, $midpoint2) = locate_midpoint ($nodeid_mx_chr1, $nodeid_mx_chr2, %{$node2pa}, %{$br_attr_nj}, $INDX_LENGTH);
	unless (defined $midpoint1) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] The subroutine 'locate_midpoint' FAILED to locate the mid-point between '$name_mx_chr1' (ID: $nodeid_mx_chr1) and '$name_mx_chr2' (ID: $nodeid_mx_chr2):  $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

    (print STDOUT "\n --- 'locate_midpoint' ENDED ---\n\n") if ($LEVEL_TEST>0);
    if ($LEVEL_TEST>1) {
	print STDOUT "{Mid-point (lower)} = $midpoint1\n";
	print STDOUT "{Mid-point (upper} = ", ((defined $midpoint2) ? $midpoint2 : 'undef'), "\n";
	print STDOUT "\n\n";
    }

	my $old_topnode = $name2id->{$TOP_NODE};
	if (defined $midpoint2) {
	    my $max_nodeid = $old_topnode;
	    foreach my $node (keys %{$node2pa}) { ($max_nodeid < $node) and ($max_nodeid = $node); }
	    unless ($old_topnode == $midpoint2) {
		unless ( move_topnode ($midpoint2, %{$node2pa}, %{$node2ch}, %{$name2id}, %{$br_attr_nj}) ) {
		    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] The subroutine 'move_topnode' FAILED to move the top node from '$old_topnode' to '$midpoint2': $! !\n";
		    return (undef, undef, undef, undef, undef);
		}
	    }
		
		# Create a new set of children of $midpoint2
		# by removing $midpoint1 from the original set #
		# (A BUG-Fixation on Nov 13 (Sun), 2011)
	    my @new_ch_midpoint2 = ();
	    foreach my $ch (@{$node2ch->{$midpoint2}}) {
		($ch == $midpoint1) or (push @new_ch_midpoint2, $ch);
	    }
	    $node2ch->{$midpoint2} = \@new_ch_midpoint2;

		# Create a new top-node between $midpoint1 and $midpoint2 #
		# (don't care about the resulting branch attributes) #
	    my $new_topnode = $max_nodeid + 1;
	    $node2ch->{$new_topnode} = [$midpoint1, $midpoint2];
	    $node2pa->{$midpoint1} = $new_topnode;
	    $node2pa->{$midpoint2} = $new_topnode;
	    $name2id->{$TOP_NODE} = $new_topnode;

		# Place the top-node exactly at the mid-point #
	    my @attr_mp1 = @{$br_attr_nj->{$midpoint1}};
	    $attr_mp1[$INDX_LENGTH] /= 2;
	    my @attr_mp2 = @attr_mp1;
	    $br_attr_nj->{$midpoint1} = \@attr_mp1;
	    $br_attr_nj->{$midpoint2} = \@attr_mp2;

	} else { # mid-point is exactly at $midpoint1 #
	    unless ($old_topnode == $midpoint1) {
		unless ( move_topnode ($midpoint1, %{$node2pa}, %{$node2ch}, %{$name2id}, %{$br_attr_nj}) ) {
		    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] The subroutine 'move_topnode' FAILED to move the top node from '$old_topnode' to '$midpoint1': $! !\n";
		    return (undef, undef, undef, undef, undef);
		}
	    }
	}
	my $new_topnode = $name2id->{$TOP_NODE};
	my @genealogy = ($node2pa, $node2ch, $name2id, $id2name, $br_attr_nj);

   (print STDOUT "\n --- finished rooting the NJ tree --\n\n") if ($LEVEL_TEST>0);
   if ($LEVEL_TEST>1) {
	print STDOUT "{ Rooted NJ tree (new top node: $new_topnode) }\n\n";
	my $option_bl = 1; # =1 if you display bootstrap at the branch, = 0 at the node.
	my $option_ni = 1; # = 1 if you display node ID, = 0 otherwise.
	unless ( br_output_tree ($new_topnode, %{$node2ch}, %{$id2name}, %{$br_attr_nj}, *STDOUT, $option_bl, $option_ni) ) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] 'br_output_tree' FAILED while testing the subroutine: $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
   }

		# Remove interior branches in the NJ tree with non-positive branch lengths #
	my @unsupported_branches = ();

	foreach my $br (sort keys %{$node2ch}) { # preliminary foreach-loop (over interior branches) #
	    my $attr = $br_attr_nj->{$br};
	    (defined $attr) or next; # Skip the top-node #
	    my $brlen = $attr->[$INDX_LENGTH];
	    ($brlen>0) and next;

	    my @attr_nj = @{$br_attr_nj->{$br}};
	    my $ext_descs = fetch_ext_offsprings_nr ($br, %{$node2ch});

	    unless (br_contract_branch ($br, %{$node2pa}, %{$node2ch}) ) {
		print STDOUT "[ERROR 'imcompl_neut_genealogy_via_nj'] The subroutine 'br_contract_branch' FAILED on the branch ID '$br' : $! !\n";
		return (undef, undef, undef, undef, undef);
	    }
	    delete $br_attr_nj->{$br};

	    push @unsupported_branches, [$br, \@attr_nj, $ext_descs];
	}			# End of the preliminary foreach-loop (over interior branches) #

   (print STDOUT "\n --- finished removing interior branches with non-positive branch lengths --\n\n") if ($LEVEL_TEST>0);
   if ($LEVEL_TEST>1) {
	print STDOUT "{ Removed Branches }\n\n";
	print STDOUT join ("\t", 'Br_ID', 'br_length(NJ)', 'IDs,of,exterior,descendants'), "\n\n";
	foreach my $br_info (@unsupported_branches) {
	    my ($br_id, $attr_nj, $ext_descs) = @{$br_info};
	    my $br_len = $attr_nj->[$INDX_LENGTH];
	    my $cnct_ext_descs = join (',', @{$ext_descs});
	    print STDOUT join ("\t", $br_id, $br_len, $cnct_ext_descs), "\n";
	}
	print STDOUT "\n";
	print STDOUT "{ Resulting reduced NJ tree (new top node: $new_topnode) }\n\n";
	my $option_bl = 1; # =1 if you display bootstrap at the branch, = 0 at the node.
	my $option_ni = 1; # = 1 if you display node ID, = 0 otherwise.
	unless ( br_output_tree ($new_topnode, %{$node2ch}, %{$id2name}, %{$br_attr_nj}, *STDOUT, $option_bl, $option_ni) ) {
	    print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] 'br_output_tree' FAILED while testing the subroutine: $! !\n";
	    return (undef, undef, undef, undef, undef);
	}
	print STDOUT "\n";
   }


		## Check consistency of each partition with the NJ tree ##

	# @{$consist_ptits[$c]} = (\$partition, \@rlv_positions, \@pars_scenarios)
	# is for the $c th partition that is consistent with the inferred genealogy;
	# @{$inconsist_ptits[$i]} = (\$partition, \@rlv_positions, \@pars_scenarios)
	# is for the $i th partition that is NOT consistent with the inferred genealogy (namely that cannot be represented by a single change along the genealogy), with
	# $partition being a string representing the partition, with
	# @rlv_positions storing all the positions showing the partition, and with
	# %{$pars_scenarios[$s]} = ($br_id => [$stat_upper, $stat_lower]), where $stat_upper/lower gives the state at the upper/lower end of the branch, describing the $s th parsimonious scenario;
	# @{$anc_inconsist_ptits[$i]} = (\$partition, \@relv_positions, \@anc_states)
	# is  for the $i th partition that is consistent with the inferred genealogy, but none of whose inferred ancestral states is $putat_anc_state, with
	# $partition being a string representing the partiton, with
	# @relf_positions storing all the positions with the partition, and with
	# @anc_states storing all the possible ancestral states;
	# @{$unsupported_branches[$u]} = ($br_id, \@br_attr_nj, \@ext_descendants)
	# is for the $u th branch NOT supported by the partitions (and therefore removed),
	# with @br_attr_nj describing the branch attribute of the NJ tree,
	# and with @ext_descendants listing the branch's descendants that are exterior nodes.

	my %br_attr;
	my @consist_ptits = my @inconsist_ptits = my @anc_inconsist_ptits = ();
	my $ct_consist_ptits = 0;


	foreach my $partition (sort keys %{$partition2positions}) { # 1st main outer foreach-loop (over partitions) #

		# %{$partition2positions} = ($partition => [\@rlv_indices, \@rlv_positions], ...)
	    my $rlv_positions = $partition2positions->{$partition}[1];
	    my $ct_rlv_positions = @{$rlv_positions};

	    my @segsite_state = split //, $partition;

		# Prepare for a list of characters and a penalty matrix #
	    my @charlist = @{$in_charlist};
	    my @pen_mtrx;
	    if (@charlist >0) {
 		@pen_mtrx = @{$in_pen_mtrx};
	    } else {
		my %char2ct;
		foreach my $char (@segsite_state) { # 1st preliminary inner foreach-loop #
		    (defined $char2ct{$char}) or ($char2ct{$char} = 0);
		    $char2ct{$char}++;
		} # End of the 1st preliminary inner foreach-loop #
		@charlist = sort keys %char2ct;
		@pen_mtrx = ();
		foreach my $char1 (@charlist) { # 2nd preliminary inner foreach-loop #
		    my @row = ();
		    foreach my $char2 (@charlist) { push @row, ($char2 eq $char1) ? 0 : 1; }
		    push @pen_mtrx, \@row;
		}			# End of the 2nd preliminary inner foreach-loop #
	    }

	    # $anc_states->[$s] is the ancestral state under the $s th scenario of the most parsimonous changes;
	    # %{$sets_parsim_chgs->[$s]} = ($br_id => [$stat_parent, $stat_node], ...), where $br_id is actually the ID of the node immediately below the branch in question,
	    # gives the $s th scenario of the most parsimonious changes at the site examined;
	    # $cts_parsim_chgs->[$s] gives the number of changes in the $s th scenario of the most parsimonious changes;
	    # $parsim_penalties->[$s] gives the total penalty on the $s th scenario of the most parsimonious changes;
	    #
			# REVISED in creation of 'ser_br_pars_out_msms.2.pl' (on Nov 10 (Thu), 2011) #
	    my ($anc_states, $sets_parsim_chgs, $cts_parsim_chgs, $parsim_penalties) = enumerate_parsim_changes_1site ($TOP_NODE, @genealogy, @{$chr_names}, @segsite_state, @charlist, @pen_mtrx);
	    unless (defined $sets_parsim_chgs) {
		print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] The subroutine 'enumerate_parsim_changes_1site' FAILED on the partition:\n";
		print STDOUT "'$partition': $! !\n";
		return (undef, undef, undef, undef, undef);
	    }
	    my $ct_scenarios = @{$sets_parsim_chgs};

	    my @consist_scenarios = my @scenarios_w_inconsist_anc = my @inconsist_scenarios = ();
	    for (my $s=0; $s<$ct_scenarios; $s++) { # 1st inner for-loop (over parsimonious scenarios) #
		if ($cts_parsim_chgs->[$s] >1) {
		    push @inconsist_scenarios, $s;
		} elsif ((defined $putat_anc_state) and ($putat_anc_state ne $anc_states->[$s])) {
		    push @scenarios_w_inconsist_anc, $s;
		} else {
		    push @consist_scenarios, $s;
		}
	    }		# End of the 1st inner for-loop (over parsimonious scenarios) #

	    if (@consist_scenarios>0) {
		$ct_consist_ptits += $ct_rlv_positions;

		my $ct_consist = @consist_scenarios;
		my @rlv_pars_scenarios = @{$sets_parsim_chgs}[@consist_scenarios];
		push @consist_ptits, [\$partition, $rlv_positions, \@rlv_pars_scenarios];

		     # add the number of relevant positions to the relevant branch length (if there are more than one scenario, evenly distribute the number among relevant branches) #
		my $incr_len = $ct_rlv_positions/$ct_consist;
		foreach my $scenario (@rlv_pars_scenarios) { # 2nd inner for-loop (over consistent parsimonious scenarios) #
		    foreach my $br_id (sort keys %{$scenario}) { # innermost for-loop (over state-changing branches) #
			if (defined $br_attr{$br_id}) {
			    $br_attr{$br_id}->[$INDX_LENGTH] += $incr_len;
			} else {
			    my @attr = ();
			    $attr[$INDX_LENGTH] = $incr_len;
			    $br_attr{$br_id} = \@attr;
			}
		    }	# End of the innermost for-loop (over state-changing branches) #
		}	# End of the 2nd inner for-loop (over consistent parsimonious scenarios) #

	    } elsif (@scenarios_w_inconsist_anc) {
		my @rlv_anc_states = @{$anc_states}[@scenarios_w_inconsist_anc]; 
		my @rlv_pars_scenarios = @{$sets_parsim_chgs}[@scenarios_w_inconsist_anc];
		push @anc_inconsist_ptits, [\$partition, $rlv_positions, \@rlv_anc_states, \@rlv_pars_scenarios];

	    } else {	# The partition is totally inconsistent with the NJ tree #
		push @inconsist_ptits, [\$partition, $rlv_positions, $sets_parsim_chgs];
	    }
	    
	}		# End of the 1st main outer foreach-loop (over partitions) #

   (print STDOUT "\n --- finished checking the consistency between the reduced NJ tree and the segregating sites ---\n\n") if ($LEVEL_TEST>0);
   (print STDOUT "#{consistent segregating sites} = $ct_consist_ptits\n\n") if ($LEVEL_TEST>0);
=set aside for later use
   if ($LEVEL_TEST>2) {
	my $ct_cons_ptits = @consist_ptits;
	print STDOUT "{ There are $ct_cons_ptits partitions consistent with the tree }\n\n";
	print STDOUT join ("\t", "'Partition'", 'Relv,Positions', 'putat,state,chg,branches'), "\n\n";
   }
=cut

	#my %br_attr;
	#my @inconsist_ptits = my @anc_inconsist_ptits =
	#my $ct_consist_ptits = 0;

	# @{$inconsist_ptits[$i]} = (\$partition, \@rlv_positions, \@pars_scenarios)
	# is for the $i th partition that is NOT consistent with the inferred genealogy (namely that cannot be represented by a single change along the genealogy), with
	# $partition being a string representing the partition, with
	# @rlv_positions storing all the positions showing the partition, and with
	# %{$pars_scenarios[$s]} = ($br_id => [$stat_upper, $stat_lower]), where $stat_upper/lower gives the state at the upper/lower end of the branch, describing the $s th parsimonious scenario;
	# @{$anc_inconsist_ptits[$i]} = (\$partition, \@relv_positions, \@anc_states, \@pars_scenarios)
	# is  for the $i th partition that is consistent with the inferred genealogy, but none of whose inferred ancestral states is $putat_anc_state, with
	# $partition being a string representing the partiton, with
	# @rlv_positions storing all the positions with the partition, and with
	# @anc_states storing all the possible ancestral states;
	# @{$unsupported_branches[$u]} = ($br_id, \@br_attr_nj, \@ext_descendants)
	# is for the $u th branch NOT supported by the partitions (and therefore removed),
	# with @br_attr_nj describing the branch attribute of the NJ tree,
	# and with @ext_descendants listing the branch's descendants that are exterior nodes.

		# Remove interior branches NOT supported by consistent partitions #
		# While rescaling the branch lengths #

	my $scale = 0;
	for (my $k=1; $k<$ct_chrm; $k++) { $scale += 1/$k; } 
	$scale /= $ct_consist_ptits;

   (print STDOUT "\n{Scaling factor for the branch lengths} = $scale\n\n") if ($LEVEL_TEST>0);
   (print STDOUT " ... Now we will remove branches NOT supported by the consistent partitions ...\n\n") if ($LEVEL_TEST>0);
   (print STDOUT join ("\t", 'Br_ID', 'br_length(NJ)', 'IDs,of,exterior,descendants'), "\n\n") if ($LEVEL_TEST>1);

	foreach my $br (sort keys %{$node2pa}) { # 2nd main foreach-loop (over branches) #

	    my $attr = $br_attr{$br};
	    if (defined $attr) {
		$attr->[$INDX_LENGTH] *= $scale;
		next;
	    }

	    unless (defined $node2ch->{$br}) { # Assign exterior branches with zero branch lengths #
		my @attr = ();
		$attr[$INDX_LENGTH] = 0;
		$br_attr{$br} = \@attr;
		next;
	    }

		# Interior branches are removed #

	    my $attr_nj = $br_attr_nj->{$br};
	    my $ext_descs = fetch_ext_offsprings_nr ($br, %{$node2ch});

	    unless ( br_contract_branch ($br, %{$node2pa}, %{$node2ch}) ) {
		print STDOUT "[ERROR 'incompl_neut_genealogy_via_nj'] 'br_contract_branch' FAILED on the interir branch '$br': $! !\n";
		return (undef, undef, undef, undef, undef);
	    }

   if ($LEVEL_TEST>1) {
	my $brlen_nj = $attr_nj->[$INDX_LENGTH];
	my $cnct_ext_descs = join (',', @{$ext_descs});
	print STDOUT join ("\t", $br, $brlen_nj, $cnct_ext_descs), "\n";
   }

	    push @unsupported_branches, [$br, $attr_nj, $ext_descs];
	}			# End of the 2nd main foreach-loop (over branches) #

   (print STDOUT " --- finished removing unsupported branches ---\n\n") if ($LEVEL_TEST>1);

 	  # Replace the NJ branch attributes with the attributes based on consistent partitions #
	   # (A FINISHING TOUCH on Nov 13 (Sun), 2011) #
	$genealogy[4] = \%br_attr;

   (print STDOUT "\n === The subroutine 'incompl_neut_genealogy_via_nj' ENDED ===\n\n") if ($LEVEL_TEST>0);

	return (\@genealogy, \@consist_ptits, \@inconsist_ptits, \@anc_inconsist_ptits, $pw_div_mtrx, \@unsupported_branches);

}		# End of the subroutine 'incompl_neut_genealogy_via_nj' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


