#!/usr/bin/perl
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# preprocess_msa_dawg.pl:
#	This Perl script pre-processes a MSA output by the molecular evolution simulator, Dawg.
#	The script was created by taking advantage of the framework of the script, 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'.
#
#	[CREATION STARTED on May 13 (Wed), 2015, ENDED on May xx (yyy), 2015.]
#
# br_infer_psm_indel_events_spt_odr_dawg.v2.pl:
#	This modified version of 'br_infer_psm_indel_events_dawg.v2.pl' infers parsimonious indel scenarios
#	while respecting the spatial order of gap-blocks in each cluster,
#	using the new module 'MyTreeMap_indels_spt_odr.pm' instead of is predecessor, 'MyTreeMap_indels.pm'.
#	(CREATED on Mar 21 (Fri), 2014)
#
# br_infer_psm_indel_events_dawg.v2.pl:
#	This modified version of 'br_infer_psm_indel_events_dawg.pl' prints out the start and end coordiates of each gap-block cluster.
#	(CREATED on Mar 9 (Sun), 2014)
#
#	REVISED on Mar 10 (Mon), 2014:
#	The script now copes with the new revision of 'MyTreeMap_indels.pm', where a new subroutine 'scan_gbcls_for_long_gaps' was created.
#
# br_infer_psm_indel_events_dawg.pl:
#	This modified version of 'br_infer_psm_indel_events.pl' can handle a control file and an output MSA of Dawg.
#	(CREATION STARTED on Feb 27 (Thu), 2014, ENDED on Feb 28 (Fri), 2014) 
#
#	REVISED on Mar 2 (Sun), 2014:
#	The script now continues the analysis even if it finds some gap-block clusters containing potential indel events
#	longer than the cut-off length.
#
#	REVISED on Mar 6 (Thu), 2014:
#	The script now continues the analysis even if encounters situations where the satellite subroutine,
#	'cal_lnlk1_indel_scn_cls', of the subroutine 'apprx1_tot_lnlk_gpattern_msa' failed for some reasons
#	(most likely because it encounters 3 or more indel events that overlap each other).
#
#
# br_infer_psm_indel_events.pl:
#	This Perl script infers parsimonious indel events resulting in a given MSA under a simple phylogeny,
#	as well as their relative probabilities.
#	The script was created by using some essential parts of 'compare_ref_vs_rec_abs_scoring.PhyML.pl'.
#	(CREATION STARTED on Nov 6 (Wed), 2013, ENDED on Nov 22 (Fri), 2013)
#

use strict;
use warnings;

use MyArrayManipu_num qw(
	diff_nmarray
	log_sum_exp
	);
use MyArrayManipu qw( binsearch_string );
use MyFile_manipu qw( open_input_file2 open_output_file list_sessions_ext );

use MyMalnIO qw(
	extract_fasta
	extract_alignment
	pr_ODEN_input
	pr_phylip_aln
	pr_clustal_like_lite
	);
	## pr_clustal_like_lite_aa
use MyMalnManipu qw( rm_gapcolumn2 );

use MyTreeManipu_II qw(
	extract_sgltree_from_file
	input_tree_file output_tree
	br_output_tree
	);

use MyTreeManipu_II_dawg qw(
	extract_sgltree_from_file_dawg
	input_tree_file_dawg
	br_output_tree_dawg
		) ;
	## read_tree_file_dawg input_subtree_dawg
	## output_tree_dawg draw_subtree_dawg
 	## br_draw_subtree_dawg
	## br_output_tree_sep_dawg br_draw_subtree_sep_dawg

use MyTreeManipu_IV qw(
	bi2multifurcated_top
	construct_neat_subtree
	construct_skeleton_subtree
	);

=for the predecessor 'compare_ref_vs_rec_abs_scoring.PhyML.pl'
use MyPhyML_parser qw(
	est_resfreqs_frm_msa print_custom_sbst_model
	extract_phyml_statistics2
	fetch_clmwise_likelihoods_phyml
	);
	## extract_phyml_statistics
=cut


	### For enumerating parsimonious indel scenarios. ###

	# A MAJOR MODIFICATION in creation of 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'. #
	# => Further MODIFICATION in creation of 'preprocess_msa_dawg.pl'. #
use MyTreeMap_indels_spt_odr qw(
	extract_cls_gblocks
	);
	##enum_psm_indel_scns_for_msa
=for the predecessor 'br_infer_psm_indel_events_dawg.v2.pl'
use MyTreeMap_indels qw(
	enum_psm_indel_scns_for_msa
	);
	## extract_cls_gblocks msa_rep_gpatterns
=cut

=for the predecessor 'br_infer_psm_indel_events_spt_odr_dawg.v2.pl'
	### For log-likelihoods of indel scenarios. ###
use MyTreeMap_indels_ML qw(
	cal_log_normal_facs_indel_distri
	cal_delta_exp_del_5p
	cal_lnlk1_indel_scn_cls
	apprx1_tot_lnlk_gpattern_msa
	);
=cut


sub partition_msa_via_usable_patterns (\@\%);
sub inspect_indel_events (\@\@$$$);
sub read_dawg_ctrl_file (\$);

=for the predecessor 'compare_ref_vs_rec_abs_scoring.PhyML.pl'
sub readin_list_segs (\$\%);
sub readin_simul_params (\$);

sub readin_dawg_tree (\$); # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.beta.pl' on Mar 15 (Tue), 2011 
sub ln_ml_sgl_seq2 (\$\%); # ADDED in creation of 'compare_apbayes_msa.bch.extJC.6.pl' on Mar 7 (Mon), 2011 #
sub add_dummy_seq (\@\@\%\%\%\%\%\@); # ADDED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 12 (Tue), 2011 #
=cut

$| = 1; # Flush the buffer of standard output everytime data flows into it.

my $program_path = $0;
my ($program_name) = ($program_path =~ m/([^\/]+)$/); 

	# INPORTED from 'ml_scoring_msa.bch.pl' on Mar 15 (Tue), 2011 #
(our $CURR_WORK_DIR = ` pwd `) =~ s/\s+$//;

	### Fixed Parameters ###

=for the predecessor 'br_infer_psm_indel_events.pl'
	# We just ignore null columns encountered,
	# because we know that the null columns in this analysis
	# resulted in indel events outside of the tree of analyzed sequences.
	# (ADDED on Nov 16 (Sat), 2013)
	#
$MyTreeMap_indels::FL_IGNORE_NULL_COLUMNS = 1;
=cut

our $DIS_LEN = 60; # Maximum length of the sequences to be displayed in a line

##our $GAP = '-'; # Character representing a gap
our $CHAR_ABS =  our $GAP  = '-';
our $CHAR_PRES = our $NGAP = 'X';

	# Characters indicating that the site is missing or that its orthology is questionable. 
our %CHAR_MISSING = ('?' => 1, 'N' => 1, 'n' => 1);

	# ID of the target sequence. # ADDED on Nov 19 (Tue), 2013. #
our $TGT_SEQ = 'seq0001'; # This is just a formality. #
##our $TGT_SEQ = 'seq0000'; # This is just a formality. #
##our $TGT_SEQ = 'Scer';

	# (ADDED on Nov 29 (Mon), 2010) #
our $TOP_NODE  = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
our $INDX_INTNODENAME = $MyTreeManipu_II_dawg::INDX_INTNODENAME;

	# (ADDED on Mar 7 (Mon), 2011) #
our $TOT_BOOT = -1; # Because bootstrap values are NOT necessary for this analysis. 
our @INDICES_ATTR = ($INDX_LENGTH, $INDX_BOOT, $TOT_BOOT);
our $COMMENT = '#';  # Indicator of the comment line in the tree file

our $INDX_STRING = undef;
	# Specifies the index for the reference to a string in the element array of a set of non-redundant columns.
	# = undef if the element itself is the reference to a string.

	### Set Parameters for the relative probabilities of parsimonious Indel Scenarios. ###

our %INDEL_MODEL2DISTRI = ( # = (indel model => distribution type, ...)
	'zipf' => 'P',
	'"PL"' => 'P',
	);


	### Parameters to be adjusted ###

	# The level of tests to examine the behavior of the script (ADDED on Sep 1 (Sun), 2012):
	# 0, nothing; 1, very brief; 2, a bit brief; 3, a bit verbose; 5, very verbose.

	# For the main part. #
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;
##our $TEST_LEVEL = 6;

	# For the subroutine 'inspect_indel_events'. #
our $TEST_LEVEL_INSP = 0;
##our $TEST_LEVEL_INSP = 1;
##our $TEST_LEVEL_INSP = 3;
##our $TEST_LEVEL_INSP = 5;

our $MIN_FREQ = 1E-06; # If the frequency of a residue is zero in the anchor MSA, this nominal frequency is used instead.

	# Essential for calculating the log-likelihoods of indel scenarios. #
	# $OPT_BD_INDELS = 0, fixed length; = 1, sum of rates; = 2, exact calculation.
#our $OPT_BD_INDELS = 0;
#our $OPT_BD_INDELS = 1;
our $OPT_BD_INDELS = 2;


	### Take in Command-Line Arguments. ###
	# (MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'.) #
	# (=> Further MODIFIED in creation of 'preprocess_msa_dawg.pl'.) #

unless (@ARGV > 2) {
##unless (@ARGV > 3) {
	print STDERR "\nUsage:  [perl] $program_name {in_msa_file} {dawg_ctrl_file} {out_file}\n";
=for the predecessor 'br_infer_psm_indel_events_dawg.pl'
	print STDERR "\nUsage:  [perl] $program_name {in_msa_file} {dawg_ctrl_file} {outfile_detail} {outfile_brief}\n";
=cut
	die "$!\n\n";
}
my ($in_msa_file, $dawg_ctrl_file, $outfile_detail) = @ARGV[0 .. 2];
##my ($in_msa_file, $dawg_ctrl_file, $outfile_detail, $outfile_brief) = @ARGV[0 .. 3];
my $intree_file = $dawg_ctrl_file;


	# Output files. #
	# (MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'.) #

my $outfile1 = $outfile_detail;
##my $outfile2 = $outfile_brief; # DISABLED in creation of 'preprocess_msa_dawg.pl'.

(my $outdir1 = $outfile1) =~ s!/?[^/]+$!!;
##(my $outdir2 = $outfile2) =~ s!/?[^/]+$!!; # DISABLED in creation of 'preprocess_msa_dawg.pl'.

($outdir1 =~ m/\S/) or ($outdir1 = '.'); 
##($outdir2 =~ m/\S/) or ($outdir2 = '.');  # DISABLED in creation of 'preprocess_msa_dawg.pl'.


	### Check input files/directories. ###

foreach my $file ($in_msa_file, $intree_file) {
    unless (( -f $file ) and ( -r $file )){
	print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE '$file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
}


foreach my $dir ($outdir1) {
##foreach my $dir ($outdir1, $outdir2) {...} # Preliminary foreach-loop (over output directories). # For the predecessor 'br_infer_psm_indel_events_dawg.pl'. #

    unless ( -d $dir ) {
	unless (0 == system ("mkdir -p $dir") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to create the directory '$dir': $? !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
    unless ( -w $dir ) {
	unless (0 == system (" chmod -R u+w $dir ") ) {
	    print STDOUT "[ERROR '$program_name'] FAILED to change the permission mode of the output directory '$dir': $?  !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die;
	}
    }
} 	# End of the preliminary foreach-loop (over output directories). #


	### Declare the Beginning of the Script ###

print STDOUT "\n<<<<<<< Program name: $program_name (path: $program_path) >>>>>>>>\n\n";

print STDOUT "Current working directory: $CURR_WORK_DIR .\n\n";

print STDOUT "Input MSA file: $in_msa_file .\n\n";

	## MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. ##
print STDOUT "Input tree and indel parameters: $intree_file .\n\n";

print STDOUT "Output files:\n";

	# MODIFIED in creation of 'preprocess_msa_dawg.pl'. #
print STDOUT "  (Preprocessed MSA) $outfile1 .\n\n\n";

##print STDOUT "   (Details)     $outfile1 ,\n"; # DISABLED in creation of 'preprocess_msa_dawg.pl'. #
##print STDOUT "  (Brief record) $outfile2 .\n\n\n"; # DISABLED in creation of 'preprocess_msa_dawg.pl'. #
##print STDOUT "Output file: $outfile .\n\n\n";


	### RETRIEVE the Parameters necessary for Indel Probabilities, as well as the guide tree. ###
	### (ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'.) ###

print STDOUT "\n<<<< As a preliminary process, read the Dawg control file '$intree_file' >>>\n\n\n";

my ($tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate)
	= read_dawg_ctrl_file ($intree_file);
	##= read_dawg_ctrl_file ($in_msa_file); # A BUG fixed on Feb 28 (Fri), 2014. #`
unless (defined $tree_string) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'read_dawg_ctrl-file' FAILED !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

our $INDEL_MODEL = $gapmodel;
our $DISTRI = $INDEL_MODEL2DISTRI{$INDEL_MODEL};

	# Total rate of insertions at a site relative to the substitution rate.
our $TOTRATE_INS = $tot_indel_rate /2;
	# Total rate of deletions at a site relative to the substitution rate.
our $TOTRATE_DEL = $tot_indel_rate /2;

 	# Exponent for the power-law distribution of insertion length (when $INDEL_MODEL = 'zipf').
our $PARAM_INS = $gap_param;
	# Exponent for the power-law distribution of deletion length (when $INDEL_MODEL = 'zipf').
our $PARAM_DEL = $gap_param;

	# Cut-off indel length at the 3'-end.
our $CUTOFF_3P = $gaplen_cutoff;
	# Cut-off indel length at the 5'-end.
our $CUTOFF_5P = $gaplen_cutoff;


our @PARAMS_INS = ($TOTRATE_INS, $DISTRI, $PARAM_INS);
our @PARAMS_DEL = ($TOTRATE_DEL, $DISTRI, $PARAM_DEL);

	### Print out the Parameters for Indel Probabilities ###

print STDOUT "\n<<<< Input Parameters for Indel Probabilities >>>>\n\n";

print STDOUT '@PARAMS_INS = ($TOTRATE_INS, $DISTRI, $PARAM_INS) = (', join (', ', @PARAMS_INS), ") ,\n";
print STDOUT '@PARAMS_DEL = ($TOTRATE_DEL, $DISTRI, $PARAM_DEL) = (', join (', ', @PARAMS_DEL), ") .\n\n";

	# ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
print STDOUT '$CUTOFF_3P = ', $CUTOFF_3P, " ,\n";
print STDOUT '$CUTOFF_5P = ', $CUTOFF_5P, " .\n\n\n";


	### Read in the input tree. ###

print STDOUT "\n<<<<< Now, read in the Input Tree >>>>>\n\n";

my %node2pa; # = ($node_id => $parent_node_id, ...) for all nodes except the top node. 
my %node2ch; # = ($node_id => \@child_node_ids, ...) for all interior nodes.
my %name2id; # = ($sequence_name => $node_id, ...) for all exterior nodes and the top-node.
my %id2name; # = ($node_id => $sequence_name, ...) for all exterior nodes.
# [my %br_attr; # = ($node_id => [$length, $bootstrap, ...], ...) for all nodes at the lowe-ends of the branches. ] --- for the predecessor 'br_infer_psm_indel_events.pl'. #
my %br_attr; # = ($node_id => [$length, $int_node_name, ...], ...) for all nodes at the lowe-ends of the branches.

	# ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
my %name2id_intnode; # = ($node_name => $node_id, ...) for all interior nodes. #
my $topnode_name; # The name of the top-node (root).

	# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #

unless ( input_tree_file_dawg (${$tree_string}, %node2pa, %node2ch, %name2id, %id2name, %br_attr, %name2id_intnode, $topnode_name) ) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input tree from the control file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

=for the predecessor 'br_infer_psm_indel_events.pl'.
unless ( extract_sgltree_from_file ($intree_file, %node2pa, %node2ch, %name2id, %id2name, %br_attr, $COMMENT) ) {

	print STDOUT "[ERROR '$program_name'] FAILED to extract the input tree from the file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
print STDOUT " --- 'extract_sgltree_from_file' performed on the file '$intree_file' ...\n\n";
=cut

my $topnode_id = $name2id{$TOP_NODE};

print STDOUT "[ Input Tree (with interior node names shown) ]:\n\n"; # MODIFIED on Mar 11 (Tue), 2014.
##print STDOUT "[ Input Tree ]:\n\n";

	# MODIFIED in creation of 'br_infer_psm_indel_events_dawg.pl'. #

unless ( br_output_tree_dawg ($topnode_id, %node2ch, %id2name, %br_attr, $topnode_name, *STDOUT, 0, 0) ) {
	print STDOUT "[ERROR '$program_name'] Subroutine 'br_output_tree_dawg' FAILED to output the input tree extracted from the control file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
##=for the predecessor 'br_infer_psm_indel_events.pl'.

	# ADDED for DEBUGGING the module 'MyTreeMap_indel.pm' (on Mar 11 (Tue), 2014). #
print "\n\n[ Input Tree (with node IDs shown) ]:\n\n";
my %br2blength;
foreach my $br (keys %br_attr) {
	my @attr = ();
	$attr[$INDX_LENGTH] = $br_attr{$br}->[$INDX_LENGTH];
	$br2blength{$br} = \@attr;
}

unless ( br_output_tree ($topnode_id, %node2ch, %id2name, %br2blength, *STDOUT, 1, 1) ) {
##unless ( output_tree ($topnode_id, %node2ch, %id2name, %br_attr, *STDOUT, 1, 1) ) {...} # --- For the predecessor 'br_infer_psm_indel_events.pl'.
	print STDOUT "[ERROR '$program_name'] Subroutine 'br_output_tree' FAILED to output the input tree extracted from the file '$intree_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
##=cut
print STDOUT "\n\n";


	### Read in the input MSA. ###

print STDOUT "\n<<<<< Now, read in the Input MSA >>>>>\n\n";

my ($in_msa, $seqids, $seq_lens) = extract_alignment ($in_msa_file);
unless (defined $in_msa) {
	print STDOUT "[ERROR '$program_name'] FAILED to extract the input MSA from the file '$in_msa_file' !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $ct_sequences = @{$seqids};

	# REMOVE the Ancestral Sequences from the Input MSA. #
	# (ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'.) #

my @msa_extseqs = my @ext_seqids = ();
my @msa_ancseqs = my @anc_seqids = (); # ADDED in creation of 'preprocess_msa_dawg.pl'. #

for (my $i=0; $i<$ct_sequences; $i++) { # Preliminary for-loop (over sequences in the input MSA). #

	my $seqid = $seqids->[$i];

		# MODIFIED in creation of 'preprocess_msa_dawg.pl'. #
	if ($seqid  =~ m/^anc/) {
	    push @anc_seqids, $seqid;
	    push @msa_ancseqs, $in_msa->[$i];
	    next;
	}
	##($seqid  =~ m/^anc/) and next;

	push @ext_seqids, $seqid;
	push @msa_extseqs, $in_msa->[$i];

}			# End of the preliminary for-loop (over sequences in the input MSA). #

@{$in_msa} = @msa_extseqs;
@{$seqids} = @ext_seqids;
$ct_sequences = @{$seqids};

unless ($ct_sequences > 1) {
	print STDOUT "[ERROR '$program_name'] The number of sequences ($ct_sequences) is TOO SMALL ! I will stop here ...\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

my $totlen_aln = length ${$in_msa->[0]};

=DISABLED on June 17 (Wed), 2015.
	## ADDED on Nov 19 (Tue), 2013. ##
my $indx_tgt;
for (my $k=0; $k<$ct_sequences; $k++) { ($seqids->[$k] eq $TGT_SEQ) and ($indx_tgt = $k); }
unless (defined $indx_tgt) {
	print STDOUT "[ERROR '$program_name'] Couldn't find the target sequence name '$TGT_SEQ' in the list of sequence names !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
=cut


print STDOUT "#{sequences in MSA} =  $ct_sequences .\n";
print STDOUT "#{columns in MSA}   = $totlen_aln .\n\n";

=DISABLED on June 17 (Wed), 2015.
print STDOUT "{Index for the target sequence '$TGT_SEQ'} =  $indx_tgt .\n\n"; # ADDED on Nov 19 (Tue), 2013. #
=cut

if ($TEST_LEVEL>6) {
##if ($TEST_LEVEL>2) {...}
    print STDOUT "\n[ Input MSA ]:\n\n";

    unless ( pr_ODEN_input (@{$seqids}, @{$in_msa}, 50, *STDOUT) ) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the input MSA !\n";
	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
    }
    print STDOUT "\n\n";
}


	### Partition the input MSA into segments, each with a particular pattern of 'usable' sequences. ###

print STDOUT "\n<<<<< Now, partition the Input MSA via usable site patterns >>>>>\n\n";

# @{$in_msa_segments->[$s]} = ($seg_start, $seg_end, \@seg_pattern, \@seg_msa) is for the $s th segment, where
#	$seg_start/end is the start/end coordinate of the segment (in the 0-based half-open convention);
#	$seg_pattern[$k] = 1 if the site in the $k th sequence is usable for the analysis,
#			 = 0 if the site is missing or its orthology is questionable (i.e., $CHAR_MISSING{$char} = 1);
#	${$seg_msa[$k]} is the $k th sequence embedded in the segment MSA.
#
my $in_msa_segments = partition_msa_via_usable_patterns (@{$in_msa}, %CHAR_MISSING);
unless (defined $in_msa_segments) {
	print STDOUT "[ERROR '$program_name'] 'partition_msa_via_usable_patterns' FAILED on the input MSA !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}
print STDOUT " --- 'partition_msa_via_usable_patterns' performed on the input MSA ...\n\n";

my $ct_segments = @{$in_msa_segments};

print STDOUT "#{Segments of particular missing site patterns} = $ct_segments .\n\n";

if ($TEST_LEVEL>3) {
    print STDOUT "[ MSA Segments ]:\n\n";

    for (my $s=0; $s<$ct_segments; $s++) { # Preliminary for-loop (over MSA segments) #

	my ($seg_start, $seg_end, $seg_pattern, $seg_msa) = @{$in_msa_segments->[$s]};

	my $cnct_seg_pattern = join ('', @{$seg_pattern});
	print STDOUT "  { The $s th segment\; Segion = \[$seg_start\, $seg_end\) }\n\n";
	print STDOUT "  Pattern (1:usable/0:missing) = '$cnct_seg_pattern' .\n\n";
	print STDOUT "  Alignment:\n\n";

   if ($TEST_LEVEL>6) {
	unless ( pr_ODEN_input (@{$seqids}, @{$seg_msa}, 50, *STDOUT) ) {
		print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the $s th segment !\n";
		print STDERR "[ERROR '$program_name'] $!\n\n";
		die ;
	}
	print STDOUT "\n\n";
   }

    }				# End of the preliminary for-loop (over MSA segments) #
    print STDOUT "\n";
}


	### MAIN ANALYSES:
	## Estimate and Compare the Relative Probabilities
	## of the Candidate (parsimonious) Indel Scenarios for the MSA segments. ###

	# MODIFIED in creation of 'preprocess_msa_dawg.pl'. #

print STDOUT "\n<<<<<<< The MAIN Pre-Processing >>>>>>>\n\n";
##print STDOUT "\n<<<<<<< The MAIN ANALYSES >>>>>>>\n\n";

my @preprocessed_seqids = (@anc_seqids, @ext_seqids);
my $ct_preprocessed_seqs = @preprocessed_seqids;
my @preprocessed_msa = ();

for (1 .. $ct_preprocessed_seqs) { my $seq = ''; push @preprocessed_msa, \$seq; }

# @msa_extseqs = my @ext_seqids = ();
# @msa_ancseqs = my @anc_seqids = (); # ADDED in creation of 'preprocess_msa_dawg.pl'. #

my $ct_anom_segs = my $ct_in_segs = my $ct_succ_segs = my $ct_fail_segs = my $ct_skipped_segs = 0;

## my $ref_coord_beg = my $rec_coord_beg = 0; # The 'beginning' coordinates of the segment in the reference MSA and reconstructed MSA, respectively. #

for (my $s = 0; $s<$ct_segments; $s++) { # Main outer for-loop (over MSA segments) #

	# @{$in_msa_segments->[$s]} = ($seg_start, $seg_end, \@seg_pattern, \@seg_msa) is for the $s th segment, where
	#	$seg_start/end is the start/end coordinate of the segment (in the 0-based half-open convention);
	#	$seg_pattern[$k] = 1 if the site in the $k th sequence is usable for the analysis,
	#			 = 0 if the site is missing or its orthology is questionable (i.e., $CHAR_MISSING{$char} = 1);
	#	${$seg_msa[$k]} is the $k th sequence embedded in the segment MSA.

	my ($seg_start, $seg_end, $seg_pattern, $seg_msa) = @{$in_msa_segments->[$s]};

	my $seg_len = $seg_end - $seg_start;
	my $cnct_seg_pattern = join ('', @{$seg_pattern});

	print STDOUT "<<<<<< The $s th Segment \; Region = \[$seg_start, $seg_end\) \; Length = $seg_len\; Usability pattern = '$cnct_seg_pattern'  >>>>>>\n\n";


    ##if ($TEST_LEVEL>1) {...} # MODIFIED on Nov 22 (Fri), 2013. # ...}
    ##if ($TEST_LEVEL>2) {...} # RE-Switched on Feb 28 (Fri), 2014. #
    if ($TEST_LEVEL>6) { # RE-Switched on Feb 28 (Fri), 2014. #
	print STDOUT "Original Alignment for the Segment:\n\n";
	unless ( pr_ODEN_input (@{$seqids}, @{$seg_msa}, 50, *STDOUT) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the segmental reference MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n\n";
    }

		# MODIFIED in creation of 'preprocess_msa_dawg.pl'. #

	my @red_seqnames = @{$seqids};
	my @red_smsa = @{$seg_msa};

	my @anc_smsa = ();

	foreach my $ancseq (@msa_ancseqs) {
	    my $anc_seg = substr (${$ancseq}, 0, $seg_len, '');
	    push @anc_smsa, \$anc_seg;
	}

	my $sbct_seqs = @red_seqnames;


			# Prepare a 'reduced' phylogenetic tree by excluding questionable sequences ##

	my %sb_name2id;
	my %sb_id2name;
	my ($sb_topnode_id, $sb_node2ch, $sb_node2pa, $sb_br_attr);

		# ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
	my %sb_name2id_intnode;
	my $sb_topnode_name;

	##my %node2pa; # = ($node_id => $parent_node_id, ...) for all nodes except the top node. 
	##my %node2ch; # = ($node_id => \@child_node_ids, ...) for all interior nodes.
	##my %name2id; # = ($sequence_name => $node_id, ...) for all exterior nodes and the top-node.
	##my %id2name; # = ($node_id => $sequence_name, ...) for all exterior nodes.
	##my %br_attr; # = ($node_id => [$length, $bootstrap, ...], ...) for all nodes at the lowe-ends of the branches.

	if ($sbct_seqs == $ct_sequences) {
		$sb_topnode_id = $topnode_id;
		$sb_node2ch = \%node2ch;
		$sb_node2pa = \%node2pa;
		$sb_br_attr = \%br_attr;
		%sb_name2id = %name2id;
		%sb_id2name = %id2name;

		# ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. #
		%sb_name2id_intnode = %name2id_intnode;
		$sb_topnode_name = $topnode_name;

	} else {

		# This is impossible. #

	    print STDOUT "[ERROR '$program_name'] Somehow, ", '$sbct_seqs ', "(=$sbct_seqs) DOES NOT EQUAL ", '$ct_sequences ', "(=$ct_sequences) !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
		# Handle Exceptional Case -- Part I + 1/2 -- #


		## (II-Pre) Pre Pre-Process:
		## Partition the MSA into blocks, each consisting of contiguous columns with the same presence/absence pattern among extant sequences.

	(print STDOUT "<<<< Step (II-Pre) Pre Pre-Process: Partition the MSA into blocks... >>>>\n\n") if ($TEST_LEVEL>0);

		# Copy the reduced MSA #
	my @cp_red_smsa = ();
	for (my $k=0; $k< $sbct_seqs; $k++) { my $cp_seg = ${$red_smsa[$k]}; push @cp_red_smsa, \$cp_seg; }

	my %gpattern2indx;
	my $ct_gpatterns = 0;

	my $set_cls_gblocks = extract_cls_gblocks (@cp_red_smsa, $CHAR_PRES, $CHAR_ABS, %gpattern2indx, $ct_gpatterns) ;
	unless (defined $set_cls_gblocks) {
	    print STDOUT "[ERROR '$program_name'] 'extract_cls_gblocks' FAILED !\n\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}

	my @set_gpatterns = ();
	foreach my $gpattern (keys %gpattern2indx) {
	    my $indx = $gpattern2indx{$gpattern};
	    $set_gpatterns[$indx] = $gpattern;
	}

	my $ct_cls = @{$set_cls_gblocks};

	my $last_end_cl = -1;
	##my $last_end_cl ; # A BUG fixed on May 14 (Thu), 2015.

	my @bf_preprocessed = (@anc_smsa, @red_smsa);

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Main middle for-loop (over clusters). # 

	    my $gblks_in_cl = $set_cls_gblocks->[$cl];
	    my $ct_gblks = @{$gblks_in_cl};

	    my $start_cl = $gblks_in_cl->[0][2];
	    my $end_cl = $gblks_in_cl->[$ct_gblks-1][3];

   (print STDOUT "<<< The $cl th cluster (#{blocks} = $ct_gblks, start = $start_cl, end = $end_cl) >>>\n\n") if ($TEST_LEVEL>1);

	    my $len_lflank = $start_cl - ($last_end_cl +1);
	    if ($len_lflank > 0) {
	    ##if ((defined $last_end_cl) and ($start_cl > $last_end_cl +1)) {...} # A BUG fixed on May 14 (Thu), 2015. #

   (print STDOUT "\n  --- Concatenate the left-flanking region, which consists of $len_lflank (gapless) columns ---\n\n") if ($TEST_LEVEL>2);

			# Concatenate the separater block. #

		for (my $i=0; $i< $ct_preprocessed_seqs; $i++) { # Auxiliary inner for-loop (over sequences). #
		    my $sbseq = substr (${$bf_preprocessed[$i]}, 0, $len_lflank, '');
		    ##my $sbseq = substr (${$bf_preprocessed[$i]}, 0, $start_cl - $last_end_cl - 1, '');
		    ${$preprocessed_msa[$i]} .= $sbseq;
		}				# End of the auxiliary inner for-loop (over sequences). #
	    }

		## (II-1) Main Pre-Process 1:
		## Remove blocks where all extant sequences have gaps (even if some ancestral sequences have non-gap characters).
		##
		## [MODIFIED in creation of 'preprocess_msa_dawg.pl'.)

   (print STDOUT "\n<< Step (II-1) Main Pre-Process 1: Remove blocks where all extant sequences have gaps... >>\n\n") if ($TEST_LEVEL>2);

	    my @ext_blkwise_msas = (); # element = [\@indices_seqs_pres, \@blockwise_msa] for each block. # @indices_seqs_pres stores the indices of the extant sequences with the "presence" state.

	    my $last_end_gb;
	    for (my $gb=0; $gb < $ct_gblks; $gb++) { # 1st main inner for-loop (over blocks). #

		my ($indx_gp, $len_gb, $start_gb, $end_gb) = @{$gblks_in_cl->[$gb]};
		my $gpattern = $set_gpatterns[$indx_gp];

   (print STDOUT "\n   <  The $gb th block (length = $len_gb; start = $start_gb; end = $end_gb; gap-pattern = '$gpattern') >\n\n") if ($TEST_LEVEL>3);

		my @blockwise_msa = ();
		for (my $i=0; $i<$ct_preprocessed_seqs; $i++) {
		    my $sbseq = substr (${$bf_preprocessed[$i]}, 0, $len_gb, '');
		    $blockwise_msa[$i] = \$sbseq;
		}

   if ($TEST_LEVEL >= 5) {

	print STDOUT "  [ Blockwise MSA ]\n\n";

	unless ( pr_ODEN_input (@preprocessed_seqids, @blockwise_msa, 50, *STDOUT) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the blockwise MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n\n";
   }

			### RESTART FROM HERE ###


		my @dec_gpattern = split //, $gpattern ;
		my @indices_seqs_pres = ();
		for (my $k=0; $k< $sbct_seqs; $k++) { ($dec_gpattern[$k] eq $CHAR_PRES) and (push @indices_seqs_pres, $k); }

    if ($TEST_LEVEL >= 5) {
	print STDOUT '   @indices_seqs_pres = (', join (', ', @indices_seqs_pres), ").\n\n\n";
    }

		##my $ct_pres = 0;
		##foreach my $state (@dec_gpattern) { ($state eq $CHAR_PRES) and ($ct_pres++); }

		if (@indices_seqs_pres > 0) {
		##if ($ct_pres > 0) {...}
			
		    push @ext_blkwise_msas, [\@indices_seqs_pres, \@blockwise_msa];
		} else {
			# If there is no 'presence' state, the block will be excluded. #

   (print STDOUT "     ... Remove this block, which is null ...\n\n") if ($TEST_LEVEL >= 5);
		}

		$last_end_gb = $end_gb;

	    }				# End of the 1st main inner for-loop (over blocks). #


	    my $ct_blkwise_msas = @ext_blkwise_msas;

	    unless ($ct_blkwise_msas>0) { # If no useful blockwise MSAs, move on to the next cluster.

   (print STDOUT "\n   --- This cluster consists only of gaps, so it will be SKIPPED ---\n\n\n") if ($TEST_LEVEL > 2);

		$last_end_cl = $end_cl;
		next;
	    }

		## (II-2) Main Pre-Process 2:
		## 2. Re-order the blocks with non-overlapping set of extant sequences with the 'presence' state.
		##
		## [MODIFIED in creation of 'preprocess_msa_dawg.pl'.)

   (print STDOUT "\n<< Step (II-2) Main Pre-Process 2: Re-order the blocks with non-overlapping set of extant sequences with the 'presence' state... >>\n\n") if ($TEST_LEVEL>2);

	    my $ct_reorderings = 1; # Initialize the counter. #

	    while ($ct_reorderings > 0) { # 2nd main inner while-loop (iteration). #

		$ct_reorderings = 0; # Reset the counter. #

		my @new_ext_blkwise_msas = (shift @ext_blkwise_msas);

		while (my $current = shift @ext_blkwise_msas) { # Inner-inner while-loop (over blockwise MSAs). #

		    my $previous = pop @new_ext_blkwise_msas;

		    my ($shared, $ex_curr, $ex_prev) = diff_nmarray (@{$current->[0]}, @{$previous->[0]}) ;

		    if ((@{$shared}>0) or ($previous->[0][0] > $current->[0][0])) {

				# Keep the order. #
			push @new_ext_blkwise_msas, $previous, $current;

		    } else {
				# Swap the orders of the two blocks. #

			push @new_ext_blkwise_msas, $current, $previous;

			$ct_reorderings++;
		    }

		} # Inner-inner while-loop (over blockwise MSAs). #

			# Update the set of extended blockwise MSAs. #
		@ext_blkwise_msas = @new_ext_blkwise_msas;

	    }				# End of the 2nd main inner while-loop (iteration). #

		## (II-Post) Post Pre-Process:
		##  Concatenate the re-ordered blockwise MSAs to the growing MSA.

   (print STDOUT "\n<< Step (II-Post) Post Pre-Process: Serially concatenate the re-ordered blocks-wise MSAs to the growing MSA. >>\n\n") if ($TEST_LEVEL>2);

	    for (my $bl=0; $bl<$ct_blkwise_msas; $bl++) { # 3rd main inner for-loop (over blockwise MSAs). #

		my $current = $ext_blkwise_msas[$bl];

   (print STDOUT "   < The $bl th blockwise MSA >\n\n") if ($TEST_LEVEL>3);

   if ($TEST_LEVEL >= 5) {

	print STDOUT "  [ Blockwise MSA ]\n\n";

	unless ( pr_ODEN_input (@preprocessed_seqids, @{$current->[1]}, 50, *STDOUT) ) {
	    print STDOUT "[ERROR '$program_name'] The subroutine 'pr_ODEN_input' FAILED on the blockwise MSA !\n";
	    print STDERR "[ERROR '$program_name'] $!\n\n";
	    die ;
	}
	print STDOUT "\n\n";
   }

		for (my $i=0; $i< $ct_preprocessed_seqs; $i++) { ${$preprocessed_msa[$i]} .= ${$current->[1][$i]}; }

	    }				# End of the 3rd main inner foreach-loop (over blockwise MSAs). #

			# Update the variable. #
	    $last_end_cl = $end_cl;

	}			# End of the main middle for-loop (over clusters). #

	# @red_seqnames = @seqids;
	# @red_smsa = @{$seg_msa};
	# @anc_smsa = ();
	#foreach my $ancseq (@msa_ancseqs) {
	#    my $anc_seg = substr (${$ancseq}, 0, $seg_len, '');
	#    push @anc_smsa, \$anc_seg;
	#}


	my $len_right_end = ($seg_len - 1) - $last_end_cl ;

	if ($len_right_end>0) { # If the segment contains a gapless segment on the right-end, concatenate it to the growing MSA. #

   (print STDOUT "<<< Concatenate the right-end region, which consists of $len_right_end (gapless) columns ... >>>\n\n\n") if ($TEST_LEVEL> 2) ;

	    for (my $i=0; $i< $ct_preprocessed_seqs; $i++) { # Auxiliary inner for-loop (over sequences). #
		my $sbseq = substr (${$bf_preprocessed[$i]}, 0, $len_right_end, '');
		${$preprocessed_msa[$i]} .= $sbseq;
	    }				# End of the auxiliary inner for-loop (over sequences). #
	}

	# ($seg_start, $seg_end, $seg_pattern, $seg_msa) = @{$in_msa_segments->[$s]};
	# $seg_len = $seg_end - $seg_start;

	$ct_succ_segs++;

}			# END of the main outer for-loop (over MSA segments) #

##close ($fh_out); # DISABLED in creation of 'preprocess_msa_dawg.pl'. #
##close ($fh_out2); # DISABLED in creation of 'preprocess_msa_dawg.pl'. #


print STDOUT "\n\n === All the Analyses FINISHED !! ===\n\n";

print STDOUT "#{Segments skipped} = $ct_skipped_segs .\n";
print STDOUT "#{Anomalous segments} = $ct_anom_segs .\n";
print STDOUT "#{Segments used as inputs} = $ct_in_segs .\n";
print STDOUT "#{Segments successfully processed} = $ct_succ_segs .\n";
print STDOUT "#{Segments failed to be processed} = $ct_fail_segs .\n\n";

unless (0 == $ct_fail_segs) {
	print STDOUT "[ERROR '$program_name'] FAILED to process $ct_fail_segs segments !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

print STDOUT "\n\n === Finally, Output the Pre-Processed MSA ===\n\n"; 

unless ( pr_clustal_like_lite (@preprocessed_seqids, @preprocessed_msa, $DIS_LEN, $outfile1) ) {
	print STDOUT "[ERROR '$program_name'] 'pr_clustal_like_lite' FAILED on the output file '$outfile1' !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

=Set aside for later use
my $fh_out = open_output_file ($outfile1, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] 'open_output_file' FAILED on the output file '$outfile1' !\n";
 	print STDERR "[ERROR '$program_name'] $!\n\n";
	die ;
}

close ($fh_out);
=cut

print STDOUT " ... Done !! ...\n\n";


	### DEFINITIONS of the Subroutines. ###


# partition_msa_via_usable_patterns:
#	This subroutine partitions an input MSA according to the
#	pattern of usability of sequences for the indel inference afterwards.
#
# INPUT: (\@in_msa, \%char_missing), where
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	%char_missing = ($character => {some defined value}, ...) for the characters
#	indicating that the site is missing or that the site's orthology is questionable.
#
# OUTPUT: \@msa_segments, where
#	@{$msa_segments[$s]} = ($seg_start, $seg_end, \@seg_pattern, \@seg_msa) is for the $s th segment, where
#	$seg_start/end is the start/end coordinate of the segment (in the 0-based half-open convention);
#	$seg_pattern[$k] = 1 if the site (including a gap) in the $k th sequence is usable for the analysis,
#			 = 0 if $char_missing{$char} is defined for the $k th sequence at the site in question.
#	${$seg_msa[$k]} is the $k th sequence embedded in the segment MSA.
#
sub partition_msa_via_usable_patterns (\@\%) {

	my ($in_msa, $char_missing) = @_;

	my $ct_seqs = @{$in_msa};
	unless ($ct_seqs>0) {
	    print STDOUT "[ERROR 'partition_msa_via_usable_patterns'] ILLEGAL INPUT: The input MSA contains NO SEQUENCE: $! !\n";
	    return undef;
	}

	my $msa_len = length ${$in_msa->[0]};

	my @msa_segments = ();

	my $last_cnct_pattern = '';

	for (my $l=0; $l<$msa_len; $l++) { # Main for-loop (over sites) #

	    my @chars = my @seg_pattern = ();

	    for (my $k=0; $k<$ct_seqs; $k++) { # Inner for-loop (over sequences) #

		my $char = substr (${$in_msa->[$k]}, 0, 1, '');
		push @chars, $char;
		push @seg_pattern, (defined $char_missing->{$char}) ? 0 : 1; 

	    }				# End of the inner for-loop (over sequences) #

	    my $cnct_pattern = join ('', @seg_pattern);

	    if ($cnct_pattern eq $last_cnct_pattern) {

			# Elongate the last segment. #

		my $last_segment = $msa_segments[$#msa_segments];

		$last_segment->[1]++;

		for (my $k=0; $k<$ct_seqs; $k++) { ${$last_segment->[3][$k]} .= $chars[$k]; }

	    } else {
			# Create a new segment. #

		my @seg_msa = ();
		foreach my $char (@chars) { push @seg_msa, \$char; }

		push @msa_segments, [$l, $l+1, \@seg_pattern, \@seg_msa];

		$last_cnct_pattern = $cnct_pattern;
	    }

	}			# End of the main for-loop (over sites) #

	return \@msa_segments;

}	# End of the subroutine 'partition_msa_via_usable_patterns' #


# inspect_indel_events:
#	This subroutine examines whether all the indel events are shorter than or equal to the cutoff-lengths.
#
# INPUT: (\@set_cls_gblocks, \@sets_psm_scns_for_cls, $cutoff_len_ins, $cutoff_len_del, $char_abs), where
#
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)
#	is for the $gb th block in the $cl th cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordiate of the block (in the 0-based convention);
#
#	@{$set_psm_scns_in_cls[$cl]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th of the parsimonious scenarios realizing the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) if the map of gap states onto the interior nodes,
#	with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id;
#	
#	$cutoff_len_ins/del is the cutoff length for the insertion/deletion events;
#
#	$char_abs is the character representing the 'absence' state.
#
# OUTPUT: \@illegal_indels, where
#	@{$illegal_indels[$il]} = ($cl, $ps, $br_id, $id, $length) is for the $il th illegal indel event, with
#	$cl specifying a cluster,
#	$ps specifying a parsimonious scenario in the cluster,
#	$br_id specifying a branch,
#	$id specifying an illegal event in the node,
#	and $length is the length of the event.
#
#	The subroutine returns undef if it fails.
#
sub inspect_indel_events (\@\@$$$) {

	my ($set_cls_gblocks, $sets_psm_scns_for_cls, $cutoff_len_ins, $cutoff_len_del, $char_abs) = @_;

    (print STDOUT ' ' x 3, "< Subroutine 'inspect_indel_events' >\n\n") if ($TEST_LEVEL_INSP>1);

	my $ct_cls = @{$set_cls_gblocks};
	unless (@{$sets_psm_scns_for_cls} == $ct_cls) {
	    print STDOUT "[ERROR 'inspect_indel_events'] ILLEGAL INPUT: The size of ", '@{$sets_psm_scns_for_cls} = (', scalar (@{$sets_psm_scns_for_cls}), ") CLASHES with the number of gap-block clusters (=$ct_cls): $! !\n";
	    return undef;
	}

	my @illegal_indels = ();

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Outer for-loop (over clusters) #

	    my $cls_gblock = $set_cls_gblocks->[$cl];
	    my $psm_scns_for_cls = $sets_psm_scns_for_cls->[$cl];
	    my $ct_psm_scns = @{$psm_scns_for_cls};

    (print STDOUT "\n", ' ' x 3, "[ The $cl th cluster has $ct_psm_scns parsimonious scenarios of indel events ]\n\n") if ($TEST_LEVEL_INSP>2);

	    for (my $ps=0; $ps<$ct_psm_scns; $ps++) { # Middle for-loop (over parsiminous scenarios) #

    (print STDOUT "\n", ' ' x 5, "\( The $ps th parsimonious scenario \)\n\n") if ($TEST_LEVEL_INSP>3);

		my ($br2indels, $node2bw_gstat) = @{$psm_scns_for_cls->[$ps]};

		foreach my $br_id (sort {$a <=> $b} keys %{$br2indels}) { # Inner-middle foreach-loop (over branches) #
		    my $indel_events = $br2indels->{$br_id};
		    my $ct_events = @{$indel_events};

    (print STDOUT ' ' x 7, "( The branch '$br_id' has $ct_events nominal indel events )\n") if ($TEST_LEVEL_INSP>3);

		    for (my $id=0; $id<$ct_events; $id++) { # Inner for-loop (over indel events) #
			my ($stat_pa, $stat_node, $indices_blk) = (defined $indel_events->[$id]) ? @{$indel_events->[$id]} : ();
			unless (defined $stat_pa) {
			    print STDOUT "[WARNING 'inspect_indel_events'] The $id th event on the branch '$br_id' is EMPTY  for the $ps th indel scenario for the $cl th cluster: $! !\n";
			    next;
			}
			my $cutoff_len = ($stat_pa eq $char_abs) ? $cutoff_len_ins : $cutoff_len_del;
			my $indel_len = 0;
			foreach my $indx_blk (@{$indices_blk}) { $indel_len += $cls_gblock->[$indx_blk][1] ; }

			unless ($indel_len <= $cutoff_len) {
			    print ' ' x 7, "\n[WARNING 'inspect_indel_events'] The $id th indel event along the $br_id th branch is EXCEEDINGLY LONG (length = $indel_len)  for the $ps th scenario in the $cl th gap-block cluster: $! !\n\n";
			    push @illegal_indels, [$cl, $ps, $br_id, $id, $indel_len];
			}

		    }				# End of the inner for-loop (over indel events) #

		} 		# End of the inner-middle foreach-loop (over branches) #

	    }			# End of the middle for-loop (over parsimonious scenarios) #

	}			# End of the outer for-loop (over clusters) #

    (print STDOUT "\n\n", ' ' x 3, "--- End of the Subroutine 'inspect_indel_events'  ---\n\n\n") if ($TEST_LEVEL_INSP>1);

	return \@illegal_indels;

}	# End of the subroutine 'inspect_indel_events' #


	## ADDED in creation of 'br_infer_psm_indel_events_dawg.pl'. ##

# read_dawg_ctrl_file:
#	This subroutine reads a Dawg control file,
#	and retrieves indel parameters and a guide tree.
#
# INPUT: \$ctrl_file, where
#	$infile is the path to a Dawg control file.
#
# OUTPUT: (\$tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate), where
#	$tree_string is a string representing the guide tree;
#	$gapmodel specifies the gap model;
#	$gap_param is the power exponent if $gapmodel = '"PL"',
#			the negative binomial parameter if $gapmodel = '"NB"';
#	$gaplen_cutoff is the cut-off for the gap-length;
#	$tot_indel_rate is the total insertion/deletion rate (per site).
#
sub read_dawg_ctrl_file (\$) {

	my ($ctrl_file) = @_;
	local $/ = "\n";

	my $fh_ctrl = open_input_file2 (${$ctrl_file});
	unless (defined $fh_ctrl) {
	    print STDOUT "[ERROR 'read_dawg_ctrl_file'] 'open_input_file2' FAILED on the control file '${$ctrl_file}': $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

	my ($tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate);

	while (my $line = <$fh_ctrl>) { # Main while-loop (over lines). #

	    ($line =~ m/^\#/) and next; # Skip a comment line. #

	    $line =~ s/\s+$//;
	    $line =~ s/\;$//; # Added to prevent an error message (concerning $tot_indel_rate) on Mar 10 (Mon), 2014. #

	    my ($option_name, $value) = split /\s*\=\s*/, $line;

	    if ($option_name eq 'Tree') {
		$tree_string = $value;

	    } elsif ($option_name eq 'GapModel') {
		$gapmodel = $value;

	    } elsif ($option_name eq 'GapParams') {

		$value =~ s/^\{//;
		$value =~ s/\}$//;
		($gap_param, $gaplen_cutoff) = split /\s*\,\s*/, $value;

	    } elsif ($option_name eq 'Lambda') {

		$tot_indel_rate = $value;
	    }

	}			# End of the main while-loop (over lines). #

	close ($fh_ctrl);

	foreach my $var ($tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate) { # Trailing foreach-loop. #
	    (defined $var) and next;
	    return (undef, undef, undef, undef, undef);
	}		# End of the trailing foreach-loop. #

	return (\$tree_string, $gapmodel, $gap_param, $gaplen_cutoff, $tot_indel_rate);

}	# End of the subroutine 'read_dawg_ctrl_file'. # 

